/*****************************************************************

  audiobuf.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Routines for the audio buffer class

#include "config.h"

#include <glib.h>
#include <pthread.h>
#include <string.h>


#include <iostream>
#include <iosfwd>

using std::cerr;
using std::ostream;

#include "gvtime.h"
#include "gvconfig.h"
#include "buffer.h"
#include "audioformat.h"
#include "audiobuf.h"
#include "audioformat.h"

#include "videoformat.h"
// #include "audiobuf.h"
// #include "audioformat.h"

#include "baseplugin.h"
#include "plugin.h"
#include "output.h"
#include "input.h"


// Audio format conversion routines

template <class T> void interleave_none_to_2(T ** src_channels,
                                             T ** dst_channels,
                                             int num_samples,
                                             int num_channels)
  {
  const T * src1 = src_channels[0];
  const T * src2 = src_channels[1];
  T * dst = dst_channels[0];
  
  int jmax = num_channels/2;
  
  for(int j = 0; j < jmax; j++)
    {
    src1 = src_channels[j*2];
    src2 = src_channels[j*2+1];
    for(int i = 0; i < num_samples; i++)
      {
      *(dst++) = *(src1++);
      *(dst++) = *(src2++);
      }
    }

  // Copy the last channel as a mono channel
  
  if(num_channels % 2)
    memcpy(dst, src2, num_samples*sizeof(T));
  }


template <class T> void interleave_2_to_none(T ** src_channels,
                                             T ** dst_channels,
                                             int num_samples,
                                             int num_channels)
  {
  T * dst1 = dst_channels[0];
  T * dst2 = dst_channels[1];
  
  const T * src = src_channels[0];
  
  int jmax = num_channels/2;
  
  for(int j = 0; j < jmax; j++)
    {
    dst1 = dst_channels[j*2];
    dst2 = dst_channels[j*2+1];
    for(int i = 0; i < num_samples; i++)
      {
      *(dst1++) = *(src++);
      *(dst2++) = *(src++);
      }
    }

  // Copy the last channel as a mono channel
  
  if(num_channels % 2)
    memcpy(dst2, src, num_samples*sizeof(T));
  }



template <class T> void interleave_none_to_all(T ** src_channels,
                                               T * dst_samples,
                                               int num_samples,
                                               int num_channels)
  {
  for(int i = 0; i < num_samples; i++)
    for(int j = 0; j < num_channels; j++)
      *(dst_samples++) = src_channels[j][i];
  }

template <class T> void interleave_all_to_none(T * src_samples,
                                               T ** dst_channels,
                                               int num_samples,
                                               int num_channels)
  {
  for(int i = 0; i < num_samples; i++)
    for(int j = 0; j < num_channels; j++)
      dst_channels[j][i] = *(src_samples++);
  }


template <class T> void interleave_2_to_all(T ** src_channels, T * dst_samples,
                                            int num_samples, int num_channels)
  {
  T * dst = dst_samples;
  
  int jmax = num_channels/2;

  for(int i = 0; i < num_samples; i++)
    {
    for(int j = 0; j < jmax; j++)
      {
      *(dst++) = src_channels[2*j][i];
      *(dst++) = src_channels[2*j+1][i];
      }
    if(num_channels % 2)
      *(dst++) = src_channels[num_channels][i];
    }
  
  }

// Mix "mono_channel" onto 2 non interleaved stereo channels
// All channels must be signed

// mix_5_to_4_s<gint8>(src->channels.s_8, src_format->samples_per_frame);


template <class T> void mix_5_to_4_s(T ** channels, int num_samples)
  {
  gint32 tmp;
  T * front_left  = channels[0];
  T * front_right = channels[1];
  T * center =      channels[4];
  
  for(int i = 0; i < num_samples; i++)
    {
    tmp = *front_left + *center;
    *(front_left++) = tmp >> 1;
    //    *(front_left++) = *center;
    tmp = *front_right + *center;
    *(front_right++) = tmp >> 1;
    //    *(front_right++) = *center;
    center++;
    }
  }

template <class T> void mix_5_to_2_s(T ** channels, int num_samples)
  {
  gint32 tmp;
  T * front_left  = channels[0];
  T * front_right = channels[1];
  T * rear_left   = channels[2];
  T * rear_right  = channels[3];
  T * center =      channels[4];

  for(int i = 0; i < num_samples; i++)
    {
    tmp = *front_left + *center + *(rear_left++);
    *(front_left++) = tmp / 3;
    tmp = *front_right + *center + *(rear_right++);
    *(front_right++) = tmp / 3;
    center++;
    }
  }

template <class T> void mix_4_to_2_s(T ** channels,
                                     int num_samples)
  {
  gint32 tmp;

  T * front_left  = channels[0];
  T * front_right = channels[1];
  T * rear_left   = channels[2];
  T * rear_right  = channels[3];

  for(int i = 0; i < num_samples; i++)
    {
    tmp = *front_left + *(rear_left++);
    *(front_left++) = tmp >> 1;
    tmp = *front_right + *(rear_right++);
    *(front_right++) = tmp >> 1;
    }
  }

static inline void swap_endian(guint16 * samples, int num_samples)
  {
  for(int i = 0; i < num_samples; i++)
    samples[i] = ((samples[i] & 0xff) << 8) | ((samples[i] & 0xff00) >> 8);
  }

template <class T> void swap_sign(T * samples, int num_samples)
  {
  int mask = 1 << (sizeof(T) * 8 - 1);
  for(int i = 0; i < num_samples; i++)
    samples[i] = samples[i] ^ mask;
  }

void GVBufferAudio::free_frame(GVAudioFrame * frame,
                               const GVAudioFormat * format)
  {
  switch(format->sample_type)
    {
    case GVAudio_8_signed:
      delete frame->samples.s_8;
      delete frame->channels.s_8;
      break;
    case GVAudio_16_signed:
      delete frame->samples.s_16;
      delete frame->channels.s_16;
      break;
    case GVAudio_8_unsigned:
      delete frame->samples.u_8;
      delete frame->channels.u_8;
      break;
    case GVAudio_16_unsigned:
      delete frame->samples.u_16;
      delete frame->channels.u_16;
      break;
    }
  }

void GVBufferAudio::alloc_frame(GVAudioFrame * frame,
                                const GVAudioFormat * format)
  {
  switch(format->sample_type)
    {
    case GVAudio_8_signed:
      frame->samples.s_8
        = new gint8[format->samples_per_frame * format->num_channels];
      frame->channels.s_8
        = new gint8*[format->num_channels];
      
      for(unsigned int i = 0; i < format->num_channels; i++)
        frame->channels.s_8[i] =
          &(frame->samples.s_8[i*(format->samples_per_frame)]);
      break;
    case GVAudio_16_signed:
      frame->samples.s_16
        = new gint16[format->samples_per_frame * format->num_channels];
      frame->channels.s_16
        = new gint16*[format->num_channels];

      for(unsigned int i = 0; i < format->num_channels; i++)
        frame->channels.s_16[i] =
          &(frame->samples.s_16[i*(format->samples_per_frame)]);
      break;
    case GVAudio_8_unsigned:
      frame->samples.u_8
        = new  guint8[format->samples_per_frame * format->num_channels];
      frame->channels.u_8
        = new guint8*[format->num_channels];

      for(unsigned int i = 0; i < format->num_channels; i++)
        frame->channels.u_8[i] =
          &(frame->samples.u_8[i*(format->samples_per_frame)]);
      break;
    case GVAudio_16_unsigned:
      frame->samples.u_16
        = new  guint16[format->samples_per_frame * format->num_channels];
      frame->channels.u_16
        = new guint16*[format->num_channels];

      for(unsigned int i = 0; i < format->num_channels; i++)
        frame->channels.s_16[i] =
          &(frame->samples.s_16[i*(format->samples_per_frame)]);
      break;
    }
  frame->valid_samples = 0;
  }

void GVBufferAudio::convert(GVAudioFrame * src, GVAudioFrame * dst)
  {
  // Do downmixing first since this reduces the amount of data
  // for later processing

  int test;

  if((test = convert_mode & GV_BUFFER_AUDIO_CONVERT_DOWNMIX_MASK))
    {
    switch(test)
      {
      case GV_BUFFER_AUDIO_CONVERT_DOWNMIX_5_TO_4:
        switch(input_format.sample_type)
          {
          case GVAudio_8_signed:
            mix_5_to_4_s<gint8>(src->channels.s_8, input_format.samples_per_frame);
            break;
          case GVAudio_16_signed:
            mix_5_to_4_s<gint16>(src->channels.s_16, input_format.samples_per_frame);
            break;
          default:
            cerr << "Cannot downmix!!\n";
          }
        break;
      case GV_BUFFER_AUDIO_CONVERT_DOWNMIX_5_TO_2:
        switch(input_format.sample_type)
          {
          case GVAudio_8_signed:
            mix_5_to_2_s<gint8>(src->channels.s_8, input_format.samples_per_frame);
            break;
          case GVAudio_16_signed:
            mix_5_to_2_s<gint16>(src->channels.s_16, input_format.samples_per_frame);
            break;
          default:
            cerr << "Cannot downmix!!\n";
          }
      case GV_BUFFER_AUDIO_CONVERT_DOWNMIX_4_TO_2:
        switch(input_format.sample_type)
          {
          case GVAudio_8_signed:
            mix_4_to_2_s<gint8>(src->channels.s_8, input_format.samples_per_frame);
            break;
          case GVAudio_16_signed:
            mix_4_to_2_s<gint16>(src->channels.s_16, input_format.samples_per_frame);
            break;
          default:
            cerr << "Cannot downmix!!\n";
          }
      }
    }

  if((test = convert_mode & GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_MASK))
    {
    
    // Interleave channels
    switch(test)
      {
      case GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_2_TO_ALL:
        switch(input_format.sample_type)
          {
          case GVAudio_8_signed:
            interleave_2_to_all<gint8>(src->channels.s_8, dst->samples.s_8,
                                       output_format.samples_per_frame,
                                       output_format.num_channels);
            break;
          case GVAudio_16_signed:
            interleave_2_to_all<gint16>(src->channels.s_16, dst->samples.s_16,
                                        output_format.samples_per_frame,
                                        output_format.num_channels);
            break;
          case GVAudio_8_unsigned:
            interleave_2_to_all<guint8>(src->channels.u_8, dst->samples.u_8,
                                        output_format.samples_per_frame,
                                        output_format.num_channels);
            break;
          case GVAudio_16_unsigned:
            interleave_2_to_all<guint16>(src->channels.u_16, dst->samples.u_16,
                                         output_format.samples_per_frame,
                                         output_format.num_channels);
            break;
          }
        break;
      case GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_NONE_TO_2:
        switch(input_format.sample_type)
          {
          case GVAudio_8_signed:
            interleave_none_to_2<gint8>(src->channels.s_8, dst->channels.s_8,
                                        output_format.samples_per_frame,
                                        output_format.num_channels);
            break;
          case GVAudio_16_signed:
            interleave_none_to_2<gint16>(src->channels.s_16, dst->channels.s_16,
                                         output_format.samples_per_frame,
                                         output_format.num_channels);
            break;
          case GVAudio_8_unsigned:
            interleave_none_to_2<guint8>(src->channels.u_8, dst->channels.u_8,
                                         output_format.samples_per_frame,
                                         output_format.num_channels);
            break;
          case GVAudio_16_unsigned:
            interleave_none_to_2<guint16>(src->channels.u_16, dst->channels.u_16,
                                          output_format.samples_per_frame,
                                          output_format.num_channels);
            break;
          }
        break;
      case GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_NONE_TO_ALL:
        switch(input_format.sample_type)
          {
          case GVAudio_8_signed:
            interleave_none_to_all<gint8>(src->channels.s_8, dst->samples.s_8,
                                          output_format.samples_per_frame,
                                          output_format.num_channels);
            break;
          case GVAudio_16_signed:
            interleave_none_to_all<gint16>(src->channels.s_16, dst->samples.s_16,
                                           output_format.samples_per_frame,
                                           output_format.num_channels);
            break;
          case GVAudio_8_unsigned:
            interleave_none_to_all<guint8>(src->channels.u_8, dst->samples.u_8,
                                           output_format.samples_per_frame,
                                           output_format.num_channels);
            break;
          case GVAudio_16_unsigned:
            interleave_none_to_all<guint16>(src->channels.u_16, dst->samples.u_16,
                                            output_format.samples_per_frame,
                                            output_format.num_channels);
            break;
          }
      case GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_ALL_TO_NONE:
        switch(input_format.sample_type)
          {
          case GVAudio_8_signed:
            interleave_all_to_none<gint8>(src->samples.s_8,
                                          dst->channels.s_8,
                                          output_format.samples_per_frame,
                                          output_format.num_channels);
            break;
          case GVAudio_16_signed:
            interleave_all_to_none<gint16>(src->samples.s_16,
                                           dst->channels.s_16,
                                           output_format.samples_per_frame,
                                           output_format.num_channels);
            break;
          case GVAudio_8_unsigned:
            interleave_all_to_none<guint8>(src->samples.u_8,
                                           dst->channels.u_8,
                                           output_format.samples_per_frame,
                                           output_format.num_channels);
            break;
          case GVAudio_16_unsigned:
            interleave_all_to_none<guint16>(src->samples.u_16,
                                            dst->channels.u_16,
                                            output_format.samples_per_frame,
                                            output_format.num_channels);
            break;
          default:
            cerr << "Cannot interleave\n";
          }
        break;
      case GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_2_TO_NONE:
        switch(input_format.sample_type)
          {
          case GVAudio_8_signed:
            interleave_2_to_none<gint8>(src->channels.s_8,
                                        dst->channels.s_8,
                                        output_format.samples_per_frame,
                                        output_format.num_channels);
            break;
          case GVAudio_16_signed:
            interleave_2_to_none<gint16>(src->channels.s_16,
                                         dst->channels.s_16,
                                         output_format.samples_per_frame,
                                         output_format.num_channels);
            break;
          case GVAudio_8_unsigned:
            interleave_2_to_none<guint8>(src->channels.u_8,
                                         dst->channels.u_8,
                                         output_format.samples_per_frame,
                                         output_format.num_channels);
            break;
          case GVAudio_16_unsigned:
            interleave_2_to_none<guint16>(src->channels.u_16,
                                            dst->channels.u_16,
                                            output_format.samples_per_frame,
                                            output_format.num_channels);
            break;
          default:
            cerr << "Cannot interleave\n";
          }
        break;
      default:
        cerr << "Cannot interleave\n";
      }
    }
  else
    dst = src;

  // Now make the width conversion

  if((test = convert_mode & GV_BUFFER_AUDIO_CONVERT_WIDTH_MASK))
    {
    g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "Cannot convert bit widths yet\n");
    return;
    }
  if(convert_mode & (GV_BUFFER_AUDIO_CONVERT_SIGN|GV_BUFFER_AUDIO_CONVERT_ENDIAN))
    {  
    switch(output_format.bytes_per_sample)
      {
      case 1:
        if(convert_mode & GV_BUFFER_AUDIO_CONVERT_SIGN)
          swap_sign<guint8>(dst->samples.u_8,
                            output_format.num_channels*output_format.samples_per_frame);
        break;
      case 2:
        if(convert_mode & GV_BUFFER_AUDIO_CONVERT_ENDIAN)
          swap_endian(dst->samples.u_16,
                      output_format.num_channels*output_format.samples_per_frame);
        if(convert_mode & GV_BUFFER_AUDIO_CONVERT_SIGN)
          swap_sign<guint16>(dst->samples.u_16,
                             output_format.num_channels*output_format.samples_per_frame);
        
      }
    }
  
  dst->valid_samples = src->valid_samples;
  }

GVBufferAudio::GVBufferAudio()
  {
  gv_config->get_int("AudioBuffer", "NumFrames", max_num_frames);
  }


GVBufferAudio::~GVBufferAudio()
  {
  gv_config->put_int("AudioBuffer", "NumFrames", max_num_frames);
  }

int GVBufferAudio::output_can_alloc()
  {
  return output->can_alloc_audio();
  }

int GVBufferAudio::input_can_alloc()
  {
  return input->can_alloc_audio();
  }

void GVBufferAudio::set_formats()
  {
  input_format = *(input->get_audio_format());
  output_format = *(output->get_audio_format());

  convert_mode = 0;
  
  // Check for different number of channels

  if(input_format.num_channels != output_format.num_channels)
    {
    switch(input_format.num_channels)
      {
      case 5:
        if(output_format.num_channels == 4)
          convert_mode |= GV_BUFFER_AUDIO_CONVERT_DOWNMIX_5_TO_4;
        else if(output_format.num_channels == 2)
          convert_mode |= GV_BUFFER_AUDIO_CONVERT_DOWNMIX_5_TO_2;
        break;
      case 4:
        if(output_format.num_channels == 2)
          convert_mode |= GV_BUFFER_AUDIO_CONVERT_DOWNMIX_4_TO_2;
        break;
      }
    if(!(convert_mode & GV_BUFFER_AUDIO_CONVERT_DOWNMIX_MASK))
      {
      g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
            "Cannot convert %d channels to %d channels\n",
            input_format.num_channels, output_format.num_channels);
      return;
      }
    }

  // Check for endian conversion
  
  if(input_format.endian_format != output_format.endian_format)
    convert_mode |= GV_BUFFER_AUDIO_CONVERT_ENDIAN;

  // Check for sampletype conversion
  
  if(input_format.sample_type != output_format.sample_type)
    {
    switch(input_format.sample_type)
      {
      case GVAudio_8_signed:
        switch(output_format.sample_type)
          {
          case GVAudio_16_signed:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_WIDTH_8_TO_16;
            break;
          case GVAudio_8_unsigned:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_SIGN;
            break;
          case GVAudio_16_unsigned:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_WIDTH_8_TO_16;
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_SIGN;
            break;
          }
        break;
      case GVAudio_16_signed:
        switch(output_format.sample_type)
          {
          case GVAudio_8_signed:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_WIDTH_16_TO_8;
            break;
          case GVAudio_8_unsigned:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_WIDTH_16_TO_8;
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_SIGN;
            break;
          case GVAudio_16_unsigned:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_SIGN;
            break;
          }
        break;
      case GVAudio_8_unsigned:
        switch(output_format.sample_type)
          {
          case GVAudio_8_signed:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_SIGN;
            break;
          case GVAudio_16_signed:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_SIGN;
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_WIDTH_8_TO_16;
            break;
          case GVAudio_16_unsigned:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_WIDTH_8_TO_16;
            break;
          }
        break;
      case GVAudio_16_unsigned:
        switch(output_format.sample_type)
          {
          case GVAudio_8_signed:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_WIDTH_16_TO_8;
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_SIGN;
            break;
          case GVAudio_16_signed:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_SIGN;
            break;
          case GVAudio_8_unsigned:
            convert_mode |= GV_BUFFER_AUDIO_CONVERT_WIDTH_16_TO_8;
            break;
          }
        break;
      }
    }

  // Check for interleaving

  if((input_format.interleave_mode != output_format.interleave_mode) &&
     (output_format.num_channels > 1) && (input_format.num_channels > 1))
    {
    switch(input_format.interleave_mode)
      {
      case GVInterleaveNone: // No interleaving, all channels are separated
        if((output_format.interleave_mode == GVInterleave2) ||
           (output_format.num_channels == 2))
          convert_mode |= GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_NONE_TO_2;
        else
          convert_mode |= GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_NONE_TO_ALL;
        break;            
      case GVInterleave2:    // 2 interleaved channels follow the next 2
        if(output_format.interleave_mode == GVInterleaveNone)
          convert_mode |= GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_2_TO_NONE;
        else if(output_format.num_channels > 2)
          convert_mode |= GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_2_TO_ALL;
        break;
      case GVInterleaveAll:   // All channels are interleaved
        if(output_format.interleave_mode == GVInterleaveNone)
          convert_mode |= GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_ALL_TO_NONE;
        else if(output_format.num_channels > 2)
          convert_mode |= GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_ALL_TO_2;
        break;
      }
//     if(!(convert_mode & GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_MASK))
//       {
//       g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
//             "Cannot convert interleaving format");
//       return;
//       }
    }

//   cerr << "Audio format conversion flags\n";

//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_ENDIAN)
//     cerr << "GV_BUFFER_AUDIO_CONVERT_ENDIAN\n";
  
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_SIGN)
//     cerr << "GV_BUFFER_AUDIO_CONVERT_SIGN\n";
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_WIDTH_8_TO_16)
//      cerr << "GV_BUFFER_AUDIO_CONVERT_WIDTH_8_TO_16\n";
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_WIDTH_16_TO_8)
//     cerr << "GV_BUFFER_AUDIO_CONVERT_WIDTH_16_TO_8\n";
  
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_DOWNMIX_5_TO_4)
//     cerr << "GV_BUFFER_AUDIO_CONVERT_DOWNMIX_5_TO_4\n";
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_DOWNMIX_5_TO_2)
//     cerr << "GV_BUFFER_AUDIO_CONVERT_DOWNMIX_5_TO_2\n";
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_DOWNMIX_4_TO_2)
//     cerr << "GV_BUFFER_AUDIO_CONVERT_DOWNMIX_4_TO_2\n";
  
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_2_TO_ALL) 
//     cerr << "GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_2_TO_ALL\n";
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_2_TO_NONE) 
//     cerr << "GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_2_TO_NONE\n";
  
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_ALL_TO_2)
//     cerr << "GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_ALL_TO_2\n";
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_ALL_TO_NONE)
//     cerr << "GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_ALL_TO_NONE\n";

//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_NONE_TO_ALL)
//      cerr << "GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_NONE_TO_ALL\n";
//   if(convert_mode & GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_NONE_TO_2)
//      cerr << "GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_NONE_TO_2\n";

  }

int GVBufferAudio::need_scratch()
  {
  return(!!(convert_mode & GV_BUFFER_AUDIO_CONVERT_INTERLEAVE_MASK));
  }

void GVBufferAudio::clear_frame(GVAudioFrame * frame,
                                const GVAudioFormat * format)
  {
  switch(format->sample_type)
    {
    case GVAudio_8_signed:
    case GVAudio_16_signed:
      memset(frame->samples.s_8, 0x00, format->buffer_size);
      break;
    case GVAudio_8_unsigned:
      memset(frame->samples.u_8, 0x80, format->buffer_size);
      break;
    case GVAudio_16_unsigned:
      for(unsigned int i = 0;
          i < format->samples_per_frame * format->num_channels; i++)
        frame->samples.s_16[i] = 0x8000;
      break;
    }
  }
