/*****************************************************************

  buffer.h

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Base class for audio and video FIFO buffers

#ifdef __GNUG__
#pragma interface
#endif

#define GV_BUFFER_FULL          (1<<0)
#define GV_BUFFER_EMPTY         (1<<1)

class GVInput;
class GVOutput;

template <class FRAME, class FORMAT> class GVBuffer
  {
protected:

  unsigned int convert_mode;

  FRAME * frames;  // array of frames

  FRAME * scratch; // for conversions
  
  FRAME * input_frame;
  FRAME * output_frame;

  pthread_mutex_t input_mutex;
  pthread_mutex_t output_mutex;

  pthread_mutex_t state_mutex;

  virtual void alloc_frame(FRAME * frame, const FORMAT * format) = 0;
  virtual void free_frame(FRAME * frame, const FORMAT * format) = 0;
  virtual void clear_frame(FRAME * frame, const FORMAT * format) = 0;
  
  FORMAT input_format;
  FORMAT output_format;
  
  // Do some format conversion
  
  virtual void convert(FRAME * src,
                       FRAME * dst) = 0; 

  virtual void set_formats() = 0;

  virtual int output_can_alloc() = 0;
  virtual int input_can_alloc() = 0;

  virtual int need_scratch() = 0;

  int max_num_frames; // Maximum number of frames
  int num_frames;     // Current number of valid frames   
  
public:

  void free();
  
  GVInput * input;
  GVOutput * output;

  int full();            // Buffer is full, no writing possible
  int empty();           // Buffer is empty, no reading possible
  
  //  int bytes_per_frame;
  
  void alloc(); // Done
  
  FRAME * lock_read();
  void unlock_read();
  
  FRAME * lock_write();
  void unlock_write();
  
  GVBuffer();
  virtual ~GVBuffer();
  void clear();  // Discard all input frames // Done

  int get_num_frames();
  float get_percentage();
  };

template <class FRAME, class FORMAT>
int GVBuffer<FRAME,FORMAT>::get_num_frames()
  {
  int ret;
  
  pthread_mutex_lock(&state_mutex);
  ret = num_frames;
  pthread_mutex_unlock(&state_mutex);
  return ret;
  }

template <class FRAME, class FORMAT>
float GVBuffer<FRAME,FORMAT>::get_percentage()
  {
  float ret;
  
  pthread_mutex_lock(&state_mutex);
  ret = (float)num_frames/(float)(max_num_frames);
  pthread_mutex_unlock(&state_mutex);
  return ret;
  }

template <class FRAME, class FORMAT>
FRAME * GVBuffer<FRAME,FORMAT>::lock_read()
  {
  pthread_mutex_lock(&output_mutex);
  return output_frame;
  }

template <class FRAME, class FORMAT>
void GVBuffer<FRAME,FORMAT>::unlock_read()
  {
  pthread_mutex_lock(&state_mutex);
  num_frames--;
  pthread_mutex_unlock(&state_mutex);

  output_frame = output_frame->next;
  pthread_mutex_unlock(&output_mutex);
  }

template <class FRAME, class FORMAT>
FRAME * GVBuffer<FRAME,FORMAT>::lock_write()
  {
  pthread_mutex_lock(&input_mutex);
  return (scratch) ? scratch : input_frame;
  }

template <class FRAME, class FORMAT>
void GVBuffer<FRAME,FORMAT>::unlock_write()
  {
  // Do some format conversion

  if(convert_mode)
    {
    if(scratch)
      {
      convert(scratch, input_frame);
      input_frame->time = scratch->time;
      }
    else
      convert(input_frame, (FRAME*)0);
    }

  input_frame = input_frame->next;
  pthread_mutex_lock(&state_mutex);
  num_frames++;
  pthread_mutex_unlock(&state_mutex);

  pthread_mutex_unlock(&input_mutex);
  }

template <class FRAME, class FORMAT>
void GVBuffer<FRAME,FORMAT>::free()
  {
  if(!frames)
    return;

  //  FRAME * tmp; 

  if(output_can_alloc())
    output->free_frames(frames, max_num_frames);
  else
    {
    for(int i = 0; i < max_num_frames; i++)
      free_frame(&(frames[i]), &output_format);
    }
  
  delete frames;

  frames = (FRAME*)0;
  output_frame = (FRAME*)0;
  input_frame = (FRAME*)0;

  if(scratch)
    {
    if(input_can_alloc())
      input->free_frame(scratch);
    else
      free_frame(scratch, &input_format);
    }
  
  scratch = (FRAME*)0;
  }
  
template <class FRAME, class FORMAT>
void GVBuffer<FRAME,FORMAT>::alloc()
  {
  free();
  set_formats();
  frames = new FRAME[max_num_frames];
  num_frames = 0;
  if(output_can_alloc())
    output->alloc_frames(frames, max_num_frames);
  else
    {
    for(int i = 0; i < max_num_frames; i++)
      alloc_frame(&(frames[i]), &output_format);
    }

  for(int i = 0; i < max_num_frames; i++)
    clear_frame(&(frames[i]), &output_format);
  
  for(int i = 0; i < max_num_frames-1; i++)
    frames[i].next = &(frames[i+1]);
  
  frames[max_num_frames-1].next = &(frames[0]);

  output_frame = frames;
  input_frame = frames;
    
  if(need_scratch())
    {
    scratch = new FRAME;
    if(input_can_alloc())
      input->alloc_frame(scratch);
    else
      alloc_frame(scratch, &input_format);
    clear_frame(scratch, &input_format);
    }
  else
    scratch = (FRAME*)0;
  }

template <class FRAME, class FORMAT>
int GVBuffer<FRAME,FORMAT>::full() // Buffer is full, no writing possible
  {
  int ret;
  pthread_mutex_lock(&state_mutex);
  ret = (num_frames == max_num_frames) ? 1 : 0;
  pthread_mutex_unlock(&state_mutex);
  return ret;
  }

template <class FRAME, class FORMAT>
int GVBuffer<FRAME,FORMAT>::empty() // Buffer is empty, no reading possible
  {
  int ret;
  pthread_mutex_lock(&state_mutex);
  ret = (num_frames == 0) ? 1 : 0;
  pthread_mutex_unlock(&state_mutex);
  return ret;
  }

template <class FRAME, class FORMAT>
void GVBuffer<FRAME,FORMAT>::clear()  // Discard all input frames
  {
  pthread_mutex_lock(&state_mutex);
  pthread_mutex_lock(&input_mutex);
  input_frame = output_frame;
  num_frames = 0;
  pthread_mutex_unlock(&input_mutex);
  pthread_mutex_unlock(&state_mutex);
  }

template <class FRAME, class FORMAT>
GVBuffer<FRAME,FORMAT>::GVBuffer()
  {
  max_num_frames = 8;
  pthread_mutex_init(&state_mutex,(pthread_mutexattr_t *)0);
  pthread_mutex_init(&input_mutex,(pthread_mutexattr_t *)0);
  pthread_mutex_init(&output_mutex,(pthread_mutexattr_t *)0);

  frames = (FRAME*)0;
  scratch = (FRAME*)0;
  input_frame = (FRAME*)0;
  output_frame = (FRAME*)0;
  }

template <class FRAME, class FORMAT>
GVBuffer<FRAME,FORMAT>::~GVBuffer()
  {
  free();
  }
