/*****************************************************************

  configitem.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Methods for GVConfigItem (Item in the configuration database

#include "config.h"

#include <glib.h>
#include "gvconfig.h"

static GList * __copy_stringlist(GList * list)
  {
  if(!list)
    return (GList*)0;

  int num = g_list_length(list);
  char * tmp;
  GList * ret = (GList*)0;

  for(int i = 0; i < num; i++)
    {
    tmp = g_strdup((char*)g_list_nth_data(list, i));
    ret = g_list_append(ret, (gpointer)tmp);
    }
  return ret;
  }

static void __delete_stringlist(GList * list)
  {
  int num = g_list_length(list);
  char * tmp;
  for(int i = 0; i < num; i++)
    {
    tmp = (char*)g_list_nth_data(list, i);
    if(tmp)
      g_free((gpointer)tmp);
    }
  g_list_free(list);
  }

GVConfigItem::GVConfigItem(const char * _name, const char * string)
  {
  used = 0;
  name = g_strdup(_name);
  type = GVConfigItemString;
  value.c_string = g_strdup(string);
  }

GVConfigItem::GVConfigItem(const char * _name, float _value)
  {
  used = 0;
  name = g_strdup(_name);

  value.c_float = _value;
  type = GVConfigItemFloat;
  }

// Need type here to distuingish between bool and int

GVConfigItem::GVConfigItem(const char * _name, int _value, GVConfigItemType _type)
  {
  used = 0;
  name = g_strdup(_name);

  if(_type == GVConfigItemBool)
    {
    _value = !!_value;
    value.c_int = _value;
    type = GVConfigItemBool;
    }
  else
    {
    type = GVConfigItemInt;
    value.c_int = _value;
    }
  }

GVConfigItem::GVConfigItem(const char * _name, guchar red, guchar green,
                           guchar blue)
  {
  used = 0;
  name = g_strdup(_name);
  value.c_color.red = red;
  value.c_color.green = green;
  value.c_color.blue = blue;
  type = GVConfigItemColor;
  }

GVConfigItem::GVConfigItem(const char * _name, GList * _stringlist)
  {
  used = 0;
  name = g_strdup(_name);
  value.c_stringlist = __copy_stringlist(_stringlist);
  type = GVConfigItemStringlist;
  }

int GVConfigItem::get_int(int & ret)
  {
  if(type == GVConfigItemInt)
    {
    used = 1;
    ret = value.c_int;
    return 1;
    }
  return 0;
  }

int GVConfigItem::get_float(float & ret)
  {
  if(type == GVConfigItemFloat)
    {
    used = 1;
    ret = value.c_float;
    return 1;
    }
  return 0;
  }

int GVConfigItem::get_string(const char ** ret)
  {
  if(type == GVConfigItemString)
    {
    used = 1;
    *ret = value.c_string;
    //    return ret;
    return 1;
    }
  return 0;
  }

int GVConfigItem::get_bool(int & ret)
  {
  if(type == GVConfigItemBool)
    {
    used = 1;
    ret = value.c_int;
    return 1;
    }
  return 0;
  }

int GVConfigItem::get_color(guchar & red, guchar & green, guchar & blue)
  {
  if(type == GVConfigItemColor)
    {
    used = 1;
    red = value.c_color.red;
    green = value.c_color.green;
    blue = value.c_color.blue;
    return 1;
    }
  return 0;
  }

int GVConfigItem::get_stringlist(GList ** ret)
  {
  if(type == GVConfigItemStringlist)
    {
    used = 1;
    if(*ret)
      __delete_stringlist(*ret);
    *ret = __copy_stringlist(value.c_stringlist);
    return 1;
    }
  return 0;
  }

void GVConfigItem::set_int(int val)
  {
  used = 1;
  free_data();
  type = GVConfigItemInt;
  value.c_int = val;
  }

void GVConfigItem::set_float(float val)
  {
  used = 1;
  free_data();

  type = GVConfigItemFloat;
  value.c_float = val;

  }

void GVConfigItem::set_string(const char * val)
  {
  used = 1;
  free_data();

  type = GVConfigItemString;
  value.c_string = g_strdup(val);;
  }

void GVConfigItem::set_stringlist(GList * list)
  {
  used = 1;
  free_data();

  type = GVConfigItemStringlist;
  value.c_stringlist = __copy_stringlist(list);
  }

void GVConfigItem::set_bool(int val)
  {
  used = 1;
  free_data();
  
  type = GVConfigItemBool;
  value.c_int = !!val;
  }

void GVConfigItem::set_color(guchar red,guchar green, guchar blue)
  {
  used = 1;
  free_data();

  type = GVConfigItemColor;
  value.c_color.red = red;
  value.c_color.green = green;
  value.c_color.blue = blue;
  }

void GVConfigItem::free_data()
  {
  if(type == GVConfigItemString)
    g_free(value.c_string);
  else if(type == GVConfigItemStringlist)
    __delete_stringlist(value.c_stringlist);
  }

GVConfigItem::~GVConfigItem()
  {
  if(type == GVConfigItemString)
    g_free(value.c_string);
  g_free(name);
  }
