/*****************************************************************

  engine.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Lots of functions for the playback engine are here

#include <glib.h>
#include <pthread.h>
#include <ctype.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <stdio.h>

#include <sys/stat.h>
#include <unistd.h>

#include <iostream>
#include <sstream>
using std::stringstream;

// #include <iosfwd>

using std::ostream;
using std::ends;
using std::cerr;


#include "config.h"
#include "gvtime.h"
#include "gvconfig.h"
#include "videoformat.h"
#include "audioformat.h"
#include "buffer.h"
#include "audiobuf.h"
#include "videobuf.h"

#include "playlist.h"

#include "baseplugin.h"
#include "plugin.h"
#include "input.h"

#include "output.h"
#include "audioout.h"
#include "videoout.h"

#include <gtk/gtk.h>
#include "configwindow.h"
#include "gvloader.h"
#include "engine.h"

#include "trackinfo.h"

GVEngine * gv_engine;

static void __log_func(const gchar   *log_domain,
                       GLogLevelFlags log_level,
                       const gchar   *message,
                       gpointer       user_data)
  {
  gv_engine->error(message);
  }

void GVEngine::plugin_attach()
  {
  pthread_mutex_lock(&plugin_mutex);
  num_plugins++;
  pthread_mutex_unlock(&plugin_mutex);
  }

void GVEngine::plugin_detach()
  {
  pthread_mutex_lock(&plugin_mutex);
  num_plugins--;
//   cerr << "Plugin_detach: Num Plugins " << num_plugins << "\n"; 
  if(!num_plugins)
    pthread_cond_broadcast(&plugin_cond);
  pthread_mutex_unlock(&plugin_mutex);
  }

void GVEngine::get_total_track_time(GVTime & time)         // Total track time
  {
  pthread_mutex_lock(&time_mutex);
  time = total_track_time;
  pthread_mutex_unlock(&time_mutex);
  }

// Current time of the track 

void GVEngine::set_current_track_time()
  {
  gdouble seconds;
  gulong microseconds;
  switch(sync_mode)
    {
    case SyncAudioOutput:
      output_audio->get_current_track_time(current_track_time);
      break;
    case SyncVideoOutput:
      output_video->get_current_track_time(current_track_time);
      break;
    case SyncInput:
      input->get_current_track_time(current_track_time);
      break;
    case SyncSoftware:
      {
      seconds = g_timer_elapsed(timer, &microseconds);
      g_timer_reset(timer);
      GVTime tmp_time;
      tmp_time.seconds = (int)(seconds);
      tmp_time.nanoseconds = microseconds * 1000;
      current_track_time += tmp_time;
      break;
      }
    default:
      break;
    }
  }

void GVEngine::get_current_track_time(GVTime & time, int exact)
  {
  pthread_mutex_lock(&time_mutex);
  if(exact)
    set_current_track_time();
  time = current_track_time;
  pthread_mutex_unlock(&time_mutex);
  }

void GVEngine::set_total_track_time(const GVTime & time)   // Total track time
  {
  pthread_mutex_lock(&time_mutex);
  total_track_time = time;
  pthread_mutex_unlock(&time_mutex);
  }

void GVEngine::add_current_track_time(const GVTime & time)
  {
  pthread_mutex_lock(&time_mutex);
  
//   cerr << "Adding "
//        << time.seconds << " Seconds "
//        << time.nanoseconds << " Nanoseconds\n";

  current_track_time += time;
  pthread_mutex_unlock(&time_mutex);
  }

// Current time of the track 

void GVEngine::set_current_track_time(const GVTime & time)
  {
  pthread_mutex_lock(&time_mutex);
  current_track_time = time;
  pthread_mutex_unlock(&time_mutex);
  }

void GVEngine::set_current_track_time(double time)
  {
  pthread_mutex_lock(&time_mutex);
  current_track_time = time;
  pthread_mutex_unlock(&time_mutex);
  }

void GVEngine::update_info()
  {
  lock_info_strings();

  if(i_input_audio_format)
    g_free(i_input_audio_format);
  if(i_input_video_format)
    g_free(i_input_video_format);

  if(i_output_audio_format)
    g_free(i_output_audio_format);
  if(i_output_video_format)
    g_free(i_output_video_format);

  if(i_input_info)
    g_free(i_input_info);
  if(i_output_audio_info)
    g_free(i_output_audio_info);
  if(i_output_video_info)
    g_free(i_output_video_info);

  if(i_track_info)
    g_free(i_track_info);

  stringstream line;

  const GVAudioFormat * audio_format;
  const GVVideoFormat * video_format;

  
  //  line.freeze(0);
  line.seekp(0);
    
  if(internal_state & GV_INT_INPUT_HAS_AUDIO)
    {  
    audio_format = input->get_audio_format();
    audio_format->print(line);
    }
  else
    line << "No Audio";
  
  line << ends;
  i_input_audio_format = g_strdup(line.str().c_str());


  //  line.freeze(0);
  line.seekp(0);

  if(internal_state & GV_INT_INPUT_HAS_VIDEO)
    {  
    video_format = input->get_video_format();
    video_format->print(line);
    }
  else
    line << "No Video";

  line << ends;
  i_input_video_format = g_strdup(line.str().c_str());

  //  line.freeze(0);
  line.seekp(0);

  if(internal_state & GV_INT_OUTPUT_AUDIO)
    {
    audio_format = output_audio->get_audio_format();
    audio_format->print(line);
    }
  else if(internal_state & GV_INT_OUTPUT_AUDIO_VIDEO)
    {
    audio_format = output_video->get_audio_format();
    audio_format->print(line);
    }
  else
    line << "No Audio";
  line << ends;
  i_output_audio_format = g_strdup(line.str().c_str());

  
  //  line.freeze(0);
  line.seekp(0);

  if(((internal_state & GV_INT_OUTPUT_AUDIO_VIDEO) ||
      (internal_state & GV_INT_OUTPUT_VIDEO)) &&
     (internal_state & GV_INT_INPUT_HAS_VIDEO))
    {
    video_format = output_video->get_video_format();
    video_format->print(line);
    }
  else
    line << "No Video";
  line << ends;
  i_output_video_format = g_strdup(line.str().c_str());

  //  line.freeze(0);
  line.seekp(0);

  //  cerr << "Updating input info...";
  
  line << "Input plugin: " << input->attributes->real_name
       << " (" << input->attributes->name << ")\n";
  if(internal_state & GV_INT_INPUT_IS_FILE)
    line << "File:\n" << input->get_filename() << "\n\n";

  input->print_info(line);
  line << ends;
  i_input_info = g_strdup(line.str().c_str());

  //  cerr << "done\n";

  
  // Update output info
  
  //  line.freeze(0);
  line.seekp(0);
  if((internal_state & GV_INT_OUTPUT_AUDIO_VIDEO) ||
     (internal_state & GV_INT_OUTPUT_VIDEO))
    {
    line << "Video Output plugin: " << output_video->attributes->real_name
         << " (" << output_video->attributes->name << ")\n";
    if(internal_state & GV_INT_OUTPUT_IS_FILE)
      line << "File:\n" << output_video->get_filename() << "\n\n";
    output_video->print_info(line);
    }
  else
    line << "Video output plugin isn't running";

  line << ends;
  i_output_video_info = g_strdup(line.str().c_str());

  //  line.freeze(0);
  line.seekp(0);

  if(internal_state & GV_INT_OUTPUT_AUDIO)
    {
    line << "Audio Output plugin: " << output_audio->attributes->real_name
         << " (" << output_audio->attributes->name << ")\n";
    if(internal_state & GV_INT_OUTPUT_IS_FILE)
      line << "File:\n" << output_audio->get_filename() << "\n\n";
    output_audio->print_info(line);
    }
  else if(internal_state & GV_INT_OUTPUT_AUDIO_VIDEO)
    line << "Audio is handled by video output";
  else
    line << "Audio output plugin isn't running";
  line << ends;
  i_output_audio_info = g_strdup(line.str().c_str());

  //  line.freeze(0);
  line.seekp(0);

  const GVTrackInfo * track_info = input->get_track_info();
  if(track_info)
    line << *track_info;
  else
    line << "No Info available";
  line << ends;
  i_track_info = g_strdup(line.str().c_str());
    
  set_state_flag(GV_ENGINE_INFO_CHANGED);
 
  unlock_info_strings();
  }

GVEngine::GVEngine()
  {
  input = (GVInput*)0;               // Input plugin
  output_audio = (GVOutputAudio*)0; // Audio output plugin
  output_video = (GVOutputVideo*)0; // Video output plugin

  i_input_audio_format = (char*)0;
  i_input_video_format = (char*)0;
  
  i_output_audio_format = (char*)0;
  i_output_video_format = (char*)0;

  i_input_info = (char*)0;
  i_output_audio_info = (char*)0;
  i_output_video_info = (char*)0;

  i_track_info = (char*)0;
  
  old_state = GVEngineStopped;
  
  // Create the buffers

  video_buffer = new GVBufferVideo;
  audio_buffer = new GVBufferAudio;

  // Create the plugin loaders

  gv_input_file_loader   = new GVPluginLoader;
  gv_input_loader        = new GVPluginLoader;
  gv_video_output_loader = new GVPluginLoader;
  gv_audio_output_loader = new GVPluginLoader;
  gv_image_reader_loader = new GVPluginLoader;

  num_plugins = 0;
  state_flags = GV_ENGINE_STATE_CHANGED;
  internal_state = 0;

  command_stack = (GList*)0;
    
  config_flags = 0;
  
  timer = (GTimer*)0;

  get_config();

  g_log_set_handler(PACKAGE, G_LOG_LEVEL_CRITICAL,
                    __log_func, (gpointer)0);
  
  input_info = (GVPluginInfo*)0;

  state = GVEngineStopped;

  pthread_cond_init(&plugin_cond,(pthread_condattr_t *)0);
  pthread_cond_init(&start_cond, (pthread_condattr_t *)0);
  
  pthread_mutex_init (&plugin_mutex,(pthread_mutexattr_t *)0);
  pthread_mutex_init (&state_mutex,(pthread_mutexattr_t *)0);
  pthread_mutex_init (&time_mutex,(pthread_mutexattr_t *)0);
  pthread_mutex_init (&command_mutex,(pthread_mutexattr_t *)0);
  pthread_mutex_init (&start_mutex,(pthread_mutexattr_t *)0);
  pthread_mutex_init (&info_mutex,(pthread_mutexattr_t *)0);
  pthread_mutex_init (&cmd_mutex,(pthread_mutexattr_t *)0);
  pthread_mutex_init (&framerate_mutex,(pthread_mutexattr_t *)0);
  pthread_mutex_init (&num_waiting_plugins_mutex, (pthread_mutexattr_t *)0);
  
  current_track_time.seconds = 0;
  current_track_time.nanoseconds = 0;

  total_track_time.seconds = 0;
  total_track_time.nanoseconds = 0;

  num_waiting_plugins = 0;
  total_waiting_plugins = 0;
  }

GVEngine::~GVEngine()
  {
  if(get_state() == GVEnginePlaying)
    {
    pthread_mutex_lock(&cmd_mutex);
    stop_cmd();  
    pthread_mutex_unlock(&cmd_mutex);
    }
  
  put_config();

  delete video_buffer;
  delete audio_buffer;

  if(input)
    {
    if(input->attributes->flags & GV_PLUGIN_IS_FILE)
      gv_input_file_loader->destroy_plugin(input);
    else
      gv_input_loader->destroy_plugin(input);
    }
  if(output_audio)
    gv_audio_output_loader->destroy_plugin(output_audio);
  if(output_video)
    gv_video_output_loader->destroy_plugin(output_video);
  
  delete gv_input_file_loader;
  delete gv_input_loader;
  delete gv_video_output_loader;
  delete gv_audio_output_loader;
  }

int GVEngine::get_state_flag(unsigned int flag)
  {
  int ret;
  pthread_mutex_lock(&state_mutex);
  ret = state_flags & flag;
  pthread_mutex_unlock(&state_mutex);
  return ret;
  }

void GVEngine::set_state_flag(unsigned int flag)
  {
  pthread_mutex_lock(&state_mutex);
  state_flags |= flag;
  pthread_mutex_unlock(&state_mutex);
  }

void GVEngine::clear_state_flag(unsigned int flag)
  {
  pthread_mutex_lock(&state_mutex);
  state_flags &= ~flag;
  pthread_mutex_unlock(&state_mutex);
  }

void GVEngine::toggle_state_flag(unsigned int flag)
  {
  pthread_mutex_lock(&state_mutex);
  state_flags ^= flag;
  pthread_mutex_unlock(&state_mutex);
  }

// void GVEngine::set_input(const char * name)
//   {
//   if((input) && !strcmp(input->attributes->name, name))
//     return;
//   GVInput * new_input;
//   }

void GVEngine::set_output_video(const char * name)
  {
  if((output_video) && !strcmp(output_video->attributes->name, name))
    return;
  GVPluginInfo * info = gv_video_output_loader->find_plugin_by_name(name);
  if(!info)
    {
    // Error
    return;
    }
  GVOutput * new_output_video
    = gv_video_output_loader->load_output(info);
  set_output_video(new_output_video);
  }

void GVEngine::set_output_audio(const char * name)
  {
  if((output_audio) && !strcmp(output_audio->attributes->name, name))
    return;
  GVPluginInfo * info = gv_audio_output_loader->find_plugin_by_name(name);
  if(!info)
    {
    // Error
    return;
    }
  GVOutputAudio * new_output_audio
    = gv_audio_output_loader->load_output_audio(info);
  set_output_audio(new_output_audio);
  }

void GVEngine::set_input(const char * name)
  {
  if((!name) || (*name == '\0'))
    {
    set_input((GVInput*)0);
    return;
    }
  if((input) && !strcmp(input->attributes->name, name))
    return;
  GVPluginInfo * info = gv_input_loader->find_plugin_by_name(name);

  if(!info)
    return;
  GVInput * new_input = gv_input_loader->load_input(info);
  set_input(new_input);
  }

void GVEngine::set_input(GVInput * new_input)
  {

//   cerr << "GVEngine::set_input: input: "
//        << ((input) ? input->attributes->name : "NULL")  
//        << " new_input: "
//        << ((new_input) ? new_input->attributes->name : "NULL") << "\n";
  
  if((new_input) && (input) && !strcmp(input->attributes->name,
                                       new_input->attributes->name))
    {
    //    cerr << "Don't need to reload plugin " << input->attributes->name << "\n";
    return;
    }

  // User wanted media files, but we aleady have them
  
  if(!(new_input) &&
     (!(input) || (input->attributes->flags & GV_PLUGIN_IS_FILE)))
    return;
  
  if(get_state() == GVEnginePlaying)
    {
//     pthread_mutex_lock(&cmd_mutex);
//     stop_cmd();  
//     pthread_mutex_unlock(&cmd_mutex);
    stop_sync();
    }
  
  // If the current plugin has files and the next doesn't, we must save
  // the playlist

  if(new_input)
    {
    if(new_input->attributes->flags & GV_PLUGIN_IS_FILE)
      {
      gv_playlist->make_private(0);
      cerr << "Playlist now public\n";
      }
    else
      {
      gv_playlist->make_private(1);
      new_input->fill_playlist();
      cerr << "Playlist now private\n";
      }
    }
  else
    {
    gv_playlist->make_private(0);
    cerr << "Playlist now public\n";
    }
  
  // If we already have an input, we must unload it

  if(input)
    {
    if(input->attributes->flags & GV_PLUGIN_IS_FILE)
      gv_input_file_loader->destroy_plugin(input);
    else
      gv_input_loader->destroy_plugin(input);
    input = (GVInput*)0;
    }
  if(new_input)
    {
    input = new_input;
    input->video_buffer = video_buffer;
    input->audio_buffer = audio_buffer;
    }
  if((input) && !(input->attributes->flags & GV_PLUGIN_IS_FILE))
    {
    input->fill_playlist();
    if(input->get_error())
      return;
    }
  }

void GVEngine::set_output_video(GVOutput * new_output)
  {
  if((new_output) && (output_video) && !strcmp(output_video->attributes->name,
                                               new_output->attributes->name))
    {
    cerr << "Don't need to reload plugin "
         << output_video->attributes->name << "\n";
    output_video->cleanup();
    return;
    }

  if(get_state() == GVEnginePlaying)
    {
//     pthread_mutex_lock(&cmd_mutex);
//     stop_cmd();  
//     pthread_mutex_unlock(&cmd_mutex);
    stop_sync();
    }

  if(output_video)
    {
    gv_video_output_loader->destroy_plugin(output_video);
    output_video = (GVOutputVideo*)0;
    }
  if(new_output)
    {
    output_video = new_output;
    output_video->video_buffer = video_buffer;
    }
  }

void GVEngine::set_output_audio(GVOutputAudio * new_output)
  {
  if((new_output) && (output_audio) && !strcmp(output_audio->attributes->name,
                                               new_output->attributes->name))
    {
//     cerr << "Don't need to reload plugin " << output_audio->attributes->name
//          << "\n";
    output_audio->cleanup();
    return;
    }

  if(get_state() == GVEnginePlaying)
    {
//     pthread_mutex_lock(&cmd_mutex);
//     stop_cmd();  
//     pthread_mutex_unlock(&cmd_mutex);
    stop_sync();
    }

  if(output_audio)
    {
    gv_audio_output_loader->destroy_plugin(output_audio);
    output_audio = (GVOutputAudio*)0;
    }
  if(new_output)
    {
    output_audio = new_output;
    output_audio->audio_buffer = audio_buffer;
    }
  }

// Load all the plugins we can find

void GVEngine::load_plugins(const char * path)
  {
  char filename[PATH_MAX];
  //  GModule * module;
  
  DIR * directory;
  struct dirent * directory_entry;
  struct stat status;

  directory = opendir(path);

  while(1)
    {
    directory_entry = readdir(directory);
    
    if(!directory_entry) // We're done
      break;

    // Plugin filenames must start with 'i' (input) or 'o' (output)
    
    if(!((directory_entry->d_name[0] == 'i') ||
         (directory_entry->d_name[0] == 'o')))
      continue;

    //    cerr << "Trying " << directory_entry->d_name << "... ";
    
    strcpy(filename, path);
    strcat(filename, G_DIR_SEPARATOR_S);
    strcat(filename, directory_entry->d_name);

    // Check if we have a regular file
    
    stat(filename, &status);
    if(!S_ISREG(status.st_mode))
      {
      //      cerr << "No Regular File\n";
      continue;
      }

    // Now check for the filename

    if(directory_entry->d_name[0] == 'i')
      {
      if(directory_entry->d_name[1] == 'f')
        {
        if(gv_input_file_loader->add_plugin(filename))
          cerr << "Found file input plugin: " << directory_entry->d_name
               << "\n";
        }
      else if(directory_entry->d_name[1] == '_')
        {
        if(gv_input_loader->add_plugin(filename))
          cerr << "Found input plugin: " << directory_entry->d_name << "\n";
        }
      else if(directory_entry->d_name[1] == 'r')
        {
        if(gv_image_reader_loader->add_plugin(filename))
          cerr << "Found imagereader plugin: "
               << directory_entry->d_name << "\n";
        }
      }
    // Video output plugin

    else if(((directory_entry->d_name[0] == 'o') &&
            (directory_entry->d_name[1] == 'v')) ||
            ((directory_entry->d_name[0] == 'o') &&
             (directory_entry->d_name[1] == '_')))
      {
      if(gv_video_output_loader->add_plugin(filename))
        cerr << "Found video output plugin: " << directory_entry->d_name << "\n";
      }

    // Audio Output Plugin
    
    else if((directory_entry->d_name[0] == 'o') &&
            (directory_entry->d_name[1] == 'a'))
      {
      if(gv_audio_output_loader->add_plugin(filename))
        cerr << "Found audio output plugin: " << directory_entry->d_name << "\n";
      }
    }

  // Set the default plugins

  if(default_input_plugin)
    set_input(default_input_plugin);
  set_output_audio(default_audio_output_plugin);
  set_output_video(default_video_output_plugin);
#if 0
  input_config.plugin_loader = gv_input_loader;
  output_video_config.plugin_loader = gv_video_output_loader;
  output_audio_config.plugin_loader = gv_audio_output_loader;
  
  input_config.create_menu();
  output_video_config.create_menu();
  output_audio_config.create_menu();  
#endif
  }

void GVEngine::completed()
  {
  set_state(GVEngineCompleting);
  }

void GVEngine::load_playlist_track()
  {
  //  GVPluginInfo * info = find_plugin(item->filename);
  //  if(!info)
  //    return;
  const GVPlaylistTrack * item = gv_playlist->get_current_track();
  GVPluginInfo * plugin_info;

  GVInput * new_input;
  
  if(!item)
    return;
  
//   cerr << "Loading new track: " << item->trackname << " from file "
//        << item->filename << " Plugin: "
//        << item->plugin << " Current plugin: "
//        << ((input) ? input->attributes->name : "NULL") << "\n";

  //  char * filename = item->filename;
  
  if(!(input) || strcmp(item->plugin, input->attributes->name))
    {
    plugin_info = gv_input_file_loader->find_plugin_by_name(item->plugin);

    if(plugin_info)
      {
      new_input = gv_input_file_loader->load_input(plugin_info);
      set_input(new_input);
      }
    else
      {
      plugin_info = gv_input_loader->find_plugin_by_name(item->plugin);
      if(plugin_info)
        {
        new_input = gv_input_loader->load_input(plugin_info);
        set_input(new_input);
        }
      }

    }
//   cerr << "Loaded: " << item->filename << " Plugin: "
//        << input->attributes->name << "\n";
  input->set_filename(item->filename);
  input->init();
  }

void GVEngine::move_playlist_track(int src, int dst)
  {
  gv_playlist->move_track(src, dst);
  }

void GVEngine::seek(float percentage)
  {
  Command * command = new Command(CommandSeek, percentage);
  push_command(command);
  }


// const GVPluginInfo * GVEngine::find_input_by_name(const char * name)
//   {
//   const GVPluginInfo * info;
//   int num;
//   num = g_list_length(input_file_plugins);
//   for(int i = 0; i < num; i++)
//     {
//     info = (GVPluginInfo*)g_list_nth_data(input_file_plugins, i);
//     if(!strcmp(info->name, name))
//       return info;
//     }
//   num = g_list_length(input_plugins);
//   for(int i = 0; i < num; i++)
//     {
//     info = (GVPluginInfo*)g_list_nth_data(input_plugins, i);
//     if(!strcmp(info->name, name))
//       return info;
//     }
//   return (const GVPluginInfo *)0;
//   }

// const GVPluginInfo * GVEngine::find_output_video_by_name(const char * name)
//   {
//   const GVPluginInfo * info;
//   int num;
//   num = g_list_length(output_video_plugins);
//   for(int i = 0; i < num; i++)
//     {
//     info = (GVPluginInfo*)g_list_nth_data(output_video_plugins, i);
//     if(!strcmp(info->name, name))
//       return info;
//     }
//   return (const GVPluginInfo *)0;
//   }

// const GVPluginInfo * GVEngine::find_output_audio_by_name(const char * name)
//   {
//   const GVPluginInfo * info;
//   int num;
//   num = g_list_length(output_audio_plugins);
//   for(int i = 0; i < num; i++)
//     {
//     info = (GVPluginInfo*)g_list_nth_data(output_audio_plugins, i);
//     if(!strcmp(info->name, name))
//       return info;
//     }
//   return (const GVPluginInfo *)0;
//   }

void GVEngine::repeat()
  {
  switch(repeat_mode)
    {
    case GVRepeatNone:
      repeat_mode = GVRepeat1;
      break;
    case GVRepeat1:
      repeat_mode = GVRepeatAll;
      break;
    case GVRepeatAll:
      repeat_mode = GVRepeatNone;
      break;
    }
  }

GVRepeatMode GVEngine::get_repeat_mode()
  {
  return repeat_mode;
  }

void GVEngine::error(const char * message)
  {
  pthread_mutex_lock(&state_mutex);
  error_message = g_strdup(message);
  pthread_mutex_unlock(&state_mutex);
  set_state(GVEngineError);
  }

char * GVEngine::get_error_message()
  {
  char * ret;
  pthread_mutex_lock(&state_mutex);
  ret = g_strdup(error_message);
  pthread_mutex_unlock(&state_mutex);
  return ret;
  }

#define SECTION "General"

void GVEngine::get_config()
  {
  int i_tmp = 1;

  config_flags = 0;
  
  gv_config->get_bool(SECTION, "AutoAdvance", i_tmp);
  if(i_tmp)
    config_flags |= GV_ENGINE_AUTO_ADVANCE;

  i_tmp = 0;
  gv_config->get_bool(SECTION, "StartPlay", i_tmp);
  if(i_tmp)
    config_flags |= GV_ENGINE_START_PLAY;

  i_tmp = 0;
  gv_config->get_bool(SECTION, "SyncSoftware", i_tmp);
  if(i_tmp)
    config_flags |= GV_ENGINE_SYNC_SOFTWARE;

  i_tmp = 1;
  gv_config->get_bool(SECTION, "DeleteIncomplete", i_tmp);
  if(i_tmp)
    config_flags |= GV_ENGINE_DELETE_INCOMPLETE;

  i_tmp = 0;
  gv_config->get_bool(SECTION, "RemoveErrorTracks", i_tmp);
  if(i_tmp)
    config_flags |= GV_ENGINE_REMOVE_ERROR_TRACKS;
  
  repeat_mode = GVRepeatAll;

  i_tmp = repeat_mode;
  gv_config->get_int(SECTION, "RepeatMode", i_tmp);
  repeat_mode = (GVRepeatMode)i_tmp;
  
  default_input_plugin = "";
  gv_config->get_string("General", "Input",
                        &default_input_plugin);

  default_audio_output_plugin = "oa_oss";
  gv_config->get_string("General", "OutputAudio",
                        &default_audio_output_plugin);

  default_video_output_plugin = "ov_x11";
  gv_config->get_string("General", "OutputVideo",
                        &default_video_output_plugin);
  use_audio = 1;
  gv_config->get_bool("General", "UseAudio", use_audio);
  
  }

void GVEngine::put_config()
  {
  gv_config->put_bool(SECTION, "AutoAdvance",
                      !!(config_flags & GV_ENGINE_AUTO_ADVANCE));
  gv_config->put_bool(SECTION, "RemoveErrorTracks",
                      !!(config_flags & GV_ENGINE_REMOVE_ERROR_TRACKS));
  gv_config->put_bool(SECTION, "StartPlay",
                      !!(config_flags & GV_ENGINE_START_PLAY));
  gv_config->put_bool(SECTION, "SyncSoftware",
                      !!(config_flags & GV_ENGINE_SYNC_SOFTWARE));
  gv_config->put_bool(SECTION, "DeleteIncomplete",
                      !!(config_flags & GV_ENGINE_DELETE_INCOMPLETE));

  gv_config->put_int(SECTION, "RepeatMode", repeat_mode);

  if(input)
    gv_config->put_string("General", "Input", input->attributes->name);
  else
    gv_config->put_string("General", "Input", (char*)0);
  
  if(output_audio)
    gv_config->put_string("General", "OutputAudio",
                          output_audio->attributes->name);
  if(output_video)
    gv_config->put_string("General", "OutputVideo",
                          output_video->attributes->name);
  gv_config->put_bool("General", "UseAudio", use_audio);

  }

void GVEngine::set_config_flag(int flag, int value)
  {
  pthread_mutex_lock(&state_mutex);
  if(value)
    config_flags |= flag;
  else
    config_flags &= ~flag;
  pthread_mutex_unlock(&state_mutex);
  }

void GVEngine::set_config_flags(int _config_flags)
  {
  pthread_mutex_lock(&state_mutex);
  config_flags = _config_flags;
  pthread_mutex_unlock(&state_mutex);
  }

int GVEngine::get_config_flags()
  {
  int ret;
  pthread_mutex_lock(&state_mutex);
  ret = config_flags;
  pthread_mutex_unlock(&state_mutex);
  return ret;
  }

GVEngineState GVEngine::get_state()
  {
  GVEngineState ret;
  pthread_mutex_lock(&state_mutex);
  ret = state;
  pthread_mutex_unlock(&state_mutex);
  return ret;
  }

GVEngineState GVEngine::get_state(GVEngineState & old_state_ret)
  {
  GVEngineState ret;
  pthread_mutex_lock(&state_mutex);
  ret = state;
  old_state_ret = old_state;
  pthread_mutex_unlock(&state_mutex);
  return ret;
  }

#if 0
static void __print_engine_state(GVEngineState state)
  {
  switch(state)
    {
    case GVEngineStopped:
      cerr << "GVEngineStopped";
      break;
    case GVEnginePaused:   
      cerr << "GVEnginePaused";
      break;
    case GVEnginePlaying:
      cerr << "GVEnginePlaying";
      break;
    case GVEngineSeeking:
      cerr << "GVEngineSeeking";
      break;
    case GVEngineCompleting:
      cerr << "GVEngineCompleting";
      break;
    case GVEngineChanging:
      cerr << "GVEngineChanging";
      break;
    case GVEngineError:
      cerr << "GVEngineError";
      break;
    }
  }
#endif

void GVEngine::set_state(GVEngineState _state)
  {
  if(get_state() == _state)
    return;
  
  pthread_mutex_lock(&state_mutex);
//   cerr << "GVEngine::set_state: Old state ";
//   __print_engine_state(state);
//   cerr << " New state ";
//   __print_engine_state(_state);
//   cerr << "\n";

  old_state = state;
  state = _state;
  state_flags |= GV_ENGINE_STATE_CHANGED;
  pthread_mutex_unlock(&state_mutex);
  }

void GVEngine::wait_for_plugins()
  {
  //  cerr << "Waiting for " << num_plugins << " plugins to exit...\n";
  pthread_mutex_lock(&plugin_mutex);
  if(num_plugins)
    pthread_cond_wait(&plugin_cond, &plugin_mutex);
  pthread_mutex_unlock(&plugin_mutex);
  //  cerr << "All Plugins exited\n";
  }

void GVEngine::wait_for_start()
  {
  pthread_mutex_lock(&num_waiting_plugins_mutex);
  num_waiting_plugins++;
  pthread_mutex_unlock(&num_waiting_plugins_mutex);
  
  pthread_mutex_lock(&start_mutex);
  pthread_cond_wait(&start_cond, &start_mutex);
  pthread_mutex_unlock(&start_mutex);
  }

void GVEngine::start_playback()
  {
  GVTime wait_time;
  wait_time.seconds = 0;
  wait_time.nanoseconds = 10000000; // 10 milliseconds

  int do_exit = 0;
  
  while(1)
    {
    pthread_mutex_lock(&num_waiting_plugins_mutex);
    if(total_waiting_plugins == num_waiting_plugins)
      do_exit = 1;
    pthread_mutex_unlock(&num_waiting_plugins_mutex);

    // We wait in any case because there can still be race conditions

    wait_time.sleep();
    if(do_exit)
      break;
    }
  
  if(sync_mode == SyncSoftware)
    {
    pthread_mutex_lock(&time_mutex);
    g_timer_start(timer);
    }
  pthread_cond_broadcast(&start_cond);

  pthread_mutex_lock(&num_waiting_plugins_mutex);
  num_waiting_plugins = 0;
  pthread_mutex_unlock(&num_waiting_plugins_mutex);
  
  if(sync_mode == SyncSoftware)
    pthread_mutex_unlock(&time_mutex);
  }

void GVEngine::set_use_audio(int _use_audio)
  {
  use_audio = !!_use_audio;
  }

void GVEngine::delete_output_files()
  {
  if(!(get_config_flags() & GV_ENGINE_DELETE_INCOMPLETE) ||
     !(internal_state & GV_INT_OUTPUT_IS_FILE))
    return;

  switch(internal_state & GV_INT_OUTPUT_MASK)
    {
    case GV_INT_OUTPUT_VIDEO:
    case GV_INT_OUTPUT_AUDIO_VIDEO:
      cerr << "Removing " << output_video->get_filename() << "\n";
      remove(output_video->get_filename());
      break;
    case GV_INT_OUTPUT_AUDIO:
      cerr << "Removing " << output_audio->get_filename() << "\n";
      remove(output_audio->get_filename());
      break;
    case GV_INT_OUTPUT_AUDIO|GV_INT_OUTPUT_VIDEO:
      cerr << "Removing " << output_audio->get_filename() << "\n";
      remove(output_audio->get_filename());
      cerr << "Removing " << output_video->get_filename() << "\n";
      remove(output_video->get_filename());
      break;
    }
  }

void GVEngine::lock_info_strings()
  {
  pthread_mutex_lock(&info_mutex);
  }

void GVEngine::unlock_info_strings()
  {
  pthread_mutex_unlock(&info_mutex);
  }

// Trigger a plugin error

void GVEngine::plugin_error(GVPlugin * plugin)
  {
  int error_code = plugin->get_error();
  switch(error_code)
    {
    case GV_PLUGIN_ERROR_FILE_INVALID:
      g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "%s: Cannot open file %s",
            plugin->attributes->name, plugin->get_filename());
      break;
    case GV_PLUGIN_ERROR_DEVICE_OPEN_FAILED:
      g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "%s: Cannot open Device",
            plugin->attributes->name);
      break;
    case GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED: // Device settings failed
      g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "%s: Device settings failed",
            plugin->attributes->name);
      break;
    case GV_PLUGIN_ERROR_NO_MEDIUM:             // No medium found
      g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "%s: No Medium found", 
            plugin->attributes->name);
      break;
    }
  }

float GVEngine::get_framerate()
  {
  float ret;
  pthread_mutex_lock(&framerate_mutex);
  ret = framerate;
  pthread_mutex_unlock(&framerate_mutex);
  return ret;
  }

void GVEngine::set_framerate(float _framerate)
  {
  pthread_mutex_lock(&framerate_mutex);
  framerate = _framerate;
  pthread_mutex_unlock(&framerate_mutex);
  }

void GVEngine::reset_input()
  {
  stop_sync();
  input->fill_playlist();
  }

void GVEngine::configure_output_video()
  {
  if(output_video)
    output_video->configure();
  }

void GVEngine::configure_output_audio()
  {
  if(output_audio)
    output_audio->configure();
  }

void GVEngine::configure_input()
  {
  if(input)
    input->configure();
  }

const GVPluginAttributes * GVEngine::get_output_video_attr()
  {
  if(!output_video)
    return (GVPluginAttributes*)0;
  return output_video->attributes;
  }

const GVPluginAttributes * GVEngine::get_output_audio_attr()
  {
  if(!output_audio)
    return (GVPluginAttributes*)0;
  return output_audio->attributes;

  }

const GVPluginAttributes * GVEngine::get_input_attr()
  {
  if(!input)
    return (GVPluginAttributes*)0;
  return input->attributes;
  }
