/*****************************************************************

  engine.h

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Playback engine

// State flags

#define GV_ENGINE_STATE_CHANGED     (1<<0)  // State changed
#define GV_ENGINE_INFO_CHANGED      (1<<1)  // State changed
#define GV_ENGINE_NO_REALTIME       (1<<2)  // No Realtime operation
#define GV_ENGINE_PRIVATE_PLAYLIST  (1<<3) // Disable some playlist options
#define GV_ENGINE_CAN_SEEK          (1<<4)  // Engine must seek
#define GV_ENGINE_ALL               0xFFFF  // All flags

// Internal flags

#define GV_INT_INPUT_HAS_AUDIO       (1<<0)  // Input has an audio stream
#define GV_INT_INPUT_HAS_VIDEO       (1<<1)  // Input has a video stream
#define GV_INT_OUTPUT_VIDEO          (1<<2)  // Video Output present
#define GV_INT_OUTPUT_AUDIO          (1<<3)  // Audio Output present
#define GV_INT_OUTPUT_AUDIO_VIDEO    (1<<4)  // Audio/Video Output present
#define GV_INT_INPUT_IS_FILE         (1<<5)  // Input reads from a file
#define GV_INT_OUTPUT_IS_FILE        (1<<6)  // Output reads from a file
#define GV_INT_INPUT_NEEDS_REALTIME  (1<<7)  // Input needs realtime
#define GV_INT_OUTPUT_NEEDS_REALTIME (1<<8)  // Output needs realtime

#define GV_INT_OUTPUT_MASK   (GV_INT_OUTPUT_VIDEO|\
                              GV_INT_OUTPUT_AUDIO|\
                              GV_INT_OUTPUT_AUDIO_VIDEO)

#define GV_INT_INPUT_MASK    (GV_INT_INPUT_HAS_AUDIO|\
                              GV_INT_INPUT_HAS_VIDEO)

#define GV_INT_FILE_MASK     (GV_INT_INPUT_IS_FILE|GV_INT_OUTPUT_IS_FILE)

#define GV_INT_REALTIME_MASK (GV_INT_INPUT_NEEDS_REALTIME|\
                              GV_INT_OUTPUT_NEEDS_REALTIME)

// Config flags

#define GV_ENGINE_AUTO_ADVANCE        (1<<0)
#define GV_ENGINE_START_PLAY          (1<<1)
#define GV_ENGINE_SYNC_SOFTWARE       (1<<2)
#define GV_ENGINE_DELETE_INCOMPLETE   (1<<3)

// Remove tracks from the playlist if they cause an error

#define GV_ENGINE_REMOVE_ERROR_TRACKS (1<<4) 

// Playback engine

class GVOutputAudio; // Audio output plugin
class GVOutput;      // Video output plugin
class GVInput;
class GVPluginLoader;
struct GVPluginAttributes;

enum GVRepeatMode {
  GVRepeatNone = 0,
  GVRepeat1    = 1,
  GVRepeatAll  = 2
};

enum GVEngineState
{
  GVEngineStopped,    
  GVEnginePaused,     
  GVEnginePlaying, 
  GVEngineSeeking,
  GVEngineCompleting,
  GVEngineChanging,
  GVEngineError
};

class GVPlugin;
class GVInput;
class GVOutputAudio;
class GVOutput;
class GVBufferAudio;
class GVBufferVideo;

class GVEngine
  {
public:

  // Info strings
  
  char * i_input_audio_format;
  char * i_input_video_format;
  
  char * i_output_audio_format;
  char * i_output_video_format;

  char * i_input_info;
  char * i_output_audio_info;
  char * i_output_video_info;

  char * i_track_info;
  
  void lock_info_strings();
  void unlock_info_strings();
  
  enum SyncMode
    {
      SyncInput,
      SyncAudioOutput,
      SyncVideoOutput,
      SyncSoftware,
      SyncNone        // Time is set by input plugin
    };
  
  // Command Type
  
  enum CommandType
    {
      CommandNext,
      CommandPrevious,
      CommandPlay,
      CommandStop,
      CommandPause,
      CommandSeek,
      CommandPlaytrack,         // Play track <i>
      CommandPlayfile,          // Play file
      CommandAddfile,           // Add file to playlist
      CommandAddfileWithPlugin, // Add file to playlist
      CommandRefreshTrack       // Refresh track <i>
    };

  // Command Structure (element on the command stack)
  
  struct Command
    {
    GVEngine::CommandType type;
    char * arg_c;
    float arg_f;
    int arg_i;
        
    Command(GVEngine::CommandType _type);

    // For seeking
    
    Command(GVEngine::CommandType _type, float _arg_f);

    // For adding files
    
    Command(GVEngine::CommandType _type, const char * _arg_c);

    // For selecting tracks

    Command(GVEngine::CommandType _type, int _arg_i);
    ~Command();
    };
  
private:
  
  GVInput * input;              // Input plugin

  void get_config();
  void put_config();

  GVOutputAudio * output_audio; // Audio output plugin
  GVOutput      * output_video; // Video output plugin

  // List of inputs and outputs
  
//   GList * input_file_plugins;
//   GList * input_plugins;

//   GList * output_video_plugins;
//   GList * output_audio_plugins;

  // Infos for inputs and outputs

  const char * default_input_plugin;
  const char * default_audio_output_plugin;
  const char * default_video_output_plugin;
  
  const GVPluginInfo * input_info;

  // State variables

  unsigned int state_flags; 
  unsigned int internal_state;
  unsigned int config_flags;

  GVEngineState state;
  GVEngineState old_state;
    
  // Time values

  GVTime total_track_time;         // Total track time
  GVTime current_track_time;       // Current time of the track 

  // Software timer

  GTimer * timer;
  SyncMode sync_mode;
  
  // Mutexes

  pthread_mutex_t info_mutex;
  
  pthread_mutex_t time_mutex;
  pthread_mutex_t state_mutex;

  pthread_mutex_t plugin_mutex;
  pthread_cond_t plugin_cond;

  pthread_mutex_t command_mutex;
  GList * command_stack;

  // The next one is locked during *_cmd functions

  pthread_mutex_t cmd_mutex;
  
  pthread_t run_thread;

  // This ensures, that start_playback won't return before all plugins
  // are waiting for start

  int total_waiting_plugins; // Number of plugins which should run
  int num_waiting_plugins;    // Number of plugins which wait for start

  pthread_mutex_t num_waiting_plugins_mutex;
  
  // Performace measurement stuff

  pthread_mutex_t framerate_mutex;
  float framerate;

public:

  float get_framerate();
  void set_framerate(float _framerate);

private:
  
  // Startup stuff

  pthread_mutex_t start_mutex;
  pthread_cond_t start_cond;
    
  int num_plugins; // Number of plugins registered

  // Load a track from the playlist and the appropriate input
  // plugin
  
  void load_playlist_track();

//   const GVPluginInfo * find_input_by_name(const char * name);
//   const GVPluginInfo * find_output_video_by_name(const char * name);
//   const GVPluginInfo * find_output_audio_by_name(const char * name);
    
  GVRepeatMode repeat_mode;

  char * error_message;

  Command * pop_command(); // Pop a command from the command stack

  void wait_for_plugins(); // Wait until all plugin threads exited

  void delete_output_files();
  
  int use_audio;

  void plugin_error(GVPlugin * plugin);

  void start_playback();
  
public:

  // Resets the input, used for updating the playlist of cd player plugins
  void reset_input();
  
  void update_info();
  
  void set_use_audio(int _use_audio);
  int get_use_audio() { return use_audio; }
  
  
  // Startup synchronization functions (called by plugins)

  void wait_for_start();
    
  void push_command(Command * _command); // Push a command on the command stack
  int get_num_commands();              // Return lenght of the command stack
  
  void run();        // Start the engine thread
  void run_func();   // Run thread loop
  
  GVEngineState get_state();
  GVEngineState get_state(GVEngineState & old_state_ret);
  
  
  void set_state(GVEngineState _state);
    
  void load_plugins(const char * path);

  void set_config_flag(int flag, int value);
  int get_config_flags();

  void set_config_flags(int _config_flags);

#if 0  
  // Configuration stuff
  class PluginConfig : public GVConfigWidget
    {
  protected:

    int from_engine;       // Plugin comes from the engine, don't destroy this
    int no_menu_callback;  // Plugin comes from the engine, don't destroy this
    
    GtkWidget ** items;

    GtkWidget * table;
    
    GtkWidget * configure_button;
    GtkWidget * plugin_menu;
    GtkWidget * option_menu;
    GtkWidget * plugin_description;

    GVPluginLoader * plugin_loader;
    
    //    GVPlugin * plugin;
    GVPluginInfo * plugin_info;
    
    void load_plugin();

    void create_menu();

    void select_item(const char * name, const char * description);

    virtual void load_plugin(GVPluginInfo * info) = 0;
    virtual void destroy_plugin() = 0;
    virtual void configure() = 0;

  public:

    void button_callback(GtkWidget * w);
    void menu_callback(GtkWidget * w);
        
    PluginConfig();
    friend class GVEngine;
    };
  
private:  

  class OutputAudioConfig : public PluginConfig 
    {
    GVOutputAudio * plugin;
  protected:
    void load_plugin(GVPluginInfo * info);
    void destroy_plugin();
    void configure();
  public:
    OutputAudioConfig();
    void apply_values();
    void get_values();
    };

  class OutputVideoConfig : public PluginConfig
    {
    GVOutput * plugin;
    GtkWidget * use_audio;
  protected:
    void load_plugin(GVPluginInfo * info);
    void destroy_plugin();
    void configure();
  public:
    OutputVideoConfig();
    void apply_values();
    void get_values();
    };

  class InputConfig : public PluginConfig
    {
    GVInput * plugin;
  protected:
    void load_plugin(GVPluginInfo * info);
    void destroy_plugin();
    void configure();

    GtkWidget * media_files_item;
  public:
    InputConfig();
    void apply_values();
    void get_values();
    };

  friend class PluginConfig;
  friend class OutputAudioConfig;
  friend class OutputVideoConfig;
  friend class InputConfig;

#endif
    
  void set_current_track_time(); 

public:

#if 0  
  OutputAudioConfig output_audio_config;
  OutputVideoConfig output_video_config;
  InputConfig input_config;

  GVConfigWidget * config_input;
  GVConfigWidget * config_video_output;
  GVConfigWidget * config_audio_output;
#endif
  
  void error(const char * message);

  char * get_error_message();
  
  void repeat();
  GVRepeatMode get_repeat_mode();

  GVBufferAudio * audio_buffer;
  GVBufferVideo * video_buffer;
  
  // Manupilate the state flags
  
  int  get_state_flag(unsigned int flag);
  void set_state_flag(unsigned int flag);
  void clear_state_flag(unsigned int flag);
  void toggle_state_flag(unsigned int flag);

  // Plugin registering functions

  void plugin_attach();
  void plugin_detach();
  
  // Time functions
  
  void get_total_track_time(GVTime & time);  // Total track time

  // Compute and get the time of the track 

  void get_current_track_time(GVTime & time, int exact = 0);

  void set_total_track_time(const GVTime & time);   // Total track time

  // Current time of the track 

  void set_current_track_time(const GVTime & time); 
  void set_current_track_time(double time); 

  // Add a certain time difference to the track time 
  
  void add_current_track_time(const GVTime & time); 

  int get_num_inputs();
  GVPluginInfo * get_input_info(int index);

  void set_output_video(GVOutput * new_output);
  void set_output_audio(GVOutputAudio * new_output);
  void set_input(GVInput * new_input);

  void set_output_video(const char * new_output_name);
  void set_output_audio(const char * new_output_name);
  void set_input(const char * new_input_name);

  void configure_output_video();
  void configure_output_audio();
  void configure_input();

  const GVPluginAttributes * get_output_video_attr();
  const GVPluginAttributes * get_output_audio_attr();
  const GVPluginAttributes * get_input_attr();
  
  GVEngine();
  
  //  GVPlaylist playlist;

  // These are the commands for the command stack

  void seek(float percentage);
  void play_track(int track);
  void refresh_track(int track);
  void play();
  void pause();
  void next();
  void previous();
  void stop();
  void stop_sync();

  void completed();
  void add_file(const char * filename);  // Add file to playlist

  // Add file to playlist with known plugin
  
  void add_file_with_plugin(const char * filename, const char * plugin_name);

  // Add file to playlist and play it immediately

  void play_file(const char * filename); 

  // These are the corresponding functions called inside the thread

private:
  
  void seek_cmd(float percentage);
  void play_track_cmd(int track);
  void refresh_track_cmd(int track);

  // end_state can be GVEnginePlaying or GVEnginePaused
  // the latter will cause gmerlin to initialize everything
  // and go into the paused status
  
  void play_cmd(GVEngineState end_state);
  void next_cmd();
  void previous_cmd();
  void stop_cmd();
  //  void completed_cmd();
  void add_file_cmd(const char * filename);  // Add file to playlist

  // Add file to playlist with known plugin
  
  void add_file_with_plugin_cmd(const char * filename,
                                const char * plugin_name); 

  void play_file_cmd(const char * filename); 
  void pause_cmd();
  
public:
      
  void move_playlist_track(int src, int dst);
  
  ~GVEngine();
  };

// The engine is a global variable

extern GVEngine * gv_engine;

