/*****************************************************************

  enginecmd.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Command stack of the engine

#include <glib.h>
#include <pthread.h>
#include <ctype.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>

#include <sys/stat.h>
#include <unistd.h>

#include <iostream>

#include "config.h"
#include "gvtime.h"
#include "gvconfig.h"
#include "videoformat.h"
#include "audioformat.h"
#include "buffer.h"
#include "audiobuf.h"
#include "videobuf.h"

#include "playlist.h"

#include "baseplugin.h"
#include "plugin.h"
#include "input.h"
#include "output.h"
#include "audioout.h"
#include "videoout.h"

#include <gtk/gtk.h>
#include "configwindow.h"
#include "engine.h"

#include "gvloader.h"

using std::cerr;

GVEngine::Command::Command(GVEngine::CommandType _type) :
  type(_type)
  {
  arg_c = (char*)0;
  arg_i = 0;
  arg_f = 0.0;
  }

// For seeking

GVEngine::Command::Command(GVEngine::CommandType _type, float _arg_f) :
  type(_type),
  arg_f(_arg_f)
  {
  arg_c = (char*)0;
  arg_i = 0;
  }

// For adding files

GVEngine::Command::Command(GVEngine::CommandType _type, const char * _arg_c) :
  type(_type)
  {
  arg_c = g_strdup(_arg_c);
  arg_f = 0.0;
  arg_i = 0;
  }

// For selecting tracks

GVEngine::Command::Command(GVEngine::CommandType _type, int _arg_i) :
  type(_type),
  arg_i(_arg_i)
  
  {
  arg_c = (char*)0;
  arg_f = 0.0;
  }


GVEngine::Command::~Command()
  {
  if(arg_c)
    g_free(arg_c);
  }
  
void GVEngine::push_command(GVEngine::Command * _command)
  {
  pthread_mutex_lock(&command_mutex);
  command_stack = g_list_append(command_stack, _command);
  pthread_mutex_unlock(&command_mutex);
  };

// Pop a command from the command stack

GVEngine::Command * GVEngine::pop_command()
  {
  GVEngine::Command * ret;
  pthread_mutex_lock(&command_mutex);
  GList * first = g_list_first(command_stack);
  if(!first)
    {
    pthread_mutex_unlock(&command_mutex);
    return (GVEngine::Command*)0;
    }
  ret = (GVEngine::Command*)(first->data);
  command_stack = g_list_remove(command_stack, first->data);
  pthread_mutex_unlock(&command_mutex);
  return ret;
  }

int GVEngine::get_num_commands()     // Return lenght of the command stack
  {
  int ret;
  pthread_mutex_lock(&command_mutex);
  ret = g_list_length(command_stack);
  pthread_mutex_unlock(&command_mutex);
  return ret;
  }

void GVEngine::play()
  {
  Command * command = new Command(CommandPlay);
  push_command(command);
  }

void GVEngine::play_cmd(GVEngineState end_state)
  {
  // Return if we are already playing
  
  if(get_state() == GVEnginePlaying)
    return;

  // Delete previous errors
    
  if(get_state() == GVEngineError)
    set_state(GVEngineStopped);
  
  // Setup the input
  
  if(!(input) || (input->attributes->flags & GV_PLUGIN_IS_FILE))
    load_playlist_track();
  else
    input->init();
  
  if(!input)
    {
    error("Load a file first");
    return;
    }

  // Do something if the input has an error

  // Non file plugins trigger an engine error if they are initialized
  // and something goes wrong
  
  if(get_state() == GVEngineError)
    return;
  
  // File inputs don't trigger an engine error, because it's not always
  // wanted, so catch their errors
  
  if(input->get_error() != GV_PLUGIN_ERROR_NONE)
    {
    // Remove tracks from the playlist or go to the next track

    if(gv_playlist->is_private())
      {
      plugin_error(input);
      return;
      }
    else
      {
      if(get_config_flags() & GV_ENGINE_REMOVE_ERROR_TRACKS)
        gv_playlist->remove_current();
      else
        gv_playlist->next();
      // Send a new play command so we can play another track
      play();
      }
    
    return;
    }
  
  
  // Now set the internal state and the number of plugins which
  // should run

  total_waiting_plugins = 1; // Input is always there
  
  internal_state = 0;  
  if(input->track_has_audio())
    internal_state |= GV_INT_INPUT_HAS_AUDIO;
  if(input->track_has_video())
    internal_state |= GV_INT_INPUT_HAS_VIDEO;

  // Set the output flags

  if(!(input->attributes->flags & GV_PLUGIN_NO_OUTPUT))
    {
    switch(internal_state & GV_INT_INPUT_MASK)
      {
      case (GV_INT_INPUT_HAS_VIDEO|GV_INT_INPUT_HAS_AUDIO):
        if((output_video->attributes->flags & GV_PLUGIN_HAS_AUDIO) &&
           use_audio)
          {
          internal_state |= GV_INT_OUTPUT_AUDIO_VIDEO;
          output_video->do_audio = 1;
          total_waiting_plugins++;
          }
        else
          {
          internal_state |= (GV_INT_OUTPUT_VIDEO|GV_INT_OUTPUT_AUDIO);
          output_video->do_audio = 0;
          total_waiting_plugins+=2;
          }
        break;
      case GV_INT_INPUT_HAS_AUDIO:
        total_waiting_plugins++;
        if((output_video->attributes->flags & GV_PLUGIN_HAS_AUDIO) &&
           use_audio)
          {
          internal_state |= GV_INT_OUTPUT_AUDIO_VIDEO;
          output_video->do_audio = 1;
          output_video->do_video = 0;
          }
        else
          internal_state |= GV_INT_OUTPUT_AUDIO;
        break;
      case GV_INT_INPUT_HAS_VIDEO:
        internal_state |= GV_INT_OUTPUT_VIDEO;
        total_waiting_plugins++;
        break;
      }
    }

//   cerr << "Total waiting plugins: " << total_waiting_plugins << "\n";
  
  // Set the file flags

  if(input->attributes->flags & GV_PLUGIN_IS_FILE)
    internal_state |= GV_INT_INPUT_IS_FILE;

  switch(internal_state & GV_INT_OUTPUT_MASK)
    {
    case GV_INT_OUTPUT_AUDIO:
      if(output_audio->attributes->flags & GV_PLUGIN_IS_FILE)
        internal_state |= GV_INT_OUTPUT_IS_FILE;
      break;
    case GV_INT_OUTPUT_VIDEO:
    case GV_INT_OUTPUT_AUDIO_VIDEO:
      if(output_video->attributes->flags & GV_PLUGIN_IS_FILE)
        internal_state |= GV_INT_OUTPUT_IS_FILE;
      break;
    case (GV_INT_OUTPUT_AUDIO|GV_INT_OUTPUT_VIDEO):
      if((output_audio->attributes->flags & GV_PLUGIN_IS_FILE) ||
         (output_video->attributes->flags & GV_PLUGIN_IS_FILE))
        internal_state |= GV_INT_OUTPUT_IS_FILE;
      break;
    }

  // Decide if we can seek or not

  //   if(input->can_seek())
  //     cerr << "input->can_seek()\n";
  
  if(!(input->can_seek()) || ((internal_state & GV_INT_INPUT_IS_FILE) &&
                              (internal_state & GV_INT_OUTPUT_IS_FILE)))
    clear_state_flag(GV_ENGINE_CAN_SEEK);
  else
    set_state_flag(GV_ENGINE_CAN_SEEK);

  // Set the realtime behaviour

  if(input->attributes->flags & GV_PLUGIN_NEEDS_REALTIME)
    internal_state |= GV_INT_INPUT_NEEDS_REALTIME;  // Input needs realtime
  
  switch(internal_state & GV_INT_OUTPUT_MASK)
    {
    case GV_INT_OUTPUT_AUDIO:
      if(output_audio->attributes->flags & GV_PLUGIN_NEEDS_REALTIME)
        internal_state |= GV_INT_OUTPUT_NEEDS_REALTIME;
      break;
    case GV_INT_OUTPUT_VIDEO:
    case GV_INT_OUTPUT_AUDIO_VIDEO:
      if(output_video->attributes->flags & GV_PLUGIN_NEEDS_REALTIME)
        internal_state |= GV_INT_OUTPUT_NEEDS_REALTIME;
      break;
    case (GV_INT_OUTPUT_AUDIO|GV_INT_OUTPUT_VIDEO):
      if(!!(output_audio->attributes->flags & GV_PLUGIN_NEEDS_REALTIME) !=
         !!(output_video->attributes->flags & GV_PLUGIN_NEEDS_REALTIME))
        {
        g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
              "Output plugins for audio (%s) and video (%s) cannot be used together",
              output_audio->attributes->real_name,
              output_video->attributes->real_name);
        input->cleanup();
        return;
        }
      else if(output_audio->attributes->flags & GV_PLUGIN_NEEDS_REALTIME)
        internal_state |= GV_INT_OUTPUT_NEEDS_REALTIME;
    }

  switch(internal_state & GV_INT_REALTIME_MASK)
    {
    case GV_INT_OUTPUT_NEEDS_REALTIME:
      cerr << "Output needs realtime\n";
      break;
    case GV_INT_INPUT_NEEDS_REALTIME:
      cerr << "Input needs realtime\n";
      break;
    case 0:
      cerr << "No realtime\n";
      break;
    case GV_INT_INPUT_NEEDS_REALTIME|GV_INT_OUTPUT_NEEDS_REALTIME:
      cerr << "Input and Output need realtime\n";
      break;
    }
  
  if(internal_state & GV_INT_REALTIME_MASK)
    clear_state_flag(GV_ENGINE_NO_REALTIME);
  else
    set_state_flag(GV_ENGINE_NO_REALTIME);
  
  // Check which sync mode we want to have
  
  if(input->attributes->flags & GV_PLUGIN_OWN_TIME)
    sync_mode = SyncNone;
  else if(config_flags & GV_ENGINE_SYNC_SOFTWARE)
    {
    sync_mode = SyncSoftware;
    pthread_mutex_lock(&time_mutex);
    input->get_total_time(total_track_time);
    current_track_time.reset();
    pthread_mutex_unlock(&time_mutex);
    }
  else
    {
    pthread_mutex_lock(&time_mutex);
    input->get_total_time(total_track_time);
    current_track_time.reset();
    pthread_mutex_unlock(&time_mutex);
    
    switch(internal_state & GV_INT_REALTIME_MASK)
      {
      case 0: // 0 means: No realtime necessary, sync with output
        switch(internal_state & GV_INT_OUTPUT_MASK)
          {
          case GV_INT_OUTPUT_AUDIO:
          case GV_INT_OUTPUT_AUDIO|GV_INT_OUTPUT_VIDEO:
            sync_mode = SyncAudioOutput;
            break;
          case GV_INT_OUTPUT_VIDEO:
          case GV_INT_OUTPUT_AUDIO_VIDEO:
            sync_mode = SyncVideoOutput;
            break;
          }
        break;
      case GV_INT_OUTPUT_NEEDS_REALTIME:
        switch(internal_state & GV_INT_OUTPUT_MASK)
          {
          case GV_INT_OUTPUT_AUDIO:
            if(output_audio->can_sync())
              sync_mode = SyncAudioOutput;
            else
              sync_mode = SyncSoftware;
            break;
          case GV_INT_OUTPUT_VIDEO:
            if(output_video->can_sync())
              sync_mode = SyncVideoOutput;
            else
              sync_mode = SyncSoftware;
            break;
          case GV_INT_OUTPUT_AUDIO|GV_INT_OUTPUT_VIDEO:
            if(output_audio->can_sync())
              sync_mode = SyncAudioOutput;
            else if(output_video->can_sync())
              sync_mode = SyncVideoOutput;
            else
              sync_mode = SyncSoftware;
            break;
          case GV_INT_OUTPUT_AUDIO_VIDEO:
            if(output_video->can_sync())
              sync_mode = SyncVideoOutput;
            else
              sync_mode = SyncSoftware;
          }
        break;
      // Output goes to a a file, try to sync with input
      case GV_INT_INPUT_NEEDS_REALTIME:
        if(input->can_sync())
          sync_mode = SyncInput;
        else
          sync_mode = SyncSoftware;
        break;

      case (GV_INT_INPUT_NEEDS_REALTIME|GV_INT_OUTPUT_NEEDS_REALTIME):
        {
        if(input->can_sync())
          sync_mode = SyncInput;
        else if((output_audio) && (output_audio->can_sync()))
          sync_mode = SyncAudioOutput;
        else if((output_video) && (output_video->can_sync()))
          sync_mode = SyncVideoOutput;
        else
          sync_mode = SyncInput;
        
//         g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
//               "Cannot have realtime in both input and output");
//         input->cleanup();
//        return;
        }
      }
    }
  
   switch(sync_mode)
     {
     case SyncVideoOutput:
       cerr << "SyncVideoOutput\n";
       break;
     case SyncAudioOutput:
       cerr << "SyncAudioOutput\n";
       break;
     case SyncInput:
       cerr << "SyncInput\n";
       break;
     case SyncSoftware:
       cerr << "SyncSoftware\n";
       break;
     case SyncNone:
       cerr << "SyncNone\n";
       break;
     }
   
   // Set realtime behaviour for the plugins
   
   if(get_state_flag(GV_ENGINE_NO_REALTIME))
    {
    input->realtime = 0;
    if(output_audio)
      output_audio->realtime = 0;
    if(output_video)
      output_video->realtime = 0;
    }
  else
    {
    input->realtime = 1;
    if(output_audio)
      output_audio->realtime = 1;
    if(output_video)
      output_video->realtime = 1;
    }
  
  // Setup the software timer
  
  if(sync_mode == SyncSoftware)
    {
    if(!timer)
      timer = g_timer_new();
    g_timer_reset(timer);
    }
  
  // Initialize the output plugins

  if(!(input->attributes->flags & GV_PLUGIN_NO_OUTPUT))
    {
    switch(internal_state & GV_INT_OUTPUT_MASK)
      {
      case GV_INT_OUTPUT_AUDIO:

        // Initialize the output        
        
        output_audio->input = input;
        output_audio->init();
        if(get_state() == GVEngineError)
          {
          input->cleanup();
          return;
          }
        // Set up the buffer(s)

        output_audio->audio_buffer = audio_buffer;
        input->audio_buffer = audio_buffer;
                
        audio_buffer->input = input;
        audio_buffer->output = output_audio;

        audio_buffer->alloc();

        if(get_state() == GVEngineError)
          return;

        // Start the output plugin(s)

        output_audio->play();
        break;

      case GV_INT_OUTPUT_VIDEO:

        // Initialize the output        
        
        output_video->input = input;
        output_video->init();

        if(get_state() == GVEngineError)
          {
          input->cleanup();
          return;
          }

        // Set up the buffer(s)

        output_video->video_buffer = video_buffer;
        input->video_buffer = video_buffer;
        
        video_buffer->input = input;
        video_buffer->output = output_video;

        video_buffer->alloc();
        if(get_state() == GVEngineError)
          return;

        // Start the output plugin(s)

        output_video->play();
        break;

      case GV_INT_OUTPUT_AUDIO|GV_INT_OUTPUT_VIDEO:

        // 2 plugins for audio and video output
        
        // Initialize the output
        output_audio->input = input;
        output_video->input = input;

        output_audio->init();
        if(get_state() == GVEngineError)
          {
          input->cleanup();
          return;
          }

        output_video->init();
        if(get_state() == GVEngineError)
          {
          input->cleanup();
          output_audio->cleanup();
          return;
          }

        // Set up the buffer(s)

        output_video->video_buffer = video_buffer;
        output_audio->audio_buffer = audio_buffer;

        input->video_buffer = video_buffer;
        input->audio_buffer = audio_buffer;
        
        video_buffer->input = input;
        video_buffer->output = output_video;

        audio_buffer->input = input;
        audio_buffer->output = output_audio;
       
        video_buffer->alloc();
        if(get_state() == GVEngineError)
          return;
        audio_buffer->alloc();
        if(get_state() == GVEngineError)
          return;

        // Start the output plugin(s)
        
        output_video->play();
        output_audio->play();
        break;
        
      case GV_INT_OUTPUT_AUDIO_VIDEO:

        // Audio and video are handled by one plugin
        
        // Initialize the output
        output_video->input = input;
        output_video->init();

        if(get_state() == GVEngineError)
          {
          input->cleanup();
          return;
          }

        if(get_state() == GVEngineError)
          return;
        
        // Set up the buffer(s)

        // Input has an audio stream
        if(internal_state &  GV_INT_INPUT_HAS_AUDIO)  
          {
          output_video->audio_buffer = audio_buffer;
          input->audio_buffer = audio_buffer;
          audio_buffer->input = input;
          audio_buffer->output = output_video;
          audio_buffer->alloc();
          if(get_state() == GVEngineError)
            return;
          }

        // Input has a video stream
        
        if(internal_state &  GV_INT_INPUT_HAS_VIDEO)  
          {
          output_video->video_buffer = video_buffer;
          input->video_buffer = video_buffer;
          video_buffer->input = input;
          video_buffer->output = output_video;
          video_buffer->alloc();
          if(get_state() == GVEngineError)
            return;
          }
        
        // Start the output plugin(s)

        output_video->play();
        break;
      }
    }

  update_info();
  input->play();
  if(end_state == GVEnginePlaying)
    {
    set_state(GVEnginePlaying);
    start_playback();
    }
  else
    set_state(GVEnginePaused);
  
  //  set_state_flag(GV_ENGINE_STATE_CHANGED);

  //  cerr << "GVEngine::play() finished\n";
  }

void GVEngine::stop()
  {
  Command * command = new Command(CommandStop);
  push_command(command);
  }

// Stop the engine and wait until it's really stopped

void GVEngine::stop_sync()
  {
  GVTime wait_time;

  wait_time.seconds = 0;
  wait_time.nanoseconds = 1000000; // 1 Millisecond
  
  stop();

  // We wait until the engine has stopped

//   while(get_state() != GVEngineStopped)
//     wait_time.sleep();

  // Now, wait until all plugins exited

  while(1)
    {
    pthread_mutex_lock(&plugin_mutex);
    if(!num_plugins)
      {
      pthread_mutex_unlock(&plugin_mutex);
      break;
      }
    else
      {
      pthread_mutex_unlock(&plugin_mutex);
      wait_time.sleep();
      }
    }
  }

void GVEngine::stop_cmd()
  {
  GVEngineState old_state = get_state();
  
  if((old_state != GVEnginePlaying) &&
     (old_state != GVEnginePaused))
    return;

  if(old_state == GVEnginePlaying)
    {
    set_state(GVEngineStopped);
    wait_for_plugins();
    }
  else if(old_state == GVEnginePaused)
    {
    set_state(GVEngineStopped);
    start_playback();
    wait_for_plugins();
    }
  delete_output_files();
  
  audio_buffer->free();
  video_buffer->free();
  
  if(sync_mode == SyncSoftware)
    {
    g_timer_reset(timer);
    g_timer_stop(timer);
    }
  current_track_time.seconds = 0;
  current_track_time.nanoseconds = 0;
  }

void GVEngine::next()
  {
  Command * command = new Command(CommandNext);
  push_command(command);
  }

void GVEngine::next_cmd()
  {
  int was_playing = (get_state() == GVEnginePlaying) ? 1 : 0;

  if(was_playing)
    {
    set_state(GVEngineChanging);
    wait_for_plugins();
    delete_output_files();
    audio_buffer->free();
    video_buffer->free();
    }
  
  gv_playlist->next();

  if(was_playing)
    {
    if(input->attributes->flags & GV_PLUGIN_CAN_CHANGE)
      {
      set_state(GVEnginePlaying);
      //      cerr << "Starting playback...";
      start_playback();
      //      cerr << "done\n";
      }
    else
      play_cmd(GVEnginePlaying);
    }
  }

void GVEngine::previous()
  {
  Command * command = new Command(CommandPrevious);
  push_command(command);
  }

void GVEngine::previous_cmd()
  {
  int was_playing = (get_state() == GVEnginePlaying) ? 1 : 0;

  if(was_playing)
    {
    set_state(GVEngineChanging);
    wait_for_plugins();
    delete_output_files();
    audio_buffer->free();
    video_buffer->free();
    }
  
  gv_playlist->previous();

  if(was_playing)
    {
    if(input->attributes->flags & GV_PLUGIN_CAN_CHANGE)
      {
      set_state(GVEnginePlaying);
      start_playback();
      }
    else
      play_cmd(GVEnginePlaying);
    }
  }

void GVEngine::pause()
  {
  GVEngineState state = get_state();
  
  if((state == GVEnginePlaying) || (state == GVEngineStopped))
    {
    Command * command = new Command(CommandPause);
    push_command(command);
//     set_state(GVEnginePaused);
//     wait_for_plugins();
    return;
    }
  else if(state == GVEnginePaused)
    {
    set_state(GVEnginePlaying);
    start_playback();
    }
  
  }

void GVEngine::play_track_cmd(int track)
  {
  int was_playing = (get_state() == GVEnginePlaying) ? 1 : 0;

  if(was_playing)
    {
    set_state(GVEngineChanging);
    wait_for_plugins();
    delete_output_files();
    audio_buffer->free();
    video_buffer->free();
    }

  gv_playlist->set_current_track(track);

  if(was_playing)
    {
    if(input->attributes->flags & GV_PLUGIN_CAN_CHANGE)
      {
      set_state(GVEnginePlaying);
      start_playback();
      }
    else
      play_cmd(GVEnginePlaying);
    }
  else
    play_cmd(GVEnginePlaying);
  }

void GVEngine::refresh_track(int track)
  {
  Command * command = new Command(CommandRefreshTrack, track);
  push_command(command);
  }

void GVEngine::refresh_track_cmd(int track)
  {
  gv_playlist->refresh_track(track);
  }

void GVEngine::play_track(int track)
  {
  Command * command = new Command(CommandPlaytrack, track);
  push_command(command);
  }

void GVEngine::play_file(const char * filename)
  {
  Command * command = new Command(CommandPlayfile, filename);
  push_command(command);
  }

void GVEngine::play_file_cmd(const char * filename)
  {
  gv_playlist->make_private(0);
  add_file_cmd(filename);
  gv_playlist->set_current_track(gv_playlist->get_num_tracks()-1);

  if(get_state() == GVEnginePlaying)
    {
    set_state(GVEngineChanging);
    wait_for_plugins();
    delete_output_files();
    audio_buffer->free();
    video_buffer->free();
    }
  play_cmd(GVEnginePlaying);
  }

void GVEngine::seek_cmd(float perc)
  {
  GVEngineState old_state = get_state();

  if((old_state != GVEnginePlaying) &&
     (old_state != GVEnginePaused))
    return;

  if(!(input->can_seek()))
    {
    cerr << "Input cannot seek, sorry\n";
    return;
    }
  
  set_state(GVEngineSeeking); // Tell the plugins to interrupt

  if(old_state == GVEnginePlaying)
    {
    wait_for_plugins();      // Wait until all plugins stopped
    if(sync_mode == SyncSoftware)
      {
      pthread_mutex_lock(&time_mutex);
      g_timer_stop(timer);
      pthread_mutex_unlock(&time_mutex);
      }
    }
  if(internal_state & GV_INT_INPUT_HAS_AUDIO)  // Input has an audio stream
    audio_buffer->clear();
  if(internal_state & GV_INT_INPUT_HAS_VIDEO)    // Input has a video stream
    video_buffer->clear();
  
  input->seek_cmd(perc);       // Seek to the new percentage

  switch(internal_state & GV_INT_OUTPUT_MASK)
    {
    case GV_INT_OUTPUT_VIDEO:
    case GV_INT_OUTPUT_AUDIO_VIDEO:
      output_video->resync();
      break;
    case GV_INT_OUTPUT_AUDIO:
      output_audio->resync();
      break;
    case (GV_INT_OUTPUT_AUDIO|GV_INT_OUTPUT_VIDEO):
      output_audio->resync();
      output_video->resync();
      break;
    }
  input->preload();

  if(old_state == GVEnginePlaying)
    {
    set_state(GVEnginePlaying);
    start_playback();        // Restart the plugins
    if(sync_mode == SyncSoftware)
      {
      pthread_mutex_lock(&time_mutex);
      g_timer_start(timer);
      pthread_mutex_unlock(&time_mutex);
      }
    }
  else
    set_state(old_state);
  }

void GVEngine::pause_cmd()
  {
  GVEngineState state = get_state();
  
  if(state == GVEnginePlaying)
    {
    set_state(GVEnginePaused);
    wait_for_plugins();
    return;
    }
  else if(state == GVEnginePaused)
    {
    set_state(GVEnginePlaying);
    start_playback();
    }
  else if(state == GVEngineStopped)
    {
    play_cmd(GVEnginePaused);
    }
  }
  
// Add a media file to the playlist

void GVEngine::add_file(const char * filename)
  {
  Command * command = new Command(CommandAddfile, filename);
  push_command(command);
  }

void GVEngine::add_file_with_plugin(const char * filename,
                                    const char * plugin_name)
  {
  char * tmp_string = g_strdup_printf("%s:%s", plugin_name, filename);
  
  Command * command = new Command(CommandAddfileWithPlugin, tmp_string);
  push_command(command);
  g_free(tmp_string);
  }

void GVEngine::add_file_cmd(const char * filename)
  {
  if(gv_playlist->is_private())
    return;
  // First search for the proper plugin

  gv_playlist->add_file(filename);
  }

void GVEngine::add_file_with_plugin_cmd(const char * filename,
                                        const char * plugin_name)
  {
  if(gv_playlist->is_private())
    return;
  // First search for the proper plugin

  gv_playlist->add_file_with_plugin(filename, plugin_name);
  }
