/*****************************************************************

  gvconfig.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Routines for the global configuration database

#include "config.h"

#include <glib.h>

#include <iostream>
#include <fstream>
#include <iomanip>

using std::ofstream;
using std::ifstream;
using std::cerr;
using std::ios;

#include <string.h>
#include "gvconfig.h"

#include <stdlib.h>
#include <ctype.h>
#include "gvutil.h"

GVConfig * gv_config;

static gint __find_func(gconstpointer a, gconstpointer  b)
  {
  return strcmp(((GVConfigSection*)a)->name, (char*)b);
  }

GVConfig::GVConfig()
  {
  sections = (GList*)0;
  }

GVConfig::~GVConfig()
  {
  GVConfigSection * section;

  int num = g_list_length(sections);
  for(int i = 0; i < num; i++)
    {
    section = (GVConfigSection*)(g_list_nth_data(sections, i));
    delete section;
    }
  g_list_free(sections);
  }

GVConfigSection * GVConfig::find_section(const char * _name)
  {
  GList * list = g_list_find_custom(sections,
                                    (gpointer)_name,
                                    __find_func);
  if(list)
    return (GVConfigSection*)(list->data);
  return (GVConfigSection*)0;
  }

void GVConfig::put_int(const char * _section, const char * _name, int value)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    {
    section = new GVConfigSection(_section);
    sections = g_list_append(sections, (gpointer)section);
    }
  section->put_int(_name, value);
  }

void GVConfig::put_bool(const char * _section, const char * _name, int value)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    {
    section = new GVConfigSection(_section);
    sections = g_list_append(sections, (gpointer)section);
    }
  section->put_bool(_name, value);
  }


void GVConfig::put_float(const char * _section, const char * _name,
                         float value)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    {
    section = new GVConfigSection(_section);
    sections = g_list_append(sections, (gpointer)section);
    }
  section->put_float(_name, value);
  }

void GVConfig::put_string(const char * _section, const char * _name,
                          const char * value)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    {
    section = new GVConfigSection(_section);
    sections = g_list_append(sections, (gpointer)section);
    }
  section->put_string(_name, value);
  }

void GVConfig::put_stringlist(const char * _section, const char * _name,
                              GList * value)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    {
    section = new GVConfigSection(_section);
    sections = g_list_append(sections, (gpointer)section);
    }
  section->put_stringlist(_name, value);
  }

void GVConfig::put_color(const char * _section, const char * _name,
                         guchar red, guchar green, guchar blue)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    {
    section = new GVConfigSection(_section);
    sections = g_list_append(sections, (gpointer)section);
    }
  section->put_color(_name, red, green, blue);
  }

  // Get functions: return 1 on success
  
int GVConfig::get_int(const char * _section, const char * _name,
                      int & value)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    return 0;
  return section->get_int(_name, value);
  }

int GVConfig::get_bool(const char * _section, const char * _name,
                       int & value)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    return 0;
  return section->get_bool(_name, value);
  }

int GVConfig::get_float(const char * _section, const char * _name,
                        float & value)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    return 0;
  return section->get_float(_name, value);
  }

int GVConfig::get_string(const char * _section, const char * _name,
                         const char ** ret)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    return 0;
  return section->get_string(_name, ret);
  }

int GVConfig::get_stringlist(const char * _section, const char * _name,
                             GList ** ret)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    return 0;
  return section->get_stringlist(_name, ret);
  }

int GVConfig::get_color(const char * _section, const char * _name,
                        guchar & red, guchar & green, guchar & blue)
  {
  GVConfigSection * section = find_section(_section);
  if(!section)
    return 1;
  return section->get_color(_name, red, green, blue);
  }

static char * begin_section_string = "BeginSection: ";
static char * end_section_string = "EndSection";

int GVConfig::read(const char * filename)
  {
  char buffer[1024];

  GVConfigSection * section;
  GVConfigItem * item;

  int begin_section_length = strlen(begin_section_string);
  int end_section_length = strlen(end_section_string);
    
  ifstream input;
  input.open(filename);

  //  char delimiter;
  *buffer = '\0';

  char * pos;
  int c_int;
  float c_float;

  if(!input)
    {
    cerr << "Couldn't open Config file " << filename << "\n";
    return 0;
    }
  while(1)   // Read Section
    {
    // Read until the first BeginSection line comes
    while(strncmp((buffer), begin_section_string,
                  begin_section_length))
      {
      input.getline(buffer, 1024);
      //      input.get(delimiter);
      if(input.eof())
        return 1;
      }

    pos = &buffer[begin_section_length];
    section = new GVConfigSection(pos);
    sections = g_list_append(sections, (gpointer)section);
    
    // Read items

    while(1)
      {
      input.getline(buffer, 1024);
      //      input.get(buffer, '\n');
      //      input.get(delimiter);
      if(input.eof())
        return 1;
      // Get the name of the item

      //

      if(!strncmp((buffer), end_section_string,
                  end_section_length))
        break;
      
      pos = strchr(buffer, ':');
      
      if(!pos)
        {
        cerr << "Garbage in config file\n";
        continue;
        }
      
      *pos = '\0';
      
      // item's name is now in buffer

      pos++;

      while(isspace(*pos) && (*pos != '\0'))
        pos++;

      // Check what could be here
      
      if(*pos == '\0')
        item = new GVConfigItem(buffer, pos);
      if(!strcmp(pos, "TRUE"))
        item = new GVConfigItem(buffer, 1, GVConfigItemBool);
      else if(!strcmp(pos, "FALSE"))
        item = new GVConfigItem(buffer, 0, GVConfigItemBool);
      else if(*pos == '[')
        {
        char * end_of_string;
        const char * start_of_string;
        pos++;
        start_of_string = pos;
        item = new GVConfigItem(buffer, (GList*)0);
        while(1)
          {
          start_of_string = gv_read_string(pos, &end_of_string);

          if(start_of_string)
            {
            item->value.c_stringlist = g_list_append(item->value.c_stringlist,
                                                     g_strdup(start_of_string));
//             cerr << "GVConfig: found stringlist item " << start_of_string
//                  << "\n";
            }
          else
            item->value.c_stringlist = g_list_append(item->value.c_stringlist, (gpointer)0);
            
          while(isspace(*end_of_string))
            end_of_string++;
          if(*end_of_string == ']')
            break;
          pos = end_of_string;
          }
        }
      
      // Colors have the format {red,green,blue}

      else if(*pos == '{')
        {
        pos++;
        guchar red, green, blue;
        red = (guchar)atoi(pos);
        pos = strchr(pos, ',');
        if(!pos)
          {
          cerr << "Garbage in config file\n";
          continue;
          }
        pos++;

        green = (guchar)atoi(pos);
        pos = strchr(pos, ',');
        if(!pos)
          {
          cerr << "Garbage in config file\n";
          continue;
          }
        pos++;

        blue = (guchar)atoi(pos);
        pos = strchr(pos, '}');
        if(!pos)
          {
          cerr << "Garbage in config file\n";
          continue;
          }
        item = new GVConfigItem(buffer, red, green , blue);
        }
      else if(read_int(pos, c_int))
        item = new GVConfigItem(buffer, c_int, GVConfigItemInt);
      else if(read_float(pos, c_float))
        item = new GVConfigItem(buffer, c_float);
      else
        item = new GVConfigItem(buffer, pos);
      section->items = g_list_append(section->items, (gpointer)item);
      
      }   // End read items loop 
    } // End read section loop
  return 1;
  }

int GVConfig::read_int(char * pos, int & ret)
  {
  char * end;
  ret = strtol(pos, &end, 10);

  if((pos == end) || (*end != '\0'))
    return 0;
  return 1;
  }

int GVConfig::read_float(char * pos, float & ret)
  {
  char * end;
  ret = strtod(pos, &end);

  if((pos == end) || (*end != '\0'))
    return 0;
  return 1;
  }


void GVConfig::write(const char * filename)
  {
  ofstream output;
  output.open(filename);

  output.setf(ios::fixed);
  output.precision(4);
  
  GVConfigSection * section;
  GVConfigItem * item;
  
  int num_sections = g_list_length(sections);
  int num_items;

  // Write the header

  output << "# Configuration File for "
         << PACKAGE << "(Version " << VERSION << ")\n";

  output << "# Everything you find here can be configured with "
         << PACKAGE << '\n';

  output << "# Editing happens at own risk ;)\n";
  
  for(int i = 0; i < num_sections; i++)
    {
    section = (GVConfigSection *)g_list_nth_data(sections, i);

    // Don't save sections starting with '_'

    if((section->name[0] == '_') || (!section->used))
      continue;
    
    output << "BeginSection: " << section->name << '\n';

    num_items = g_list_length(section->items);
    
    for(int j = 0; j < num_items; j++)
      {
      item = (GVConfigItem*)g_list_nth_data(section->items, j);

      if(!item->used)
        continue;
      output << item->name << ": ";

      switch(item->type)
        {
        case GVConfigItemBool:
          if(item->value.c_int)
            output<< "TRUE";
          else
            output << "FALSE";
          break;
        case GVConfigItemInt:
          output << item->value.c_int;
          break;
        case GVConfigItemFloat:
          output << item->value.c_float;
          break;
       case GVConfigItemString:
         if(item->value.c_string)
           output << item->value.c_string;
         break;
       case GVConfigItemStringlist:
         if(item->value.c_stringlist)
           {
           int l = g_list_length(item->value.c_stringlist);
           char * c_tmp;
           output << "[";
           for(int i = 0; i < l; i++)
             {
             c_tmp = (char*)g_list_nth_data(item->value.c_stringlist, i);
             gv_write_string(output, c_tmp);
             if(i == l - 1)
               output << "]";
             else
               output << " ";
             }
           }
         break;
        case GVConfigItemColor:
         output << '{' << (int)(item->value.c_color.red) << ", " 
                << (int)(item->value.c_color.green) << ", "
                << (int)(item->value.c_color.blue) << '}';
          break;
        }
      output << '\n';
      }
    output << "EndSection\n\n";
    }
  }
