/*****************************************************************

  gvloader.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include "config.h"
#include <ctype.h>
#include <string.h>

#include <glib.h>

#include <dlfcn.h>

#include <pthread.h>
#include <iostream>

using std::cerr;
using std::ostream;

#include "gvtime.h"
#include "gvconfig.h"
#include "videoformat.h"
#include "audioformat.h"
#include "buffer.h"
#include "audiobuf.h"
#include "videobuf.h"

// #include "playlist.h"

#include "baseplugin.h"
#include "plugin.h"
#include "input.h"

#include "output.h"
#include "audioout.h"
#include "videoout.h"
#include "imagereader.h"

#include "gvloader.h"

GVPluginInfo::GVPluginInfo(GVBasePlugin * plugin, const char * _filename)
  {
  name = g_strdup(plugin->attributes->name);
  real_name = g_strdup(plugin->attributes->real_name);

  if(plugin->attributes->suffixes)
    suffixes = g_strdup(plugin->attributes->suffixes);
  else
    suffixes = (char*)0;

  description = g_strdup(plugin->attributes->description);
  
  filename = g_strdup(_filename);
  flags = plugin->attributes->flags;
  }

GVPluginInfo::~GVPluginInfo()
  {
  g_free(filename);
  g_free(name);
  if(suffixes)
    g_free(suffixes);
  g_free(description);
  }

int GVPluginInfo::check_suffix(const char * suffix)
  {
  char * start = suffixes;
  char * end = suffixes;

  while((*end != ' ') && (*end != '\0'))
    end++;

  while(*start != '\0')
    {
    if(!strncmp(suffix, start, end - start))
      return 1;
    start = end;
    if(*end == '\0')
      break;
    else
      {
      start++;
      end++;
      while((*end != ' ') && (*end != '\0'))
        end++;
      }
    }
  return 0;
  }

GVPluginLoader * gv_input_file_loader   = (GVPluginLoader*)0;
GVPluginLoader * gv_input_loader        = (GVPluginLoader*)0;
GVPluginLoader * gv_video_output_loader = (GVPluginLoader*)0;
GVPluginLoader * gv_audio_output_loader = (GVPluginLoader*)0;
GVPluginLoader * gv_image_reader_loader = (GVPluginLoader*)0;

// Loader Class

const char * GVPluginLoader::get_plugin_name(int index)
  {
  const GVPluginInfo * info = (const GVPluginInfo*)(g_list_nth_data(plugins, index));
  if(!info)
    return (const char *)0;
  else
    return info->name;
  }

const char * GVPluginLoader::get_plugin_real_name(int index)
  {
  const GVPluginInfo * info = (const GVPluginInfo*)(g_list_nth_data(plugins, index));
  if(!info)
    return (const char *)0;
  else
    return info->real_name;
  }

const GVPluginInfo * GVPluginLoader::get_plugin_info(int index)
  {
  return (const GVPluginInfo*)(g_list_nth_data(plugins, index));
  }

// Loader functions

GVBasePlugin * GVPluginLoader::load_plugin(const char * filename)
  {
  GVBasePlugin * ret;
  GVBasePlugin * (*LoadPluginFunc)(void) = (GVBasePlugin * (*)(void))0; 

  void * module = dlopen(filename, RTLD_LAZY);
  
  if(!module)
    return (GVBasePlugin*)0;

//   void (*init_func)(void);

//   if((init_func = (void(*)(void))dlsym(module, "_init")))
//     init_func();
    
  if(!(LoadPluginFunc = (GVBasePlugin *(*)(void))dlsym(module, "get_plugin")))
    {
    cerr << "No Symbol get_plugin found, " << dlerror() << "\n";
    dlclose(module);
    return (GVBasePlugin*)0;
    }
  if(!LoadPluginFunc)
    {
    cerr << "Symbol get_plugin is NULL, " << "\n";
    return (GVBasePlugin*)0;
    }
  
  ret = LoadPluginFunc();
  ret->module = (gpointer)module;

//   cerr << "Loaded " << filename << ", attributes: "
//        << (const void*)ret->attributes << "\n";
  
  return ret;
  }

void GVPluginLoader::destroy_plugin(GVBasePlugin * plugin)
  {
  if(!plugin)
    {
    cerr << "Tried to Destroy NULL plugin\n";
    return;
    }
  void (*DestroyPluginFunc)(GVBasePlugin*) = (void (*)(GVBasePlugin*))0;

  //  cerr << "Destroying " << plugin->attributes->name << "\n";

  void * module = plugin->module;
  DestroyPluginFunc = (void (*)(GVBasePlugin*))dlsym(module, "destroy_plugin");
 
  DestroyPluginFunc(plugin);

//   void (*fini_func)(void);

//   if((fini_func = (void(*)(void))dlsym(module, "_fini")))
//     fini_func();
  dlclose(module);
  }


GVInput * GVPluginLoader::load_input(GVPluginInfo * info)
  {
  return (GVInput*)(load_plugin(info->filename));
  }

GVOutput * GVPluginLoader::load_output(GVPluginInfo * info)
  {
  return (GVOutput*)(load_plugin(info->filename));
  }

GVOutputAudio * GVPluginLoader::load_output_audio(GVPluginInfo * info)
  {
  return (GVOutputAudio*)(load_plugin(info->filename));
  }

GVImageReader * GVPluginLoader::load_image_reader(GVPluginInfo * info)
  {
  return (GVImageReader*)(load_plugin(info->filename));
//   cerr << "Loaded Imagereader from " << info->filename << " attributes: "
//        << (const void*)ret->attributes << "\n";
  //  return ret;

  //  return (GVImageReader*)(load_plugin(info->filename));
  }

// Find plugins

GVPluginInfo * GVPluginLoader::find_plugin_by_name(const char * name)
  {
  GVPluginInfo * info;
  for(int i = 0; i < num_plugins; i++)
    {
    info = (GVPluginInfo*)g_list_nth_data(plugins, i);
    if(!strcmp(info->name, name))
      return info;
    }
  return (GVPluginInfo *)0;
  }

GVPluginInfo * GVPluginLoader::find_plugin_by_suffix(const char * filename)
  {
  GVPluginInfo * info;

  const char * pos_c;
  char * pos;
  
  // Check if there is a suffix
  
  if((pos_c = strrchr(filename, '.')))
    {
    pos_c++;
    if(pos_c == '\0')
      return (GVPluginInfo*)0;
    }
  else
    return (GVPluginInfo*)0;

  char * suffix = g_strdup(pos_c);

  pos = suffix;
  while(*pos != '\0')
    {
    *pos = tolower(*pos);
    pos++;
    }
  
  for(int i = 0; i < num_plugins; i++)
    {
    info = (GVPluginInfo*)g_list_nth_data(plugins, i);
    if(info->check_suffix(suffix))
      {
      g_free(suffix);
      return info;
      }
    }

  g_free(suffix);
  return (GVPluginInfo*)0;
  }

GVPluginLoader::GVPluginLoader()
  {
  num_plugins = 0;
  plugins = (GList*)0;
  }

// Add a plugin to the list, returns true on success

int GVPluginLoader::add_plugin(const char * module_filename)
  {
  GVBasePlugin * plugin = load_plugin(module_filename);
  if(!plugin)
    return 0;

  GVPluginInfo * info = new GVPluginInfo(plugin, module_filename);
        
  plugins = g_list_append(plugins, (gpointer)info);
  num_plugins++;

  destroy_plugin(plugin);
    
  return 1;
  }
