/*****************************************************************

  gvtime.h

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Time type
// This has 2 fields: seconds and nanoseconds
// If a time is negative, only the seconds are negative!!

#define GV_NANOSEC_PER_SEC 1000000000

struct GVTime
  {
  gint32 seconds;
  gint32 nanoseconds;

  GVTime& operator=(const GVTime& time)
    {
      seconds = time.seconds;
      nanoseconds = time.nanoseconds;
      return *this;
    }

  GVTime& operator=(double time)
    {
      seconds = (int)time;
      time -= seconds;
      time *= GV_NANOSEC_PER_SEC;
      nanoseconds = (int)time;
      return *this;
    }
  
  GVTime& operator-=(const GVTime& time)
    {
    nanoseconds -= time.nanoseconds;
    if(nanoseconds < 0)
      {
      nanoseconds += GV_NANOSEC_PER_SEC;
      seconds--;
      }
    seconds -= time.seconds;
    return *this;
    }

  GVTime& operator+=(const GVTime& time)
    {
    nanoseconds += time.nanoseconds;
    if(nanoseconds >= GV_NANOSEC_PER_SEC)
      {
      nanoseconds -= GV_NANOSEC_PER_SEC;
      seconds++;
      }
    seconds += time.seconds;
    return *this;
    }
  
  double get_double() const
    {
    return (double)seconds +
      (double)(nanoseconds)/(double)(GV_NANOSEC_PER_SEC);
    }
  
  void set_from_samples(guint64 samples, unsigned int samplerate)
    {
    gint64 tmp;
    seconds = samples/samplerate;
    tmp = samples % samplerate;
    tmp *= GV_NANOSEC_PER_SEC;
    tmp /= samplerate;
    nanoseconds = (gint)(tmp);
    }

  void set_from_frames(unsigned int frames, double framerate)
    {
    // We do this in double precision here

    double time = frames / framerate;

    seconds = (int)(time);
    time -= seconds;
    time *= GV_NANOSEC_PER_SEC;
    nanoseconds = (int)(time);
    }

  guint64 get_samples(unsigned int samplerate);
  int get_frames(double framerate);
  
  
  int operator<(const GVTime & time) const
    {
    if(seconds < time.seconds)
      return 1;
    else if(seconds > time.seconds)
      return 0;
    else
      return (nanoseconds < time.nanoseconds);
    }

  int operator>(const GVTime & time) const
    {
    if(seconds > time.seconds)
      return 1;
    else if(seconds < time.seconds)
      return 0;
    else
      return (nanoseconds > time.nanoseconds);
    } 

  int operator==(const GVTime & time) const
    {
    if((seconds == time.seconds) && (nanoseconds == time.nanoseconds))
      return 1;
    else return 0;
    }
  
  GVTime& operator/=(int fac)
    {
    gint64 old_seconds = seconds;
    
    nanoseconds/=fac;
    seconds /= fac;

    old_seconds %= fac;
    old_seconds *= GV_NANOSEC_PER_SEC;
    old_seconds /= fac;
    
    nanoseconds += (gint)(old_seconds);
    return *this;
    }

  void sleep(); // Sleep for the specified time
  void print(char *,char *); // print the time in seconds

  void reset()
    {
    seconds = 0;
    nanoseconds = 0;
    }

  void set_undefined()
    {
    seconds = G_MININT;
    nanoseconds = 0;
    }

  int undefined() const
    {
    return ((seconds == G_MININT) && (nanoseconds == 0));
    }
  
  };
