/*****************************************************************

  input.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Common Methods for input plugins

#include "config.h"

#include <gtk/gtk.h>
#include <pthread.h>
#include <iostream.h>
#include <string.h>

#include "gvtime.h"

#include "videoformat.h"
#include "audioformat.h"
#include "buffer.h"

#include "videobuf.h"
#include "audiobuf.h"

#include "playlist.h"
#include "baseplugin.h"
#include "plugin.h"
#include "input.h"

#include "output.h"
#include "audioout.h"
#include "videoout.h"

#include "configwindow.h"
#include "engine.h"

#include "gvconfig.h"

// Load audio and video frames until the buffers are full

void GVInput::preload()
  {
  if(attributes->flags & (GV_PLUGIN_NO_OUTPUT|GV_PLUGIN_IS_RECORDER))
    return;
  
  int audio_full = ((flags & GV_PLUGIN_TRACK_HAS_AUDIO) && do_audio) ? 0 : 1;
  int video_full = ((flags & GV_PLUGIN_TRACK_HAS_VIDEO) && do_video) ? 0 : 1;

  GVAudioFrame * audio_frame;
  GVVideoFrame * video_frame;
    
  while(!(audio_full) || !(video_full))
    {
    if(!audio_full && !audio_buffer->full())
      {
      audio_frame = audio_buffer->lock_write();

      audio_frame->time.set_undefined();
      fill_audio_buf(audio_frame);
      audio_samples_written += audio_frame->valid_samples;
      if(audio_frame->time.undefined())
        audio_frame->time.set_from_frames((unsigned int)audio_samples_written,
                                          (unsigned int)(audio_format.samplerate));
      audio_buffer->unlock_write();
      }
    else
      audio_full = 1;
    
    if(!video_full && !video_buffer->full())
      {
      video_frame = video_buffer->lock_write();
      video_frame->time.set_undefined();
      fill_video_buf(video_frame);
      video_frames_written++;
      if(video_frame->time.undefined())
        video_frame->time.set_from_frames((unsigned int)video_frames_written,
                                          (video_format.framerate));
      video_frame->skip = 0;
      video_buffer->unlock_write();
      }
    else
      video_full = 1;
    }
  }

int GVInput::colorspace_supported(GVColorspace colorspace)
  {
  return 0;
  }

void GVInput::playback_func()
  {
  //  cerr << attributes->name << ": playback_func() started\n";
  //  pthread_mutex_lock(&destroy_mutex);

  gv_engine->plugin_attach();

  if((flags & GV_PLUGIN_TRACK_HAS_VIDEO) && do_video)
    gv_engine->set_framerate(video_format.framerate);
    
  frames_to_skip = 0;
  skipped_frames = 0;
  
  video_frames_written = 0;
  audio_samples_written = 0;
  
  GVAudioFrame * audio_frame;
  GVVideoFrame * video_frame;

  audio_completed = ((flags & GV_PLUGIN_TRACK_HAS_AUDIO) && do_audio) ? 0 : 1;
  video_completed = ((flags & GV_PLUGIN_TRACK_HAS_VIDEO) && do_video) ? 0 : 1;

//   cerr << "Preloadling...";
//   preload();
//   cerr << "done\n";

  GVTime wait_time,short_time;

  int audio_full = (flags & GV_PLUGIN_TRACK_HAS_AUDIO) ? 0 : 1;
  int video_full = (flags & GV_PLUGIN_TRACK_HAS_VIDEO) ? 0 : 1;

  int have_audio = !!(flags & GV_PLUGIN_TRACK_HAS_AUDIO);
  int have_video = !!(flags & GV_PLUGIN_TRACK_HAS_VIDEO);
  
  // Set the waittime
  if(realtime)
    {
    if(!(flags & GV_PLUGIN_TRACK_HAS_AUDIO))
      wait_time = video_format.frame_time ;
    else if(!(flags & GV_PLUGIN_TRACK_HAS_VIDEO))
      wait_time = audio_format.frame_time  ;
    else // audio and video are present, take the smaller one of both
      wait_time = (audio_format.frame_time < video_format.frame_time) ?
        audio_format.frame_time : video_format.frame_time ;
    short_time = wait_time;
    short_time /= 5;
    }
  else
    {
    wait_time.seconds = 0;
    wait_time.nanoseconds = 100000; // 100 microseconds should be enough

    short_time.seconds = 0;
    short_time.nanoseconds = 100000; // 100 microseconds should be enough
    }

  //  GVEngineState state;

  internal_video_time.reset();
  internal_audio_time.reset();
  internal_time.reset();

  // Check, if we are recording
  
  int limited_recording = 0;

  if((flags & GV_PLUGIN_IS_RECORDER) && !(total_time.undefined()))
    limited_recording = 1;
  
  int skip_flag;
  
  preload();
    
  gv_engine->wait_for_start();
  
  while(1)
    {
    // Engine stopped

    if(check_engine_state())
      break;
    
    if((have_audio) && (!audio_completed))
      {
      if(audio_buffer->full())
        audio_full = 1;
      else
        {
        audio_full = 0;
        audio_frame = audio_buffer->lock_write();
        audio_frame->time.set_undefined();
        fill_audio_buf(audio_frame);
        audio_samples_written+= audio_frame->valid_samples;
        internal_audio_time.set_from_frames((unsigned int)(audio_samples_written),
                                            (unsigned int)(audio_format.samplerate));

        if(audio_frame->time.undefined())
          audio_frame->time = internal_audio_time;

        if(limited_recording)
          {
          if(audio_samples_written >= audio_format.num_samples)
            {
            audio_frame->valid_samples
              -= audio_samples_written - audio_format.num_samples;
            audio_completed = 1;
            }
          }
        audio_buffer->unlock_write();
        }
      }
    if((have_video) && !(video_completed))
      {
      if(video_buffer->full())
        video_full = 1;
      else
        {
        video_full = 0;
	// skipping frame if necessary
        pthread_mutex_lock(&skip_mutex);
        if(frames_to_skip)
          {
          skip_flag = 1;
          frames_to_skip = skip((int)frames_to_skip);
          cerr << attributes->name << ": Skipped " << frames_to_skip << " video frames\n";
          skipped_frames += frames_to_skip;
          gv_engine->set_framerate(video_format.framerate *
                                 (float)(video_frames_written-skipped_frames)/
                                 (float)video_frames_written);
          video_frames_written += frames_to_skip;
          frames_to_skip = 0;
          }
        else
          skip_flag = 0;
        
        pthread_mutex_unlock(&skip_mutex);

        video_frame = video_buffer->lock_write();
	video_frame->time.set_undefined(); 
        fill_video_buf(video_frame);

        video_frame->skip = skip_flag;
        
        video_frames_written++;

        if (video_frame->time.undefined())
          {
          internal_video_time.set_from_frames(video_frames_written,
                                              video_format.framerate);
          video_frame->time = internal_video_time;
          }
        else
          internal_video_time = video_frame->time;
        
	video_buffer->unlock_write();

        if(limited_recording)
          {
          if(internal_video_time > total_time)
            video_completed = 1;
          }
        }
      }

    if(!have_audio)
      {
      pthread_mutex_lock(&internal_time_mutex);
      internal_time = internal_video_time;
      pthread_mutex_unlock(&internal_time_mutex);
      }
    else if(!have_video)
      {
      pthread_mutex_lock(&internal_time_mutex);
      internal_time = internal_audio_time;
      pthread_mutex_unlock(&internal_time_mutex);
      }
    else
      {
      pthread_mutex_lock(&internal_time_mutex);
      internal_time = (internal_video_time > internal_audio_time)
        ? internal_video_time : internal_audio_time;
      pthread_mutex_unlock(&internal_time_mutex);
      }
    
    if((video_completed) && (audio_completed))
      break;

    if((audio_full) && (video_full))
      wait_time.sleep();
    else
      short_time.sleep();
    }
  cleanup();
  gv_engine->plugin_detach();
  // BP: This must be "&&", because video_completed is always 1
  // for audio only files  
  if((video_completed) && (audio_completed))
    gv_engine->set_state(GVEngineCompleting);
  }

void GVInput::seek_cmd(float perc)
  {
  int have_video = !!((flags & GV_PLUGIN_TRACK_HAS_VIDEO) && do_video);
  int have_audio = !!((flags & GV_PLUGIN_TRACK_HAS_AUDIO) && do_audio);

  if(have_video)
    video_frames_written = (int)(video_format.num_frames*perc + 0.5);
  if(have_audio)
    audio_samples_written = (int)(audio_format.num_samples*perc + 0.5);
  
  // Let the derived class search...
  
  seek(perc);

  if(have_video)
    internal_video_time.set_from_frames(video_frames_written,
                                        video_format.framerate);

  if(have_audio)
    internal_audio_time.set_from_frames(audio_samples_written,
                                        audio_format.samplerate);
  if(!have_audio)
    {
    pthread_mutex_lock(&internal_time_mutex);
    internal_time = internal_video_time;
    pthread_mutex_unlock(&internal_time_mutex);
    }
  else if(!have_video)
    {
    pthread_mutex_lock(&internal_time_mutex);
    internal_time = internal_audio_time;
    pthread_mutex_unlock(&internal_time_mutex);
    }
  else
    {
    pthread_mutex_lock(&internal_time_mutex);
    internal_time = (internal_video_time > internal_audio_time)
      ? internal_video_time : internal_audio_time;
    pthread_mutex_unlock(&internal_time_mutex);
    }
  //  preload();        // Preload the buffers, so we can fire up again

  pthread_mutex_lock(&internal_time_mutex);
  gv_engine->set_current_track_time(internal_time); 
  pthread_mutex_unlock(&internal_time_mutex);

  cerr << "Seeked, audio_time: " << internal_audio_time.get_double()
       << " video_time: " << internal_video_time.get_double()
       << " internal_time: " << internal_time.get_double() << "\n";
  
  }

void GVInput::fill_playlist()
  {
  
  }

// Cleanup function (called before selecting a new file)

void GVInput::fill_audio_buf(GVAudioFrame * frame)
  {
  
  }

void GVInput::fill_video_buf(GVVideoFrame * frame)
  {
  
  }

GVInput::GVInput(const GVPluginAttributes * _attributes) :
  GVPlugin(_attributes)
  {
  pthread_mutex_init(&skip_mutex,(pthread_mutexattr_t *)0);
  best_colorspace = GVColorspaceRGB24;
  flags = 0;
  do_audio = 1;
  do_video = 1;
  device = (const char*)0;
  devices = (GList*)0;

  // We set this here, so recording plugins don't have to care  
  
  total_time.set_undefined();
  record_time.set_undefined();
  }

GVInput::~GVInput()
  {
  if(devices)
    {
    int num = g_list_length(devices);
    for(int i = 0; i < num; i++)
      {
      char * c_tmp = (char*)g_list_nth_data(devices, i);
      g_free(c_tmp);
      }
    g_list_free(devices);
    }
  }

void GVInput::alloc_frame(GVVideoFrame * frame)
  {

  }

void GVInput::free_frame(GVVideoFrame * frame)
  {
  }

void GVInput::alloc_frame(GVAudioFrame * frame)
  {
  }

void GVInput::free_frame(GVAudioFrame * frame)
  {

  }

void GVInput::seek(float perc)
  {
  }

int GVInput::skip(int num_frames)
  {
  // This is currently used only by the singlepicture plugin,
  // I think, no skipping is neccesary here since it's done
  // automatically
  return num_frames;
  }


void GVInput::set_colorspace(GVColorspace colorspace)
  {
  if(!colorspace_supported(colorspace))
    {
    g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "Illegal colorspace");
    return;
    }
  video_format.colorspace = colorspace;
  video_format.init();
  }

void GVInput::skip_video_frames(int num_frames)
  {
  pthread_mutex_lock(&skip_mutex);
  frames_to_skip += (guint) num_frames;
  pthread_mutex_unlock(&skip_mutex);
  }

const GVTrackInfo * GVInput::get_track_info()
  {
  return (const GVTrackInfo*)0;
  }

// Multiple devices support

static gint __compare_func(gconstpointer  a,
                           gconstpointer  b)
  {
  if(!a)
    return 1;
  return strcmp((const char*)a, (const char*)b);
  }

char * GVInput::get_device_string(char * devicefile)
  {
  return (char*)0;
  }

void GVInput::add_device(const char * _device)
  {
  // Check if a device with this name is already here

  char * c_tmp = g_strdup(_device);
  char * tmp_devicename;

  // Device is already here

  gpointer test_ptr = g_list_find_custom(devices, c_tmp, __compare_func);
  
  if(test_ptr)
    {
    g_free(c_tmp);
    device_index = g_list_index(devices, test_ptr)/2;
    return;
    }
  else
    {
    devices = g_list_append(devices, (gpointer)c_tmp);
    if(attributes->flags & GV_PLUGIN_HAS_DEVICESTRING)
      {
      tmp_devicename = get_device_string(c_tmp);
      devices = g_list_append(devices, (gpointer)tmp_devicename);
      }
    else
      devices = g_list_append(devices, (gpointer)0);
    device_index = g_list_index(devices, c_tmp);
    }
  }

void GVInput::remove_device(const char * _device)
  {
  char * c_tmp = g_strdup(_device);
  
  GList * test_list = g_list_find_custom(devices, c_tmp, __compare_func);
  if(!test_list)
    {
    g_free(c_tmp);
    return;
    }
  
  int index = g_list_index(devices, test_list->data);
  
  g_free(test_list->data);
  devices = g_list_remove(devices, test_list->data);
  
  test_list = g_list_nth(devices, index);

  if(test_list->data)  
    g_free(test_list->data);
  
  devices = g_list_remove(devices, test_list->data);
  
  g_free(c_tmp);
  }

void GVInput::set_device(const char * _device)
  {
  char * c_tmp = g_strdup(_device);
  const char * old_device;
  
  GList * test_list = g_list_find_custom(devices, c_tmp, __compare_func);
  if(!test_list)
    {
    g_free(c_tmp);
    return;
    }
  
  device_index = g_list_index(devices, test_list->data)/2;
  old_device = device;
  device = (const char*)g_list_nth_data(devices, device_index * 2);
  
  g_free(c_tmp);

  if(strcmp(_device, old_device))
    gv_engine->reset_input();
  
  }

#define SECTION attributes->name

void GVInput::get_devices_config(const char * _default_device)
  {
  char * default_device = g_strdup(_default_device);

  devices = g_list_append(devices, default_device);
  devices = g_list_append(devices, get_device_string(default_device));
  
  gv_config->get_stringlist(SECTION, "Devices", &devices);

  device_index = 0;
  gv_config->get_int(SECTION, "DeviceIndex", device_index);
  device = (const char*)g_list_nth_data(devices, 2*device_index);
  }

void GVInput::put_devices_config()
  {
  gv_config->put_int(SECTION, "DeviceIndex", device_index);
  gv_config->put_stringlist(SECTION, "Devices", devices);
  }

void GVInput::get_recording_config()
  {
  int limit_recording_time = 0;
  int seconds;

  seconds = 60;
  gv_config->get_int(attributes->name, "RecordingTime", seconds);
  record_time.nanoseconds = 0;
  record_time.seconds = seconds;
  
  gv_config->get_bool(attributes->name, "LimitRecordingTime",
                      limit_recording_time);

  if(!limit_recording_time)
    record_time.set_undefined();
  }

void GVInput::put_recording_config()
  {
  if(record_time.undefined())
    {
    gv_config->put_bool(attributes->name, "LimitRecordingTime", 0);
    gv_config->put_int(attributes->name, "RecordingTime", 60);
    }
  else
    {
    gv_config->put_bool(attributes->name, "LimitRecordingTime", 1);
    gv_config->put_int(attributes->name, "RecordingTime", record_time.seconds);
    }
  }

// return 1 if the string could, be an url

int GVInput::is_url()
  {
  char * pos = strchr(filename, ':');
  if(!pos)
    return 0;
  pos++;
  if(*pos != '/')
    return 0;
  pos++;
  if(*pos != '/')
    return 0;
  return 1;
  }

void GVInput::set_trackname()
  {
  if(trackname)
    delete trackname;

  if(is_url())
    trackname = g_strdup(filename);
  else
    {
    const char * pos1;
    const char * pos2;
    
    if(!(pos1 = strrchr(filename, G_DIR_SEPARATOR)))
      pos1 = filename;
    else
      pos1++;
    
    if(!(pos2 = strrchr(filename, '.')))
      pos2 = &(filename[strlen(filename)]);
    
    trackname = new char[pos2 - pos1 + 1];
    strncpy(trackname, pos1, pos2-pos1);
    trackname[pos2-pos1] = '\0';
    }
  }
