/*****************************************************************

  playlist.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Non Gtk related playlist routines

#include "config.h"

#include <glib.h>
#include <pthread.h>
//#include <fstream.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>

#include <iostream>
#include <sstream>
using std::stringstream;

using std::ostream;

#include "gvtime.h"
#include "gvconfig.h"
#include "videoformat.h"
#include "audioformat.h"
#include "buffer.h"
#include "audiobuf.h"
#include "videobuf.h"

#include "playlist.h"

#include "baseplugin.h"
#include "plugin.h"
#include "input.h"

#include "output.h"
#include "audioout.h"
#include "videoout.h"

#include <gtk/gtk.h>
#include "configwindow.h"
#include "engine.h"

#include "gvloader.h"

#include "globalconfig.h"

#include "trackinfo.h"

using std::cerr;

GVPlaylist * gv_playlist = (GVPlaylist*)0;

GVPlaylistTrack::GVPlaylistTrack(const char * _filename,
                                 const char * _trackname,
                                 GVTime & _time, int _flags,
                                 const char * _plugin,
                                 gpointer _data)
  {
  if(_filename)
    filename = g_strdup(_filename);
  else
    filename = (char*)0;
  if(_trackname)
    trackname = g_strdup(_trackname);
  else
    trackname = (char*)0;
  if(_plugin)
    plugin = g_strdup(_plugin);
  else
    plugin = (char*)0;
    
  time = _time;
  flags = _flags;
  data = _data;
  }

GVPlaylistTrack::GVPlaylistTrack(const char * _filename)
  {
  if(_filename)
    filename = g_strdup(_filename);
  else
    filename = (char*)0;
  trackname = (char*)0;
  plugin = (char*)0;
    
  time.reset();
  flags = 0;
  data = (gpointer)0;
  }

GVPlaylistTrack::~GVPlaylistTrack()
  {
  if(filename)
    g_free(filename);
  if(trackname)
    g_free(trackname);
  }

void GVPlaylist::read_tmp()
  {
  read(tmp_filename);
  }

void GVPlaylist::write_tmp()
  {
  write(tmp_filename);
  }

int GVPlaylist::get_current_track_index()
  {
  if(!current_track)
    return -1;
  else  
    return g_list_index(current_tracks, (gpointer)current_track);
  }

void GVPlaylist::append_track(const char * _filename, const char * _trackname,
                              GVTime & _time, int _flags,
                              const char * _plugin,
                              gpointer _data)
  {
  GVPlaylistTrack * new_item = new GVPlaylistTrack(_filename, _trackname,
                                                 _time, _flags, _plugin,
                                                 _data);

  current_tracks = g_list_append(current_tracks, (gpointer)new_item);

  if(!num_tracks)
    {
    current_track = new_item;
    current_track->flags |= GV_PLAYLIST_ITEM_IS_CURRENT;
    }
  num_tracks++;
  set_state_flag(GV_PLAYLIST_CONTENTS_CHANGED);
  }

int GVPlaylist::get_num_tracks()
  {
  return num_tracks;
  }
  
const GVPlaylistTrack * GVPlaylist::get_track(int index)
  {
  pthread_mutex_lock(&state_mutex);
  const GVPlaylistTrack* ret =
    (const GVPlaylistTrack*)g_list_nth_data(current_tracks, index);
  pthread_mutex_unlock(&state_mutex);
  
  return ret;
  }

// Return the total time in the playlist

void GVPlaylist::get_total_time(GVTime & ret) 
  {
  const GVPlaylistTrack * item;
  ret.seconds = 0;
  ret.nanoseconds = 0;
  for(int i = 0; i < num_tracks; i++)
    {
    item = get_track(i);
    if(!(item->time.undefined()))
      ret += item->time;
    }
  }

// Return the total time BEFORE the current track

void GVPlaylist::get_total_time_before(GVTime & ret) 
  {
  const GVPlaylistTrack * item;
  ret.reset();
  for(int i = 0; i < num_tracks; i++)
    {
    item = get_track(i);
    if(item != current_track)
      {
      if(!(item->time.undefined()))
        ret += item->time;
      }
    else
      break;
    }
  }

GVPlaylist::GVPlaylist()
  {
  test_input = (GVInput*)0;

  current_tracks = (GList*)0;
  num_tracks = 0;
  current_track = (GVPlaylistTrack *)0;
  
  pthread_mutex_init (&state_mutex,(pthread_mutexattr_t *)0);

  // Is set in main.cc
  
  const char * home_dir = (char*)0;
  gv_config->get_string("_", "HomeDir", &home_dir);
    
  tmp_filename  = g_strdup_printf("%stmp_playlist", home_dir);
  save_filename = g_strdup_printf("%splaylist", home_dir);
  
  state_flags = 0;
  
  private_playlist = 0;
  read(save_filename);
  }

GVPlaylist::~GVPlaylist()
  {
  write(save_filename);
  clear();
  }

void GVPlaylist::clear()
  {
  if(gv_engine)
    gv_engine->stop();
  for(int i = 0; i < num_tracks; i++)
    {
    GVPlaylistTrack * item =
      (GVPlaylistTrack *)g_list_nth_data(current_tracks, 0);
    current_tracks = g_list_remove(current_tracks, (gpointer)item);
    delete item;
    }
  num_tracks = 0;
  current_track = (GVPlaylistTrack*)0;
  set_state_flag(GV_PLAYLIST_CONTENTS_CHANGED);
  }

void GVPlaylist::remove_selected()
  {
  int current_deleted = 0;
  for(int i = 0; i < num_tracks; i++)
    {
    GVPlaylistTrack * item =
      (GVPlaylistTrack*)g_list_nth_data(current_tracks, i);
    if(item->flags & GV_PLAYLIST_ITEM_IS_SELECTED)
      {
      if(item->flags & GV_PLAYLIST_ITEM_IS_CURRENT)
        current_deleted = 1;
      remove_track(item);
      i--;
      }
    }
  if(current_deleted)
    {
    gv_engine->stop();
    current_track = (num_tracks) ?
      (GVPlaylistTrack*)g_list_nth_data(current_tracks, 0) :
    (GVPlaylistTrack*)0;
    }
  set_state_flag(GV_PLAYLIST_CONTENTS_CHANGED);
  }

void GVPlaylist::refresh_selected()
  {
  for(int i = 0; i < num_tracks; i++)
    {
    GVPlaylistTrack * item =
      (GVPlaylistTrack*)g_list_nth_data(current_tracks, i);

    // We don't refresh the current track

    if((item->flags & GV_PLAYLIST_ITEM_IS_SELECTED) &&
       !(item->flags & GV_PLAYLIST_ITEM_IS_CURRENT))
      {
      if(!refresh_track(item))
        remove_track(item);
      }
    }
  }


void GVPlaylist::remove_current()
  {
  int current_index = get_current_track_index();

  if(current_index == -1)
    return;
    
  gpointer item = g_list_nth_data(current_tracks, current_index);
  
  current_tracks = g_list_remove(current_tracks, item);

  if(num_tracks - 1)
    {
    // Last track was current
    if(current_index == num_tracks - 1)
      set_current_track(current_index - 1);
    else
      set_current_track(current_index);
    }
  else
    current_track = (GVPlaylistTrack*)0;
  
  num_tracks--;
  set_state_flag(GV_PLAYLIST_CONTENTS_CHANGED);
  }

void GVPlaylist::select_track(int track, int select)
  {
  GVPlaylistTrack * item
    = (GVPlaylistTrack *)g_list_nth_data(current_tracks, track);
  if(select)
    item->flags |= GV_PLAYLIST_ITEM_IS_SELECTED;
  else
    item->flags &= ~GV_PLAYLIST_ITEM_IS_SELECTED;
  }

void GVPlaylist::set_current_track(int track)
  {
  pthread_mutex_lock(&state_mutex);
  if(!current_track)
    {
    pthread_mutex_unlock(&state_mutex);
    return;
    }
  current_track->flags &= ~GV_PLAYLIST_ITEM_IS_CURRENT;

  current_track = (GVPlaylistTrack*)g_list_nth_data(current_tracks, track);

  current_track->flags |= GV_PLAYLIST_ITEM_IS_CURRENT;
  state_flags |= GV_PLAYLIST_CURRENT_CHANGED;
  pthread_mutex_unlock(&state_mutex);
  }

void GVPlaylist::next()
  {
  pthread_mutex_lock(&state_mutex);
  if(!current_track)
    {
    pthread_mutex_unlock(&state_mutex);
    return;
    }
  current_track->flags &= ~GV_PLAYLIST_ITEM_IS_CURRENT;

  int index = g_list_index(current_tracks, (gpointer)current_track);
  index++;

  if(index >= num_tracks)
    index = 0;
  current_track = (GVPlaylistTrack*)g_list_nth_data(current_tracks, index);

  current_track->flags |= GV_PLAYLIST_ITEM_IS_CURRENT;
  state_flags |= GV_PLAYLIST_CURRENT_CHANGED;
  pthread_mutex_unlock(&state_mutex);
  }

void GVPlaylist::previous()
  {
  pthread_mutex_lock(&state_mutex);
  if(!current_track)
    {
    pthread_mutex_unlock(&state_mutex);
    return;
    }
  current_track->flags &= ~GV_PLAYLIST_ITEM_IS_CURRENT;

  int index = g_list_index(current_tracks, (gpointer)current_track);
  index--;

  if(index < 0)
    index = num_tracks - 1;
  current_track = (GVPlaylistTrack*)g_list_nth_data(current_tracks, index);

  current_track->flags |= GV_PLAYLIST_ITEM_IS_CURRENT;
  state_flags |= GV_PLAYLIST_CURRENT_CHANGED;
  pthread_mutex_unlock(&state_mutex);
  }

void GVPlaylist::move_track(int src, int dst)
  {
  pthread_mutex_lock(&state_mutex);
  gpointer item = g_list_nth_data(current_tracks, src);
  current_tracks = g_list_remove(current_tracks, item);
  current_tracks = g_list_insert(current_tracks, item, dst);

  state_flags |= GV_PLAYLIST_CONTENTS_CHANGED;
  
  pthread_mutex_unlock(&state_mutex);
  }

static int __compare_func(gconstpointer a, gconstpointer b)
  {
  return strcmp(((const GVPlaylistTrack*)a)->trackname,
                ((const GVPlaylistTrack*)b)->trackname);
                
  }

void GVPlaylist::sort()
  {
  current_tracks = g_list_sort(current_tracks, __compare_func);
  }


void GVPlaylist::make_private(int _private_playlist)
  {
  if(!!_private_playlist == !!private_playlist)
    return;

  if(!(private_playlist) && (_private_playlist))
    {
    write_tmp();
    private_playlist = 1;
    return;
    }
  else if((private_playlist) && !(_private_playlist))
    {
    read_tmp();
    private_playlist = 0;
    return;
    }
  }

int GVPlaylist::is_private()
  {
  return private_playlist;
  }

void GVPlaylist::set_state_flag(int flag)
  {
  pthread_mutex_lock(&state_mutex);
  state_flags |= flag;
  pthread_mutex_unlock(&state_mutex);
  }

int GVPlaylist::get_state_flags()
  {
  int ret;
  pthread_mutex_lock(&state_mutex);
  ret = state_flags;
  pthread_mutex_unlock(&state_mutex);
  return ret;
  }

void GVPlaylist::clear_state_flag(int flag)
  {
  pthread_mutex_lock(&state_mutex);
  state_flags &= ~flag;
  pthread_mutex_unlock(&state_mutex);
  }

GVPlaylistAlbum * GVPlaylist::find_album(GVPlaylistAlbum * parent,
                                         const GVPlaylistAlbum * album)
  {
  GVPlaylistAlbum * ret;
  GVPlaylistAlbum * tmp_album;
  
  for(int i = 0; i < parent->num_albums; i++)
    {
    if((ret = (GVPlaylistAlbum*)g_list_nth_data(parent->albums, i)) == album)
      return ret;
    }
  for(int i = 0; i < parent->num_albums; i++)
    {
    tmp_album = (GVPlaylistAlbum*)g_list_nth_data(parent->albums, i);
    if((ret = find_album(tmp_album, album)))
      return ret;
    }
  return (GVPlaylistAlbum*)0;
  }

GVPlaylistAlbum * GVPlaylist::find_album(const GVPlaylistAlbum * album)
  {
  if(album == root_album)
    return root_album;
  return find_album(root_album, album);
  }

void GVPlaylist::select_current_album(const GVPlaylistAlbum * album)
  {
  GVPlaylistAlbum * new_album = find_album(album);
  if(!new_album)
    return;
  num_tracks = g_list_length(new_album->tracks);
  current_album = new_album;
  }

void GVPlaylist::select_display_album(const GVPlaylistAlbum * album)
  {
  
  }

void GVPlaylist::set_track_name(int index, const char * new_name)
  {
  GVPlaylistTrack * track = (GVPlaylistTrack*)g_list_nth_data(current_tracks,
                                                              index);
  if(track->trackname)
    g_free(track->trackname);
  track->trackname = g_strdup(new_name);
  set_state_flag(GV_PLAYLIST_CONTENTS_CHANGED);
  }

void GVPlaylist::remove_track(GVPlaylistTrack * track)
  {
  current_tracks = g_list_remove(current_tracks, (gpointer)track);
  delete track;
  num_tracks--;
  }

// Refresh a track, return 0 on error

int GVPlaylist::refresh_track(GVPlaylistTrack * track)
  {
  if(is_private())
    return 1;
  GVPluginInfo * info;
  
  if(!(track->flags & GV_PLAYLIST_ITEM_KEEP_PLUGIN) || !(track->plugin))
    info = gv_input_file_loader->find_plugin_by_suffix(track->filename);
  else
    info = gv_input_file_loader->find_plugin_by_name(track->plugin);
    
  std::stringstream line;
  
  // Return silently if no plugin was found
  
  if(!info)
    {
    cerr << "No Plugin found for " << track->filename << "\n";
    //    remove_track(track);
    return 0;
    }

  if((test_input) && strcmp(test_input->attributes->name, info->name))
    {
    gv_input_file_loader->destroy_plugin(test_input);
    test_input = gv_input_file_loader->load_input(info);
    }
  else if(!test_input)
    test_input = gv_input_file_loader->load_input(info);
  
  test_input->set_filename(track->filename);
  test_input->init();
  
  if(test_input->get_error())
    {
    cerr << "Couldn't open " << track->filename << " with "
         << test_input->attributes->name << "\n";
    test_input->cleanup();
    return 0;
    }

  // Set the trackname

  const GVTrackInfo * track_info = test_input->get_track_info();

  if(track->trackname)
    g_free(track->trackname);
  
  if((gv_global_config->use_track_info) && (track_info))
    {
    line.seekp(0);
    //    line.freeze(0);
    track_info->print_string(line);
    track->trackname = g_strdup(line.str().c_str());
    }
  else
    {
    test_input->set_trackname();
    track->trackname = g_strdup(test_input->get_trackname());
    }

  // Set the new filename in the case, the input changed it
  // (That's true for URL redirectors (.mp3 files), in which case
  // the original URL is stored). We also set the "keep plugin" flag then

  if(strcmp(test_input->get_filename(), track->filename))
    {
    g_free(track->filename);
    track->filename = g_strdup(test_input->get_filename());
    track->flags |= GV_PLAYLIST_ITEM_KEEP_PLUGIN;
    }
  
  // Set the flags (keep flags which aren't set here)
  
  track->flags &= ~(GV_PLAYLIST_ITEM_HAS_VIDEO|GV_PLAYLIST_ITEM_HAS_AUDIO);

  if(test_input->track_has_video())
    track->flags |= GV_PLAYLIST_ITEM_HAS_VIDEO;
  if(test_input->track_has_audio())
    track->flags |= GV_PLAYLIST_ITEM_HAS_AUDIO;
 
  test_input->get_total_time(track->time);

  // Change the new plugin name if it changed

  if(!(track->flags & GV_PLAYLIST_ITEM_KEEP_PLUGIN) || !(track->plugin))
    {
    if(track->plugin)
      g_free(track->plugin);
    track->plugin = g_strdup(test_input->attributes->name);
    }
  
  //  gv_playlist->append_track(filename, test_input->get_trackname(), time,
  //                            track_flags, test_input->attributes->name);

  test_input->cleanup();
  set_state_flag(GV_PLAYLIST_CONTENTS_CHANGED);
  return 1;
  }

void GVPlaylist::refresh_track(int track_number)
  {
  //  cerr << "Refreshing track " << track_number << "...\n";
  GVPlaylistTrack * track
    = (GVPlaylistTrack*)(g_list_nth_data(current_tracks, track_number));
  refresh_track(track);
  //  cerr << "done with track" << track_number << ".\n";
  }

void GVPlaylist::add_file(const char * filename)
  {
//   cerr << "GVPlaylist::add_file: filename = "
//        << filename << "\n";
  GVPlaylistTrack * new_item = new GVPlaylistTrack(filename);

  if(!refresh_track(new_item))
    {
    delete new_item;
    return;
    }
  else
    {
    current_tracks = g_list_append(current_tracks, (gpointer)new_item);
    if(!num_tracks)
      {
      current_track = new_item;
      current_track->flags |= GV_PLAYLIST_ITEM_IS_CURRENT;
      }
    num_tracks++;
    }
  }

void GVPlaylist::add_file_with_plugin(const char * filename,
                                      const char * plugin_name)
  {
  //  GV_PLAYLIST_ITEM_KEEP_PLUGIN

  GVPlaylistTrack * new_item = new GVPlaylistTrack(filename);

  new_item->flags |= GV_PLAYLIST_ITEM_KEEP_PLUGIN;
  new_item->plugin = g_strdup(plugin_name);
    
  if(!refresh_track(new_item))
    {
    delete new_item;
    return;
    }
  else
    {
    current_tracks = g_list_append(current_tracks, (gpointer)new_item);
    if(!num_tracks)
      {
      current_track = new_item;
      current_track->flags |= GV_PLAYLIST_ITEM_IS_CURRENT;
      }
    num_tracks++;
    }
  
  }
