/*****************************************************************

  playlist.h

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Playlist (non gtk part!!)

#define GV_PLAYLIST_ITEM_HAS_VIDEO   (1<<0)
#define GV_PLAYLIST_ITEM_HAS_AUDIO   (1<<1)
#define GV_PLAYLIST_ITEM_IS_SELECTED (1<<2)
#define GV_PLAYLIST_ITEM_IS_CURRENT  (1<<3)
#define GV_PLAYLIST_ITEM_NOSAVE      (1<<4)
#define GV_PLAYLIST_ITEM_KEEP_PLUGIN (1<<5)
#define GV_PLAYLIST_ITEM_IS_URL      (1<<6) // URL

class GVPluginInfo;

struct GVPlaylistTrack
  {
  char *filename;              // Name of the file
  char *trackname;             // Name of the track
  char * plugin;               // Name of the plugin
  GVTime time;                 // Total time of the track
  int flags;                   // Flags (audio/video/current/selected)

  gpointer data;      // Private data for some plugins
  
  GVPlaylistTrack(const char * _filename, const char * _trackname,
                  GVTime & _time, int _flags, const char * _plugin,
                  gpointer _data = (gpointer)0);

  GVPlaylistTrack(const char * _filename);

  ~GVPlaylistTrack();
  };

#define GV_ALBUM_IS_PRIVATE  (1<<0)
#define GV_ALBUM_IS_CURRENT  (1<<1)
#define GV_ALBUM_IS_SELECTED (1<<2)

// Extended track info

struct GVExtendedTrackInfo
  {
  char * artist;
  char * title;
  char * genre;
  char * comment;
  int year;
  };

struct GVPlaylistAlbum
  {
  int flags; 
  gchar * name;

  GVPluginInfo * plugin_info;
  
  // Todo: Eventually add year and genre here (For ID3 configuration)
  
  GVPlaylistAlbum * parent; // NULL if root album
  GList * tracks;           // GVPlaylistTrack
  GList * albums;           // GVPlaylistAlbum

  int num_tracks;
  int num_albums;
  
  void clear();
  GVPlaylistAlbum(const char * _name, GVPluginInfo * info);
  };

#define GV_PLAYLIST_CURRENT_CHANGED  (1<<0) // Current Track Changed
#define GV_PLAYLIST_CONTENTS_CHANGED (1<<1) // Contents Changed

class GVPlaylist
  {
  GList * albums;
  GList * current_tracks;

  GVPlaylistAlbum * current_album; // Currently played album
  GVPlaylistAlbum * display_album; // Currently displayed album
  GVPlaylistAlbum * root_album;
  
  int num_tracks;
  
  GVPlaylistTrack * current_track;

  pthread_mutex_t state_mutex;

  char * tmp_filename;
  char * save_filename;

  int private_playlist; // Playlist is private, items are no files

  int state_flags;

  void set_state_flag(int flag);

  GVPlaylistAlbum * find_album(const GVPlaylistAlbum * album);
  GVPlaylistAlbum * find_album(GVPlaylistAlbum * parent,
                               const GVPlaylistAlbum * album);

  GVInput * test_input;     // Input plugin for adding files to
                            // the playlist
  
public:

  const GVPlaylistAlbum * get_root_album() { return root_album; }
  
  void select_current_album(const GVPlaylistAlbum * album);
  void select_display_album(const GVPlaylistAlbum * album);
  
  int get_state_flags();

  void clear_state_flag(int flag);
  
  void make_private(int _private_playlist);
  int is_private();
  
  void sort();
  
  const GVPlaylistTrack * get_current_track() { return current_track; }

  void set_current_track(int track);

  void move_track(int src, int dst);
  
  
  void next();
  void previous();
  
  void select_track(int index, int select);
    
  void append_track(const char * _filename, const char * _trackname,
                    GVTime & time, int _flags,
                    const char * _plugin,
                    gpointer data = (gpointer)0);

  // Append a file

  void add_file(const char * filename);
  void add_file_with_plugin(const char * filename, const char * plugin_name);

  int refresh_track(GVPlaylistTrack * track);
  void remove_track(GVPlaylistTrack * track);
  
  int get_num_tracks();
  int get_current_track_index();
  
  const GVPlaylistTrack * get_track(int index);

  void set_track_name(int index, const char * new_name);
  
  void get_total_time(GVTime & ret); // Return the total time in the playlist

  // Return the total time in the playlist before the current track
  
  void get_total_time_before(GVTime & ret);
  
  
  GVPlaylist();
  ~GVPlaylist();
  void clear();
  void remove_selected();  
  void refresh_selected();  

  void refresh_track(int track_number);
  
  void remove_current();
  
  
  void read(const char * filename);
  void write(const char * filename);

  void read_tmp();
  void write_tmp();

  };

extern GVPlaylist * gv_playlist;
