#ifndef AVIFILE_IAVIPLAYER_H
#define AVIFILE_IAVIPLAYER_H

/********************************************************

    AVI player, ver. 0.6.1
    Copyright 2001 Eugene Kuznetsov  (divx@euro.ru)

*********************************************************/

#include "infotypes.h"
#include "except.h"
#include "subtitle.h"


typedef void (*KILLHANDLER)(int, void* p = 0);
typedef int (*AUDIOFUNC)(void*, int, void* p = 0);

class CImage;
class PlayerWidget;
class IRtConfig;
class VideoRenderer;
class StreamInfo;
/**
 *
 *   This class is meant to contain all code that is
 * not related to graphical interface of AVI/ASF player: you provide
 * an interface ( buttons, windows, etc. ), this class
 * provides you with video data and performs audio playback.
 *
 *   NEVER call playback-controlling functions ( even indirectly
 * through event processing calls, such as qApp->processEvents
 * in Qt ) from its callbacks.
 *
 *    Usage:
 *
 *   Call CreateAviPlayer(), specifying file name of AVI/ASF file, bit depth
 * of physical screen and optional file name of subtitle file.
 * Catch FatalError if player can't be constructed. Set two callbacks:
 * kill handler ( will be called when player stops ) and draw callback
 * ( called each frame during playback and frequently if paused. Takes
 * non-zero pointer to video data or NULL if frame sholdn't be drawn ).
 *  Passed URL may refer to remote ( MMS://... ) ASF file. Such files are
 * opened asynchronously. It means that constructor only performs sanity
 * checking on the file name, starts actual opening process ( in another
 * thread ) and returns. You should periodically call isOpened() until
 * it returns true, which means that opening process has finished. After that,
 * you can determine if it was successful by calling isValid(). For local
 * files isOpened() and isValid() will return 'true' immediately after
 * successful construction.
 *   URL can point to ASX ( redirector ) file. For such file isRedirector()
 * will return 'true'. If this is the case, read list of redirected URLs
 * with getURLs(). After that, IAviPlayer object may be destroyed because
 * you can't do anything else with it.
 *  Prepare to draw video and call start(). Player immediately starts
 * playing. Its state can be controlled by calls play()/pause()/reseek(),
 * player can be stopped by stop() and started again by start().
 * Once again: don't try to call these functions from draw callback or
 * your program will burn in hell forever.
 *  When playing ASF files over the network and AVI files over slow
 * connections, player may occasionally get into 'buffering' state.
 * It means that player stops to output data and waits until its internal
 * buffers are refilled. You can determine if it's in buffering state
 * by calling GetState(). It'll put buffer fill percentage into variable
 * pointed by its argument if buffering takes place ( here 1 corresponds
 * to full buffers and 0 to empty buffers ).
 *  Player can be destroyed with 'delete'. Deleting automatically stops
 * playback and frees all resources.
 *
 */
class IAviPlayer
{
public:
    enum Property
    {
        PROPERTY_END,		// 0 - marks last property
	AUDIO_RENDERER,		// char*
	VIDEO_RENDERER,		// char*
	SUBTITLE_RENDERER,	// char*
	VIDEO_CODECS,		// char*
	AUDIO_CODECS,		// char*
	USE_YUV,                // bool
        AUTOREPEAT,             // bool
	DEFAULT_AUDIO_STREAM,   // int (0..65536)
	SUBTITLE_FONT,		// char*
        SUBTITLE_ASYNC_TIME_MS,	// int
	USE_HTTP_PROXY,         // bool
	HTTP_PROXY,             // char*
	VIDEO_PRESERVE_ASPECT,	// bool
	VIDEO_BUFFERED,         // bool
	VIDEO_DIRECT,           // bool
        VIDEO_DROPPING,         // bool
	VIDEO_QUALITY_AUTO,	// bool
	DISPLAY_FRAME_POS,      // bool
	AUDIO_RESAMPLING,	// bool
	AUDIO_RESAMPLING_RATE,	// int
	AUDIO_PLAYING_RATE,	// int
	AUDIO_MASTER_TIMER,	// bool
	AUDIO_VOLUME,           // int (0..10000 - 1000 - is standard level)
	ASYNC_TIME_MS,		// int
        AUDIO_STREAM,		// char*

	// fill new entires here
        // add appropriate line into avifileprop.cpp

	LAST_PROPERTY,

	// these could be only queried
	QUERY_AVG_DROP = 10000,	// int		average drop in %
	QUERY_AVG_QUALITY,      // int 		average image quality in %
        QUERY_AUDIO_RENDERERS,  //
        QUERY_VIDEO_RENDERERS,  //
	QUERY_SUBTITLE_RENDERERS,//
	QUERY_VIDEO_WIDTH,	// int
	QUERY_VIDEO_HEIGHT,	// int
        QUERY_EOF,              // bool
    };

    enum State
    {
	Opening,
	Buffering,
	Playing,
	Stopped,
	Paused,
	Invalid
    };

    virtual ~IAviPlayer();
    virtual HRESULT SetColorSpace(fourcc_t csp, bool test_only) =0;

    virtual void SetKillHandler(KILLHANDLER handler, void* argument=0) =0;
    virtual void SetAudioFunc(AUDIOFUNC func, void* argument=0) =0;

    virtual double Reseek(double timepos) =0;
    virtual HRESULT ReseekExact(double timepos) =0;  // nonzero if unable
    virtual HRESULT PageUp() =0;		// to next keyframe
    virtual HRESULT PageDown() =0;		// to prev keyframe ( the one before frame drawn last time )
    virtual HRESULT NextFrame() =0;		// to next frame ( mostly useful in pause mode )
    virtual void Pause(bool state) =0;
    virtual void Play() =0;                     // start playing
    virtual void Start() =0;
    virtual void Stop() =0;
    virtual bool IsOpened() const =0;
    virtual bool IsRedirector() const =0;
    virtual bool IsValid() const =0;
    virtual bool IsPaused() const =0;		// true -> paused
    virtual bool IsPlaying() const =0;		// true -> playing || paused
    virtual bool IsStopped() const =0;		// true -> !playing && !paused

    virtual State GetState(double* percent) =0; // player state
    virtual bool GetURLs(avm::vector<avm::string>& urls) const =0;

    virtual int GetWidth() const =0;		// video width
    virtual int GetHeight() const =0;		// video height
    virtual const char* GetFileName() const =0; // name of played stream
    virtual double GetFps() const =0;           // frames per second
    virtual framepos_t GetFramePos() const =0;  // playing frame position
    //  GetTime() would be better name here...
    virtual double GetLengthTime() const = 0;	// length of longer stream
    virtual double GetPos() const =0;           // playing position in sec

    virtual bool HasSubtitles() const =0;
    virtual HRESULT InitSubtitles(const char* filename) =0;
    virtual const subtitle_line_t* GetCurrentSubtitles() =0;

    virtual double GetAudioLengthTime() const =0;
    virtual const char* GetAudioFormat() const =0;
    virtual StreamInfo* GetAudioStreamInfo() const =0;

    virtual double GetVideoLengthTime() const =0;
    virtual const char* GetVideoFormat() const =0;
    virtual StreamInfo* GetVideoStreamInfo() const =0;


    // reading and setting list of video renderers - should work in runtime later
    virtual const avm::vector<VideoRenderer*>& GetVideoRenderers() const =0;
    virtual HRESULT SetVideoRenderers(avm::vector<VideoRenderer*>) =0;
    //to be added!!!  virtual const std::vector<AudioRenderer*>& GetAudioRenderers() const =0;
    //to be added!!!  virtual int SetAudioRenderers(vector<AudioRenderer*>) =0;

    // Obsolete: Audio/video sync correction,  argument in seconds
    virtual void SetAsync(float asynctime) =0;
    virtual float GetAsync() const =0;

    virtual const CodecInfo& GetCodecInfo() const =0;

    virtual IRtConfig* GetRuntimeConfig() const =0;
    virtual void Restart() =0;

    virtual void SetVideo(bool enabled) =0;
    virtual bool GetVideo() const =0;

    virtual void SetAudio(bool enabled) =0;
    virtual bool GetAudio() const =0;

    // General interface to control various things on runtime
    // - list of properties with arguments is finished with 0
    // WARNING: passed type has to match its type!!!
    virtual HRESULT Set(...) = 0;
    // bool is passed as 4 bytes - so it's not problem here
    virtual HRESULT Get(...) const = 0;
};
/*
 This object has interface very similar to IAviPlayer, but
 it does drawing by itself. You do not need to provide
 pointer to kill handler, but you can still do it. Attempts
 to install draw callback will be ignored.
 Create an object with this interface by calling CreateAviPlayer2().
 Its arguments:
 parent - pointer to interface which handles a few specific
 key & mouse events in movie window.
 Optional. Pass it if you want to do this handling by yourself.
 dpy - pointer to opened X display.
 Optional. If it's non-NULL,
 movie window will be created on this display. If it's NULL,
 movie window won't be created at all.
 filename & subname - same as when constructing IAviPlayer.
 */

class IAviPlayer2: public IAviPlayer
{
public:
    virtual ~IAviPlayer2() {}
    virtual int Refresh()					=0;
    virtual int Resize(int& new_w, int& new_h)			=0;
    virtual int ToggleFullscreen(bool maximize = false)		=0;
};

IAviPlayer* CreateAviPlayer(const char* filename, int bitdepth,
			    const char* subfile = 0) throw(FatalError);

//void* instead of Display*
IAviPlayer2* CreateAviPlayer2(PlayerWidget*, void*, const char* filename,
			      const char* subname) throw(FatalError);

extern "C" double GetAvifileVersion();

#endif //AVIFILE_IAVIPLAYER_H
