#ifndef AVIFILE_UTILS_H
#define AVIFILE_UTILS_H

#include "default.h"
#include "formats.h"
/**
 *  Utility functions.
 *    plain C functions are here
 */


/**
 *  These two functions return number of ticks ( processor cycles )
 *   since power-on. Return value of first one wraps around each
 *   few seconds ( depends on processor ), but calculations with
 *   32-bit 'int' are a bit faster than with 64-bit 'int64_t'.
 *  They can be used for very precise and fast measurement of time
 *   intervals - call to localcount() takes less than 0.2us on 500
 *   MHz processor, and its resolution is measured in nanoseconds.
 *   On systems with properly configured Linux kernel they can be 
 *   used as 3-4 times faster alternative to gettimeofday() 
 *   because they do not require a trip to kernel and back.
 *   On other systems they will probably be even more efficient.
 *  For processors without time-stamp counter feature ( a few
 *   old Pentium-compatible models by Cyrix and maybe others )
 *   these functions are emulated using gettimeofday().
 */

extern uint_t (*localcount)(void);
extern int64_t (*longcount)(void);

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* memory lock for internal usage only - will be removed in future! */
void avm_memory_lock_init(void);
void avm_memory_lock(void);
void avm_memory_unlock(void);

const char* avm_fcc_name(fourcc_t fcc);
const char* avm_wave_format_name(short fmt); /* not sure - might be special type */
char* avm_wave_format(char* buf, uint_t s, const WAVEFORMATEX* wf);

/* wrapping functions */
/* A thread-save usec sleep */
int avm_usleep(unsigned long);

/* some system doesn't have setenv/unsetenv */
int avm_setenv(const char *name, const char *value, int overwrite);
void avm_unsetenv(const char *name);

static inline uint16_t avm_read_be16(const void* p) {
	const uint8_t* c = (const uint8_t*) p;
	return c[0] | c[1] << 8;
}

static inline uint32_t avm_read_be32(const void* p) {
        const uint8_t* c = (const uint8_t*) p;
	return c[0] | c[1] << 8 | c[2] << 16 | c[3] << 24;
}

static inline uint16_t avm_read_le16(const void* p) {
	const uint8_t* c = (const uint8_t*) p;
	return c[0] << 8 | c[1];
}

static inline uint32_t avm_read_le32(const void* p) {
        const uint8_t* c = (const uint8_t*) p;
	return c[0] << 24 | c[1] << 16 | c[2] << 8 | c[3];
}

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* AVIFILE_UTILS_H */
