#include "fillplugins.h"
#include "audiodecoder.h"
#include "plugin.h"
#include "except.h"
#include "utils.h"
#include "fourcc.h"

extern "C" {
#include "mad.h"
}

#include <stdio.h>
#include <ctype.h>

// code here inspired & cut&pasted from gmerlin sources

static int mpeg_bitrates[16][6] =
{
  {   0,         0,      0,      0,      0,      0 },
  {  32000,  32000,  32000,  32000,  32000,   8000 },
  {  64000,  48000,  40000,  64000,  48000,  16000 },
  {  96000,  56000,  48000,  96000,  56000,  24000 },
  { 128000,  64000,  56000, 128000,  64000,  32000 },
  { 160000,  80000,  64000, 160000,  80000,  64000 },
  { 192000,  96000,  80000, 192000,  96000,  80000 },
  { 224000, 112000,  96000, 224000, 112000,  56000 },
  { 256000, 128000, 112000, 256000, 128000,  64000 },
  { 288000, 160000, 128000, 288000, 160000, 128000 },
  { 320000, 192000, 160000, 320000, 192000, 160000 },
  { 352000, 224000, 192000, 352000, 224000, 112000 },
  { 384000, 256000, 224000, 384000, 256000, 128000 },
  { 416000, 320000, 256000, 416000, 320000, 256000 },
  { 448000, 384000, 320000, 448000, 384000, 320000 },
  {      0,      0,      0,      0,      0,      0 }
};

static int mpeg_samplerates[3][3] =
{
  { 44100, 48000, 32000 }, // MPEG1
  { 22050, 24000, 16000 }, // MPEG2
  { 11025, 12000, 8000 }   // MPEG2.5
};

#define MPEG_ID_MASK        0x00180000
#define MPEG_MPEG1          0x00180000
#define MPEG_MPEG2          0x00100000
#define MPEG_MPEG2_5        0x00000000

#define MPEG_LAYER_MASK     0x00060000
#define MPEG_LAYER_III      0x00020000
#define MPEG_LAYER_II       0x00040000
#define MPEG_LAYER_I        0x00060000
#define MPEG_PROTECTION     0x00010000
#define MPEG_BITRATE_MASK   0x0000F000
#define MPEG_FREQUENCY_MASK 0x00000C00
#define MPEG_PAD_MASK       0x00000200
#define MPEG_PRIVATE_MASK   0x00000100
#define MPEG_MODE_MASK      0x000000C0
#define MPEG_MODE_EXT_MASK  0x00000030
#define MPEG_COPYRIGHT_MASK 0x00000008
#define MPEG_HOME_MASK      0x00000004
#define MPEG_EMPHASIS_MASK  0x00000003

#define LAYER_I_SAMPLES       384
#define LAYER_II_III_SAMPLES 1152

struct AudioInfoMp3
{
    enum { MPEG1, MPEG2, MPEG2_5 } mode;
    enum { STEREO, JSTEREO, DUAL_CHANNEL, MODE_MONO } stereo_mode;
    int vbr;
    int layer;
    int bitrate;
    int start_offset;
    int samplerate;
    int samples_per_frame;
    int num_channels;
    int frame_size;

    // need 40 bytes
    int Init(const unsigned char* buf)
    {
	uint32_t header = avm_read_le32(buf);
        start_offset = 40;

	// CHECK if it's mpeg & Get the sampling frequency

	int frequency_index = (header & MPEG_FREQUENCY_MASK) >> 10;

	switch (header & MPEG_ID_MASK)
	{
	case MPEG_MPEG1:
	    mode = MPEG1;
	    samplerate = mpeg_samplerates[0][frequency_index];
	    break;
	case MPEG_MPEG2:
	    mode = MPEG2;
	    samplerate = mpeg_samplerates[1][frequency_index];
	    break;
	case MPEG_MPEG2_5:
	    mode = MPEG2_5;
	    samplerate = mpeg_samplerates[2][frequency_index];
	    break;
	}

        // Get stereo mode index
	switch ((header & MPEG_MODE_MASK) >> 6)
	{
	case 0:
	    stereo_mode = STEREO;
	    break;
	case 1:
	    stereo_mode = JSTEREO;
	    break;
	case 2:
	    stereo_mode = DUAL_CHANNEL;
	    break;
	case 3:
	    stereo_mode = MODE_MONO;
	    break;
	}
	num_channels = (stereo_mode == MODE_MONO) ? 1 : 2;

	// Get layer
	switch (header & MPEG_LAYER_MASK)
	{
	case MPEG_LAYER_I:
	    layer = 1;
	    samples_per_frame = LAYER_I_SAMPLES;
	    break;
	case MPEG_LAYER_II:
	    layer = 2;
	    samples_per_frame = LAYER_II_III_SAMPLES;
	    break;
	case MPEG_LAYER_III:
	    layer = 3;
	    samples_per_frame = LAYER_II_III_SAMPLES;
	    break;
	}

	if (mode == MPEG2)
	    samples_per_frame /= 2;
	// Is this true??
	else if (mode == MPEG2_5)
	    samples_per_frame /= 4;


        // Check for VBR stream (Xing header)
        int b = 4;
	if (header & MPEG_ID_MASK)
	    b += (stereo_mode != MODE_MONO) ? 32 : 17; // mpeg1
        else
	    b += (stereo_mode != MODE_MONO) ? 17 : 9; // mpeg2
	uint32_t xing = avm_read_be32(buf + b);
	vbr = (mmioFOURCC('X', 'i', 'n', 'g') == xing);

	bitrate = (vbr) ? 0 : GetBitrate(header);
        frame_size = GetFrameSize(header);
	if (vbr)
	    start_offset += frame_size + 4;

        if (frame_size)
	    printf("Mad MPEG header  VBR:%d  Layer:%d  Bitrate:%d  SampleRate:%d\n"
		   "SamplesPerFrame:%d  NumChannels:%d  StereoMode:%d  Offset:%d\n",
		   vbr, layer, bitrate, samplerate, samples_per_frame,
		   num_channels, stereo_mode, start_offset);

        return start_offset;
    }

    int GetBitrate(uint32_t header)
    {
	int32_t bitrate_index = (header & MPEG_BITRATE_MASK) >> 12;
	return mpeg_bitrates[bitrate_index][layer + ((mode == MPEG1) ? -1 : 2)];
    }

    int GetFrameSize(uint32_t header)
    {
	if ((header & 0xffe00000) != 0xffe00000)
	    return 0;

	int tmp_bitrate = GetBitrate(header);

	if (!tmp_bitrate)
	    return 0;

	int ret;

	if (layer == 1)
	{
	    ret = (12 * tmp_bitrate / samplerate);
	    if (header & MPEG_PAD_MASK)
		ret++;
	    ret *= 4;
	}
	else
	{
	    ret = 144 * tmp_bitrate / samplerate;
	    if (header & MPEG_PAD_MASK)
		ret++;
	}
	if (mode == MPEG2)
	    ret /= 2;
	// For mpeg 2.5, this is not tested
	else if (mode == MPEG2_5)
	    ret /= 4;

	return ret;
    }
};


#define _
static char const *error_str(enum mad_error error)
{
  static char str[17];

  switch (error) {
  case MAD_ERROR_BUFLEN:
  case MAD_ERROR_BUFPTR:
    /* these errors are handled specially and/or should not occur */
    break;

  case MAD_ERROR_NOMEM:		 return _("not enough memory");
  case MAD_ERROR_LOSTSYNC:	 return _("lost synchronization");
  case MAD_ERROR_BADLAYER:	 return _("reserved header layer value");
  case MAD_ERROR_BADBITRATE:	 return _("forbidden bitrate value");
  case MAD_ERROR_BADSAMPLERATE:	 return _("reserved sample frequency value");
  case MAD_ERROR_BADEMPHASIS:	 return _("reserved emphasis value");
  case MAD_ERROR_BADCRC:	 return _("CRC check failed");
  case MAD_ERROR_BADBITALLOC:	 return _("forbidden bit allocation value");
  case MAD_ERROR_BADSCALEFACTOR: return _("bad scalefactor index");
  case MAD_ERROR_BADFRAMELEN:	 return _("bad frame length");
  case MAD_ERROR_BADBIGVALUES:	 return _("bad big_values count");
  case MAD_ERROR_BADBLOCKTYPE:	 return _("reserved block_type");
  case MAD_ERROR_BADSCFSI:	 return _("bad scalefactor selection info");
  case MAD_ERROR_BADDATAPTR:	 return _("bad main_data_begin pointer");
  case MAD_ERROR_BADPART3LEN:	 return _("bad audio data length");
  case MAD_ERROR_BADHUFFTABLE:	 return _("bad Huffman table select");
  case MAD_ERROR_BADHUFFDATA:	 return _("Huffman data overrun");
  case MAD_ERROR_BADSTEREO:	 return _("incompatible block_type for JS");
  }

  sprintf(str, "error 0x%04x", error);
  return str;
}


class MAD_Decoder : public IAudioDecoder
{
    AudioInfoMp3 ai;
    struct mad_stream stream;
    struct mad_frame frame;
    struct mad_synth synth;
    bool m_bInitialized;

public:
    MAD_Decoder(const CodecInfo&, const WAVEFORMATEX*);
    ~MAD_Decoder();
    int Convert(const void*, uint_t, void*, uint_t, uint_t*, uint_t*);
    //int GetOutputFormat(WAVEFORMATEX* destfmt);
};

MAD_Decoder::MAD_Decoder(const CodecInfo& info, const WAVEFORMATEX* wf)
    :IAudioDecoder(info, wf), m_bInitialized(false)
{
    mad_stream_init(&stream);
    mad_frame_init(&frame);
    mad_synth_init(&synth);
}

MAD_Decoder::~MAD_Decoder()
{
    mad_synth_finish(&synth);
    mad_frame_finish(&frame);
    mad_stream_finish(&stream);
}

#if 0
int MAD_Decoder::GetOutputFormat(WAVEFORMATEX* destfmt)
{
    if (!destfmt)
	return -1;
    *destfmt = in_fmt;

    destfmt->wBitsPerSample = 16;
    destfmt->wFormatTag = 0x2000;
    destfmt->nAvgBytesPerSec = 192000;  // after conversion
    destfmt->nBlockAlign = MAD_BLOCK_SIZE;
    destfmt->nSamplesPerSec = destfmt->nAvgBytesPerSec / destfmt->nChannels
	/ (destfmt->wBitsPerSample / 8);

    /*
     destfmt->nBlockAlign = destfmt->nChannels * destfmt->wBitsPerSample / 8;
     destfmt->nAvgBytesPerSec = destfmt->nSamplesPerSec * destfmt->nBlockAlign;
     destfmt->cbSize = 0;
     */
    char b[200];
    avm_wave_format(b, sizeof(b), &in_fmt);
    printf("src %s\n", b);
    avm_wave_format(b, sizeof(b), destfmt);
    printf("dst %s\n", b);

    return 0;
}
#endif

int MAD_Decoder::Convert(const void* in_data, uint_t in_size,
			 void* out_data, uint_t out_size,
			 uint_t* size_read, uint_t* size_written)
{
    int frame_size;
    uint_t in_size_init = in_size;

    if (!m_bInitialized)
    {
	ai.Init((const unsigned char*)in_data);
	frame_size = ai.frame_size;
	if (frame_size != 0)
	    m_bInitialized = true;
    }
    else
    {
	uint32_t header = avm_read_le32((const unsigned char*) in_data);
	frame_size = ai.GetFrameSize(header);
    }

    if (frame_size == 0)
    {
	if (size_read)
	    *size_read = 1;
	if (size_written)
	    *size_written = 0;
        return 0;
    }

#if 0
    // checking if the sample continues correctly
    const unsigned int *p = (const unsigned int*) (in_data);

    for (unsigned i = 0; i < frame_size/(sizeof(int)) + 10; i++)
    {
	if (!(i % 4))
	    printf("\n%4d ", i * 4);
	//printf("  0x%4.4x", abs(p[i] - p[i-2]) & 0xffff);
	//printf("  0x%02x-%c", p[i], (isprint(p[i]) ? p[i] : '-'));
	printf("  0x%08x", avm_read_le32((const unsigned char*)(p + i)));
    }
    printf("\n");
    avm_usleep(100000);
#endif

    mad_stream_buffer(&stream, (const unsigned char*) in_data, 2 * frame_size);
    //mad_stream_sync(&stream);
    if (mad_frame_decode(&frame, &stream) == -1)
    {
	//printf("mad_frmae_decode: stream error: %s (harmless after seeking)\n",
	//       error_str(stream.error));
	mad_frame_mute(&frame);
    }
    else
    {
	mad_synth_frame(&synth, &frame);

	// Convert mad_fixed to int16_t
	int16_t* samples = (int16_t*) out_data;

#if 0
	printf("pos:%d  fsize:%d ch:%d spf:%d -> %d     PTR:%p\n",
	       out_size, frame_size, ai.num_channels, ai.samples_per_frame,
	       ai.samples_per_frame * ai.num_channels * 2, in_data);
	printf("Synth %d  ch:%d  l:%d\n", synth.pcm.samplerate, synth.pcm.channels,
	       synth.pcm.length);

	printf("Timer %d %f   mode: %d  \n",
	       frame.header.duration.seconds, 1/(double)frame.header.duration.fraction,
	       frame.header.mode);
#endif
	for(int i = 0; i < synth.pcm.channels; i++)
	{
	    for(int j = 0; j < synth.pcm.length; j++)
	    {
		int sample = synth.pcm.samples[i][j] >> (MAD_F_FRACBITS + 1 - 16);
		if (sample > 32767)
		    sample = 32767;
		else if (sample < -32768)
		    sample = -32768;

		samples[j * ai.num_channels + i] = (int16_t)(sample);
		//      cerr << "Sample: " << samples[i][j] << "\n";
	    }
	}
    }
    if (size_read)
	*size_read = frame_size + in_size_init - in_size - 0;
    if (size_written)
	*size_written = 2 * synth.pcm.channels * synth.pcm.length;

    return 0;
}



// PLUGIN loading part
static avm::vector<CodecInfo> audiocodecs;

extern "C" const avm::vector<CodecInfo>& RegisterPlugin()
{
    if (audiocodecs.size())
	return audiocodecs;
    libmad_FillPlugins(audiocodecs);
    return audiocodecs;
}

extern "C" int GetPluginVersion() { return PLUGIN_API_VERSION; }

#define __MODULE__ "MAD audio decoder"
extern "C" IAudioDecoder* CreateAudioDecoder(const CodecInfo& info, const WAVEFORMATEX* format)
{
    IAudioDecoder* decoder = 0;
    try
    {
	switch (info.fourcc)
	{
	case 0x50:
	case 0x55:
	    decoder = new MAD_Decoder(info, format);
	    break;
	default:
	    throw FATAL("Audio format ID %d unsupported", format->wFormatTag);
	}
	printf("Audio in %s format\n", avm_wave_format_name(info.fourcc));
    }
    catch (FatalError& e)
    {
	e.Print();
    }
    return decoder;
}
