/*****************************************************************

  i_avifile.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Avifile input plugin (for AVI and ASF files).
// Can use win32 sodecs in /usr/lib/win32/

#include <iostream>

using std::ostream;
using std::cerr;

#include <avifile/include/avifile.h>
#include <avifile/include/utils.h>
#include <avifile/include/fourcc.h>
#include <avifile/include/except.h>
#include <avifile/include/StreamInfo.h>

#include <glib.h>

#include <string.h>

#include "../gv.h"

#include "i_avifile.h"

#define AUDIO_STREAM 0
#define VIDEO_STREAM 0

static GVPluginAttributes _attributes = {
  "i_avifile",
  "AVI/ASF Input Plugin",
  "Plugin for AVI/ASF files \nUses win23 codecs in /usr/lib/win32",
  "avi asf",
  GV_PLUGIN_HAS_AUDIO|GV_PLUGIN_HAS_VIDEO|GV_PLUGIN_IS_FILE|GV_PLUGIN_CAN_SEEK
};

GVInputAvifile::GVInputAvifile() :
  GVInput(&_attributes)
  {
  do_audio = 1;
  do_video = 1;
  avifile = (IAviReadFile*)0;
  audio_stream = (IAviReadStream*)0;
  video_stream = (IAviReadStream*)0;
  audio_stream_info = (StreamInfo*)0;
  video_stream_info = (StreamInfo*)0;
  }

int GVInputAvifile::colorspace_supported(GVColorspace colorspace)
  {
  switch(colorspace)
    {
    case GVColorspaceBGR24:
    case GVColorspaceBGR32:
    case GVColorspaceBGR16:
      return 1;
      break;
    case GVColorspaceYUV_420_P:
      // Re-Enable this if avifile supports planar formats
      if(colorspace_mask & IVideoDecoder::CAP_YV12)
        return 1;
      return 0;
      break;
    case GVColorspaceYUY2:
      if(colorspace_mask & IVideoDecoder::CAP_YUY2)
        return 1;
      break;
    default:
      return 0;
    }
  return 0;
  }

void GVInputAvifile::set_colorspace(GVColorspace colorspace)
  {
  switch(colorspace)
    {
    case GVColorspaceBGR24:
      video_stream->GetDecoder()->SetDestFmt(24, 0);
      break;
    case GVColorspaceBGR32:
      video_stream->GetDecoder()->SetDestFmt(32, 0);
      break;
    case GVColorspaceBGR16:
      video_stream->GetDecoder()->SetDestFmt(16, 0);
      break;
    case GVColorspaceYUV_420_P:
      //      cerr << "Colorspace: YUV_420_P\n";
      video_stream->GetDecoder()->SetDestFmt(0, RIFFINFO_YV12);
      break;
    case GVColorspaceYUY2:
      //      cerr << "Colorspace: YUY2\n";
      video_stream->GetDecoder()->SetDestFmt(0, RIFFINFO_YUY2);
      break;
    default:
      g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
            "Illegal colorspace selected");
      break;
    }
  video_format.colorspace = colorspace;
  video_format.init();
  }

void GVInputAvifile::init()
  {
  GVTime audio_time;
  GVTime video_time;

  audio_time.reset();
  video_time.reset();
    
  int failed;
  
  flags = 0;
  error = GV_PLUGIN_ERROR_NONE;
  
  // Open the file
  try
    {
    avifile = CreateIAviReadFile(filename);
    }
  catch(FatalError& avi_error)
    {
    if(avifile) delete avifile;
    error = GV_PLUGIN_ERROR_FILE_INVALID;
    return;
    }

  total_time.reset();

  // Check for video
  
  if((num_video_tracks = avifile->VideoStreamCount()))
    {
    video_stream = avifile->GetStream(VIDEO_STREAM, AviStream::Video);
    failed = 0;
    try
      {
      if(video_stream->StartStreaming())
        failed = 1;
      }
    catch(FatalError& avi_error)
      {
      avi_error.Print();
      failed = 1;
      }
    if(!failed)
      {
      video_stream_info = video_stream->GetStreamInfo();
      
      BITMAPINFOHEADER bitmapinfoheader;
      video_stream->GetVideoFormatInfo((void*)&(bitmapinfoheader),
                                       sizeof(BITMAPINFOHEADER));
      

      video_format.width = bitmapinfoheader.biWidth;
      
      video_format.height = bitmapinfoheader.biHeight;
      
      //       video_format.width = video_stream_info->GetVideoWidth();
      //       video_format.height = video_stream_info->GetVideoHeight();

      video_format_tag = bitmapinfoheader.biCompression;
      //      video_format_tag = video_stream_info->GetFourCC();
      
      colorspace_mask
        = (int)(video_stream->GetDecoder()->GetCapabilities());
      //      cerr << "Colorspace Mask: " << colorspace_mask << "\n";

      video_format.framerate = 1.0 / (video_stream->GetFrameTime());

      video_format.num_frames = video_stream_info->GetStreamFrames();

      // Calculate the framerate from total time and number of frames
      // This must be done since the framerate of asf files can be wrong
      // resulting in very huge total times.
            
      float test_framerate =
        video_format.num_frames / video_stream_info->GetLengthTime();
      
      if((test_framerate > 1.01 * video_format.framerate) ||
         (test_framerate < 0.99 * video_format.framerate))
        {
        video_time = video_stream_info->GetLengthTime();
        video_format.framerate = test_framerate;
        }
      else
        {
        video_time.set_from_frames(video_format.num_frames,
                                   video_format.framerate);
        }
      
      video_format.flags = 0;
      flags |= GV_PLUGIN_TRACK_HAS_VIDEO;
      //      video_stream->SeekToTime(0);

      // Set the best colorspace

      best_colorspace = GVColorspaceRGB24;
      
      if(colorspace_mask & IVideoDecoder::CAP_YV12)
        best_colorspace = GVColorspaceYUV_420_P;
      else if(colorspace_mask & IVideoDecoder::CAP_YUY2)
        best_colorspace = GVColorspaceYUY2;
      else if(colorspace_mask == 0) // No YUV supported, try RGB
        {
        switch(bitmapinfoheader.biBitCount)
          {
          case 16:
            best_colorspace = GVColorspaceBGR16;
            break;
          case 24:
            best_colorspace = GVColorspaceBGR24;
            break;
          }
        }
      }
    }
  else
    video_stream = (IAviReadStream*)0;

  audio_format_tag = 0;
  
  // Check for audio
  if((num_audio_tracks = avifile->AudioStreamCount()))
    {
    audio_stream = avifile->GetStream(AUDIO_STREAM, AviStream::Audio);

    failed = 0;
    try
      {
      if(audio_stream->StartStreaming())
        failed = 1;
      }
    catch(FatalError& avi_error)
      {
      avi_error.Print();
      failed = 1;
      }
   
    if(!failed)
      {
      int num_channels;
      int samplerate;
      int bytes_per_sample;

      WAVEFORMATEX waveformatex;
      char *extended = (char*)0;

      audio_stream_info = audio_stream->GetStreamInfo();
      audio_stream->GetAudioFormatInfo(&waveformatex, &extended);
      
      if(extended)
        delete extended;

      audio_format_tag = waveformatex.wFormatTag;
      
      num_channels = waveformatex.nChannels;
      samplerate = waveformatex.nSamplesPerSec;

      bytes_per_sample = 2;
      audio_format.sample_type = GVAudio_16_signed;

      switch(waveformatex.wBitsPerSample)
        {
        case 8:
          audio_format.sample_type = GVAudio_8_unsigned;
          bytes_per_sample = 1;
          break;
        case 16:
          audio_format.sample_type = GVAudio_16_signed;
          bytes_per_sample = 2;
          break;
        }

//       cerr << "**** audio_stream->GetFrameTime() : "
//            << audio_stream->GetFrameTime() << "\n";
      
      
      flags |= GV_PLUGIN_TRACK_HAS_AUDIO;

      audio_format.interleave_mode = GVInterleaveAll;
      //      audio_stream->Seek(0);

      //      int audio_frame_size = audio_stream->GetFrameSize()/2;
//       int audio_frame_size
//         = (int)(audio_stream->GetFrameTime()*samplerate+0.5);

//       unsigned int avi_samples_per_frame = audio_frame_size;

//       // check, how much frames to read

      //      unsigned int gmerlin_samples_per_frame; // How much would we like?

      size_t bytes_to_read = audio_stream->GetFrameSize();

      cerr << "GetFrameSize " << bytes_to_read << "\n";
      cerr << "GetStreamFrames: " << audio_stream_info->GetStreamFrames()
           << "\n";
      cerr << "num_samples: "
           << (bytes_to_read*audio_stream_info->GetStreamFrames()) /
        (num_channels * bytes_per_sample) << "\n";
      size_t min_bytes_to_read;

      if(video_stream)
        {
        min_bytes_to_read = (int)(1.0 / (video_format.framerate) * samplerate *
                                  bytes_per_sample * num_channels * 2.0);
        }
      else
        min_bytes_to_read = bytes_per_sample * num_channels * 4096;
            
      if(bytes_to_read < min_bytes_to_read)
        bytes_to_read = min_bytes_to_read;

      audio_format.samples_per_frame = bytes_to_read /
        (num_channels * bytes_per_sample);
            
//       if(flags & GV_PLUGIN_TRACK_HAS_VIDEO)
//         audio_format.samples_per_frame =
//           (int)((1.0/video_format.framerate)*samplerate+0.5)*2;
//       else
//         audio_format.samples_per_frame = 4096;

//       if(avi_samples_per_frame > gmerlin_samples_per_frame)
//         {
//         audio_format.samples_per_frame = avi_samples_per_frame;
//         frames_to_read = 1;
//         }
//       else
//         {
//         frames_to_read = gmerlin_samples_per_frame/avi_samples_per_frame;

//         audio_format.samples_per_frame=(frames_to_read*avi_samples_per_frame);
//         }

//       cerr << "Frames to read: " << frames_to_read << "\n";
      
      
      
      // Intialize the format
      
      audio_format.num_channels = num_channels;
      audio_format.samplerate = samplerate;
      
      audio_time = audio_stream_info->GetLengthTime();
      
      audio_format.num_samples =
        audio_time.get_samples(audio_format.samplerate);
      
      audio_format.init();
      }
    }
  else
    audio_stream = (IAviReadStream*)0;

  total_time = (video_time > audio_time) ? video_time : audio_time;
  }

void GVInputAvifile::fill_audio_buf(GVAudioFrame * frame)
  {
  uint_t bytes_read;
//   cerr << "Avifile: Fill audio buf, " << audio_format.samples_per_frame
//        << " samples...";
  uint_t frames_read;
  audio_stream->ReadFrames((void*)(frame->samples.s_16),
                           (uint_t)audio_format.buffer_size,
                           (uint_t)audio_format.buffer_size,
                           frames_read,
                           bytes_read);
  
  //  cerr << "Read " << frames_read << " Frames " << bytes_read << "Bytes\n";
  
  frame->valid_samples = bytes_read /
    (audio_format.bytes_per_sample * audio_format.num_channels);

  // reading frame time from input file

//   frame->time=audio_stream->GetTime();

//   cerr << "Audio frame time: " << frame->time.get_double() << "\n";

  if(audio_stream->Eof())
    audio_completed = 1;
  }

void GVInputAvifile::fill_video_buf(GVVideoFrame * frame)
  {
  //  cerr << "Avifile: Fill video buf...";
  video_stream->ReadFrame();
  unsigned char *data = video_stream->GetFrame()->Data();
  memcpy(frame->pixels, data, video_stream->GetFrameSize());
  //  cerr << "done\n";

  // reading frame time from input file
  frame->time=video_stream->GetTime();
  if(video_stream->Eof())
    video_completed = 1;

  //  cerr << "Video frame time: " << frame->time.get_double() << "\n";
  }

void GVInputAvifile::cleanup()
  {
  if(audio_stream)
    audio_stream->StopStreaming();
  if(video_stream)
    video_stream->StopStreaming();
    
  if(avifile)
    {
    delete avifile;
    avifile = (IAviReadFile*)0;
    }

  if(audio_stream_info)
    {
    delete audio_stream_info;
    audio_stream_info = (StreamInfo*)0;
    }
  if(video_stream_info)
    {
    delete video_stream_info;
    video_stream_info = (StreamInfo*)0;
    }
  }

GVInputAvifile::~GVInputAvifile()
  {
  }


void GVInputAvifile::seek(float percentage)
  {
  int have_video = !!((flags & GV_PLUGIN_TRACK_HAS_VIDEO) && do_video);
  int have_audio = !!((flags & GV_PLUGIN_TRACK_HAS_AUDIO) && do_audio);
  
  double seek_time;

//   cerr << "GVInputAvifile::seek: percentage: "
//        << percentage << " Seek time: " <<  seek_time << "\n";

  // Set the video time

  if(have_video)
    {
    video_stream->Seek((Unsigned)(video_format.num_frames*percentage+0.5));
    seek_time = video_stream->GetTime();
    }
  else
    seek_time = total_time.get_double();

  if(have_audio)
    audio_stream->SeekTime(seek_time);
  
  double real_percentage = seek_time / total_time.get_double();

//   cerr << "Real percentage: " << real_percentage << "\n";
  
  if(have_video)
    video_frames_written
      = (int)(video_format.num_frames*real_percentage + 0.5);
  if(have_audio)
    audio_samples_written =
      (int)(audio_format.num_samples*real_percentage + 0.5);

//   cerr << "Audio samples written: " << audio_samples_written <<
//     " Video frames written: " << video_frames_written << "\n";
  }


void GVInputAvifile::print_info(ostream & output)
  {
  if(audio_stream_info)
    {
    output << "Audio Codec: " << avm_wave_format_name((short)audio_format_tag)
           << "\n";
    //           << audio_stream_info << "\n";
    }
  if(video_stream_info)
    {
    output << "Video Codec: " << avm_fcc_name((fourcc_t)video_format_tag)
           << "\n";
      //           << video_stream_info << "\n";
    }
  }

int GVInputAvifile::skip(int num_frames)
  {
//   Unsigned pos = video_stream->GetPos();
//   Unsigned nextKeyPos = video_stream->GetNextKeyFrame((int)pos + num_frames);
//   int skipped_frames = (int)nextKeyPos - (int)pos;
   
  
//   if (skipped_frames <= 0) // with some movies GetNextKeyFrame does not give
// 			// a consistent value ????
// 	{
// 	cerr << "GVInputAvifile::skip : skipped_frames <= 0 : " << skipped_frames << "\n";
// 	return 0;
// 	}
//   if (skipped_frames == 1 )  // one frame is not worth
// 			// I ma not sure it is very useful !!!
// 	{
// 	video_stream->ToNextKeyFrame();
//         nextKeyPos = video_stream->GetNextKeyFrame(pos+(Unsigned)2);
//         skipped_frames = (int)nextKeyPos - (int)pos;
// 	if (skipped_frames <= 0) 
//           {
//           cerr << "GVInputAvifile::skip : skipped_frames <= 0\n";
//           return 0;
//           }
// 	} 
  //  video_stream->SeekToNextKeyFrame();
//   //  cerr << "skipping " << (int)nextKeyPos - (int)pos << "video frames\n";
  //  return (int)nextKeyPos - (int)pos;

//   for(int i = 0; i < num_frames; i++)
//     video_stream->SkipFrame();
//   return num_frames;

  Unsigned pos = video_stream->GetPos();
  cerr << "Skipping to position " << pos + num_frames
       << "Time: " << (pos + num_frames)/video_format.framerate << "\n";

  
  //  video_stream->SkipTo((pos + num_frames)/video_format.framerate);

  GVTime tmp_time;
  gv_engine->get_current_track_time(tmp_time, 0);

  video_stream->SkipTo(tmp_time.get_double());
  return num_frames;
  }

#define PLUGIN_NAME GVInputAvifile
#include "interface.cc"

