/*****************************************************************

  i_mpeg.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Mpeg decoder for gmerlin
// Uses the libmpeg3 library (http://heroines.sourceforge.net)

#include <iostream>

using std::ostream;
using std::cerr;

#include <glib.h>

#include "../gv.h"



extern "C" {

#include "libmpeg3/libmpeg3.h"
#include "libmpeg3/mpeg3protos.h"

int mpeg3demux_read_ifo(mpeg3_t *file, mpeg3_demuxer_t *demuxer,
                        int read_timecodes);
}


#include "i_mpeg.h"

void GVInputMpeg::cleanup()
  {
  if(mpeg_file)
    {
    mpeg3_close(mpeg_file);
    mpeg_file = (mpeg3_t*)0;
    }
  }

static GVPluginAttributes _attributes = 
  {
    "i_mpeg",
    "Mpeg Input Plugin",
    "MPEG Plugin for MPEG1/MPEG2 video,\nMPEG1 audio layer I,II,III and VOB files",
    "mpg m1v m2v vob ifo",
    GV_PLUGIN_HAS_AUDIO|
    GV_PLUGIN_HAS_VIDEO|
    GV_PLUGIN_IS_FILE|
    GV_PLUGIN_CAN_SEEK
  };

GVInputMpeg::GVInputMpeg() :
  GVInput(&_attributes)
  {
  mpeg_file = (mpeg3_t*)0;
  }

int GVInputMpeg::colorspace_supported(GVColorspace colorspace)
  {
  switch(colorspace)
    {
    //     case GVColorspaceRGB24:
    //     case GVColorspaceRGB16:
    //     case GVColorspaceBGR24:
    //    case GVColorspaceBGR32:
    //    case GVColorspaceBGR16:
    case GVColorspaceYUV_420_P:
      return 1;
    default:
      return 0;
    }
  }

#define AUDIO_STREAM 0
#define VIDEO_STREAM 0

void GVInputMpeg::init()
  {
  error = GV_PLUGIN_ERROR_NONE;
  
  flags = 0;
  
  GVTime audio_time;
  GVTime video_time;
  
  double mpeg_time = 0.0; // For Ifo files
  
  // First check if the file is valid
  
  if(!mpeg3_check_sig(filename))
    {
    cerr << "Check sig failed\n";
    error = GV_PLUGIN_ERROR_FILE_INVALID;
    return;
    }
  
  if(!(mpeg_file = mpeg3_open(filename)))
    {
    cerr << "mpeg3_open failed\n";
    error = GV_PLUGIN_ERROR_FILE_INVALID;
    return;
    }
  
  // Read the toc of an ifo file
  
  if(mpeg_file->is_ifo_file)
    {
    mpeg3_demuxer_t *demuxer = mpeg3_new_demuxer(mpeg_file, 0, 0, -1);
    mpeg3io_open_file(mpeg_file->fs);
    mpeg3demux_read_ifo(mpeg_file, demuxer, 1);
    mpeg3io_close_file(mpeg_file->fs);
    mpeg3demux_timecode_t *timecode;
    mpeg3_title_t *title;
    //    double end_time;
    for(int i = 0; i < demuxer->total_titles; i++)
      {
      title = (demuxer->titles[i]);
//       cerr << "File: " << demuxer->titles[i]->fs->path << "\n";
      for(int j = 0; j < title->timecode_table_size; j++)
        {
        timecode = &(title->timecode_table[j]);
//         cerr << "Timecode " << j
//              << " Start time: " << timecode->start_time
//              << " End time: "   << timecode->end_time << "\n";
        mpeg_time = timecode->end_time;
        }
      }
    }
  else
    mpeg_time = 0.0;
  
  // Check for audio streams

  num_audio_tracks = mpeg3_total_astreams(mpeg_file);
  //  cerr << "num_audio_tracks: " << num_audio_tracks << "\n";
  if(num_audio_tracks)
    {
    audio_format.num_channels = mpeg3_audio_channels(mpeg_file, AUDIO_STREAM);
//     cerr << "Found Audio: " << num_audio_tracks << " Tracks, "
//          << audio_format.num_channels << " Channels in track 0\n";

    // This is a dirty hack, because currently, gmerlin supports
    // only up to 5 channels. The multichannel support will,
    // however, be reworked in one of the next versions, and
    // it will support 6 channels then (also on Creative 5.1 cards)
    // so for now, you will have to live without the LFE channel.
    
    if(audio_format.num_channels == 6)
      audio_format.num_channels = 5;
        
    audio_format.samplerate = mpeg3_sample_rate(mpeg_file, AUDIO_STREAM);

    if(mpeg_file->is_ifo_file)
      {
      audio_format.num_samples
        = (unsigned int)(audio_format.samplerate * mpeg_time+0.5);
      audio_time = mpeg_time;
      }
   else
     {
     audio_format.num_samples = mpeg3_audio_samples(mpeg_file, AUDIO_STREAM);
     audio_time.set_from_frames((unsigned int)(audio_format.num_samples),
                                (unsigned int)(audio_format.samplerate));
     }

    //     cerr << "Audio Time: " << audio_time.seconds << " Seconds\n";
    flags |= GV_PLUGIN_TRACK_HAS_AUDIO;
    audio_format.interleave_mode = GVInterleaveNone;
    
    //    cerr << "Audio Track: " << audio_format.num_channels << " Channels\n"
    //         << "Samplerate: " << audio_format.samplerate << " Hz, "
    //         << audio_format.num_samples << " Samples\n";
    audio_format.sample_type = GVAudio_16_signed;
    }
  else
    audio_time.reset();
    
  // Check for video streams

  num_video_tracks = mpeg3_total_vstreams(mpeg_file);
  //  cerr << "num_video_tracks: " << num_video_tracks << "\n";
  
  if(num_video_tracks)
    {
    video_format.width = mpeg3_video_width(mpeg_file, VIDEO_STREAM);
    video_format.height = mpeg3_video_height(mpeg_file, VIDEO_STREAM);
    video_format.framerate = mpeg3_frame_rate(mpeg_file, VIDEO_STREAM);
    video_format.num_frames = mpeg3_video_frames(mpeg_file, VIDEO_STREAM);

    //    if(mpeg3_colormodel(mpeg_file, VIDEO_STREAM) == MPEG3_YUV420P) 
      best_colorspace = GVColorspaceYUV_420_P;
      //    else
      //      best_colorspace = GVColorspaceYUV_422_P;

    video_format.colorspace = best_colorspace;
    
    video_format.flags = 0;

    if(mpeg_file->is_ifo_file)
      {
      video_format.num_frames
        = (unsigned int)(video_format.framerate * mpeg_time+0.5);
      video_time = mpeg_time;
      }
    else
      {
      video_format.num_frames = mpeg3_video_frames(mpeg_file, VIDEO_STREAM);
      video_time.set_from_frames(video_format.num_frames,
                                 video_format.framerate);
      }
    flags |= GV_PLUGIN_TRACK_HAS_VIDEO;
    }
  else
    video_time.reset();
  
  if(!(flags & (GV_PLUGIN_TRACK_HAS_VIDEO|GV_PLUGIN_TRACK_HAS_AUDIO)))
    {
    cerr << "File " << filename << " contains no video and no audio\n";
    error = GV_PLUGIN_ERROR_FILE_INVALID;
    mpeg3_close(mpeg_file);
    mpeg_file = (mpeg3_t*)0;
    return;
    }
  
  total_time = (video_time > audio_time) ? video_time : audio_time;
  
  if(flags & GV_PLUGIN_TRACK_HAS_VIDEO)
    audio_format.samples_per_frame =
      (int)((1.0 / video_format.framerate) *
            audio_format.samplerate + 0.5);
  else
    audio_format.samples_per_frame = 4096;
  
  audio_format.init();
  video_format.init();
  searched = 0;
  }

void GVInputMpeg::fill_audio_buf(GVAudioFrame * frame)
  {
  //  cerr << "Fill Audio Buf!\n";

  //  int res;
  
  //  int num_samples;
  
  mpeg3_read_audio(mpeg_file, (float *)0, frame->channels.s_16[0],
                   0, audio_format.samples_per_frame, AUDIO_STREAM);
  //  cerr << "res: " << res << "\n"; 

  // Channels (for 5.1 streams)
  // 0: Left
  // 1: Center
  // 2: Right
  // 3: Rear Left
  // 4: Rear Right

  if(audio_format.num_channels == 2)
    {
    mpeg3_reread_audio(mpeg_file, (float*)0, frame->channels.s_16[1],
                       1, audio_format.samples_per_frame, AUDIO_STREAM);
    }
  else if(audio_format.num_channels == 5)
    {
    // Read center channel
    mpeg3_reread_audio(mpeg_file, (float*)0, frame->channels.s_16[4],
                       1, audio_format.samples_per_frame, AUDIO_STREAM);
    
    // Read right channel
    mpeg3_reread_audio(mpeg_file, (float*)0, frame->channels.s_16[1],
                       2, audio_format.samples_per_frame, AUDIO_STREAM);
    
    // Read left rear channel
    mpeg3_reread_audio(mpeg_file, (float*)0, frame->channels.s_16[2],
                       3, audio_format.samples_per_frame, AUDIO_STREAM);
    
    // Read right rear channel
    mpeg3_reread_audio(mpeg_file, (float*)0, frame->channels.s_16[3],
                       4, audio_format.samples_per_frame, AUDIO_STREAM);
    }

  if(!mpeg_file->is_ifo_file)
    {
    if(audio_samples_written + audio_format.samples_per_frame
       < audio_format.num_samples)
      frame->valid_samples = audio_format.samples_per_frame;
    else
      {
      audio_completed = 1;
      frame->valid_samples = audio_format.num_samples
        - audio_samples_written;
      cerr << "audio_completed, valid samples: " << frame->valid_samples << "\n";

      }
    }
  else
    {
    frame->valid_samples = audio_format.samples_per_frame;

    if(mpeg3_end_of_audio(mpeg_file, AUDIO_STREAM))
      {
      cerr << "audio_completed\n";
      audio_completed = 1;
      }
    }
  // Check for the end of the stream
#if 0
  if(searched)
    if(mpeg3_end_of_audio(mpeg_file, AUDIO_STREAM))
      {
      cerr << "audio_completed\n";
      audio_completed = 1;
      }
  else
    {
    if((audio_frames_written+1)*audio_format.samples_per_frame >=
       audio_format.num_samples)
      {
      cerr << "audio_completed\n";
      audio_completed = 1;
      }
    }
#endif
  }

void GVInputMpeg::fill_video_buf(GVVideoFrame * frame)
  {
  int res;
  
  //  cerr << "GVInputMpeg::fill_video_buf\n";
  switch(video_format.colorspace)
    {
    case GVColorspaceBGR16:
      res = mpeg3_read_frame(mpeg_file, frame->pixels_rows,
                             0, 0, video_format.width, video_format.height,
                             video_format.width, video_format.height,
                             MPEG3_RGB565, VIDEO_STREAM);
      break;
    case GVColorspaceRGB24:
      res = mpeg3_read_frame(mpeg_file, frame->pixels_rows,
                             0, 0, video_format.width, video_format.height,
                             video_format.width, video_format.height,
                             MPEG3_RGB888, VIDEO_STREAM);
      break;
    case GVColorspaceBGR32:
      res = mpeg3_read_frame(mpeg_file, frame->pixels_rows,
                             0, 0, video_format.width, video_format.height,
                             video_format.width, video_format.height,
                             MPEG3_BGRA8888, VIDEO_STREAM);
      break;
    case GVColorspaceYUV_420_P:
      res = mpeg3_read_yuvframe(mpeg_file, (char*)frame->y,
                                (char*)frame->u,
                                (char*)frame->v,
                                0, 0, video_format.width, video_format.height,
                                VIDEO_STREAM);
      break;
    }
  
  if(!mpeg_file->is_ifo_file)
    {
    if(video_frames_written + 1 >= video_format.num_frames)
      {
      cerr << "video completed\n";
      video_completed = 1;
      }
    }
  else
    {
    if(mpeg3_end_of_video(mpeg_file, VIDEO_STREAM))
      {
      cerr << "video_completed\n";
      video_completed = 1;
      }
    }
  }


void GVInputMpeg::seek(float percentage)
  {
  //  cerr << "Mpeg: Searching...";
  mpeg3_seek_percentage(mpeg_file, percentage);
  
  searched = 1;
  //  cerr << "done\n";
  }

void GVInputMpeg::print_info(ostream & output)
  {
  output << "Type of the File: ";
  
  if(mpeg_file->is_transport_stream)
    output << "Transport Stream\n";
  if(mpeg_file->is_program_stream)
    output << "Program Stream\n";
  if(mpeg_file->is_ifo_file)
    output << "Ifo File\n";
  if(mpeg_file->is_audio_stream)
    output << "Audio Stream\n";
  if(mpeg_file->is_video_stream)
    output << "Video Stream\n";

  // Now print some audio and video infos

  if(flags & GV_PLUGIN_TRACK_HAS_AUDIO)
    {
    mpeg3audio_t * audio_decoder = mpeg_file->atrack[AUDIO_STREAM]->audio;
    output << "Audio: ";
    switch(audio_decoder->format)
      {
      case AUDIO_AC3:
        output << "AC3\n";
        break;
      case AUDIO_MPEG:
        if((audio_decoder->layer != 3) ||
           (audio_decoder->layer == 3) &&
           (audio_decoder->samples_per_frame == 1152))
          output << "MPEG 1 Layer " << audio_decoder->layer << "\n";
        else
          {
          output << "MPEG 2 Layer " << audio_decoder->layer << "\n";
          //          output << "audio_decoder->samples_per_frame"
          }
        break;
      case AUDIO_PCM:
        output << "PCM\n";
        break;
      }
    if(audio_decoder->format == AUDIO_AC3)
      output << "Bitrate: " << (audio_decoder->bitrate) << " kb/s\n";
    else
      output << "Bitrate: " << (audio_decoder->bitrate)/1000 << " kb/s\n";

    //    output << "Samples per frame: " << audio_decoder->samples_per_frame << "\n";
    output << "Copyright: "
           << ((audio_decoder->copyright) ? "Yes, " : "No, ")
           << "Original: "
           << ((audio_decoder->original) ? "Yes " : "No ") << "\n";
    }

  if(flags & GV_PLUGIN_TRACK_HAS_VIDEO)
    {
    mpeg3video_t * video_decoder = mpeg_file->vtrack[AUDIO_STREAM]->video;
    output << "Video: ";
    output << "Bitrate: " << (video_decoder->bitrate) << " kb/s\n";
    //    output << "Samples per frame: " << audio_decoder->samples_per_frame << "\n";
    }
  }

int GVInputMpeg::skip(int num_frames)
  {
  //  cerr << "GVInputMpeg::skip " << num_frames << "\n";
  mpeg3_drop_frames(mpeg_file, num_frames, VIDEO_STREAM);
  return num_frames;
  }


#define PLUGIN_NAME GVInputMpeg
#include "interface.cc"
