/*****************************************************************

  i_oss.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <iostream>

using std::cerr;
using std::ostream;

#include "../gv.h"
#include "i_oss.h"
#include "oss_common.h"

#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "../gvconfig.h"
#include "../gtkutil.h"

static GVPluginAttributes _attributes = 
  {
    "i_oss",
    "OSS Recorder",
    "Plugin for recording from OSS soundcards",
    (char*)0,
    GV_PLUGIN_HAS_AUDIO|
    GV_PLUGIN_NEEDS_REALTIME|
    GV_PLUGIN_IS_DEVICE|
    GV_PLUGIN_IS_RECORDER
  };

void GVInputOSS::fill_playlist()
  {
  //  cerr << "GVInputOSS::fill_playlist()\n";

  gv_playlist->clear();
  
  gv_playlist->append_track((char*)0, "Recording Stream",
                            record_time,
                            GV_PLAYLIST_ITEM_HAS_AUDIO|
                            GV_PLAYLIST_ITEM_NOSAVE,
                            attributes->name);
  
  }

void GVInputOSS::reset()
  {
  if(fd != -1)
    {
    close(fd);
    fd = -1;
    }
  }

GVInputOSS::GVInputOSS() :
  GVInput(&_attributes)
  {
  get_config();
  fd = -1;
  flags = GV_PLUGIN_TRACK_HAS_AUDIO | GV_PLUGIN_TRACK_IS_STREAM;

  config_format.endian_format = GVNativeEndian;
  config_window = (ConfigWindow*)0;
  config_format.samples_per_frame = 4096;
  }

void GVInputOSS::fill_audio_buf(GVAudioFrame * frame)
  {
  if(fd == -1)
    open_device();

  int num_bytes = read(fd, (void*)(frame->channels.s_8[0]), read_size);
  frame->valid_samples = num_bytes /
    (audio_format.num_channels*audio_format.bytes_per_sample);
  }
  
void GVInputOSS::init()
  {
  audio_format = config_format;

  if(audio_format.num_channels == 2)
    audio_format.interleave_mode = GVInterleave2;
  else
    audio_format.interleave_mode = GVInterleaveNone;

  total_time = record_time;

  audio_format.num_samples = total_time.get_samples(audio_format.samplerate);
  audio_format.init();

  read_size = audio_format.bytes_per_sample *
    audio_format.num_channels *
    audio_format.samples_per_frame;
  //  cerr << "read_size: " << read_size << "\n";

  if(!open_device())
    return;
  close(fd);
  fd = -1;
  }

int GVInputOSS::open_device()
  {
  //  cerr << "GVInputOSS::open_device(), device: " << device << "...";
  if((fd = open(device, O_RDONLY, 0)) == -1)
    {
    g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "Cannot open device %s", device);
    error = GV_PLUGIN_ERROR_DEVICE_OPEN_FAILED;
    return 0; 
    }

  int sample_format = oss_get_sample_format(audio_format);
  
  if(ioctl(fd, SNDCTL_DSP_SETFMT, &sample_format) == -1)
    {
    g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "Unsupported sample format");
    error = GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED;
    return 0; 
    }

  int num_channels = audio_format.num_channels;
  
  if(ioctl(fd, SNDCTL_DSP_CHANNELS, &num_channels) == -1)
    {
    g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
          "Device %s doesn't support %d channel sound",
          device, audio_format.num_channels);
    error = GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED;
    return 0;
    }

  int samplerate = audio_format.samplerate;
  
  if(ioctl(fd, SNDCTL_DSP_SPEED, &samplerate) == -1)
    {
    g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
          "Device %s doesn't support a samplerate of %d Hz",
          device, audio_format.samplerate);
    error = GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED;
    return 0;
    }
  //  cerr << "Done\n";
  return 1;
  }

void GVInputOSS::configure()
  {
  if(!config_window)
    config_window = new ConfigWindow(this);
  config_window->run();
  }

void GVInputOSS::cleanup()
  {
  if(fd != -1)
    {
    close(fd);
    fd = -1;
    }
  }

GVInputOSS::~GVInputOSS()
  {
  cleanup();
  put_config();
  if(config_window)
    delete config_window;
  }

#define SECTION attributes->name

void GVInputOSS::get_config()
  {
  int samplerate = 44100;
  get_devices_config("/dev/dsp");
  get_recording_config();
  gv_config->get_int(SECTION, "Samplerate", samplerate);

  config_format.samplerate = samplerate;
  
  int bits = 16;
  gv_config->get_int(SECTION, "Bits", bits);

  switch(bits)
    {
    case 8:
      config_format.sample_type = GVAudio_8_signed;
      break;
    case 16:
      config_format.sample_type = GVAudio_16_signed;
      break;
    }

  int num_channels = 2;
  gv_config->get_int(SECTION, "Channels", num_channels);
  config_format.num_channels = num_channels;
  }

void GVInputOSS::put_config()
  {
  put_devices_config();
  put_recording_config();
  gv_config->put_int(SECTION, "Samplerate", config_format.samplerate);

  switch(config_format.sample_type)
    {
    case GVAudio_8_signed:
    case GVAudio_8_unsigned:
      gv_config->put_int(SECTION, "Bits", 8);
      break;
    case GVAudio_16_unsigned:
    case GVAudio_16_signed:
      gv_config->put_int(SECTION, "Bits", 16);
    }
  gv_config->put_int(SECTION, "Channels", config_format.num_channels);
  }

static GVConfigMenuItemInt oss_channels[] =
  {
    { 1,    "Mono" },
    { 2,  "Stereo" }
  };

static int oss_bits[] =
  {
    { 8 },
    { 16 }
  };

GVInputOSS::ConfigWindow::ConfigWindow(GVInputOSS * _plugin) :
  GVConfigWindow(_plugin->attributes->real_name),
  record_time(_plugin),
  device_widget(_plugin),
  samplerate("Samplerate", 4000, 96000, 0),
  channels("Mode"),
  bits("Bits")
  {
  plugin = _plugin;
  GtkWidget * notebook = gtk_notebook_new();
  
  GtkWidget * table = gvgtk_table_new(3, 2, 0);

  channels.create_menu(oss_channels,
                       sizeof(oss_channels)/sizeof(GVConfigMenuItemInt));
  bits.create_menu(oss_bits, sizeof(oss_bits)/sizeof(int));

  samplerate.attach(table, 0, 0);
  channels.attach(table, 1, 0);
  bits.attach(table, 2, 0);
  
  gtk_widget_show(table);

  GtkWidget * tab_label;
  tab_label = gtk_label_new("Recording Options");
  gtk_widget_show(tab_label);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), table, tab_label);

  tab_label = gtk_label_new("Time");
  gtk_widget_show(tab_label);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), record_time.widget,
                           tab_label);

  tab_label = gtk_label_new("Devices");
  gtk_widget_show(tab_label);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), device_widget.widget,
                           tab_label);

  gtk_widget_show(notebook);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), notebook);
  }

void GVInputOSS::ConfigWindow::get_values()
  {
  device_widget.get_values();
  record_time.get_values();

  channels.set((int)(plugin->config_format.num_channels));

  int num_bits = 0;

  switch(plugin->config_format.sample_type)
    {
    case GVAudio_8_signed:
    case GVAudio_8_unsigned:
      num_bits = 8;
      break;
    case GVAudio_16_signed:
    case GVAudio_16_unsigned:
      num_bits = 16;
      break;
    }
  bits.set(num_bits);
  samplerate.set_value(plugin->config_format.samplerate);

  }

void GVInputOSS::ConfigWindow::apply_values()
  {
  device_widget.apply_values();
  record_time.apply_values();

  int num_bits = bits.get_int();
  
  if(num_bits == 8)
    plugin->config_format.sample_type = GVAudio_8_signed;
  else if(num_bits == 16)
    plugin->config_format.sample_type = GVAudio_16_signed;

  plugin->config_format.samplerate = (int)(samplerate.get_value());

  }

#define PLUGIN_NAME GVInputOSS
#include "interface.cc"
