/*****************************************************************

  i_quicktime.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#ifdef __GNUG__
#pragma implementation "i_quicktime.h"
#endif


#include <iostream>

using std::ostream;
using std::cerr;

#include <glib.h>
#include <string.h>
#include "../gv.h"

#include "quicktime4linux/quicktime.h"
// #include "quicktime4linux/colormodels.h"

#include "i_quicktime.h"

#include "quicktime_common.h"

void GVInputQuicktime::cleanup()
  {
  if(quicktime_file)
    {
    quicktime_close(quicktime_file);
    quicktime_file = (quicktime_t*)0;
    }
  }

static GVPluginAttributes _attributes = 
  {
    "i_quicktime",
    "Quicktime4linux Input Plugin",
    "Input plugin for quicktime files\n\
See the documentation at\nhttp://heroines.sourceforge.net\nfor supported codecs",
    "mov",
    GV_PLUGIN_HAS_AUDIO|
    GV_PLUGIN_HAS_VIDEO|
    GV_PLUGIN_IS_FILE|
    GV_PLUGIN_CAN_SEEK
  };

GVInputQuicktime::GVInputQuicktime() :
  GVInput(&_attributes)
  {
  quicktime_file = (quicktime_t*)0;
  video_format.flags = 0;
  }

#define AUDIO_STREAM 0
#define VIDEO_STREAM 0

void GVInputQuicktime::init()
  {
  error = GV_PLUGIN_ERROR_NONE;
  
  flags = 0;

  GVTime audio_time;
  GVTime video_time;

  // First check if the file is valid

  if(!quicktime_check_sig(filename))
    {
    error = GV_PLUGIN_ERROR_FILE_INVALID;
    return;
    }

  if(!(quicktime_file = quicktime_open(filename, 1, 0)))
    {
    error = GV_PLUGIN_ERROR_FILE_INVALID;
    return;
    }

  quicktime_info_present = 0;
//   if(!quicktime_read_info(quicktime_file))
//     {
//     cerr << "No MOOV-Atom present\n";
//     quicktime_info_present = 0;
//     }
//   else
//     {
//     cerr << "found MOOV-Atom\n";
//     quicktime_info_present = 1;
//     }
    
  // Check for audio streams

  if((num_audio_tracks = quicktime_audio_tracks(quicktime_file)))
    {
    if(!quicktime_supported_audio(quicktime_file, AUDIO_STREAM))
      {
      cerr << "No codec \"" << quicktime_audio_compressor(quicktime_file,
                                                          AUDIO_STREAM)
           << "\" found.\n";
      audio_time.reset();
      }
    else
      {
      audio_format.num_channels =
        quicktime_track_channels(quicktime_file, AUDIO_STREAM);
      audio_format.samplerate =
        quicktime_sample_rate(quicktime_file, AUDIO_STREAM);
      audio_format.num_samples = quicktime_audio_length(quicktime_file,
                                                        AUDIO_STREAM);
      audio_time.set_from_frames((unsigned int)(audio_format.num_samples),
                                 (unsigned int)(audio_format.samplerate));
      flags |= GV_PLUGIN_TRACK_HAS_AUDIO;
      audio_format.interleave_mode = GVInterleaveNone;
      
      //    cerr << "Audio Track: " << audio_format.num_channels << " Channels\n"
      //         << "Samplerate: " << audio_format.samplerate << " Hz, "
      //         << audio_format.num_samples << " Samples\n";
      //     cerr << "Audio Compressor: "
      //          << quicktime_audio_compressor(quicktime_file, AUDIO_STREAM)
      //          << "\n";
      audio_format.sample_type = GVAudio_16_signed;
      }
    }
  else
    audio_time.reset();
  
  // Check for video streams

  if((num_video_tracks = quicktime_video_tracks(quicktime_file)))
    {
    if(!quicktime_supported_video(quicktime_file, VIDEO_STREAM))
      {
      cerr << "No codec \"" << quicktime_video_compressor(quicktime_file,
                                                          VIDEO_STREAM)
           << "\" found.\n";
      video_time.reset();
      }
    else
      {
      video_format.width =
        quicktime_video_width(quicktime_file, VIDEO_STREAM);
      video_format.height =
        quicktime_video_height(quicktime_file, VIDEO_STREAM);
      video_format.framerate =
        quicktime_frame_rate(quicktime_file, VIDEO_STREAM);
      video_format.num_frames =
        quicktime_video_length(quicktime_file, VIDEO_STREAM);

      best_colorspace = GVColorspaceRGB24;
      
      char * video_compressor =
        quicktime_video_compressor(quicktime_file, VIDEO_STREAM);

      // Try some colorspaces
      
      if(!strcmp(video_compressor, QUICKTIME_YUV2))
        best_colorspace = GVColorspaceYUY2;
      else if(!strcmp(video_compressor, QUICKTIME_RAW) ||
              !strcmp(video_compressor, QUICKTIME_PNG))
        {
        int video_depth = quicktime_video_depth(quicktime_file, VIDEO_STREAM);
        switch(video_depth)
          {
          case 24:
            best_colorspace = GVColorspaceRGB24;
            break;
          case 32:
            best_colorspace = GVColorspaceRGBA32;
            break;
          default:
            break;
          }
        }
      else if(quicktime_reads_cmodel(quicktime_file,
                                     get_quicktime_colorspace(GVColorspaceYUV_420_P),
                                     VIDEO_STREAM))
        best_colorspace = GVColorspaceYUV_420_P;
      else if(quicktime_reads_cmodel(quicktime_file,
                                     get_quicktime_colorspace(GVColorspaceYUY2),
                                     VIDEO_STREAM))
        best_colorspace = GVColorspaceYUY2;
      else
        best_colorspace = GVColorspaceRGB24;

      video_time.set_from_frames(video_format.num_frames,
                                 video_format.framerate);
      
      flags |= GV_PLUGIN_TRACK_HAS_VIDEO;
      //      quicktime_set_video_position(quicktime_file, 0, VIDEO_STREAM);
      }

    }
  else
    video_time.reset();
  
  if(!(flags & (GV_PLUGIN_TRACK_HAS_VIDEO|GV_PLUGIN_TRACK_HAS_AUDIO)))
    {
    cerr << "File " << filename << " contains no video and no audio\n";
    error = GV_PLUGIN_ERROR_FILE_INVALID;
    quicktime_close(quicktime_file);
    quicktime_file = (quicktime_t*)0;
    return;
    }
  
  total_time = (video_time > audio_time) ? video_time : audio_time;

  if(flags & GV_PLUGIN_TRACK_HAS_VIDEO)
    audio_format.samples_per_frame =
      (int)((1.0 / video_format.framerate) * audio_format.samplerate + 0.5);
  else
    audio_format.samples_per_frame = 4096;

  audio_format.init();

  video_format.colorspace = best_colorspace;

  video_format.init();

  quicktime_set_preload(quicktime_file, 2000000);
  quicktime_set_cpus(quicktime_file, 1);
  quicktime_set_video_position(quicktime_file, 0, 0);
  }

void GVInputQuicktime::fill_audio_buf(GVAudioFrame * frame)
  {
  //  cerr << "Fill Audio Buf!\n";

  //  int res;
  
  //  int offset = audio_format.samples_per_frame;

  //  int num_samples;

  int result;
  
  for(unsigned int i = 0; i < audio_format.num_channels; i++)
    {
    quicktime_set_audio_position(quicktime_file, audio_samples_written,
                                 AUDIO_STREAM);
    result = quicktime_decode_audio(quicktime_file,
                                    frame->channels.s_16[i], (float *)0,  
                                    audio_format.samples_per_frame, i);
    }

  if(audio_samples_written +
     audio_format.samples_per_frame < audio_format.num_samples)
     frame->valid_samples = audio_format.samples_per_frame;
   else
     {
     //     cerr << "audio_completed\n";
     audio_completed = 1;
     frame->valid_samples = audio_format.num_samples - audio_samples_written;
     }
  //  cerr << "Read Audio: Result: " << result << "\n";
  
  }

int GVInputQuicktime::colorspace_supported(GVColorspace colorspace)
  {
//   if(colorspace == GVColorspaceYUV_420_P)
//     return 0;

//    if(colorspace != GVColorspaceRGB24)
//      return 0;
//    return 1;

  if(colorspace == best_colorspace)
    return 1;
  else
    return !!(quicktime_reads_cmodel(quicktime_file,
                                     get_quicktime_colorspace(colorspace),
                                     VIDEO_STREAM));
  }


void GVInputQuicktime::fill_video_buf(GVVideoFrame * frame)
  {
//   cerr << "GVInputQuicktime::fill_video_buf: colorspace: "
//        << get_quicktime_colorspace(video_format.colorspace) << "\n";
//   cerr << "Size of the picture: " << video_format.width << " x "
//        << video_format.height << "\n";
  int result;
  switch(video_format.colorspace)
    {
    case GVColorspaceRGB24:
      result =
        quicktime_decode_video(quicktime_file,
                               frame->pixels_rows,
                               VIDEO_STREAM);
      break;
    default:
      result =
        quicktime_decode_scaled(quicktime_file, 
                                0, // Location of input frame to take picture
                                0,
                                video_format.width,
                                video_format.height,
                                video_format.width, 
                                video_format.height,
                                get_quicktime_colorspace(video_format.colorspace),
                                frame->pixels_rows, 
                                VIDEO_STREAM);
      break;
    }
  if(video_frames_written + 1 == video_format.num_frames)
    video_completed = 1;
  }


void GVInputQuicktime::seek(float percentage)
  {
  cerr << "Quicktime: Searching...";
  if(flags & GV_PLUGIN_TRACK_HAS_AUDIO)
    {
    quicktime_set_audio_position(quicktime_file,
                                 (gint64)((float)(audio_format.num_samples)*
                                       percentage+0.5),
                                 AUDIO_STREAM);
    audio_samples_written = quicktime_audio_position(quicktime_file,
                                                     AUDIO_STREAM);
    }
  
  if(flags & GV_PLUGIN_TRACK_HAS_VIDEO)
    {
    quicktime_set_video_position(quicktime_file,
                                 (gint64)((float)(video_format.num_frames)*
                                          percentage+0.5),
                                 VIDEO_STREAM);
    video_frames_written = quicktime_video_position(quicktime_file,
                                                    VIDEO_STREAM);
    }
  cerr << "done\n";
  }

void GVInputQuicktime::print_info(ostream & output)
  {
  if(flags & GV_PLUGIN_TRACK_HAS_AUDIO)
    {
    output << "Audio Tracks: " << num_audio_tracks;
    output << ", Audio Codec: "
           << quicktime_audio_compressor(quicktime_file, AUDIO_STREAM) << "\n";
    }
  else
    output << "No Audio\n";
 
  if(flags & GV_PLUGIN_TRACK_HAS_VIDEO)
    {
    output << "Video Tracks: " << num_video_tracks;
    output << ", Video Codec: "
           << quicktime_video_compressor(quicktime_file, VIDEO_STREAM) << "\n";
    }
  else
    output << "No Video\n";
    
  if(quicktime_info_present)
    {
    output << "File Info: " << quicktime_get_info(quicktime_file) << "\n";
    output << "Copyright: " << quicktime_get_copyright(quicktime_file) << "\n";
    output << "Name: "      << quicktime_get_name(quicktime_file) << "\n";
    }
  else
    output << "No further Information present\n";

  }

GVInputQuicktime::~GVInputQuicktime()
  {
  cleanup();
  }

int GVInputQuicktime::skip(int num_frames)
  {
  int position =  quicktime_video_position(quicktime_file, VIDEO_STREAM);
  position += num_frames;
  quicktime_set_video_position(quicktime_file, position, VIDEO_STREAM);
  return num_frames;
  }


#define PLUGIN_NAME GVInputQuicktime
#include "interface.cc"
