/*****************************************************************

  i_singlepic.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Base class for single picture plugins

#include <iostream>
#include <fstream>

using std::cerr;
using std::ostream;
using std::ifstream;

#include "../gv.h"
#include "../globalconfig.h"

#include "../gvloader.h"

#include "i_singlepic.h"

#include <string.h>
#include <ctype.h>

#include <stdio.h>
#include <stdlib.h>

const GVPluginAttributes _attributes = 
  {
    "i_singlepic",
    "Single picture input Plugin",
    "Input plugin for single pictures\n\
Loads al image formats for which imagereaders exist",
    "jpg png tga",
    GV_PLUGIN_HAS_VIDEO|
    GV_PLUGIN_IS_FILE|
    GV_PLUGIN_CAN_SEEK
  };

GVInputSinglepic::GVInputSinglepic() :
  GVInput(&_attributes)
  {
  filename_mask = (char*)0;
  current_filename = (char*)0;
  flags = GV_PLUGIN_TRACK_HAS_VIDEO;
  image_reader = (GVImageReader*)0;
  }

GVInputSinglepic::~GVInputSinglepic()
  {
  if(filename_mask)
    delete filename_mask;
  if(current_filename)
    delete current_filename;
  }

void GVInputSinglepic::init()
  {
  error = 0;
  if(filename_mask)
    {
    delete filename_mask;
    filename_mask = (char*)0;
    }
  
  if(current_filename)
    {
    delete current_filename;
    current_filename = (char*)0;
    }

  // Now find an image loder

  GVPluginInfo * info =
    gv_image_reader_loader->find_plugin_by_suffix(filename);
  
  if(!info)
    {
    error = GV_PLUGIN_ERROR_FILE_INVALID;
    return;
    }
  
  image_reader = gv_image_reader_loader->load_image_reader(info);
  
  video_format.framerate = gv_global_config->singlepic_framerate;
  
  // Find the files to read

  char * first_nondigit;
  char * first_digit; 

  first_nondigit = &(filename[strlen(filename)-1]);
  
  while(!isdigit(*first_nondigit))
    first_nondigit--;

  first_digit = first_nondigit;

  while(isdigit(*first_digit))
    first_digit--;

  first_digit++;
  first_nondigit++;

//   cerr << "Filename: " << filename << "\n";
//   cerr << "First digit: " << first_digit << "\n";
//   cerr << "First nondigit: " << first_nondigit << "\n";
  
  
  num_digits = first_nondigit - first_digit;

  // read the offset

  offset = atoi(first_digit);

  // Allocate space

  current_filename = new char[strlen(filename)+1];
  filename_mask = new char[strlen(filename)+3];

  // Set up the mask
  
  strncpy(filename_mask, filename, first_digit - filename);

  char * pos = filename_mask + (int)(first_digit - filename);

  sprintf(pos, "%%0%dd", num_digits);
  strcat(pos, first_nondigit);

//   cerr << "Mask: " << filename_mask << "\n";

  // This sets up the width and height

  if(!image_reader->init(filename, 1))
    return;

  best_colorspace = image_reader->get_best_colorspace();
  
  video_format = *(image_reader->get_video_format());
  
  video_format.colorspace = best_colorspace;
  video_format.framerate = gv_global_config->singlepic_framerate;
  video_format.init();

  //  video_format.print(cerr);
    
  // Now check how many files we can read

  int index = 0;

  ifstream input;

  video_format.num_frames = 1;
  
  while(1)
    {
    sprintf(current_filename, filename_mask, index + offset + 1);
//     cerr << "Trying " << current_filename << "\n";
    input.open(current_filename);
    if(!input)
      break;
    input.close();
    video_format.num_frames++;
    index++;
    }
  total_time.set_from_frames(video_format.num_frames,
                             video_format.framerate);
  }

void GVInputSinglepic::fill_video_buf(GVVideoFrame * frame)
  {
  sprintf(current_filename, filename_mask, video_frames_written + offset);

  if(video_frames_written)
    image_reader->init(current_filename, 0);

  image_reader->load_image(frame);
      
  if(video_frames_written >= video_format.num_frames - 1)
    {
    video_completed = 1;
    image_reader->cleanup(1);
    }
  else
    image_reader->cleanup(0);
  }

// Create the trackname
// If you have a files test00001.jpg .. test00010.jpg, the trackname
// will be test00001-00010

void GVInputSinglepic::set_trackname()
  {  char * first_nondigit; 

  char * mask = new char[10];
  char * pos;

  char * pure_filename = strrchr(filename, G_DIR_SEPARATOR);
  if(!pure_filename)
    pure_filename = filename;
  else
    pure_filename++;
  
  if(trackname)
    g_free(trackname);

  trackname = (char*)g_malloc(strlen(pure_filename)+1+num_digits);
  
  first_nondigit = &(pure_filename[strlen(pure_filename)-1]);
  
  while(!isdigit(*first_nondigit))
    first_nondigit--;

  first_nondigit++;

  sprintf(mask, "%%0%dd", num_digits);

  strncpy(trackname, pure_filename, first_nondigit - pure_filename);

  pos = &(trackname[first_nondigit - pure_filename]);

  *pos = '-';
  pos++;
  
  sprintf(pos, mask, offset + video_format.num_frames);

  delete mask;
  
  }

void GVInputSinglepic::cleanup()
  {
  if(filename_mask)
    {
    delete filename_mask;
    filename_mask = (char*)0;
    }
  if(current_filename)
    {
    delete current_filename;
    current_filename = (char*)0;
    }
  if(image_reader)
    {
    gv_image_reader_loader->destroy_plugin(image_reader);
    image_reader = (GVImageReader*)0;
    }
  }

int GVInputSinglepic::colorspace_supported(GVColorspace colorspace)
  {
  if(image_reader)
    return image_reader->colorspace_supported(colorspace);
  else
    return 0;
  }

#define PLUGIN_NAME GVInputSinglepic
#include "interface.cc"
