/*****************************************************************

  i_test.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#ifdef __GNUG__
#pragma implementation "i_test.h"
#endif

#include <iostream>
using std::ostream;

#include "../gv.h"
#include "../gvconfig.h"
#include "../gtkutil.h"

#include <string.h>

#include "i_test.h"


// Test plugin for  gmerlin
// This produces a colorbar in all colorspaces
// Maybe, audio test signals will also be implemented


// Colorbar constants 

// The color order is as follows:
// white, yellow, cyan, green, magenta, reg. blue, black

static guchar __rgb24_pixels[8][3] =
{ 
  { 0xFF, 0xFF, 0xFF },
  { 0xFF, 0xFF, 0x00 },
  { 0x00, 0xFF, 0xFF },
  { 0x00, 0xFF, 0x00 },
  { 0xFF, 0x00, 0xFF },
  { 0xFF, 0x00, 0x00 },
  { 0x00, 0x00, 0xFF },
  { 0x00, 0x00, 0x00 }
};

static guchar __rgb16_pixels[8][2] = 
{
  {0xff, 0xff},
  {0xff, 0x07},
  {0xe0, 0xff},
  {0xe0, 0x07},
  {0x1f, 0xf8},
  {0x1f, 0x00},
  {0x00, 0xf8},
  {0x00, 0x00},
};

static guchar __bgr16_pixels[8][2] = 
{
  {0xff, 0xff},
  {0xe0, 0xff},
  {0xff, 0x07},
  {0xe0, 0x07},
  {0x1f, 0xf8},
  {0x00, 0xf8},
  {0x1f, 0x00},
  {0x00, 0x00},
};

static guchar __yuv_pixels[8][3] = 
{
  {0xff, 0x80, 0x80},
  {0xe1, 0x00, 0x94},
  {0xb2, 0xab, 0x00},
  {0x95, 0x2b, 0x15},
  {0x69, 0xd4, 0xea},
  {0x4c, 0x54, 0xff},
  {0x1d, 0xff, 0x6b},
  {0x00, 0x80, 0x80},
};

static GVPluginAttributes _attributes = 
  {
    "i_test",
    "Test Input Plugin",
    "Creates a colorbar in various colorspaces",
    (char*)0,
    GV_PLUGIN_HAS_AUDIO|
    GV_PLUGIN_HAS_VIDEO
  };

GVInputTest::GVInputTest() :
  GVInput(&_attributes)
  {
  video_format.flags = 0;
  config_window = (ConfigWindow*)0;
  get_config();
  
  }

GVInputTest::~GVInputTest()
  {
  if(config_window)
    delete config_window;
  put_config();
  }


void GVInputTest::fill_audio_buf(GVAudioFrame * frame)
  {
  
  }

void GVInputTest::fill_video_buf(GVVideoFrame * frame)
  {
  memcpy(frame->pixels, video_frame.pixels, video_format.buffer_size);
  }
  
void GVInputTest::configure()
  {
  if(!config_window)
    config_window = new ConfigWindow(this);
  config_window->run();
  }

void GVInputTest::cleanup()
  {
  
  }

int GVInputTest::colorspace_supported(GVColorspace c)
  {
  return (c == colorspace);
  }

void GVInputTest::get_config()
  {
  framerate = 25.0;
  width = 320;
  height = 240;

  gv_config->get_int(attributes->name, "Width", width);
  gv_config->get_int(attributes->name, "Height", height);
  gv_config->get_float(attributes->name, "Framerate", framerate);

  int i_tmp = (int)GVColorspaceRGB24;
  gv_config->get_int(attributes->name, "Colorspace", i_tmp);
  colorspace = (GVColorspace)i_tmp;
  
  }

void GVInputTest::put_config()
  {
  gv_config->put_int(attributes->name, "Width", width);
  gv_config->put_int(attributes->name, "Height", height);
  gv_config->put_int(attributes->name, "Colorspace", colorspace);
  gv_config->put_float(attributes->name, "Framerate", framerate);
  }

// Fill the playlist with various test signals

void GVInputTest::fill_playlist() 
  {
  gv_playlist->clear();
  }

void GVInputTest::init()
  {
  // Setup the video frame

  video_format.width = width;
  video_format.height = height;
  video_format.framerate = framerate;
  video_format.colorspace = colorspace;

  video_format.num_frames = 10000;
  flags = GV_PLUGIN_TRACK_HAS_VIDEO;
  best_colorspace = colorspace;
    
  video_format.init();
    
  video_frame.alloc(&video_format);

  int * start;

  start = new int[9];

  start[0] = 0;
  start[8] = width;
  
  for(int i = 1; i < 8; i++)
    start[i] = (i*width) / 8;
    
  switch(colorspace)
    {
    case GVColorspaceRGB16:
      for(int i = 0; i < height; i++)
        {
        for(int j = 0; j < 8; j++)
          {
          for(int k = start[j]; k < start[j+1]; k++)
            {
            video_frame.pixels_rows[i][2*k]   = __rgb16_pixels[j][0];
            video_frame.pixels_rows[i][2*k+1] = __rgb16_pixels[j][1];
            }
          }
        }
      break;
    case GVColorspaceBGR16:
      for(int i = 0; i < height; i++)
        {
        for(int j = 0; j < 8; j++)
          {
          for(int k = start[j]; k < start[j+1]; k++)
            {
            video_frame.pixels_rows[i][2*k]   = __bgr16_pixels[j][0];
            video_frame.pixels_rows[i][2*k+1] = __bgr16_pixels[j][1];
            }
          }
        }
      break;
    case GVColorspaceRGB24:

      for(int i = 0; i < height; i++)
        {
        for(int j = 0; j < 8; j++)
          {
          for(int k = start[j]; k < start[j+1]; k++)
            {
            video_frame.pixels_rows[i][3*k]   = __rgb24_pixels[j][0];
            video_frame.pixels_rows[i][3*k+1] = __rgb24_pixels[j][1];
            video_frame.pixels_rows[i][3*k+2] = __rgb24_pixels[j][2];
            }
          }
        }
      break;
    case GVColorspaceBGR24:
      for(int i = 0; i < height; i++)
        {
        for(int j = 0; j < 8; j++)
          {
          for(int k = start[j]; k < start[j+1]; k++)
            {
            video_frame.pixels_rows[i][3*k]   = __rgb24_pixels[j][2];
            video_frame.pixels_rows[i][3*k+1] = __rgb24_pixels[j][1];
            video_frame.pixels_rows[i][3*k+2] = __rgb24_pixels[j][0];
            }
          }
        }
      break;
    case GVColorspaceRGB32:
      for(int i = 0; i < height; i++)
        {
        for(int j = 0; j < 8; j++)
          {
          for(int k = start[j]; k < start[j+1]; k++)
            {
            video_frame.pixels_rows[i][4*k]   = __rgb24_pixels[j][0];
            video_frame.pixels_rows[i][4*k+1] = __rgb24_pixels[j][1];
            video_frame.pixels_rows[i][4*k+2] = __rgb24_pixels[j][2];
            }
          }
        }
      break;
    case GVColorspaceBGR32:
      for(int i = 0; i < height; i++)
        {
        for(int j = 0; j < 8; j++)
          {
          for(int k = start[j]; k < start[j+1]; k++)
            {
            video_frame.pixels_rows[i][4*k]   = __rgb24_pixels[j][2];
            video_frame.pixels_rows[i][4*k+1] = __rgb24_pixels[j][1];
            video_frame.pixels_rows[i][4*k+2] = __rgb24_pixels[j][0];
            }
          }
        }
      break;
      
    // 4:2:0 planar YUV, pixels[0]: y, pixels[1]: u, pixels[2]: v
    case GVColorspaceYUV_420_P:
      {
      guchar * y0 = video_frame.y;
      guchar * y1 = video_frame.y + width;
      guchar * u = video_frame.u;
      guchar * v = video_frame.v;

      for(int i = 0; i < height/2; i++)
        {
        for(int j = 0; j < 8; j++)
          {
          for(int k = start[j]; k < start[j+1]; k++)
            {
            *(y0++) = __yuv_pixels[j][0];
            *(y1++) = __yuv_pixels[j][0];
            if(!(k & 1))
              {
              *(u++)  = __yuv_pixels[j][1];
              *(v++)  = __yuv_pixels[j][2];
              }
            }
          }
        y0 += width;
        y1 += width;
        }
      }
      break;
    case GVColorspaceYUV_422_P:
      {
      guchar * y = video_frame.y;
      guchar * u = video_frame.u;
      guchar * v = video_frame.v;

      for(int i = 0; i < height; i++)
        {
        for(int j = 0; j < 8; j++)
          {
          for(int k = start[j]; k < start[j+1]; k++)
            {
            *(y++) = __yuv_pixels[j][0];
            if(!(k & 1))
              {
              *(u++)  = __yuv_pixels[j][1];
              *(v++)  = __yuv_pixels[j][2];
              }
            }
          }
        }
      }
      break;
    // 4:2:2 packed YUV, Ordering |Y0|U0|Y1|V0| ....
    case GVColorspaceYUY2:
      for(int i = 0; i < height; i++)
        {
        for(int j = 0; j < 8; j++)
          {
          for(int k = start[j]; k < start[j+1]; k++)
            {
            if(!(k & 1))
              {
              video_frame.pixels_rows[i][2*k]   = __yuv_pixels[j][0];
              video_frame.pixels_rows[i][2*k+1] = __yuv_pixels[j][1];
              }
            else
              {
              video_frame.pixels_rows[i][2*k] = __yuv_pixels[j][0];
              video_frame.pixels_rows[i][2*k+1] = __yuv_pixels[j][2];
              }
            }
          }
        }
      break;
    }
  delete start;
  }

static GVConfigMenuItemInt __colorspaces[] =
{
  { (int)GVColorspaceRGB16,     "RGB16" }, // 565
  { (int)GVColorspaceBGR16,     "BGR16" }, // 565
  { (int)GVColorspaceRGB24,     "RGB24" }, // 888
  { (int)GVColorspaceBGR24,     "BGR24" },
  { (int)GVColorspaceRGB32,     "RGB32" },
  { (int)GVColorspaceBGR32,     "BGR32" },
  { (int)GVColorspaceYUV_420_P, "YUV 420 Planar" },
  { (int)GVColorspaceYUV_422_P, "YUV 422 Planar" },
  { (int)GVColorspaceYUY2,      "YUV2" }
};

GVInputTest::ConfigWindow::ConfigWindow(GVInputTest * _plugin) :
  GVConfigWindow(_plugin->attributes->real_name),
  colorspace("Colorspace"),
  width("Width", 4, 1024, 0),
  height("Height", 4, 1024, 0),
  framerate("Framerate", 1.0, 60.0, 2)
  {
  plugin = _plugin;

  table = gvgtk_table_new(4, 2, 0);

  colorspace.create_menu(__colorspaces,
                         sizeof(__colorspaces)/sizeof(GVConfigMenuItemInt));

  colorspace.attach(table, 0, 0);
  width.attach(table, 1, 0);
  height.attach(table, 2, 0);
  framerate.attach(table, 3, 0);
  
  gtk_widget_show(table);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), table);
  }

GVInputTest::ConfigWindow::~ConfigWindow()
  {
  
  }

void GVInputTest::ConfigWindow::get_values()
  {
  colorspace.set(plugin->colorspace);
  width.set_value(plugin->width);
  height.set_value(plugin->height);
  framerate.set_value(plugin->framerate);
  }

void GVInputTest::ConfigWindow::apply_values()
  {
  plugin->colorspace = (GVColorspace)(colorspace.get_int());
  plugin->width = (int)(width.get_value());
  plugin->height = (int)(height.get_value());
  plugin->framerate = framerate.get_value();
  }

#define PLUGIN_NAME GVInputTest
#include "interface.cc"
