/*****************************************************************

  if_vorbis.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <iostream>

using std::ostream;

#include "../gv.h"
#include "../trackinfo.h"
#include <vorbis/vorbisfile.h>
#include "if_vorbis.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static GVPluginAttributes _attributes = 
  {
    "i_vorbis",
    "Ogg Vorbis Input Plugin",
    "See http://www.vorbis.com for more info",
    "ogg",
    GV_PLUGIN_HAS_AUDIO|GV_PLUGIN_CAN_SEEK|GV_PLUGIN_IS_FILE
  };

GVInputVorbis::GVInputVorbis() :
  GVInput(&_attributes)
  {
  audio_format.endian_format = GVNativeEndian;
  audio_format.sample_type = GVAudio_16_signed;
  audio_format.interleave_mode = GVInterleaveAll;
  flags = GV_PLUGIN_TRACK_HAS_AUDIO;
  track_info = (GVTrackInfo*)0;
  }

// Ogg vorbis input plugin

void GVInputVorbis::fill_audio_buf(GVAudioFrame * frame)
  {
  int bitstream = -1;
  long bytes_read = ov_read(&vorbisfile,
                            (char*)(frame->samples.s_8),
                            audio_format.samples_per_frame*audio_format.num_channels * 2,
                            audio_format.endian_format == GVBigEndian ? 1 : 0,
                            2,
                            1,
                            &bitstream);
  if(bytes_read < 0)
    {
    gv_engine->set_state(GVEngineError);
    frame->valid_samples = 0;
    }
  else if(bytes_read == 0)
    {
    audio_completed = 1;
    frame->valid_samples = 0;
    }
  else 
    frame->valid_samples = bytes_read/(2*audio_format.num_channels);
  }

const GVTrackInfo * GVInputVorbis::get_track_info()
  {
  return track_info;
  }

void GVInputVorbis::seek(float percentage)
  {
  ov_pcm_seek(&vorbisfile,
              (gint64)(audio_format.num_samples * percentage + 0.5));
  audio_samples_written = ov_pcm_tell(&vorbisfile);
  }

static char * _artist_key = "ARTIST=";
static char * _album_key = "ALBUM=";
static char * _title_key = "TITLE=";
// static char * _version_key = "VERSION=";
static char * _track_number_key = "TRACKNUMBER=";
// static char * _organization_key = "ORGANIZATION=";
static char * _genre_key = "GENRE=";
// static char * _description_key = "DESCRIPTION=";
// static char * _date_key = "DATE=";
// static char * _location_key = "LOCATION=";
// static char * _copyright_key = "COPYRIGHT=";

void GVInputVorbis::init()
  {
  vorbis_comment * comment;
  vorbis_info * info;

  FILE * file = fopen(filename, "r");

  if(!file)
    {
    error = GV_PLUGIN_ERROR_OPEN_FILE;
    return;
    }

  if(ov_open(file, &vorbisfile, (char*)0, 0))
    {
    fclose(file);
    error = GV_PLUGIN_ERROR_FILE_INVALID;
    return;
    }

  // Set up the audio format

  info = ov_info(&vorbisfile, -1);
  
  audio_format.samplerate = info->rate;
  audio_format.num_channels = info->channels;

  audio_format.num_samples = ov_pcm_total(&vorbisfile, -1);

  audio_format.samples_per_frame = 4096 / (audio_format.num_channels * 2);
    
  audio_format.init();

  total_time.set_from_frames((unsigned int)audio_format.num_samples,
                             (unsigned int)audio_format.samplerate);
  
  // TODO Set up the track info

  comment = ov_comment(&vorbisfile, -1);
  int key_len;
  int comment_added = 0;

    
  if(comment->comments)
    {
    track_info = new GVTrackInfo;
    for(int i = 0; i < comment->comments; i++)
      {
      key_len = strlen(_artist_key);
      if(!strncasecmp(comment->user_comments[i], _artist_key,
                      key_len))
        {
        track_info->set_artist(&(comment->user_comments[i][key_len]));
        continue;
        }
      key_len = strlen(_album_key);
      if(!strncasecmp(comment->user_comments[i], _album_key,
                      key_len))
        {
        track_info->set_album(&(comment->user_comments[i][key_len]));
        continue;
        }

      key_len = strlen(_title_key);
      if(!strncasecmp(comment->user_comments[i], _title_key,
                      key_len))
        {
        track_info->set_title(&(comment->user_comments[i][key_len]));
        continue;
        }
      
      key_len = strlen(_genre_key);
      if(!strncasecmp(comment->user_comments[i], _genre_key,
                      key_len))
        {
        track_info->set_genre(&(comment->user_comments[i][key_len]));
        continue;
        }

      key_len = strlen(_track_number_key);
      if(!strncasecmp(comment->user_comments[i], _track_number_key,
                      key_len))
        {
        track_info->track_number = atoi(&(comment->user_comments[i][key_len]));
        continue;
        }

      if(!(comment_added) && !strchr(comment->user_comments[i], '='))
        {
        track_info->set_comment(comment->user_comments[i]);
        comment_added = 1;
        continue;
        }
      }
    }
  }



void GVInputVorbis::cleanup()
  {
  ov_clear(&vorbisfile);
  if(track_info)
    {
    delete track_info;
    track_info = (GVTrackInfo*)0;
    }
  }

// Another stupidity:
// Some ogg files have their bitrate in bits/sec, others in kilobits/sec.

// This function shows every bitrate correct

static void __print_bitrate(ostream & output, long bitrate)
  {
  if(bitrate > 1000)
    output << bitrate/1000;
  else
    output << bitrate;
  }

void GVInputVorbis::print_info(ostream & output)
  {
  const vorbis_info * vinfo = ov_info(&vorbisfile, -1);
  if(!vinfo)
    {
    output << "No Info available";
    return;
    }
  output << "Version: " << vinfo->version << "\n";

  if((vinfo->bitrate_lower == vinfo->bitrate_nominal) &&
     (vinfo->bitrate_upper == vinfo->bitrate_nominal))
    {
    output << "Fixed Bitrate: ";
    __print_bitrate(output, vinfo->bitrate_nominal);
    output << " kb/s";
    }
  else
    {
    output << "Variable Bitrate: ";

    output << "Nominal: ";
    __print_bitrate(output, vinfo->bitrate_nominal);
    output << " kb/s, ";

    if(vinfo->bitrate_lower == -1)
      output << "Lower limit unknown, ";
    else
      {
      output << "Lower: ";
      __print_bitrate(output, vinfo->bitrate_lower);
      output << " kb/s, ";
      }
    if(vinfo->bitrate_lower == -1)
      output << "Upper limit unknown\n";
    else
      {
      output << "Upper: ";
      __print_bitrate(output, vinfo->bitrate_upper);
      output << " kb/s\n";
      }
    }
  }

#define PLUGIN_NAME GVInputVorbis
#include "interface.cc"
