/*****************************************************************

  ir_tga.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Reader for TGA images (used for reading animations with alpha channel
// produced by blender)

#ifdef __GNUG__
#pragma implementation "ir_tga.h"
#endif

#include <pthread.h>
#include <string.h>
#include <gtk/gtk.h>

#include <iostream>
#include <fstream>
using std::ifstream;
using std::cerr;
using std::ios;

#include "../gv.h"


// #include "../imagereader.h"
#include "ir_tga.h"

// #include "configwindow.h"
// #include "globalconfig.h"

#include <stdio.h>

static GVPluginAttributes _attributes = 
  {
    "ir_tga",
    "TGA loader",
    "Plugin for loading tga images",
    "tga",
    GV_PLUGIN_HAS_VIDEO|
    GV_PLUGIN_IS_FILE
  };

GVImageReaderTGA::GVImageReaderTGA() :
  GVImageReader(&_attributes)
  {
  colormap = (TgaColor*)0;
  }


guint16 GVImageReaderTGA::read_ushort()
  {
  guint16 ret;

  guint8 read_byte = 0x00;

  input.get((char&)read_byte);
  ret = read_byte;

  input.get((char&)read_byte);

  ret |= (read_byte << 8);
  return ret;
  }

gint16 GVImageReaderTGA::read_short()
  {
  gint16 ret;

  gint8 read_byte = 0x00;

  input.get((char&)read_byte);
  ret = read_byte;

  input.get((char&)read_byte);

  ret |= (read_byte << 8);
  return ret;
  }

guint32 GVImageReaderTGA::read_ulong()
  {
  guint32 ret;

  guint8 read_byte = 0x00;

  input.get((char&)read_byte);
  ret = read_byte;

  input.get((char&)read_byte);
  ret |= (read_byte << 8);

  input.get((char&)read_byte);
  ret |= (read_byte << 16);

  input.get((char&)read_byte);
  ret |= (read_byte << 24);
  return ret;
  }

gint32 GVImageReaderTGA::read_long()
  {
  gint32 ret;

  gint8 read_byte = 0x00;

  input.get((char&)read_byte);
  ret = read_byte;

  input.get((char&)read_byte);
  ret |= (read_byte << 8);

  input.get((char&)read_byte);
  ret |= (read_byte << 16);

  input.get((char&)read_byte);
  ret |= (read_byte << 24);
  return ret;
  }

guint8 GVImageReaderTGA::read_ubyte()
  {
  guint8 ret = 0x00;
  input.get((char&)ret);
  return ret;
  }

gint8  GVImageReaderTGA::read_byte()
  {
  gint8 ret = 0x00;
  input.get((char&)ret);
  return ret;
  }

GVImageReaderTGA::~GVImageReaderTGA()
  {
  cleanup(1);
  }

static const char * new_tga_signature = "TRUEVISION-XFILE.";

#define IMAGE_TYPE_NONE                     0
#define IMAGE_TYPE_UNCOMPRESSED_COLORMAP    1
#define IMAGE_TYPE_UNCOMPRESSED_TRUECOLOR   2
#define IMAGE_TYPE_UNCOMPRESSED_BLACK_WHITE 3

#define IMAGE_TYPE_RLE_COLORMAP    9
#define IMAGE_TYPE_RLE_TRUECOLOR   10
#define IMAGE_TYPE_RLE_BLACK_WHITE 11

int GVImageReaderTGA::init(const char * filename, int first = 1)
  {
  input.open(filename);
  if(!input)
    return 0;

//   if(first)
//     {
    // Check if we have a new or old TGA file
    
  input.seekg(-18, ios::end);

  char * signature = new char[18];
  input.read(signature, 18);
  if(!strcmp(signature, new_tga_signature))
    new_tga = 1;
  else
    new_tga = 0;
  delete signature;

  if(first)
    cerr << (new_tga ? "New TGA" : "Old TGA") << "\n";
    
  input.seekg(ios::beg);

  int id_length = read_ubyte();
  int colormap_included = read_ubyte();
  if(colormap_included > 1)
    return 0;

  if(first)
    cerr << (colormap_included ? "Found Colormap" : "No Colormap") << "\n";
        
  image_type = read_ubyte();

  // Return if we don't know the image type

  if(first)
    {
    switch(image_type)
      {
      case IMAGE_TYPE_UNCOMPRESSED_COLORMAP:
        cerr << "IMAGE_TYPE_UNCOMPRESSED_COLORMAP (not supported)\n";
        return 0;
        break;
      case IMAGE_TYPE_UNCOMPRESSED_TRUECOLOR:
        cerr << "IMAGE_TYPE_UNCOMPRESSED_TRUECOLOR\n";
        break;
      case IMAGE_TYPE_UNCOMPRESSED_BLACK_WHITE:
        cerr << "IMAGE_TYPE_UNCOMPRESSED_BLACK_WHITE\n";
        return 0;
        break;
      case IMAGE_TYPE_RLE_COLORMAP:
        cerr << "IMAGE_TYPE_RLE_COLORMAP\n";
        return 0;
        break;
      case IMAGE_TYPE_RLE_TRUECOLOR:
        cerr << "IMAGE_TYPE_RLE_TRUECOLOR\n";
        break;
      case IMAGE_TYPE_RLE_BLACK_WHITE:
        cerr << "IMAGE_TYPE_RLE_BLACK_WHITE\n";
        break;
      default:
        return 0;
        break;
      }
    }
    
  // Read the colomap information
    
  colormap_first_index = read_ushort();
  colormap_length = read_ushort();
  colormap_entry_size = read_ubyte();

  if(colormap_included)
    colormap = new TgaColor[colormap_length];
    
    // X origin and y-origin are ignored

  read_ushort();
  read_ushort();

    // Width and height
  if(first)
    {
    video_format.width  = read_ushort();
    video_format.height = read_ushort();
    cerr << "Size: " << video_format.width << "x"
         << video_format.height << "\n";
    }
  else
    {
    read_ushort();
    read_ushort();
    }

  if(first)
    {
    pixel_depth = read_ubyte();
    cerr << "Pixel depth " << pixel_depth << "\n";
    }
  else
    read_ubyte();
    
  int image_descriptor = read_ubyte();

  if(first)
    cerr << "image_descriptor: " << image_descriptor << "\n";

  if(first)
    {
    alpha_bits = image_descriptor & 0x0f;
    if(!(alpha_bits) && (pixel_depth == 32))
      alpha_bits = 8;
    bottom_to_top = !(image_descriptor & (1<<5));
    right_to_left = !!(image_descriptor & (1<<4));
    cerr << "Image direction: "
         << (bottom_to_top ? "Bottom to top, " : "Top to bottom, ")
         << (right_to_left ? "right to Left\n" : "left to right\n");
    }
    
  // Skip Image ID
    
  if(id_length)
    input.seekg(id_length, ios::cur);
    
    // Skip or read the colormap
    
  if(colormap_included)
    {
    switch(image_type)
      {
      case IMAGE_TYPE_UNCOMPRESSED_COLORMAP:
      case IMAGE_TYPE_RLE_COLORMAP:
        // TODO: Read the colormap
        cerr << "Colormapped Images not supported right now\n";
        return 0;                    
      case IMAGE_TYPE_UNCOMPRESSED_TRUECOLOR:
      case IMAGE_TYPE_UNCOMPRESSED_BLACK_WHITE:
      case IMAGE_TYPE_RLE_TRUECOLOR:
      case IMAGE_TYPE_RLE_BLACK_WHITE:
        // Skip the colormap
        input.seekg(colormap_length * (colormap_entry_size+7)/8,
                    ios::cur);
        break;
      }
    }
    
  if(first)
    {
    cerr << "Alpha Bits: " << alpha_bits << "\n";
    if(alpha_bits)
      best_colorspace = GVColorspaceRGBA32;
    else
      best_colorspace = GVColorspaceRGB24;
    video_format.colorspace = best_colorspace;
    video_format.num_frames = 1;
    video_format.init();
    }
  return 1;
  }

// Return true if colorspace is supported

int GVImageReaderTGA::colorspace_supported(GVColorspace colorspace)
 {
  if(colorspace == best_colorspace)
    return 1;
  else
    return 0;
  }

// Load image into the video frame structure. The frame must be
// allocated with the proper size by the caller!!

void GVImageReaderTGA::load_image(GVVideoFrame * frame)
  {
  int i_start;
  int i_end;
  int i_step;

  if(bottom_to_top)
    {
    i_start = video_format.height-1;
    i_end   = -1;
    i_step = -1;
    }
  else
    {
    i_start = 0;
    i_end = video_format.height;
    i_step = 1;
    }

  if(right_to_left)
    {
    switch(image_type)
      {
      case IMAGE_TYPE_UNCOMPRESSED_COLORMAP:
      case IMAGE_TYPE_UNCOMPRESSED_TRUECOLOR:
      case IMAGE_TYPE_UNCOMPRESSED_BLACK_WHITE:
        for(int i = i_start; i != i_end; i += i_step)
          read_scanline_right_left_uncompressed(frame->pixels_rows[i]);
        break;        
      case IMAGE_TYPE_RLE_COLORMAP:
      case IMAGE_TYPE_RLE_TRUECOLOR:
      case IMAGE_TYPE_RLE_BLACK_WHITE:
        for(int i = i_start; i != i_end; i += i_step)
          read_scanline_right_left_rle(frame->pixels_rows[i]);
        break;
      }
    }
  else
    {
    switch(image_type)
      {
      case IMAGE_TYPE_UNCOMPRESSED_COLORMAP:
      case IMAGE_TYPE_UNCOMPRESSED_TRUECOLOR:
      case IMAGE_TYPE_UNCOMPRESSED_BLACK_WHITE:
        for(int i = i_start; i != i_end; i += i_step)
          read_scanline_left_right_uncompressed(frame->pixels_rows[i]);
        break;        
      case IMAGE_TYPE_RLE_COLORMAP:
      case IMAGE_TYPE_RLE_TRUECOLOR:
      case IMAGE_TYPE_RLE_BLACK_WHITE:
        for(int i = i_start; i != i_end; i += i_step)
          read_scanline_left_right_rle(frame->pixels_rows[i]);
        break;
      }
    }
  }

// Cleanup all allocated resources. If the argument "last" is 0,
// don't clean resources, which can be reused by series of pictures
// in singlepicture plugins
  
void GVImageReaderTGA::cleanup(int last)
  {
  input.close();

  if(last)
    {
    if(colormap)
      {
      delete [] colormap;
      colormap = (TgaColor*)0;
      }
    }
  }

// Read a pixel, can be a colormap entry or a pixel in the
// image

void GVImageReaderTGA::read_pixel_raw(guchar * pixel, int depth)
  {
  switch(depth)
    {
    case 8:
      pixel[0] = read_ubyte();
      pixel[1] = pixel[0];
      pixel[2] = pixel[0];
      break;
    case 24:
      pixel[0] = read_ubyte();
      pixel[1] = read_ubyte();
      pixel[2] = read_ubyte();
      break;
    case 32:      // 32 bit images are asssumed to be RGBA
      pixel[0] = read_ubyte();
      pixel[1] = read_ubyte();
      pixel[2] = read_ubyte();
      pixel[3] = read_ubyte();
      break;
    default:
      break;
    }
  }

// Read a pixel with colormap lookup

void GVImageReaderTGA::read_pixel(guchar * pixel)
  {
  switch(image_type)
    {
    case IMAGE_TYPE_UNCOMPRESSED_TRUECOLOR:
    case IMAGE_TYPE_RLE_TRUECOLOR:
    case IMAGE_TYPE_UNCOMPRESSED_BLACK_WHITE:
    case IMAGE_TYPE_RLE_BLACK_WHITE:
      read_pixel_raw(pixel, pixel_depth);
      break;
    case IMAGE_TYPE_UNCOMPRESSED_COLORMAP:
    case IMAGE_TYPE_RLE_COLORMAP:
      break;
    }
  }

void
GVImageReaderTGA::read_scanline_left_right_uncompressed(guchar * out_pixels)
  {
  //  cerr << "read_scanline_left_right_uncompressed\n";
  for(int i = 0; i < video_format.width; i++)
    {
    read_pixel(out_pixels);
    out_pixels+=video_format.bytes_per_pixel;
    }
  }

void
GVImageReaderTGA::read_scanline_right_left_uncompressed(guchar * out_pixels)
  {
  //  cerr << "read_scanline_right_left_uncompressed\n";
  out_pixels += video_format.width*video_format.bytes_per_pixel;
  for(int i = 0; i < video_format.width; i++)
    {
    out_pixels-=video_format.bytes_per_pixel;
    read_pixel(out_pixels);
    }
  }

void
GVImageReaderTGA::read_scanline_left_right_rle(guchar * out_pixels)
  {
  int num_read = 0;
  int num_pixels;
  int rle_header;
  guchar tmp_pixel[4];
  //  cerr << "read_scanline_left_right_rle\n";
  
  while(1)
    {
    rle_header = read_ubyte();
    num_pixels = (rle_header & 0X7F) + 1;

    if(rle_header & 0x80) // Run lenght packet
      {
      read_pixel(tmp_pixel);
      for(int i = 0; i < num_pixels; i++)
        {
        memcpy(out_pixels, tmp_pixel, video_format.bytes_per_pixel);
        num_read++;
        out_pixels+=video_format.bytes_per_pixel;
        }
      }
    else // Raw packet
      {
      for(int i = 0; i < num_pixels; i++)
        {
        read_pixel(out_pixels);
        num_read++;
        out_pixels+=video_format.bytes_per_pixel;
        }
      }
    if(num_read >= video_format.width)
      break;
    }
  }

void
GVImageReaderTGA::read_scanline_right_left_rle(guchar * out_pixels)
  {
  int num_read = 0;
  int num_pixels;
  int rle_header;
  guchar tmp_pixel[4];

  out_pixels += video_format.width * video_format.bytes_per_pixel;
  //  cerr << "read_scanline_right_left_rle\n";
  
  while(1)
    {
    rle_header = read_ubyte();
    num_pixels = (rle_header & ~0X80) + 1;

    if(rle_header & 0x80) // Run lenght packet
      {
      read_pixel(tmp_pixel);
      for(int i = 0; i < num_pixels; i++)
        {
        out_pixels-=video_format.bytes_per_pixel;
        memcpy(out_pixels, tmp_pixel, video_format.bytes_per_pixel);
        num_read++;
        }
      }
    else // Raw packet
      {
      for(int i = 0; i < num_pixels; i++)
        {
        out_pixels-=video_format.bytes_per_pixel;
        read_pixel(out_pixels);
        num_read++;
        }
      }
    if(num_read >= video_format.width)
      break;
    }
  }

#define PLUGIN_NAME GVImageReaderTGA
#include "interface.cc"
