/*****************************************************************

  decoder.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <string.h>
#include <stdlib.h>

#include <iostream.h>

#include <glib.h>
#include "source.h"
#include "fifo.h"
#include "input.h"

#include "toc.h"

#include "audio.h"
#include "video.h"
#include "demuxer.h"

#include "decoder.h"

#include "mad.h"

#include "../../trackinfo.h"


GVMpegDecoder::GVMpegDecoder()
  {
  track_info = (GVTrackInfo*)0;
  input = (GVMpegInput*)0;
  audio_decoder = (GVMpegAudioDecoder*)0;
  video_decoder = (GVMpegVideoDecoder*)0;
  }

void GVMpegDecoder::set_input(GVMpegInput * _input)
  {
  input = _input;
  }

#define RIFF_DATA_HEADER 0x64617461

// Read the size of the next RIFF chunk

#define READ_CHUNK_SIZE ptr=input->fifo.read_data(4);\
                        chunk_size=ptr[3];\
                        chunk_size<<=8;\
                        chunk_size|=ptr[2];\
                        chunk_size<<=8;\
                        chunk_size|=ptr[1];\
                        chunk_size<<=8;\
                        chunk_size|=ptr[0];\
                        bytes_to_skip+=4

void GVMpegDecoder::skip_riff_header()
  {
  guint32 chunk_type;
  guint32 chunk_size;

  guchar * ptr;
  
  // There are 12 bytes until the first chunk

  guint bytes_to_skip = 12;
  input->fifo.read_data(12);

  // Read all chunks until the "data" chunk appears, which is our mp3 file

  chunk_type = input->fifo.read_32bits();
  bytes_to_skip += 4;
  
  while(chunk_type != RIFF_DATA_HEADER)
    {
    // read the chunk size
    READ_CHUNK_SIZE;
    input->fifo.read_data(chunk_size);
    bytes_to_skip+=chunk_size;
    
    chunk_type = input->fifo.read_32bits();
    bytes_to_skip += 4;
    }

  // Here, we have the data header, read the size and we are done
  
  READ_CHUNK_SIZE;
  input->add_start_offset(bytes_to_skip);

  cerr << "Found RIFF header, skipping " << bytes_to_skip << " Bytes\n";
  }

void GVMpegDecoder::seek(float percentage)
  {
  if(!(input->can_seek()))
    return;

  if(percentage > 1.0)
    percentage = 1.0;
  else if(percentage < 0.0)
    percentage = 0.0;
  
  switch(stream_type)
    {
    case GVMpegElementalAudio:  // Elemental audio stream
      audio_decoder->seek(percentage);
      break;
    case GVMpegElementalVideo:  // Elemental video stream
      video_decoder->seek(percentage);
      break;
    case GVMpegSystem:          // System stream (audio and video)
      demuxer->seek(percentage);
      break;
    case GVMpegIfo:             // Ifo file (points to several VOB files)
      cerr << "Seeking not supported yet";
    }
  }

// The following macros are for guint32 arguments

#define IS_MPEG_AUDIO_HEADER(h) ((h&0xFFE00000)==0xFFE00000)
#define IS_RIFF_HEADER(h)       (h==0x52494646)
#define IS_VIDEO_HEADER(h)      (h==0x000001b3)

#define IS_SYSTEM_HEADER(h)     (h==0x000001bb)
#define IS_PACK_START_CODE(h)   (h==0x000001ba)

// Check for audio stream

int GVMpegDecoder::check_audio()
  {
  guint32 header = input->fifo.get_32bits();

  if(IS_RIFF_HEADER(header))
    skip_riff_header();

  header = input->fifo.get_32bits();
  
  if(IS_MPEG_AUDIO_HEADER(header))
    return 1;
  else
    return 0;
  }

// TODO

int GVMpegDecoder::check_video()  // Check for video stream
  {
  guint32 header = input->fifo.get_32bits();
  return IS_VIDEO_HEADER(header);
  }

// TODO

int GVMpegDecoder::check_system() // Check for system stream
  {
  guint32 header = input->fifo.get_32bits();
  return IS_PACK_START_CODE(header);
  }

// TODO

int GVMpegDecoder::check_ac3()    // Check for IFO file
  {
  return 0;
  }

// Initialize decoder, return 1 on success

#define MAX_BYTES_TO_SKIP 1024

int GVMpegDecoder::init()
  {
  if(!input)
    return 0;
    
  //  input->init();

  // Skip any junk which could come before the data,
  // which occur in some nihilistic files

  //  skip_zeros();

  int num_skip = 0;
  int ok = 0;
    
  while(1)
    {
    if(check_audio())
      {
      stream_type = GVMpegElementalAudio;
      ok = 1;
      break;
      }
    else if(check_video())
      {
      stream_type = GVMpegElementalVideo;
      ok = 1;
      break;
      }
    else if(check_system())
      {
      stream_type = GVMpegSystem;
      ok = 1;
      break;
      }
    else if(check_ac3())
      {
      stream_type = GVMpegElementalAC3;
      ok = 1;
      break;
      }
    input->fifo.read_data(1);
    num_skip++;
    if(num_skip >= MAX_BYTES_TO_SKIP)
      break;
    }
  
  if(!ok)
    return 0;

  num_audio_streams = 0;
  num_video_streams = 0;
  
  switch(stream_type)
    {
    case GVMpegElementalAudio:
      cerr << "Detected mpeg audio stream\n";

      num_audio_streams = 1;
      audio_decoder = new GVMpegAudioDecoderMad(&(input->fifo));
      audio_decoder->init();

      // Connect the input to the decoder and build the toc
      audio_decoder->input = input;
      if(input->can_seek())
        audio_decoder->build_toc();
      // ID3Tags are read from the input
      track_info = input->get_track_info();
      break;

    case GVMpegElementalVideo:
      cerr << "Detected mpeg video stream\n";

      num_video_streams = 1;
      video_decoder = new GVMpegVideoDecoder(&(input->fifo));

      if(!video_decoder->init())
        return 0;

      // Connect the input to the decoder and build the toc
      video_decoder->input = input;
      if(input->can_seek())
        video_decoder->build_toc();
      break;
    case GVMpegSystem:
      cerr << "Detected mpeg system stream\n";
      demuxer = new GVMpegDemuxer(&(input->fifo));
      if(!demuxer->init())
        return 0;
      return 0;
      break;
    case GVMpegElementalAC3:
      return 0;
      break;
    default:
      // All other stuff isn't supported yet
      break;
    }
  return 1;
  }

void GVMpegDecoder::cleanup()
  {
  if(audio_decoder)
    {
    audio_decoder->cleanup();
    delete audio_decoder;
    audio_decoder = (GVMpegAudioDecoder*)0;
    }
  if(video_decoder)
    {
    video_decoder->cleanup();
    delete video_decoder;
    video_decoder = (GVMpegVideoDecoder*)0;
    }
//   if(input)
//     {
//     input->close();
//     input->cleanup();
//     }
  }
