/*****************************************************************

  mad.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <iostream.h>
#include <glib.h>
#include "mpeg.h"

#include <string.h>
#include "mad.h"
#include "toc.h"

extern "C" {
#include "../avifile/plugins/libmp3lame_audioenc/lame3.70/VbrTag.h"
}

#define BYTES_TO_SKIP 2

static int mpeg_bitrates[16][6] = {
  {   0,         0,      0,      0,      0,      0 },
  {  32000,  32000,  32000,  32000,  32000,   8000 },
  {  64000,  48000,  40000,  64000,  48000,  16000 },
  {  96000,  56000,  48000,  96000,  56000,  24000 },
  { 128000,  64000,  56000, 128000,  64000,  32000 },
  { 160000,  80000,  64000, 160000,  80000,  64000 },
  { 192000,  96000,  80000, 192000,  96000,  80000 },
  { 224000, 112000,  96000, 224000, 112000,  56000 },
  { 256000, 128000, 112000, 256000, 128000,  64000 },
  { 288000, 160000, 128000, 288000, 160000, 128000 },
  { 320000, 192000, 160000, 320000, 192000, 160000 },
  { 352000, 224000, 192000, 352000, 224000, 112000 },
  { 384000, 256000, 224000, 384000, 256000, 128000 },
  { 416000, 320000, 256000, 416000, 320000, 256000 },
  { 448000, 384000, 320000, 448000, 384000, 320000 },
  {      0,      0,      0,      0,      0,      0 }
};

static int mpeg_samplerates[3][3] = {
  { 44100, 48000, 32000 }, // MPEG1
  { 22050, 24000, 16000 }, // MPEG2
  { 11025, 12000, 8000 }   // MPEG2.5
};

static void print_mad_error(mad_error error)
  {
  switch(error)
    {
    case MAD_ERROR_BUFLEN:
      cerr << "input buffer too small (or EOF)";
      break;
    case MAD_ERROR_BUFPTR:
      cerr << "invalid (null) buffer pointer";
      break;
    case MAD_ERROR_NOMEM:
      cerr << "not enough memory";
      break;
    case MAD_ERROR_LOSTSYNC:
      cerr << "lost synchronization";
      break;
    case MAD_ERROR_BADLAYER:
      cerr << "reserved header layer value";
      break;
    case MAD_ERROR_BADBITRATE:
      cerr << "forbidden bitrate value";
      break;
    case MAD_ERROR_BADSAMPLERATE:
      cerr << "reserved sample frequency value";
      break;
    case MAD_ERROR_BADEMPHASIS:
      cerr << "reserved emphasis value";
      break;
    case MAD_ERROR_BADCRC:
      cerr << "CRC check failed";
      break;
    case MAD_ERROR_BADBITALLOC:
      cerr << "forbidden bit allocation value";
      break;
    case MAD_ERROR_BADSCALEFACTOR:
      cerr << "bad scalefactor index";
      break;
    case MAD_ERROR_BADFRAMELEN:
      cerr << "bad frame length";
      break;
    case MAD_ERROR_BADBIGVALUES:
      cerr << "bad big_values count";
      break;
    case MAD_ERROR_BADBLOCKTYPE:
      cerr << "reserved block_type";
      break;
    case MAD_ERROR_BADSCFSI:
      cerr << "bad scalefactor selection info";
      break;
    case MAD_ERROR_BADDATAPTR:
      cerr << "bad main_data_begin pointer";
      break;
    case MAD_ERROR_BADPART3LEN:
      cerr << "bad audio data length";
      break;
    case MAD_ERROR_BADHUFFTABLE:
      cerr << "bad Huffman table select";
      break;
    case MAD_ERROR_BADHUFFDATA:
      cerr << "Huffman data overrun";
      break;
    case MAD_ERROR_BADSTEREO:
      cerr << "incompatible block_type for JS";
      break;
    }
  }

#define MPEG_ID_MASK        0x00180000
#define MPEG_MPEG1          0x00180000
#define MPEG_MPEG2          0x00100000
#define MPEG_MPEG2_5        0x00000000

#define MPEG_LAYER_MASK     0x00060000
#define MPEG_LAYER_III      0x00020000
#define MPEG_LAYER_II       0x00040000
#define MPEG_LAYER_I        0x00060000
#define MPEG_PROTECTION     0x00010000
#define MPEG_BITRATE_MASK   0x0000F000
#define MPEG_FREQUENCY_MASK 0x00000C00
#define MPEG_PAD_MASK       0x00000200
#define MPEG_PRIVATE_MASK   0x00000100
#define MPEG_MODE_MASK      0x000000C0
#define MPEG_MODE_EXT_MASK  0x00000030
#define MPEG_COPYRIGHT_MASK 0x00000008
#define MPEG_HOME_MASK      0x00000004
#define MPEG_EMPHASIS_MASK  0x00000003

#define LAYER_I_SAMPLES       384
#define LAYER_II_III_SAMPLES 1152

int GVMpegAudioDecoderMad::get_frame_size()
  {
  guint32 header = fifo->get_32bits();

  if((header & 0xFFE00000) != 0xFFE00000)
    return 0;
    
  int bitrate_index = (header & MPEG_BITRATE_MASK) >> 12;

  int tmp_bitrate = 0;
  
  switch(audio_info.audio_mode)
    {
    case GVMpegAudioMPEG1:
      switch(audio_info.layer)
        {
        case 1:
          tmp_bitrate = mpeg_bitrates[bitrate_index][0];
          break;
        case 2:
          tmp_bitrate = mpeg_bitrates[bitrate_index][1];
          break;
        case 3:
          tmp_bitrate = mpeg_bitrates[bitrate_index][2];
          break;
        }
      break;
    case GVMpegAudioMPEG2:
    case GVMpegAudioMPEG2_5:
      {
      switch(audio_info.layer)
        {
        case 1:
          tmp_bitrate = mpeg_bitrates[bitrate_index][3];
          break;
        case 2:
          tmp_bitrate = mpeg_bitrates[bitrate_index][4];
          break;
        case 3:
          tmp_bitrate = mpeg_bitrates[bitrate_index][5];
          break;
        }
      }
    default: // This won't happen, but keeps gcc quiet
      break;
    }

  if(!tmp_bitrate)
    return 0;
  
  // cerr << "Protection bit set\n";
    
//     cerr << "No protection bit\n";
//   else

  int ret;
  
  if(audio_info.layer == 1)
    {
    ret = (12 * tmp_bitrate / audio_info.samplerate);
    if(header & MPEG_PAD_MASK)
      ret++;
    ret *= 4;
    }
  else
    {
    ret = 144 * tmp_bitrate / audio_info.samplerate;
    if(header & MPEG_PAD_MASK)
      ret++;
    }
  if(audio_info.audio_mode == GVMpegAudioMPEG2)
    ret /= 2;

  // For mpeg 2.5, this is not tested
  
  else if(audio_info.audio_mode == GVMpegAudioMPEG2_5)
    ret /= 4;

  guchar * ptr = fifo->get_data(ret + BYTES_TO_SKIP + 2);

  if(!ptr)
    return ret;
  
  ptr += ret;

  int skipped_bytes = 0;
  
  while(skipped_bytes <= BYTES_TO_SKIP)
    {
    if(((ptr[0] & 0xff) == 0xFF) && ((ptr[1] & 0xe0) == 0xE0))
      {
      if(skipped_bytes)
        cerr << "GVMpegAudioDecoderMad: Frame was "
             << skipped_bytes << " byte(s) too long\n";
      return ret;
      }
    else
      {
      ret++;
      ptr++;
      skipped_bytes++;
      }
    }
    
  return 0;
  }

void GVMpegAudioDecoderMad::init()
  {
  mad_stream_init(&stream);
  mad_frame_init(&frame);
  mad_synth_init(&synth);

  guchar * ptr = fifo->get_data(40);

  if(CheckVbrTag(ptr))
    {
    //    cerr << "Found VBR header\n";
    audio_info.vbr = 1;
    }
  else
    {
    audio_info.vbr = 0;
    //    cerr << "Fixed bitrate\n";
    }
  
  guint32 header = fifo->get_32bits();

  // CHECK if it's mpeg

  switch(header & MPEG_ID_MASK)
    {
    case MPEG_MPEG1:
      audio_info.audio_mode = GVMpegAudioMPEG1;
      break;
    case MPEG_MPEG2:
      audio_info.audio_mode = GVMpegAudioMPEG2;
      break;
    case MPEG_MPEG2_5:
      audio_info.audio_mode = GVMpegAudioMPEG2_5;
      break;
    }

  //  audio_info.vbr = 0;
  
  // Get layer
  
  switch(header & MPEG_LAYER_MASK)
    {
    case MPEG_LAYER_I:
      audio_info.layer = 1;
      break;
    case MPEG_LAYER_II:
      audio_info.layer = 2;
      break;
    case MPEG_LAYER_III:
      audio_info.layer = 3;
      break;
    }
    
  // Get the bitrate
  
  int bitrate_index = (header & MPEG_BITRATE_MASK) >> 12;

  if(!audio_info.vbr)
    {
    switch(audio_info.audio_mode)
      {
      case GVMpegAudioMPEG1:
        switch(audio_info.layer)
          {
          case 1:
            audio_info.bitrate = mpeg_bitrates[bitrate_index][0];
            break;
          case 2:
            audio_info.bitrate = mpeg_bitrates[bitrate_index][1];
            break;
        case 3:
          audio_info.bitrate = mpeg_bitrates[bitrate_index][2];
          break;
        }
        break;
      case GVMpegAudioMPEG2:
      case GVMpegAudioMPEG2_5:
        switch(audio_info.layer)
          {
          case 1:
            audio_info.bitrate = mpeg_bitrates[bitrate_index][3];
            break;
          case 2:
            audio_info.bitrate = mpeg_bitrates[bitrate_index][4];
            break;
          case 3:
            audio_info.bitrate = mpeg_bitrates[bitrate_index][5];
            break;
          }
      default: // This won't happen, but keeps gcc quiet
        break;
      }
    }
  else audio_info.bitrate = 0;
  
  // Get the sampling frequency

  int frequency_index = (header & MPEG_FREQUENCY_MASK) >> 10;

  switch(audio_info.audio_mode)
    {
    case GVMpegAudioMPEG1:
      audio_info.samplerate = mpeg_samplerates[0][frequency_index];
      break;
    case GVMpegAudioMPEG2:
      audio_info.samplerate = mpeg_samplerates[1][frequency_index];
      break;
    case GVMpegAudioMPEG2_5:
      audio_info.samplerate = mpeg_samplerates[2][frequency_index];
      break;
    }
  
  int mode_index = (header & MPEG_MODE_MASK) >> 6;

  switch(mode_index)
    {
    case 0:
      audio_info.stereo_mode = GVMpegStereoModeStereo;
      break;
    case 1:
      audio_info.stereo_mode = GVMpegStereoModeJStereo;
      break;
    case 2:
      audio_info.stereo_mode = GVMpegStereoModeDualChannel;
      break;
    case 3:
      audio_info.stereo_mode = GVMpegStereoModeMono;
      break;
    }

  audio_info.num_channels
    = (audio_info.stereo_mode == GVMpegStereoModeMono) ? 1 : 2;

  //  audio_info.frame_size = get_frame_size();
  
  //  cerr << "Frame size " << audio_info.frame_size << "\n";  

  switch(audio_info.layer)
    {
    case 1:
      audio_info.samples_per_frame = LAYER_I_SAMPLES;
      break;
    case 2:
    case 3:
      audio_info.samples_per_frame = LAYER_II_III_SAMPLES;
    }

  if(audio_info.audio_mode == GVMpegAudioMPEG2)
    audio_info.samples_per_frame /= 2;

  // Is this true??
  
  if(audio_info.audio_mode == GVMpegAudioMPEG2_5)
    audio_info.samples_per_frame /= 4;

  if(audio_info.vbr)
    {
    start_offset = get_frame_size();
    fifo->read_data(start_offset);
    }
  else
    start_offset = 0;
  
  decoded_samples = 0;
  audio_info.num_samples = 0;
  }

void GVMpegAudioDecoderMad::cleanup()
  {
//   if(stream_buffer)
//     {
//     delete stream_buffer;
//     stream_buffer = (guchar*)0;
//     }
  if(toc)
    {
    delete toc;
    toc = (GVMpegToc*)0;
    }
  mad_stream_finish(&stream);
  mad_frame_finish(&frame);
  mad_synth_finish(&synth);
  //  stream_buffer_size = 0;
  }

int GVMpegAudioDecoderMad::decode_frame(gint16 ** samples)
  {
  // For the case, that this is the last frame, we need a buffer to
  // satisfy mad's need for frame_size + MAD_BUFFER_GUARD bytes
  
  guchar * stream_buffer = (guchar*)0;

  //  cerr << "GVMpegAudioDecoderMad::decode_frame\n";
//   if(source->eof())
//     {
//     cerr << "EOF\n";
//     return 0;
//     }
  
  int frame_size = get_frame_size();

  if(!frame_size)
    {
    cerr << "No valid frame\n";
    return 0;
    }
  
  //  cerr << "Decode Frame: frame_size: " << frame_size << "\n";

  int bytes_available;

  fifo->get_data(frame_size + MAD_BUFFER_GUARD, &bytes_available);
  
  if((bytes_available >= 0) && (bytes_available < frame_size))
    {
    cerr << "Stream over\n";

    // There is still data left, but it's less than an entire frame
    return 0;
    }
  
  if((bytes_available >= 0) &&
     (bytes_available < frame_size + MAD_BUFFER_GUARD))
    {
    // The last frame must be copied to an internal buffer
    
    stream_buffer = new guchar[frame_size + MAD_BUFFER_GUARD];

    guchar * source_ptr = fifo->read_data(frame_size);
    memcpy(stream_buffer, source_ptr, frame_size);
    memset(&(stream_buffer[frame_size]), 0x00, MAD_BUFFER_GUARD); 
    mad_stream_buffer(&stream, source_ptr, frame_size + MAD_BUFFER_GUARD);
    }
  else
    {
    // Pass the original data pointer from the fifo

    guchar * source_ptr = fifo->get_data(frame_size + MAD_BUFFER_GUARD);
    mad_stream_buffer(&stream, source_ptr, frame_size + MAD_BUFFER_GUARD);

    // Advance the read pointer of the source
    fifo->read_data(frame_size);
    }
  
  //  mad_stream_sync(&stream);
  
  //  cerr << "Mad stream buffer: " << (void*)stream.buffer << "\n";
  
 
  if(mad_frame_decode(&frame, &stream) == -1)
    {
    cerr << "mad_frame_decode: stream.error: ";
    print_mad_error(stream.error);
    cerr << " (harmless after seeking)\n";

    mad_frame_mute(&frame);
    
//     if(stream_buffer)
//       delete stream_buffer;
//     return 0;
    }
  else
    mad_synth_frame(&synth, &frame);
    
  // Convert mad_fixed to gint16
  
  for(int i = 0; i < audio_info.num_channels; i++)
    {
    for(int j = 0; j < audio_info.samples_per_frame; j++)
      {
      if (synth.pcm.samples[i][j] >= MAD_F_ONE)
        synth.pcm.samples[i][j] = MAD_F_ONE - 1;
      else if (synth.pcm.samples[i][j] < -MAD_F_ONE)
        synth.pcm.samples[i][j] = -MAD_F_ONE;

      int sample = synth.pcm.samples[i][j] >> (MAD_F_FRACBITS + 1 - 16);
      
      samples[i][j] = (gint16)(sample);
      //      cerr << "Sample: " << samples[i][j] << "\n";
      }
    }

  //  mad_stream_finish(&stream);
  
  decoded_samples += audio_info.samples_per_frame;

  if(stream_buffer)
    delete stream_buffer;

  return audio_info.samples_per_frame;
  }

void GVMpegAudioDecoderMad::seek(float percentage)
  {
  guint frame_size;
  
  gint64 frame_to_seek = (int)(percentage * toc->num_frames);
  
  int int_percentage = (int)(percentage * 100.0);
  
   cerr << "Seeking to Frame " << frame_to_seek << " of "
        << toc->num_frames << "\n";
//   cerr << "Next frame in toc: "
//        << toc->toc_entry[int_percentage].frame_number << "\n";
  
  input->seek(toc->toc_entry[int_percentage].file_position, GVMpegSeekStart);
  
  gint64 frame_num = toc->toc_entry[int_percentage].frame_number;
  
//   cerr << "Skipping " << frame_to_seek - frame_num << " Frames\n";
  
  while(frame_num < frame_to_seek)
    {
    frame_size = get_frame_size();
//     cerr << "Frame num: " << frame_num
//          << " Frame 2 seek: " << frame_to_seek
//          << " Skipping " << frame_size << " Bytes\n";

    fifo->read_data(frame_size);
    frame_num++;
    }

  decoded_samples = audio_info.samples_per_frame * frame_to_seek;

//   mad_stream_finish(&stream);

//   mad_stream_sync(&stream);
//   mad_frame_finish(&frame);
//   mad_synth_finish(&synth);

//   mad_stream_init(&stream);
//   mad_frame_init(&frame);
//   mad_synth_init(&synth);

  //  frame.header.flags |= MAD_FLAG_INCOMPLETE;

  }

// The toc can be built, if
// 1. The deocoder is connected directly to the input
//    (Elemental Audio Stream)
// 2. The input is seekable
// None of these conditions are checked here!

#define FRAMES_TO_ALLOC 1000

// Number of bytes to skip until we find the next frame

void GVMpegAudioDecoderMad::build_toc()
  {
  int frame_size;

  //  gint64 bytes_left;
  
  toc = new GVMpegToc;
 
  // First Run: Count the frames

  toc->num_frames = 0;

  //  cerr << "Building toc, first run\n";

  guint64 max_num_frames = FRAMES_TO_ALLOC;

  guint64 * file_positions = new guint64[FRAMES_TO_ALLOC];

  GList * toc_list = (GList*)0;

  guint64 file_position = start_offset;

  while(!fifo->eof())
    {
    frame_size = get_frame_size();

    if(!frame_size)
      {
      // frame_size == 0 means, that there is no valid mpeg audio frame
//       cerr << "Found " << source->get_available_read_bytes()
//            << " Bytes nonstandard garbage at the end of the file\n";
      break;
      }
    
    //    bytes_left = source->get_available_read_bytes();

    int bytes_left;

    fifo->get_data(frame_size, &bytes_left);
    
    if(bytes_left < frame_size)
      {
//       cerr << "Found " << bytes_left << " Bytes of an incomplete mpeg frame ("
//            << "expected " << frame_size << " bytes)\n";
      break;
      }

    if(toc->num_frames + 1 > max_num_frames)
      {
      // Put the array into the list
      
      toc_list = g_list_append(toc_list, (gpointer)file_positions);

      // Allocate a new one

      file_positions = new guint64[FRAMES_TO_ALLOC];
      max_num_frames += FRAMES_TO_ALLOC;
      }

    file_positions[toc->num_frames % FRAMES_TO_ALLOC] = file_position;
    
    file_position += frame_size;
    
    toc->num_frames++;
    
    fifo->read_data(frame_size);
    }
  
  toc_list = g_list_append(toc_list, (gpointer)file_positions);
  
  // Second run: Fill the toc

  guint64 next_frame;

  guint64 * position_ptr;
  
  for(int i = 0; i < 100; i++)
    {
    // Get the next frame for the TOC

    next_frame = (guint64)(toc->num_frames*i/100.0);

    position_ptr = (guint64*)(g_list_nth_data(toc_list,
                                            next_frame/FRAMES_TO_ALLOC));
    
    toc->toc_entry[i].frame_number = next_frame;
    toc->toc_entry[i].file_position =
      position_ptr[next_frame % FRAMES_TO_ALLOC]; 
    }
  input->seek(start_offset, GVMpegSeekStart);   

  audio_info.num_samples = audio_info.samples_per_frame * toc->num_frames;

  // Free the list

  int imax = toc->num_frames / FRAMES_TO_ALLOC;
  for(int i = 0; i < imax; i++)
    {
    position_ptr = (guint64*)(g_list_nth_data(toc_list, i));
    delete position_ptr;
    }
  g_list_free(toc_list);
  
  //  cerr << "Built toc: " << toc->num_frames << " Frames\n";
  }

