#ifdef __GNUG__
#pragma implementation "mpeg_common.h"
#endif

#include <iostream>

using std::ostream;
using std::cerr;

#include "../gv.h"
#include "libgmerlin_mpeg/mpeg.h"
#include "mpeg_common.h"


void GVInputMpegCommon::print_info(ostream & output)
  {
  print_info_common(output);
  }


void GVInputMpegCommon::print_info_common(ostream & output)
  {
  switch(mpeg_decoder.get_stream_type())
    {
    case GVMpegElementalAudio:  // Elemental audio stream
      output << "Elemental Audio stream\n";
      break;
    case GVMpegElementalVideo:  // Elemental video stream
      output << "Elemental Video stream\n";
      break;
    case GVMpegSystem:          // System stream (audio and video)
      output << "System stream\n";
      break;
    case GVMpegIfo:             // Ifo file (points to several VOB files)
      output << "Ifo file\n";
      break;
    }

  if(flags & GV_PLUGIN_TRACK_HAS_AUDIO)
    {
    if(flags & GV_PLUGIN_TRACK_HAS_VIDEO)
      output << "Audio:\n";

    const GVMpegAudioInfo * audio_info
      = mpeg_decoder.audio_decoder->get_audio_info();

    switch(audio_info->audio_mode)
      {
      case GVMpegAudioMPEG1:
        output << "MPEG 1 audio layer " << audio_info->layer;
        break;
      case GVMpegAudioMPEG2:
        output << "MPEG 2 audio layer " << audio_info->layer;
        break;
      case GVMpegAudioMPEG2_5:
        output << "MPEG 2.5 audio layer " << audio_info->layer;
        break;
      }
    if(audio_info->vbr)
      output << " (Variable bitrate, ";
    else
      output << " (Bitrate: " << audio_info->bitrate/1000 << " kb/s ";

    switch(audio_info->stereo_mode)
      {
      case GVMpegStereoModeMono:
        output << "Mono";
        break;
      case GVMpegStereoModeStereo:
        output << "Stereo";
        break;
      case GVMpegStereoModeJStereo:
        output << "Joint Stereo";
        break;
      case GVMpegStereoModeDualChannel:
        output << "Dual Channel";
        break;
      }
    output << ")";

    }

  // TODO: video info

  if(flags & GV_PLUGIN_TRACK_HAS_VIDEO)
    {
    if(flags & GV_PLUGIN_TRACK_HAS_AUDIO)
      output << "Video:\n";
    }
  
  }

const GVTrackInfo * GVInputMpegCommon::get_track_info()
  {
  return mpeg_decoder.get_track_info();
  }

void GVInputMpegCommon::fill_audio_buf(GVAudioFrame * frame)
  {
  frame->valid_samples
    = mpeg_decoder.audio_decoder->decode_frame(frame->channels.s_16);
  if(!frame->valid_samples)
    audio_completed = 1;
  }

void GVInputMpegCommon::fill_video_buf(GVVideoFrame * frame)
  {
  mpeg_decoder.video_decoder->decode_frame(frame->y, frame->u, frame->v);
  if(mpeg_decoder.video_decoder->eof())
    video_completed = 1;
  }

GVInputMpegCommon::GVInputMpegCommon(const GVPluginAttributes * _attributes) :
  GVInput(_attributes)
  {
  // Set some values which will never change

  audio_format.interleave_mode = GVInterleaveNone;
  audio_format.sample_type = GVAudio_16_signed;

  best_colorspace = GVColorspaceYUV_420_P;
  }

int GVInputMpegCommon::colorspace_supported(GVColorspace colorspace)
  {
  return (int)((colorspace == GVColorspaceYUV_420_P));
  }

void GVInputMpegCommon::cleanup_common()
  {
  mpeg_decoder.cleanup();
  cerr << "Video frames written: " << video_frames_written << "\n";
  }

void GVInputMpegCommon::init_common()
  {
  flags = 0;

  GVTime audio_time;
  GVTime video_time;
  
  if(!mpeg_decoder.init())
    {
    error = GV_PLUGIN_ERROR_FILE_INVALID;
    return;
    }

  // TODO: Check for number of streams
  
  switch(mpeg_decoder.get_stream_type())
    {
    case GVMpegElementalAudio:  // Elemental audio stream
      flags = GV_PLUGIN_TRACK_HAS_AUDIO;
      break;
    case GVMpegElementalVideo:  // Elemental video stream
      flags = GV_PLUGIN_TRACK_HAS_VIDEO;
      break;
    case GVMpegSystem:          // System stream (audio and video)
      flags = GV_PLUGIN_TRACK_HAS_VIDEO|GV_PLUGIN_TRACK_HAS_AUDIO;
      break;
    case GVMpegIfo:             // Ifo file (points to several VOB files)
      flags = GV_PLUGIN_TRACK_HAS_VIDEO|GV_PLUGIN_TRACK_HAS_AUDIO;
      break;
    }

  // Initialize audio
  
  if(flags & GV_PLUGIN_TRACK_HAS_AUDIO)
    {
    const GVMpegAudioInfo * audio_info
      = mpeg_decoder.audio_decoder->get_audio_info();

    audio_format.samplerate = audio_info->samplerate;
    audio_format.num_channels = audio_info->num_channels;
    audio_format.samples_per_frame = audio_info->samples_per_frame;
    audio_format.num_samples = audio_info->num_samples;
    audio_format.init();

    if(audio_format.num_samples)
      audio_time.set_from_samples(audio_format.num_samples,
                                  audio_format.samplerate);
    else
      audio_time.set_undefined();
    }
  else
    audio_time.reset();

  // Initialize video
  
  if(flags & GV_PLUGIN_TRACK_HAS_VIDEO)
    {
    const GVMpegVideoInfo * video_info
      = mpeg_decoder.video_decoder->get_video_info();
    video_format.framerate = video_info->framerate;
    //    video_format.framerate = 1.0;

    video_format.width = video_info->width;
    video_format.height = video_info->height;

    // TODO: Get the number of frames!!

    video_format.num_frames = video_info->num_frames;

    if(video_format.num_frames)
      video_time.set_from_frames(video_format.num_frames,
                                 video_format.framerate);
    else
      video_time.set_undefined();
    video_format.colorspace = GVColorspaceYUV_420_P;
    video_format.init();
    
    }
  else
    video_time.reset();

  if(audio_time.undefined())
    total_time.set_undefined();
  else if(video_time.undefined())
    total_time.set_undefined();
  else
    total_time = (video_time > audio_time) ? video_time : audio_time;
  }

void GVInputMpegCommon::seek(float percentage)
  {
  mpeg_decoder.seek(percentage);
  }
