/*****************************************************************

  o_quicktime.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Quicktime file writer

#ifdef __GNUG__
#pragma implementation "o_openquicktime.h"
#endif

#include <iostream>

using std::ostream;

#include <glib.h>
#include <string.h>

#include "../gv.h"

#include <openquicktime/openquicktime.h>
#include <openquicktime/colormodels.h>

#include "o_openquicktime.h"

#include "../gvconfig.h"

#include "../gtkutil.h"

#include "openquicktime_common.h"

static GVPluginAttributes _attributes = 
  {
    "o_openquicktime",
    "Openquicktime Output Plugin",
    "Formatwriter for Quicktime files. See\n\
http://openquicktime.sourceforge.net\n\
for what this can do",
    "mov",
    GV_PLUGIN_HAS_AUDIO|GV_PLUGIN_HAS_VIDEO|GV_PLUGIN_IS_FILE
  };

#define MP3_FIXED_BITRATE_CODE "MP3_FIXED_BITRATE"

GVOutputOpenquicktime::GVOutputOpenquicktime() :
  GVOutput(&_attributes)
  {
  config_window = (ConfigWindow*)0;
  quicktime_file = (quicktime_t*)0;

  video_fourcc = (char*)0;
  audio_fourcc = (char*)0;

  quicktime_info = (char*)0;
  quicktime_copyright = (char*)0;
  
  get_config();
  }

GVOutputOpenquicktime::~GVOutputOpenquicktime()
  {
  cleanup();
  put_config();

  if(config_window)
    delete config_window;

  if(video_fourcc)
    g_free(video_fourcc);
  if(audio_fourcc)
    g_free(audio_fourcc);
  if(quicktime_info)
    g_free(quicktime_info);
  if(quicktime_copyright)
    g_free(quicktime_copyright);
  }

void GVOutputOpenquicktime::configure()
  {
  if(!config_window)
    config_window = new ConfigWindow(this);
  config_window->run();
  }

void GVOutputOpenquicktime::init()
  {
  // Open the file

  flags = 0;
  
  if(!create_outfilename("mov"))
    return;
  
  quicktime_file = quicktime_open(filename, 0, 1);

  //  quicktime_set_copyright(quicktime_file, quicktime_copyright);

  create_trackname();

  //  quicktime_set_name(quicktime_file, trackname);
  //  quicktime_set_info(quicktime_file, quicktime_info);
  
  if(input->track_has_audio() && do_audio)
    {
    audio_format = *(input->get_audio_format());
    
    //    cerr << "audio_fourcc: " << audio_fourcc << "\n";
    
    int audio_bits = (audio_format.bytes_per_sample*8);
    //    cerr << "audio_bits: " << audio_bits << "\n";

    if(!strcmp(audio_fourcc, MP3_FIXED_BITRATE_CODE))
      {
      char tmp_audio_fourcc[5];
      tmp_audio_fourcc[0] = 'm';
      tmp_audio_fourcc[1] = 's';
      tmp_audio_fourcc[2] = 0;
      tmp_audio_fourcc[3] = 0x55;
      tmp_audio_fourcc[4] = 0;
      quicktime_set_audio(quicktime_file, audio_format.num_channels,
                          audio_format.samplerate, audio_bits,
                          tmp_audio_fourcc);
      }
    else
      quicktime_set_audio(quicktime_file, audio_format.num_channels,
                          audio_format.samplerate, audio_bits, audio_fourcc);
    
    audio_format.endian_format = GVNativeEndian;

    // The next stuff comes from quicktime4linux
    // It will be here until openquicktime supports these
    // audio codecs
    
    if(!strcmp(audio_fourcc, QUICKTIME_RAW))
      {
      audio_format.sample_type = GVAudio_8_unsigned;
      audio_format.interleave_mode = GVInterleaveAll;
      write_raw_audio = 1;
      }
    else if(!strcmp(audio_fourcc, QUICKTIME_TWOS))
      {
      audio_format.sample_type = (audio_bits == 16) ?
        GVAudio_16_signed : GVAudio_8_signed;
      audio_format.endian_format = GVBigEndian;
      audio_format.interleave_mode = GVInterleaveAll;
      write_raw_audio = 1;
      }
    else
      {
      audio_format.interleave_mode = GVInterleaveNone;
      write_raw_audio = 0;
      }
    if(!quicktime_supported_audio(quicktime_file, 0))
      {
      g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
            "Cannot encode audio, check settings");
      return;
      }
    audio_format.init();

    flags |= GV_PLUGIN_TRACK_HAS_AUDIO;
    }

  if(input->track_has_video() && do_video)
    {
    video_format = *(input->get_video_format());
    quicktime_set_video(quicktime_file, 1,
                        video_format.width, video_format.height,
                        video_format.framerate, video_fourcc); 

    if(!strcmp(video_fourcc, QUICKTIME_JPEG))
      {
      if((video_format.width % 16)||(video_format.height % 16))
        {
        g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
              "Width and height must be multiples of 16 for codec %s",
              QUICKTIME_JPEG);
        cleanup();
        return;
        }
      quicktime_set_video_param (quicktime_file, 0,
                                 "quality", (void*)(&jpeg_quality));
      }
    
    //    quicktime_set_depth(quicktime_file, 24, 0);
    
    if(!quicktime_supported_video(quicktime_file, 0))
      {
      g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
            "Cannot encode video, check settings");
      cleanup();
      return;
      }

    // Select a colormodel

    GVColorspace colorspace = input->get_best_colorspace();
    int quicktime_colormodel = get_quicktime_colorspace(colorspace);

    int video_ok = 1;
    if(quicktime_writes_cmodel(quicktime_file,
                               quicktime_colormodel, 0))
      {
      video_format.colorspace = colorspace;
      input->set_colorspace(colorspace);
//       quicktime_set_cmodel(quicktime_file,
//                            quicktime_colormodel);
      }
    else
      {
      // Input should have it's best colorspace
      input->set_colorspace(colorspace);

      if(quicktime_writes_cmodel(quicktime_file,
                                 get_quicktime_colorspace(GVColorspaceYUV_420_P), 0))
        video_format.colorspace = GVColorspaceYUV_420_P;

      else if(quicktime_writes_cmodel(quicktime_file,
                                      get_quicktime_colorspace(GVColorspaceYUV_422_P), 0))
        video_format.colorspace = GVColorspaceYUV_422_P;
      
      else if(quicktime_writes_cmodel(quicktime_file,
                                      get_quicktime_colorspace(GVColorspaceYUY2), 0))
        video_format.colorspace = GVColorspaceYUY2;
      else if(quicktime_writes_cmodel(quicktime_file,
                                      get_quicktime_colorspace(GVColorspaceRGB24), 0))
        video_format.colorspace = GVColorspaceRGB24;

      else
        video_ok = 0;

      }

    if(video_ok)
      {
      video_format.init();
      flags |= GV_PLUGIN_TRACK_HAS_VIDEO;
      }
    else
      g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
            "Cannot Encode Video, check settings");
    }
  
  }

void GVOutputOpenquicktime::cleanup()
  {
  if(quicktime_file)
    {
   quicktime_close(quicktime_file);
    quicktime_file = (quicktime_t*)0;
    }
  }

void GVOutputOpenquicktime::put_frame(GVAudioFrame * frame)
  {
  if(write_raw_audio)
    quicktime_write_audio(quicktime_file, (char*)(frame->samples.s_8),
                          frame->valid_samples, 0); 
  else
    quicktime_encode_audio(quicktime_file,
                           frame->channels.s_16, (float **)0,
                           frame->valid_samples); 
  }

void GVOutputOpenquicktime::put_frame(GVVideoFrame * frame)
  {
  quicktime_encode_video(quicktime_file,
                         get_quicktime_colorspace(video_format.colorspace),
                         frame->pixels_rows, 0);
  }

#define SECTION attributes->name

void GVOutputOpenquicktime::get_config()
  {
  const char * test = QUICKTIME_JPEG;

  gv_config->get_string(SECTION, "VideoCodec", &test);

  if(video_fourcc)
    g_free(video_fourcc);
  video_fourcc = g_strdup(test);

  test = QUICKTIME_TWOS;

  gv_config->get_string(SECTION, "AudioCodec", &test);

  if(audio_fourcc)
    g_free(audio_fourcc);
  audio_fourcc = g_strdup(test);

  test = "Don't know :-)";

  gv_config->get_string(SECTION, "Copyright", &test);

  if(quicktime_copyright)
    g_free(quicktime_copyright);
  quicktime_copyright = g_strdup(test);

  test = "Created by Gmerlin";

  gv_config->get_string(SECTION, "Info", &test);

  if(quicktime_info)
    g_free(quicktime_info);
  quicktime_info = g_strdup(test);

  jpeg_quality = 75;

  gv_config->get_int(SECTION, "JPEGQuality", jpeg_quality);
  }

void GVOutputOpenquicktime::put_config()
  {
  gv_config->put_string(SECTION, "VideoCodec", video_fourcc);
  gv_config->put_string(SECTION, "AudioCodec", audio_fourcc);
  gv_config->put_string(SECTION, "Copyright", quicktime_copyright);
  gv_config->put_string(SECTION, "Info", quicktime_info);
  gv_config->put_int(SECTION, "JPEGQuality", jpeg_quality);
  }

// Config Window

static GVConfigMenuItemString __quicktime_video_codecs [] =
{
//  { QUICKTIME_RAW, "Raw RGB"},
  { QUICKTIME_JPEG, "JPEG"},
  { QUICKTIME_3IVX, "3ivx"}
    //  { QUICKTIME_PNG, "PNG" },
//  { QUICKTIME_MJPA, "MJPA" },
//  { QUICKTIME_YUV2, "YUV 4:2:2" },
//  { QUICKTIME_YUV4, "YUV 4:2:0 (Not compatible!)" }
};

static GVConfigMenuItemString __quicktime_audio_codecs [] =
{
//  { QUICKTIME_RAW,  "RAW" },
//  { QUICKTIME_IMA4, "IMA4" },
//  { QUICKTIME_TWOS, "Twos compliment" },
//  { QUICKTIME_ULAW, "ulaw" }
{ QUICKTIME_MP3, "Mp3 (Variable Bitrate)"},
{ MP3_FIXED_BITRATE_CODE, "Mp3 (Fixed Bitrate)"}
};

GVOutputOpenquicktime::ConfigWindow::ConfigWindow(GVOutputOpenquicktime * _plugin) :
  GVConfigWindow(_plugin->attributes->real_name),
  audio_codec("Audio Codec"),
  video_codec("Video Codec"),
  jpeg_quality("JPEG Quality", 1.0, 100.0, 0),
  quicktime_copyright("Copyright"),
  quicktime_info("Info")
  {
  plugin = _plugin;
  table = gvgtk_table_new(5, 2, 0);
    
  audio_codec.create_menu(__quicktime_audio_codecs,
                          sizeof(__quicktime_audio_codecs)/
                          sizeof(GVConfigMenuItemString));
  
  video_codec.create_menu(__quicktime_video_codecs,
                          sizeof(__quicktime_video_codecs)/
                          sizeof(GVConfigMenuItemString));
  
  audio_codec.attach(table, 0, 0);
  video_codec.attach(table, 1, 0);
  jpeg_quality.attach(table, 2, 0);
  
  quicktime_copyright.attach(table, 3, 0);
  quicktime_info.attach(table, 4, 0);

  gtk_widget_show(table);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), table);

  }

void GVOutputOpenquicktime::ConfigWindow::get_values()
  {
  audio_codec.set(plugin->audio_fourcc);
  video_codec.set(plugin->video_fourcc);
  jpeg_quality.set_value(plugin->jpeg_quality);
    
  quicktime_copyright.set_string(plugin->quicktime_copyright);
  quicktime_info.set_string(plugin->quicktime_info);
    
  }

void GVOutputOpenquicktime::ConfigWindow::apply_values()
  {
  if(plugin->audio_fourcc)
    g_free(plugin->audio_fourcc);
  plugin->audio_fourcc = g_strdup(audio_codec.get_string());

  if(plugin->video_fourcc)
    g_free(plugin->video_fourcc);
  plugin->video_fourcc = g_strdup(video_codec.get_string());
    
  if(plugin->quicktime_copyright)
    g_free(plugin->quicktime_copyright);
  plugin->quicktime_copyright = g_strdup(quicktime_copyright.get_string());

  if(plugin->quicktime_info)
    g_free(plugin->quicktime_info);
  plugin->quicktime_info = g_strdup(quicktime_info.get_string());

  plugin->jpeg_quality = jpeg_quality.get_value_int();
  }

GVOutputOpenquicktime::ConfigWindow::~ConfigWindow()
  {
  
  }

#define PLUGIN_NAME GVOutputOpenquicktime
#include "interface.cc"
