/*****************************************************************

  oa_esd.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// ESounD output driver

// Most of the esd stuff was taken from libao-0.6.0, which can be found
// at the Ogg/Vorbis download site

#include <iostream> // print_info()

using std::ostream;
using std::cerr;

#include "../gv.h"
#include "../gvconfig.h"
#include "../configwindow.h"

#include "../gtkutil.h"

#include "oa_esd.h"

#include <esd.h>
#include <unistd.h> // write


static GVPluginAttributes _attributes =
  {
    "oa_esd",
    "EsounD output driver",
    "Audio output driver for the\n\
Enlighted Sound Daemon",
    NULL,
    GV_PLUGIN_HAS_AUDIO|
    GV_PLUGIN_NEEDS_REALTIME
  };

#define SECTION attributes->name

void GVOutputAudioEsound::get_config()
  {
  const char * tmp_host = "";
  gv_config->get_string(SECTION, "Host", &tmp_host);
  set_filename(tmp_host);
  }

void GVOutputAudioEsound::put_config()
  {
  gv_config->put_string(SECTION, "Host", filename);
  }

GVOutputAudioEsound::GVOutputAudioEsound() :
  GVOutputAudio(&_attributes)
  {
  esd_socket = -1;
  esd_pause = 0;
  config_window = (ConfigWindow*)0;
  get_config();
  }

GVOutputAudioEsound::~GVOutputAudioEsound()
  {
  if(config_window)
    delete config_window;
  put_config();
  }

void GVOutputAudioEsound::configure()
  {
  if(!config_window)
    config_window = new ConfigWindow(this);
  config_window->run();
  }

void GVOutputAudioEsound::init()
  {
  // Set the things, we should always set

  int esd_bits = 0;
  int esd_channels;
  int esd_mode = ESD_STREAM;
  int esd_func = ESD_PLAY;

  audio_format.interleave_mode = GVInterleave2;
  
  const GVAudioFormat * input_format = input->get_audio_format();

  audio_format.num_samples = input_format->num_samples;
  audio_format.samples_per_frame = input_format->samples_per_frame;
  
  switch(input_format->sample_type)
    {
    case GVAudio_8_signed:
    case GVAudio_8_unsigned:
      audio_format.sample_type = GVAudio_8_signed;
      esd_bits = ESD_BITS8;
      break;
    case GVAudio_16_signed:
    case GVAudio_16_unsigned:
      audio_format.sample_type = GVAudio_16_signed;
      esd_bits = ESD_BITS16;
      break;
    }

  if(input_format->num_channels > 1)
    {
    audio_format.num_channels = 2;
    esd_channels = ESD_STEREO;
    }
  else
    {
    audio_format.num_channels = 1;
    esd_channels = ESD_MONO;
    }

  audio_format.samplerate = input_format->samplerate;

  esd_format = esd_bits | esd_channels | esd_mode | esd_func;  
  audio_format.init();
  open();
  }

void GVOutputAudioEsound::reset()
  {
  if(esd_socket < 0)
    return;

  // Close and open again

  cerr << "Resetting esd socket...";
    
  // We wait until nothing is inside any buffers
    
   GVTime delay_time;
   delay_time.seconds = 0;
   delay_time.nanoseconds = 100000000; // 100 ms

//   int delay = esd_get_latency(esd_socket);
  
//   while(delay > 0)
//     {
//     cerr << "Delay: " << delay << "\n";
   delay_time.sleep();
//     delay = esd_get_latency(esd_socket);
//     }
  close(esd_socket);
  esd_socket = -1;
  esd_pause = 1;
  
  cerr << "done\n";
  }

int GVOutputAudioEsound::open()
  {
  if(*filename == '\0')
    {
    esd_socket = esd_play_stream(esd_format, audio_format.samplerate,
                                 (char*)0,
                                 "gmerlin output");
    if(esd_socket < 0)
      {
      g_log(PACKAGE,
            G_LOG_LEVEL_CRITICAL,
            "Cannot connect to local esound daemon");
      error = GV_PLUGIN_ERROR_CONNECT;
      return 0;
      }
    }
  else
    {
    esd_socket = esd_play_stream(esd_format,
                                 audio_format.samplerate,
                                 filename,
                                 "gmerlin output");
    if(esd_socket < 0)
      {
      g_log(PACKAGE,
            G_LOG_LEVEL_CRITICAL, "Cannot connect to esound host %s",
            filename);
      error = GV_PLUGIN_ERROR_CONNECT;
      return 0;
      }
    }

  return 1;
  }


void GVOutputAudioEsound::put_frame(GVAudioFrame * frame)
  {
  if(esd_pause)
    {
    cerr << "Reopening esd socket...";
    open();
    esd_pause = 0;
    cerr << "done\n";
    }

  write(esd_socket, frame->samples.s_8,
        frame->valid_samples * audio_format.bytes_per_sample *
        audio_format.num_channels);
  }

void GVOutputAudioEsound::cleanup()
  {
  if(esd_socket > 0)
    {
    esd_close(esd_socket);
    esd_socket = -1;
    }
  }

void GVOutputAudioEsound::print_info(ostream & output)
  {
  if(*filename == '\0')
    output << "Streaming to local esd";
  else
    output << "Streaming to esd at " << filename;
  }

GVOutputAudioEsound::ConfigWindow::ConfigWindow(GVOutputAudioEsound * _plugin) :
  GVConfigWindow(_plugin->attributes->real_name),
  host("Remote host (leave empty for local)")
  {
  plugin = _plugin;
  GtkWidget * table;
  table = gvgtk_table_new(1, 2, 0);
  host.attach(table, 0, 0);
  gtk_widget_show(table);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), table);   
  gtk_widget_show(mainbox);   
  }

GVOutputAudioEsound::ConfigWindow::~ConfigWindow()
  {
  
  }

void GVOutputAudioEsound::ConfigWindow::get_values()
  {
  host.set_string(plugin->get_filename());
  }

void GVOutputAudioEsound::ConfigWindow::apply_values()
  {
  plugin->set_filename(host.get_string());
  }

#define PLUGIN_NAME GVOutputAudioEsound
#include "interface.cc"
