/*****************************************************************

  oa_oss.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// OSS Audio driver

#include <iostream>

using std::ostream;
using std::cerr;

#include "../gv.h"
#include "oa_oss.h"
#include "oss_common.h"

#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "../gvconfig.h"
#include "../gtkutil.h"

static GVPluginAttributes _attributes =
  {
    "oa_oss",
    "OSS Audio Driver",
    "Audio driver for OSS Soundcards",
    NULL,
    GV_PLUGIN_HAS_AUDIO|
    GV_PLUGIN_NEEDS_REALTIME|
    GV_PLUGIN_IS_DEVICE|     // Plugin reads from a device
    GV_PLUGIN_CAN_SYNC       // Plugin has a working set_time function
  };

GVOutputAudioOSS::GVOutputAudioOSS() :
  GVOutputAudio(&_attributes)
  {
  config_window = (ConfigWindow*)0;
  
  pthread_mutex_init (&fd_mutex,(pthread_mutexattr_t *)0);

  const char * _devicename = "/dev/dsp";
  gv_config->get_string(attributes->name, "Device1", &_devicename);
  device1 = g_strdup(_devicename);

  _devicename = "/dev/dsp1";
  gv_config->get_string(attributes->name, "Device2", &_devicename);
  device2 = g_strdup(_devicename);
  
  _devicename = "/dev/dsp2";
  gv_config->get_string(attributes->name, "Device3", &_devicename);
  device3 = g_strdup(_devicename);

  mirror_rear = 0;
  
  gv_config->get_bool(attributes->name, "RearStereo", mirror_rear);
  
  fd1 = -1;
  fd2 = -1;
  fd3 = -1;
  }

void GVOutputAudioOSS::configure()
  {
  if(!config_window)
    config_window = new ConfigWindow(this);
  config_window->run();
  }


GVOutputAudioOSS::~GVOutputAudioOSS()
  {
  cleanup();

  gv_config->put_string(attributes->name, "Device1", device1);
  gv_config->put_string(attributes->name, "Device2", device2);
  gv_config->put_string(attributes->name, "Device3", device3);

  gv_config->put_bool(attributes->name, "RearStereo", mirror_rear);
  
  if(config_window)
    delete config_window;
  }

// AUDIO TIMING METHOD
//#define TIME_ORIGINAL //hard with translation
//#define TIME_DIRECT_HARD // hard direct
#define TIME_INPUTANDBUFFER // input + buffer
void GVOutputAudioOSS::get_current_track_time(GVTime & old_time)
  {
  if(fd1 == -1)
    return;

#ifdef TIME_ORIGINAL 
  count_info bytes_played;
  int bytes_difference;
  
  pthread_mutex_lock(&fd_mutex);
  ioctl(fd1, SNDCTL_DSP_GETOPTR, &bytes_played);
  pthread_mutex_unlock(&fd_mutex);

  if(bytes_played.bytes < old_bytes_played)
    old_bytes_played -= G_MAXINT;

  //  cerr << bytes_played.bytes << " bytes played\n";

  if (bytes_played.bytes - old_bytes_played > 10000 * audio_format.bytes_per_sample * audio_format.num_channels) 
	{
	bytes_difference =0;
        //         cerr << "bytes played " << bytes_played.bytes << " old_bytes_played " << old_bytes_played << "\n"; 
	}
  else 
	{
	bytes_difference = bytes_played.bytes - old_bytes_played;
	}
  old_bytes_played = bytes_played.bytes;
    
  // Divide by samplerate and number of channels

  bytes_difference/=(audio_format.bytes_per_sample);
  if(audio_format.num_channels > 1)
    bytes_difference/=2;
  GVTime diff_time;
  
  diff_time.set_from_frames((unsigned int)bytes_difference,
                                     (unsigned int)audio_format.samplerate);
  
  // We can only deliver relative values
  old_time += diff_time;
#endif

#ifdef TIME_DIRECT_HARD
 count_info bytes_played;
 double thetime;
 int thevalue; // temporary byte value
 
  pthread_mutex_lock(&fd_mutex);
  ioctl(fd1, SNDCTL_DSP_GETOPTR, &bytes_played);
  pthread_mutex_unlock(&fd_mutex);


//  cerr << bytes_played.bytes << " bytes played\n";

  if ((int)bytes_played.bytes - old_bytes_played > 10000 * audio_format.bytes_per_sample * audio_format.num_channels)
        {
        thevalue = old_bytes_played;
         cerr << "bytes played " << bytes_played.bytes << " old_bytes_played " << old_bytes_played << "\n";
        }
  else
        {
        thevalue  = bytes_played.bytes;
        }
  old_bytes_played = thevalue;

  thetime  = (double)thevalue/audio_format.bytes_per_sample/audio_format.samplerate;
  if(audio_format.num_channels > 1)
    thetime/=2.0;
  old_time = thetime;

#endif
#ifdef TIME_INPUTANDBUFFER
  // to do  add the buffer delay
  int unplayed_bytes;
  int unplayed_samples;
  GVTime inBuffer;
  
  // old_bytes_played is used to store the last value in case of error
  // BP: This happens, when output is already closed while the video
  // output still plays frames. We should return 0 then

  pthread_mutex_lock(&fd_mutex);
  if (ioctl(fd1, SNDCTL_DSP_GETODELAY, &unplayed_bytes)== -1)
    {
    unplayed_bytes=old_bytes_played;
//     cerr << "SNDCTL_DSP_GETODELAY error: taking old value "
//          << unplayed_bytes << "\n";
    }
  pthread_mutex_unlock(&fd_mutex);
  
  old_bytes_played = unplayed_bytes;
  old_time = lastPutTime;
  
  unplayed_samples = unplayed_bytes / (audio_format.bytes_per_sample);
  if(audio_format.num_channels > 1)
    unplayed_samples /= 2; 
  
  inBuffer.set_from_samples((guint64)(unplayed_samples), audio_format.samplerate);
  //  inBuffer.print("Audio inBuffer"," ");
  pthread_mutex_lock(&internal_time_mutex);
  old_time = internal_time;
  pthread_mutex_unlock(&internal_time_mutex);
  old_time -= inBuffer;
#endif 
  //old_time.print("Output A "," "); 
  }

void GVOutputAudioOSS::init()
  {
  open_device(1);
  }

void GVOutputAudioOSS::reset()
  {
  close_device();
  open_device(0);
  }

// Open the device and set parameters
// If devices for more than 2 channels can't be openend,
// it's no error. The number of channels is decreased and
// the audio buffer will do downmixing

void GVOutputAudioOSS::open_device(int init)
  {
  const GVAudioFormat * input_format;
  if(init)
    input_format = input->get_audio_format();
  else
    input_format = &audio_format;
  
  // Open audio device
  
  if((fd1 = open(device1, O_WRONLY, 0)) ==-1)
    {
    g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "Cannot open device %s", device1);
    error = GV_PLUGIN_ERROR_DEVICE_OPEN_FAILED;
    return;
    }

  if((input_format->num_channels > 2) || (mirror_rear))
    {
    if((fd2 = open(device2, O_WRONLY, 0)) == -1)
      {
      // Use downmixing      
      }
    }
  if(input_format->num_channels > 4)
    {
    if((fd3 = open(device3, O_WRONLY, 0)) == -1)
      {
      // Use downmixing
      }
    }
  
  // Try to set the output format
  int sample_format = oss_get_sample_format(*input_format);

  if(((fd1 != -1) && ioctl(fd1, SNDCTL_DSP_SETFMT, &sample_format) == -1) ||
     ((fd2 != -1) && ioctl(fd2, SNDCTL_DSP_SETFMT, &sample_format) == -1) ||
     ((fd3 != -1) && ioctl(fd3, SNDCTL_DSP_SETFMT, &sample_format) == -1))
    {
    // Format not supported

    g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "Cannot set sampletype");
    error = GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED;
    return;
    }
  else if(init)
    {
    //    cerr << "Setting Format: success\n"; 
    audio_format.endian_format = input_format->endian_format;
    audio_format.sample_type = input_format->sample_type;
    }
  
  // Oss is always interleaved

  if(init)
    audio_format.interleave_mode = GVInterleave2;
  
  // Set the number of channels

  int channels = 1;

  switch(input_format->num_channels)
    {
    case 1: // Mono
      channels = 1;
      if(ioctl(fd1, SNDCTL_DSP_CHANNELS, &channels) == -1)
        {
        g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
              "Device %s doesn't support 1 channel sound",
              device1);
        error = GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED;
        return;
        }
      if(fd2 != -1) // Mono can also be sent to rear
        {
        if(ioctl(fd2, SNDCTL_DSP_CHANNELS, &channels) == -1)
          {
          // This is no error, just close device
          close(fd2);
          fd2 = -1;
          }
        }
      break;
    case 2: // Stereo
      channels = 2;
      if(ioctl(fd1, SNDCTL_DSP_CHANNELS, &channels) == -1)
        {
        g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
              "Device %s doesn't support 2 channel sound",
              device1);
        error = GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED;
        return;
        }
      if(fd2 != -1)
        {
        channels = 2;
        if(ioctl(fd2, SNDCTL_DSP_CHANNELS, &channels) == -1)
          {
          close(fd2);
          fd2 = -1;
          return;
          }
        }
      break;
    case 4: // 4 channel (quadrophony)
      channels = 2;
      if(ioctl(fd1, SNDCTL_DSP_CHANNELS, &channels) == -1)
        {
        g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
              "Device %s doesn't support 2 channel sound", device1);
        error = GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED;
        return;
        }
      channels = 2;
      if(ioctl(fd2, SNDCTL_DSP_CHANNELS, &channels) == -1)
        {
        close(fd2);
        fd2 = -1;
        }
      break;
    case 5: // 5 channel
      channels = 2;
      if(ioctl(fd1, SNDCTL_DSP_CHANNELS, &channels) == -1)
        {
        g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
              "Device %s doesn't support 2 channel sound", device1);
        error = GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED;
        return;
        }
      if(fd2 != -1)
        {
        channels = 2;
        if(ioctl(fd2, SNDCTL_DSP_CHANNELS, &channels) == -1)
          {
          close(fd2);
          fd2 = -1;
          }
        }
      if(fd3 != -1)
        {
        channels = 1;
        if(ioctl(fd3, SNDCTL_DSP_CHANNELS, &channels) == -1)
          {
          close(fd3);
          fd3 = -1;
          }
        }
      break;
    default:
      g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
            "Don't know how to play %d channel sound",
            input_format->num_channels);
      error = GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED;
      return;
    }
  
  // Set the sampling speed

  int samplerate = input_format->samplerate;
  
  if(((fd1 != -1) && ioctl(fd1, SNDCTL_DSP_SPEED, &samplerate) == -1) ||
     ((fd2 != -1) && ioctl(fd2, SNDCTL_DSP_SPEED, &samplerate) == -1) ||
     ((fd3 != -1) && ioctl(fd3, SNDCTL_DSP_SPEED, &samplerate) == -1))
    {
    error = GV_PLUGIN_ERROR_DEVICE_SETTINGS_FAILED;
    g_log(PACKAGE, G_LOG_LEVEL_CRITICAL, "Couldn't set samplerate");
    return;
    }
  //  cerr << "Setting Sampling Rate: success\n"; 
  if(init)
    audio_format.samplerate = samplerate;
  
  if(init)
    {
    audio_format.samples_per_frame = input_format->samples_per_frame;
    audio_format.num_samples = input_format->num_samples;
    audio_format.init();

    // Now set the number of channels we could get

    switch(input_format->num_channels)
      {
      case 1:
        audio_format.num_channels = 1;
        break;
      case 2:
        audio_format.num_channels = 2;
        break;
      case 4:
        if(fd2 != -1)
          audio_format.num_channels = 4;
        else
          audio_format.num_channels = 2;
        break;
      case 5:
        if(fd3 == -1) 
          {
          if(fd2 == -1)
            audio_format.num_channels = 2;
          else
            audio_format.num_channels = 4;
          }
        else
          audio_format.num_channels = 5;
        break;
      }
    audio_format.init();
    }
  old_bytes_played = 0;
  lastPutTime=(double) 0.0;
  }

void GVOutputAudioOSS::put_frame(GVAudioFrame * frame)
  {
  
  lastPutTime = frame->time;
  int write_size = frame->valid_samples * audio_format.bytes_per_sample;
  if(audio_format.num_channels > 1)
    write_size *= 2;
  
  switch(audio_format.num_channels)
    {
    case 1:
    case 2:
      pthread_mutex_lock(&fd_mutex);
      write(fd1,(void*)(frame->channels.s_8[0]), write_size);
      if((mirror_rear) && (fd2 != -1))
        write(fd2,(void*)(frame->channels.s_8[0]), write_size);
      pthread_mutex_unlock(&fd_mutex);
      break;
    case 4:
      pthread_mutex_lock(&fd_mutex);
      write(fd1,(void*)(frame->channels.s_8[0]), write_size);
      if(fd2 != -1)
        write(fd2,(void*)(frame->channels.s_8[2]), write_size);
      pthread_mutex_unlock(&fd_mutex);
      break;
    case 5:
      pthread_mutex_lock(&fd_mutex);
      write(fd1,(void*)(frame->channels.s_8[0]), write_size);
      if(fd2 != -1)
        write(fd2,(void*)(frame->channels.s_8[2]), write_size);
      if(fd3 != -1)
        write(fd3,(void*)(&frame->channels.s_8[4]), write_size/2);
      pthread_mutex_unlock(&fd_mutex);
      break;
    }
  
  }

void GVOutputAudioOSS::close_device()
  {
  pthread_mutex_lock(&fd_mutex);
  if(fd1 == -1)
    {
    pthread_mutex_unlock(&fd_mutex);
    return;
    }
  //  cerr << "GVOutputAudioOSS::close_device()\n";
  close(fd1);
  fd1 = -1;

  if(fd2 != -1)
    {
    close(fd2);
    fd2 = -1;
    }

  if(fd3 != -1)
    {
    close(fd3);
    fd3 = -1;
    }
  old_bytes_played = 0;
  pthread_mutex_unlock(&fd_mutex);
  }

void GVOutputAudioOSS::cleanup()
  {
  close_device();
  }

void GVOutputAudioOSS::print_info(ostream & output)
  {
  switch(audio_format.num_channels)
    {
    case 1:
      output << "Playing 1 channel to " << device1;
      if(fd2 != -1)
        output << " and " << device2 << " (mirrored)\n";
      output << "\n";
      break;
    case 2:
      output << "Playing 2 channels to " << device1;
      if(fd2 != -1)
        output << " and " << device2 << " (mirrored)\n";
      output << "\n";
      break;
    case 4:
      output << "Playing front channels to " << device1 << "\n";
      output << "Playing rear channels to " << device2 << "\n";
      break;
    case 5:
      output << "Playing front channels to " << device1 << "\n";
      output << "Playing rear channels to " << device2 << "\n";
      output << "Playing center channel to " << device3 << "\n";
      break;
    }
  }


GVOutputAudioOSS::ConfigWindow::ConfigWindow(GVOutputAudioOSS * _plugin) :
  GVConfigWindow(_plugin->attributes->real_name),
  devicename1("Front Device", 0),
  devicename2("Surround Device", 0),
  devicename3("Center/Subwoofer device", 0)
  {
  plugin = _plugin;

  mirror_rear =
    gtk_check_button_new_with_label("Use Rear for Mono and Stereo");

  gtk_widget_show(mirror_rear);

  table = gvgtk_table_new(4, 3, 0);

  devicename1.attach(table, 0, 0);
  devicename2.attach(table, 1, 0);
  devicename3.attach(table, 2, 0);

  gtk_table_attach_defaults(GTK_TABLE(table), mirror_rear, 0, 3, 3, 4);
    
  gtk_widget_show(table);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), table);
  }

GVOutputAudioOSS::ConfigWindow::~ConfigWindow()
  {
  //  gtk_widget_destroy(table);
  }

void GVOutputAudioOSS::ConfigWindow::get_values()
  {
  devicename1.set_filename(plugin->device1);
  devicename2.set_filename(plugin->device2);
  devicename3.set_filename(plugin->device3);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mirror_rear),
                               plugin->mirror_rear);
  }

void GVOutputAudioOSS::ConfigWindow::apply_values()
  {
  if(plugin->device1)
    g_free(plugin->device1);
  plugin->device1 = g_strdup(devicename1.get_filename());

  if(plugin->device2)
    g_free(plugin->device2);
  plugin->device2 = g_strdup(devicename2.get_filename());

  if(plugin->device3)
    g_free(plugin->device3);
  plugin->device3 = g_strdup(devicename3.get_filename());

  plugin->mirror_rear =
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mirror_rear));
  }

#define PLUGIN_NAME GVOutputAudioOSS
#include "interface.cc"
