/*****************************************************************

  oa_sndfile.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// sndfile output plugin

#ifdef __GNUG__
#pragma implementation "oa_sndfile.h"
#endif

#include <iostream>

using std::ostream;
using std::cerr;

#include "../gv.h"
#include "../gvconfig.h"
#include "../gtkutil.h"

#include "libsndfile/src/sndfile.h"
#include "oa_sndfile.h"
#include <string.h>

static GVConfigMenuItemInt sndfile_formats[] =
{
  { SF_FORMAT_WAV,  "Microsoft WAV format (big endian)" },
  { SF_FORMAT_AIFF, "Apple/SGI AIFF format (little endian)" }, 
  { SF_FORMAT_AU,   "Sun/NeXT AU format (big endian)"},
  { SF_FORMAT_AULE, "DEC AU format (little endian)"},
  { SF_FORMAT_RAW,  "RAW PCM data" },
  { SF_FORMAT_PAF,  "Ensoniq PARIS file format" },
  { SF_FORMAT_SVX,  "Amiga IFF / SVX8 / SV16 format" },
  { SF_FORMAT_NIST, "Sphere NIST format" }
};

static GVConfigMenuItemInt sndfile_subformats[] =
  {
    { SF_FORMAT_PCM,       "PCM" },
    { SF_FORMAT_FLOAT,     "32 bit floats" },
    { SF_FORMAT_ULAW,      "U-Law encoded" },
    { SF_FORMAT_ALAW,      "A-Law encoded" },
    { SF_FORMAT_IMA_ADPCM, "IMA ADPCM" },
    { SF_FORMAT_MS_ADPCM,  "Microsoft ADPCM" },
    { SF_FORMAT_PCM_BE,    "Big endian PCM data" },
    { SF_FORMAT_PCM_LE,    "Little endian PCM data" },
    { SF_FORMAT_PCM_S8,    "Signed 8 bit PCM" },
    { SF_FORMAT_PCM_U8,    "Unsigned 8 bit PCM" },
    { SF_FORMAT_SVX_FIB,   "SVX Fibonacci Delta encoding" },
    { SF_FORMAT_SVX_EXP,   "SVX Exponential Delta encoding" },
    { SF_FORMAT_GSM610,    "GSM 6.10 encoding" },
    { SF_FORMAT_G721_32,   "32kbs G721 ADPCM encoding" },
    { SF_FORMAT_G723_24,   "24kbs G723 ADPCM encoding" }
  };

static GVPluginAttributes _attributes = 
  {
    "oa_sndfile",
    "Sndfile Output Plugin",
    "Formatwriter for WAV, AU and SND files",
    "wav au snd",
    GV_PLUGIN_HAS_AUDIO|GV_PLUGIN_IS_FILE
  };

GVOutputAudioSndfile::GVOutputAudioSndfile() :
  GVOutputAudio(&_attributes)
  {
  config_window = (ConfigWindow*)0;
  file = (SNDFILE*)0;
  get_config();
  }

GVOutputAudioSndfile::ConfigWindow::~ConfigWindow()
  {
  gtk_widget_destroy(table);
  }


GVOutputAudioSndfile::ConfigWindow::ConfigWindow(GVOutputAudioSndfile *
                                                 _plugin) :
  GVConfigWindow(_plugin->attributes->real_name),
  format_menu("Formats"),
  subformat_menu("Subformats")
  {
  format_menu.create_menu(sndfile_formats,
                          sizeof(sndfile_formats)/sizeof(GVConfigMenuItemInt));
  subformat_menu.create_menu(sndfile_subformats,
                             sizeof(sndfile_subformats)/
                             sizeof(GVConfigMenuItemInt));

  table = gvgtk_table_new(2, 2, 0);

  format_menu.attach(table, 0, 0);
  subformat_menu.attach(table, 1, 0);
  
  gtk_widget_show(table);
  
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), table);
  
  plugin = _plugin;
  }

void GVOutputAudioSndfile::ConfigWindow::get_values()
  {
  int format = plugin->file_info.format & SF_FORMAT_TYPEMASK;
  int subformat = plugin->file_info.format & SF_FORMAT_SUBMASK;
  
  format_menu.set(format);
  subformat_menu.set(subformat);
  }

void GVOutputAudioSndfile::ConfigWindow::apply_values()
  {
  int format = format_menu.get_int();
  int subformat = subformat_menu.get_int();
  //   cerr << "Format: " << format << "\n";
  //   cerr << "SubFormat: " << subformat << "\n";
  plugin->file_info.format = format|subformat;
  }

void GVOutputAudioSndfile::configure()
  {
  if(!config_window)
    config_window = new ConfigWindow(this);
  config_window->run();
  }

GVOutputAudioSndfile::~GVOutputAudioSndfile()
  {
  put_config();
  if(config_window)
    delete config_window;
  }

void GVOutputAudioSndfile::init()
  {
  const GVAudioFormat * input_format = input->get_audio_format();

  audio_format = *input_format;
  
  file_info.samplerate = (unsigned int)(audio_format.samplerate);
  file_info.samples =    (unsigned int)(audio_format.num_samples);
  file_info.channels =   (unsigned int)(audio_format.num_channels);
  
  audio_format.sample_type = GVAudio_16_signed;
  file_info.pcmbitwidth = 16;

  file_info.sections = 0;
  file_info.seekable = 0;
  
  if(!sf_format_check(&file_info))
    {
    g_log(PACKAGE, G_LOG_LEVEL_CRITICAL,
          "Audio Format mismatch, check settings");
    return;
    }

  // Now think about a proper suffix

  const char * suffix = "";

  switch(file_info.format & SF_FORMAT_TYPEMASK)
    {
    case SF_FORMAT_WAV:      /* Microsoft WAV format (big endian). */
      suffix = "wav";
      break;
    case SF_FORMAT_AIFF:     /* Apple/SGI AIFF format (little endian). */
      suffix = "aiff";
      break;
    case SF_FORMAT_AU:       /* Sun/NeXT AU format (big endian). */
      suffix = "au";
      break;
    case SF_FORMAT_AULE:     /* DEC AU format (little endian). */
      suffix = "au";
      break;
    case SF_FORMAT_RAW:      /* RAW PCM data. */
      suffix = "raw";
      break;
    case SF_FORMAT_PAF:      /* Ensoniq PARIS file format. */
      suffix = "paf";
      break;
    case SF_FORMAT_SVX:      /* Amiga IFF / SVX8 / SV16 format. */
      suffix = "svx";
      break;
    case SF_FORMAT_NIST:     /* Sphere NIST format. */
      suffix = "nist";
      break;
    default:
      cerr << "Unkown format " << file_info.format << "\n";
    }

  if(!create_outfilename(suffix))
    return;
  
  file = sf_open_write(filename, &file_info);
  if(!file)
    sf_perror(file);
  
  //  file = sf_open_write(filename, &file_info);
  
  // We want only signed 16 interleaved

  audio_format.sample_type = GVAudio_16_signed;
  audio_format.interleave_mode = GVInterleaveAll;
  
  audio_format.init();
  }

void GVOutputAudioSndfile::cleanup()
  {
  if(file)
    {
    sf_close(file);
    file = (SNDFILE*)0;
    }
  }

void GVOutputAudioSndfile::get_config()
  {
  int format = SF_FORMAT_WAV|SF_FORMAT_PCM;
  if(gv_config->get_int(attributes->name, "Format", format))
    file_info.format = (unsigned int)format;
  }

void GVOutputAudioSndfile::put_config()
  {
  gv_config->put_int(attributes->name, "Format", file_info.format);
  }

void GVOutputAudioSndfile::put_frame(GVAudioFrame * frame)
  {
  cerr << "Writing Audio frame...";
  sf_writef_short(file, frame->samples.s_16, frame->valid_samples);
  cerr << "done\n";
  }

#define PLUGIN_NAME GVOutputAudioSndfile
#include "interface.cc"
