/*****************************************************************

  oa_vorbis.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Ogg Vorbis encoder for gmerlin

#ifdef __GNUG__
#pragma implementation "oa_vorbis.h"
#endif

#include <iostream>
#include <fstream>
#include <sstream>

using std::ofstream;
using std::ostream;
using std::stringstream;
using std::ios;
using std::ends;
 
#include <stdlib.h>
#include <time.h>

#include "../gv.h"
#include "../gvconfig.h"
#include "../gtkutil.h"

#include "../trackinfo.h"

extern "C" {
#include <vorbis/vorbisenc.h>
#include <vorbis/vorbisfile.h>
}

#include "oa_vorbis.h"
// #include <string.h>

static GVPluginAttributes _attributes = 
  {
    "oa_vorbis",
    "Ogg Vorbis encoder",
    "Encoder for Ogg Vorbis files",
    "ogg",
    GV_PLUGIN_HAS_AUDIO|GV_PLUGIN_IS_FILE
  };

GVOutputAudioVorbis::GVOutputAudioVorbis() :
  GVOutputAudio(&_attributes)
  {
  config_window = (ConfigWindow*)0;
  get_config();
  }

static char * _artist_key = "ARTIST=";
static char * _album_key = "ALBUM=";
static char * _title_key = "TITLE=";
static char * _track_number_key = "TRACKNUMBER=";
static char * _genre_key = "GENRE=";
// static char * _organization_key = "ORGANIZATION=";
// static char * _version_key = "VERSION=";
// static char * _description_key = "DESCRIPTION=";
// static char * _date_key = "DATE=";
// static char * _location_key = "LOCATION=";
// static char * _copyright_key = "COPYRIGHT=";

static void __comment_add(vorbis_comment * vc, const char * _comment)
  {
  char * comment = g_strdup(_comment);
  vorbis_comment_add(vc, comment);
  g_free(comment);
  }

void GVOutputAudioVorbis::build_comment()
  {
  vorbis_comment_init(&vc);
  
  const GVTrackInfo * input_info = input->get_track_info();

  int i_tmp;
  const char * c_tmp;
  stringstream line;

  c_tmp = default_track_info.get_title(input_info, build_comment_flags);
  if(c_tmp)
    {
    line.seekp(0);
//    line.freeze(0);
    line << _title_key << c_tmp << ends;
    __comment_add(&vc, line.str().c_str());
    }

  c_tmp = default_track_info.get_album(input_info, build_comment_flags);
  if(c_tmp)
    {
    line.seekp(0);
//    line.freeze(0);
    line << _album_key << c_tmp << ends;
    __comment_add(&vc, line.str().c_str());
    }

  c_tmp = default_track_info.get_artist(input_info, build_comment_flags);
  if(c_tmp)
    {
    line.seekp(0);
//    line.freeze(0);
    line << _artist_key << c_tmp << ends;
    __comment_add(&vc, line.str().c_str());
    }
  
  c_tmp = default_track_info.get_genre(input_info, build_comment_flags);
  if(c_tmp)
    {
    line.seekp(0);
//    line.freeze(0);
    line << _genre_key << c_tmp << ends;
    __comment_add(&vc, line.str().c_str());
    }

  i_tmp = default_track_info.get_track_number(input_info, build_comment_flags);
  if(i_tmp > 0)
    {
    line.seekp(0);
//    line.freeze(0);
    line << _track_number_key << i_tmp << ends;
    __comment_add(&vc, line.str().c_str());
    }

  c_tmp = default_track_info.get_comment(input_info, build_comment_flags);
  if(c_tmp)
    {
    line.seekp(0);
//    line.freeze(0);
    line << c_tmp << ends;
    __comment_add(&vc, line.str().c_str());
    //    cerr << "Added Comment: " << line.str().c_str() << "\n";
    }
  
  
  }

void GVOutputAudioVorbis::init()
  {
  if(!create_outfilename("ogg"))
    return;

  audio_format = *input->get_audio_format();

  // the Vorbis encoder likes non interleaved samples

  audio_format.interleave_mode = GVInterleaveNone;

  audio_format.sample_type = GVAudio_16_signed;
  audio_format.endian_format = GVNativeEndian;
  
  vorbis_info_init(&vi);
  vorbis_encode_init(&vi,audio_format.num_channels,audio_format.samplerate,
                     -1, bitrate, -1);

  build_comment();

  vorbis_analysis_init(&vd,&vi);
  vorbis_block_init(&vd,&vb);
  
  /* set up our packet->stream encoder */
  /* pick a random serial number; that way we can more likely build
     chained streams just by concatenation */
  srand(time(NULL));
  ogg_stream_init(&os,rand());

  output.open(filename);

  ogg_packet header;
  ogg_packet header_comm;
  ogg_packet header_code;

  vorbis_analysis_headerout(&vd,&vc,&header,&header_comm,&header_code);
  ogg_stream_packetin(&os,&header); /* automatically placed in its own
                                         page */
  ogg_stream_packetin(&os,&header_comm);
  ogg_stream_packetin(&os,&header_code);

  /* We don't have to write out here, but doing so makes streaming 
   * much easier, so we do, flushing ALL pages. This ensures the actual
   * audio data will start on a new page
   */
  while(1)
    {
    int result=ogg_stream_flush(&os,&og);
    if(result==0)break;
    output.write((char*)og.header,og.header_len);
    output.write((char*)og.body,og.body_len);
    }
  audio_format.init();
  }

void GVOutputAudioVorbis::put_frame(GVAudioFrame * frame)
  {
  int eos = 0;

  float **buffer=vorbis_analysis_buffer(&vd,frame->valid_samples);
  
  for(unsigned int j = 0; j < audio_format.num_channels; j++)
    for(unsigned int i=0;i<frame->valid_samples; i++)
      buffer[j][i]= frame->channels.s_16[j][i]/32768.0;
  
  /* tell the library how much we actually submitted */

  vorbis_analysis_wrote(&vd,frame->valid_samples);

  while(vorbis_analysis_blockout(&vd,&vb)==1)
    {
    /* analysis */
    vorbis_analysis(&vb,&op);
      
    /* weld the packet into the bitstream */
    ogg_stream_packetin(&os,&op);

    /* write out pages (if any) */
    while(!eos)
      {
      int result=ogg_stream_pageout(&os,&og);
      if(result==0)
        break;
      output.write((char*)og.header,og.header_len);
      output.write((char*)og.body,og.body_len);

      /* this could be set above, but for illustrative purposes, I do
         it here (to show that vorbis does know where the stream ends) */
        
      if(ogg_page_eos(&og))
        eos=1;
      }
    }
  }

void GVOutputAudioVorbis::cleanup()
  {
  // Do a last encoding step

  vorbis_analysis_wrote(&vd,0);

  while(vorbis_analysis_blockout(&vd,&vb)==1)
    {
    
    /* analysis */
    vorbis_analysis(&vb,&op);
      
    /* weld the packet into the bitstream */
    ogg_stream_packetin(&os,&op);

    /* write out pages (if any) */
    int eos = 0;
    while(!eos)
      {
      int result=ogg_stream_pageout(&os,&og);
      if(result==0)
        break;
      output.write((char*)og.header,og.header_len);
      output.write((char*)og.body,og.body_len);

        /* this could be set above, but for illustrative purposes, I do
           it here (to show that vorbis does know where the stream ends) */
        
      if(ogg_page_eos(&og))
        eos=1;

      }
    }

  output.close();

  ogg_stream_clear(&os);
  vorbis_block_clear(&vb);
  vorbis_dsp_clear(&vd);
  vorbis_comment_clear(&vc);
  vorbis_info_clear(&vi);
  }

GVOutputAudioVorbis::~GVOutputAudioVorbis()
  {
  if(config_window)
    delete config_window;
  put_config();
  }

void GVOutputAudioVorbis::configure()
  {
  if(!config_window)
    config_window = new ConfigWindow(this);
  config_window->run();
  }

#define SECTION attributes->name

void GVOutputAudioVorbis::get_config()
  {
  bitrate = 128;
  gv_config->get_int(SECTION, "Bitrate", bitrate);

  build_comment_flags = GV_TRACK_INFO_FORCE_COMMENT;

  gv_config->get_int(SECTION, "BuildCommentFlags",
                     build_comment_flags);
  
  const char * tmp;
  tmp = (const char*)0;
  
  gv_config->get_string(SECTION, "DefaultArtist", &tmp);
  default_track_info.set_artist(tmp);
  
  gv_config->get_string(SECTION, "DefaultAlbum", &tmp);
  default_track_info.set_album(tmp);

  tmp = "Made with gmerlin";
  gv_config->get_string(SECTION, "DefaultComment", &tmp);
  default_track_info.set_comment(tmp);

  tmp = (const char*)0;
  gv_config->get_string(SECTION, "DefaultGenre", &tmp);
  default_track_info.set_genre(tmp);

  tmp = (const char*)0;
  gv_config->get_string(SECTION, "DefaultTitle", &tmp);
  default_track_info.set_title(tmp);

  default_track_info.track_number = 0;
  gv_config->get_int(SECTION, "DefaultTrackNumber",
                     default_track_info.track_number);
    
  }

void GVOutputAudioVorbis::put_config()
  {
  gv_config->put_int(SECTION, "Bitrate", bitrate);

  gv_config->put_int(SECTION, "BuildCommentFlags",
                     build_comment_flags);


  gv_config->put_string(SECTION, "DefaultArtist", default_track_info.artist);
  
  gv_config->put_string(SECTION, "DefaultAlbum", default_track_info.album);

  gv_config->put_string(SECTION, "DefaultComment", default_track_info.comment);

  gv_config->put_string(SECTION, "DefaultGenre", default_track_info.genre);

  gv_config->put_string(SECTION, "DefaultTitle", default_track_info.title);

  gv_config->put_int(SECTION, "DefaultTrackNumber", default_track_info.track_number);
  }

static int vorbis_bitrates[] =
  {
    112,
    128,
    160,
    192,
    256,
    350
  };

GVOutputAudioVorbis::ConfigWindow::ConfigWindow(GVOutputAudioVorbis *
                                                _plugin) :
  GVConfigWindow(_plugin->attributes->real_name),
  default_artist("Default Artist"),
  default_album("Default Album"),
  default_title("Default Title"),
  default_track_number("Default Tracknumber", 0.0, 1000.0, 0),
  default_genre("Default Genre"),
  default_comment("Default Comment"),
  bitrate("Average bitrate")
  {
  bitrate.create_menu(vorbis_bitrates, sizeof(vorbis_bitrates)/sizeof(int));

  use_default_artist = gtk_check_button_new_with_label("Use this always");
  use_default_album = gtk_check_button_new_with_label("Use this always");
  use_default_title = gtk_check_button_new_with_label("Use this always");
  use_default_track_number = gtk_check_button_new_with_label("Use this always");
  use_default_genre = gtk_check_button_new_with_label("Use this always");
  use_default_comment = gtk_check_button_new_with_label("Use this always");

  gtk_widget_show(use_default_artist);
  gtk_widget_show(use_default_album);
  gtk_widget_show(use_default_title);
  gtk_widget_show(use_default_track_number);
  gtk_widget_show(use_default_genre);
  gtk_widget_show(use_default_comment);
  
  GtkWidget * table = gvgtk_table_new(8, 3, 0);
  bitrate.attach(table, 0, 0);

  default_artist.attach(table, 1, 0);
  gtk_table_attach_defaults(GTK_TABLE(table), use_default_artist, 2, 3, 1, 2);
  
  default_album.attach(table, 2, 0);
  gtk_table_attach_defaults(GTK_TABLE(table), use_default_album, 2, 3, 2, 3);

  default_title.attach(table, 3, 0);
  gtk_table_attach_defaults(GTK_TABLE(table), use_default_title, 2, 3, 3, 4);

  default_track_number.attach(table, 4, 0);
  gtk_table_attach_defaults(GTK_TABLE(table), use_default_track_number,
                            2, 3, 4, 5);

  default_genre.attach(table, 5, 0);
  gtk_table_attach_defaults(GTK_TABLE(table), use_default_genre, 2, 3, 5, 6);

  default_comment.attach(table, 6, 0);
  gtk_table_attach_defaults(GTK_TABLE(table), use_default_comment, 2, 3, 6, 7);

  gtk_widget_show(table);

  gtk_box_pack_start_defaults(GTK_BOX(mainbox), table);
  plugin = _plugin;
  }

GVOutputAudioVorbis::ConfigWindow::~ConfigWindow()
  {

  }

void GVOutputAudioVorbis::ConfigWindow::get_values()
  {
  bitrate.set(plugin->bitrate);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(use_default_artist),
                               !!(plugin->build_comment_flags & GV_TRACK_INFO_FORCE_ARTIST));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(use_default_album),
                               !!(plugin->build_comment_flags & GV_TRACK_INFO_FORCE_ALBUM));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(use_default_title),
                               !!(plugin->build_comment_flags & GV_TRACK_INFO_FORCE_TITLE));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(use_default_track_number),
                               !!(plugin->build_comment_flags & GV_TRACK_INFO_FORCE_TRACK));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(use_default_genre),
                               !!(plugin->build_comment_flags & GV_TRACK_INFO_FORCE_GENRE));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(use_default_comment),
                               !!(plugin->build_comment_flags & GV_TRACK_INFO_FORCE_COMMENT));

  
  default_artist.set_string(plugin->default_track_info.artist);
  default_album.set_string(plugin->default_track_info.album);
  default_title.set_string(plugin->default_track_info.title);
  default_comment.set_string(plugin->default_track_info.comment);

  //  default_year.set_value(plugin->default_track_info.year);
  default_track_number.set_value(plugin->default_track_info.track_number);
  default_genre.set_string(plugin->default_track_info.genre);
  }

void GVOutputAudioVorbis::ConfigWindow::apply_values()
  {
  plugin->bitrate = bitrate.get_int();

  plugin->build_comment_flags = 0;
  
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(use_default_artist)))
    plugin->build_comment_flags |= GV_TRACK_INFO_FORCE_ARTIST;
  
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(use_default_album)))
    plugin->build_comment_flags |= GV_TRACK_INFO_FORCE_ALBUM;
  
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(use_default_title)))
    plugin->build_comment_flags |= GV_TRACK_INFO_FORCE_TITLE;
    
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(use_default_track_number)))
    plugin->build_comment_flags |= GV_TRACK_INFO_FORCE_TRACK;
  
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(use_default_genre)))
    plugin->build_comment_flags |= GV_TRACK_INFO_FORCE_GENRE;

  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(use_default_comment)))
    plugin->build_comment_flags |= GV_TRACK_INFO_FORCE_COMMENT;

  
  plugin->default_track_info.set_artist(default_artist.get_string());
  plugin->default_track_info.set_album(default_album.get_string());
  plugin->default_track_info.set_title(default_title.get_string());
  plugin->default_track_info.set_genre(default_genre.get_string());
  plugin->default_track_info.set_comment(default_comment.get_string());

//   plugin->default_track_info.year
//     = (int)(default_year.get_value()+0.5);
  plugin->default_track_info.track_number
    = (int)(default_track_number.get_value()+0.5);

  }

#define PLUGIN_NAME GVOutputAudioVorbis
#include "interface.cc"
