/*****************************************************************

  ov_x11.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// X11 video driver

#include <iostream>

using std::cerr;
using std::ostream;

#include "ov_x11.h"

extern "C" {
#include <X11/Xmd.h>
#include <X11/extensions/dpms.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>

#include <stdio.h>
#include <math.h>
}

#include "../configwindow.h"
#include "../gvconfig.h"


static GVPluginAttributes _attributes =
  {
    "ov_x11",
    "X11 Video Driver",
    "X11 Video driver with Xvideo and XShm support",
    NULL,
    GV_PLUGIN_HAS_VIDEO|
    GV_PLUGIN_CAN_ALLOC| // Plugin must alloc it's own frames
    GV_PLUGIN_NEEDS_REALTIME
  };

GVOutputVideoX11::GVOutputVideoX11() :
  GVOutputVideo(&_attributes)
  {
  shm_image = (XImage*)0;
  shminfo = (XShmSegmentInfo*)0;
  config_window = (ConfigWindow*)0;
  get_config();
  }

GVOutputVideoX11::~GVOutputVideoX11()
  {
  put_config();
  }


void GVOutputVideoX11::configure()
  {
  if(!config_window)
    config_window = new ConfigWindow(this);
  config_window->run();
  }

void GVOutputVideoX11::alloc_frames(GVVideoFrame * frames, int num)
  {
  win->lock();
  switch(video_mode)
    {
    case GV_X11_VIDEO_MODE_XSHM:
      alloc_frames_xshm(frames, num);
      break;
    case GV_X11_VIDEO_MODE_XV_SHM:
      alloc_frames_xv(frames, num);
      break;
    }
  win->unlock();
  }

void GVOutputVideoX11::free_frames(GVVideoFrame * frames, int num)
  {
  switch(video_mode)
    {
    case GV_X11_VIDEO_MODE_XSHM:
      free_frames_xshm(frames, num);
      break;
    case GV_X11_VIDEO_MODE_XV_SHM:
      free_frames_xv(frames, num);
      break;
    }
  }


int GVOutputVideoX11::can_alloc_video()
  {
  if(video_mode == GV_X11_VIDEO_MODE_GDKRGB)
    return 0;
  return 1;
  }

void GVOutputVideoX11::get_drawing_coords()
  {
  int width, height;

  win->get_size(width, height);

  if((width != window_width) || (height != window_height))
    {
    window_width = width;
    window_height = height;
    
    switch(video_mode)
      {
      case GV_X11_VIDEO_MODE_XV_SHM:
        {
        float dst_ratio = (float)(window_width)/(float)(window_height);
        float src_ratio = (force_aspect_ratio) ? aspect_ratio :
          video_format.aspect_ratio;
        
//          cerr << "Window Width: " << window_width
//               << "Window Height: " <<  window_height << "\n";
        
//          cerr << "dst_ratio: " << dst_ratio
//               << " src_ratio: " << src_ratio << "\n";
        
        // No scaling necessary
                
        if(((window_width == video_format.width) &&
            (window_height == video_format.height))
           || (fabs(dst_ratio - src_ratio) < 0.1))
          {
          draw_x = 0;
          draw_y = 0;
          draw_width = window_width;
          draw_height = window_height;
          }
        // Make black stripes on the top and bottom
        else if(dst_ratio < src_ratio)
          {
          draw_width = window_width;
          draw_height = (int)((float)draw_width / src_ratio);
          draw_x = 0;
          draw_y = (window_height - draw_height)/2;
          }
        // Make black stripes on the left and right
        else
          {
          draw_height = window_height;
          draw_width = (int)((float)draw_height * src_ratio);
          draw_x = (window_width - draw_width)/2;
          draw_y = 0;
          }
//         cerr << "draw_x: " << draw_x << " draw_y: " << draw_y
//              << " draw_width: " << draw_width
//              << " draw_height: " << draw_height << "\n";
        
        break;
        }
      case GV_X11_VIDEO_MODE_XSHM:
      case GV_X11_VIDEO_MODE_GDKRGB:
        
        break;
      }
    }
  }

void GVOutputVideoX11::init()
  {
  win = gv_video_window;
  
  win->lock();
  win->show();
  win->set_active(1);

  dpms = 0;
  if(DPMSCapable(win->x_display))
    {
    CARD16 val1;
    BOOL val2;
    DPMSInfo(win->x_display, &val1, &val2);

    if(val2)
      {
      cerr << "Disabling DPMS...";
      DPMSDisable(win->x_display);
      dpms = 1;
      cerr << "done\n";
      }

    

    }
  
  //  cerr << "GVOutputVideoX11::init() started\n";
  
  const GVVideoFormat * input_format = input->get_video_format();

  video_format.width = input_format->width;
  video_format.height = input_format->height;
  //  video_format.colorspace = input_format->colorspace;
  video_format.framerate = input_format->framerate;
  video_format.num_frames = input_format->num_frames;
  video_format.flags = input_format->flags;
  video_format.init();

  // Check for shm and xv support

  int have_shm_support = check_shm_support();
  int have_xv_support = 0;

  if(check_xv)
    {
    // First try the best colorspace from the input plugin

    GVColorspace best_colorspace = input->get_best_colorspace();

    // Try the best colorspace
    
    if(best_colorspace == GVColorspaceYUV_420_P)
      {
      if((have_xv_support = check_xv_support(GVColorspaceYUV_420_P)))
        {
        input->set_colorspace(GVColorspaceYUV_420_P);
        video_format.colorspace = GVColorspaceYUV_420_P;
        }
      }
    else if(best_colorspace == GVColorspaceYUY2)
      {
      if((have_xv_support = check_xv_support(GVColorspaceYUY2)))
        {
        input->set_colorspace(GVColorspaceYUY2);
        video_format.colorspace = GVColorspaceYUY2;
        }
      }

    // Now try if we can get this with conversion
    
    else if(input->colorspace_supported(GVColorspaceYUV_420_P))
      {
      if((have_xv_support = check_xv_support(GVColorspaceYUV_420_P)))
        {
        input->set_colorspace(GVColorspaceYUV_420_P);
        video_format.colorspace = GVColorspaceYUV_420_P;
        }
      }
    else if(input->colorspace_supported(GVColorspaceYUY2))
      {
      if((have_xv_support = check_xv_support(GVColorspaceYUY2)))
        {
        input->set_colorspace(GVColorspaceYUY2);
        video_format.colorspace = GVColorspaceYUY2;
        }
      }
    else if(force_xv)
      {
      input->set_colorspace(best_colorspace);

      if((have_xv_support = check_xv_support(GVColorspaceYUV_420_P)))
        video_format.colorspace = GVColorspaceYUV_420_P;
      else if((have_xv_support = check_xv_support(GVColorspaceYUY2)))
        video_format.colorspace = GVColorspaceYUY2;
      }

    }
  
  if((have_shm_support) && (have_xv_support))
     {
     //     cerr << "Have SHM + Xv!\n";
     video_mode = GV_X11_VIDEO_MODE_XV_SHM;
     }
   else if(have_shm_support)
    {
    //    cerr << "Have SHM!\n";
    video_mode = GV_X11_VIDEO_MODE_XSHM;
    }
  else
    video_mode = GV_X11_VIDEO_MODE_GDKRGB;

  // Check which depth we have

   //  int width, height, x, y;

   render_depth = win->get_depth();
 
  //  cerr << "Render Depth: " << render_depth << "\n";

  //  cerr << "Default Depth " << DefaultDepth(win->x_display, 0);

  if(!force_aspect_ratio)
    win->set_original_size(video_format.width, video_format.height);
  else
    win->set_original_size(video_format.width,
                           (int)((float)(video_format.width)/aspect_ratio));

  window_width = 0;
  window_height = 0;

  get_drawing_coords();

  GVColorspace input_colorspace = input->get_best_colorspace();
    
  switch(video_mode)
    {
    // Best, not supported by every Xfree86 driver
    case GV_X11_VIDEO_MODE_XV_SHM:  
      init_xv();
      break;
    case GV_X11_VIDEO_MODE_XSHM:
      init_xshm();
      switch(shm_image->bits_per_pixel)
        {
        case 24:
          if((shm_image->red_mask == 0xff0000) && 
             (shm_image->green_mask == 0xff00) &&
             (shm_image->blue_mask == 0xff))
            {
            if(input->colorspace_supported(GVColorspaceBGR24))
              input_colorspace = GVColorspaceBGR24;
            video_format.colorspace = GVColorspaceBGR24;
            }
          else if((shm_image->blue_mask == 0xff0000) && 
             (shm_image->green_mask == 0xff00) &&
             (shm_image->red_mask == 0xff))
            {
            if(input->colorspace_supported(GVColorspaceRGB24))
              input_colorspace = GVColorspaceRGB24;
            video_format.colorspace = GVColorspaceRGB24;
            }
          break;
        case 32:
          // Check for colorspace RGB32
          if((shm_image->red_mask == 0xff) && 
             (shm_image->green_mask == 0xff00) &&
             (shm_image->blue_mask == 0xff0000))
            {
            if(input->colorspace_supported(GVColorspaceRGB32))
              input->set_colorspace(GVColorspaceRGB32);
            video_format.colorspace = GVColorspaceRGB32;
            }
          // Check for colorspace BGR32
          if((shm_image->red_mask == 0xff0000) && 
             (shm_image->green_mask == 0xff00) &&
             (shm_image->blue_mask == 0xff))
            {
            if(input->colorspace_supported(GVColorspaceBGR32))
              input_colorspace = GVColorspaceBGR32;
            video_format.colorspace = GVColorspaceBGR32;
            }
          break;
        case 16:
          if((shm_image->red_mask == 63488) &&
             (shm_image->green_mask == 2016) &&
             (shm_image->blue_mask == 31))
            {
            if(input->colorspace_supported(GVColorspaceBGR16))
              input_colorspace = GVColorspaceBGR16;
            video_format.colorspace = GVColorspaceBGR16;
            }
          else if((shm_image->blue_mask == 63488) &&
                  (shm_image->green_mask == 2016) &&
                  (shm_image->red_mask == 31))
            {
            if(input->colorspace_supported(GVColorspaceRGB16))
              input_colorspace = GVColorspaceRGB16;
            video_format.colorspace = GVColorspaceRGB16;
            }
          break;
        }
      input->set_colorspace(input_colorspace);
      break;
    case GV_X11_VIDEO_MODE_GDKRGB:
      if(input->colorspace_supported(GVColorspaceRGB24))
        input->set_colorspace(GVColorspaceRGB24);
      video_format.colorspace = GVColorspaceRGB24;
      init_gdkrgb();
      break;
    }
  //  cerr << "RenderDepth: " << render_depth << "\n";
  video_format.init();
  win->unlock();
  //  cerr << "GVOutputVideoX11::init() finished\n";
  
  }

void GVOutputVideoX11::put_frame(GVVideoFrame * frame)
  {
  get_drawing_coords();
  cleaned_up = 0;
  switch(video_mode)
    {
    case GV_X11_VIDEO_MODE_XV_SHM:
      win->lock();
      put_frame_xv(frame);
      win->unlock();
      break;
    case GV_X11_VIDEO_MODE_XSHM:
      win->lock();
      put_frame_xshm(frame);
      win->unlock();
      break;
    case GV_X11_VIDEO_MODE_GDKRGB:
      win->lock();
      put_frame_gdkrgb(frame);
      win->unlock();
      break;
    }
  }
  
void GVOutputVideoX11::cleanup()
  {
  //  cerr << "GVOutputVideoX11::cleanup()\n";

  win->lock();

  switch(video_mode)
    {
    case GV_X11_VIDEO_MODE_XV_SHM:
      if(last_window)
        {
        gv_XvStopVideo(win->x_display, xv_port, last_window);
        gv_XvUngrabPort(win->x_display, xv_port, CurrentTime);
        //        cerr << "Xv Port ungrabbed\n";
        last_window = 0;
        }
      break;
    case GV_X11_VIDEO_MODE_XSHM:
      break;
    case GV_X11_VIDEO_MODE_GDKRGB:
      break;
    }

  if(dpms)
    {
    cerr << "Enabling DPMS...";
    DPMSEnable(win->x_display);
    cerr << "done\n";
    }
  win->unlock();
  win->set_active(0);
  }

void GVOutputVideoX11::init_gdkrgb()
  {
  gdk_rgb_init();
  win->set_size(video_format.width, video_format.height);
  }

void GVOutputVideoX11::put_frame_gdkrgb(GVVideoFrame * frame)
  {
  switch(video_format.colorspace)
    {
    case GVColorspaceRGB24:
      //      cerr << "Put rgb24 Video Frame (GdkRgb)\n";
      //      gdk_draw_rgb_image(win->gdk_window, win->gdk_gc, 0, 0,
      //                         video_format.width, video_format.height,
      //                         GDK_RGB_DITHER_NONE, frame->pixels, video_format.width*3);
      break;
    default:
      break;
    }
  gdk_flush();
  }

// Configuration stuff

static GVConfigMenuItemDouble x11_aspect_ratios[] =
  {
  { 4.0/3.0, "4:3" },
  { 16.0/9.0, "16:9" }
  };

#define SECTION attributes->name

void GVOutputVideoX11::get_config()
  {
  force_aspect_ratio = 0;
  aspect_ratio = 16.0/9.0;
  check_xv = 1;
  const char * test = "16:9";
  
  gv_config->get_string(SECTION, "AspectRatio", &test);
  for(unsigned int i = 0;
      i < sizeof(x11_aspect_ratios)/sizeof(GVConfigMenuItemDouble);
      i++)
    {
    if(!strcmp(test, x11_aspect_ratios[i].label))
      {
      aspect_ratio = x11_aspect_ratios[i].d;
      break;
      }
    }
  gv_config->get_bool(SECTION, "ForceAspectRatio", force_aspect_ratio);
  gv_config->get_bool(SECTION, "CheckXV", check_xv);

  force_xv = 1;
  gv_config->get_bool(SECTION, "ForceXV", force_xv);

  }

void GVOutputVideoX11::put_config()
  {
  for(unsigned int i = 0;
      i < sizeof(x11_aspect_ratios)/sizeof(GVConfigMenuItemDouble);
      i++)
    {
//     cerr << "aspect_ratio: " << aspect_ratio << " x11_aspect_ratios[i].d: "
//          << x11_aspect_ratios[i].d << "\n";
    if(aspect_ratio - x11_aspect_ratios[i].d < 0.01)
      {
      gv_config->put_string(SECTION,
                           "AspectRatio",
                           x11_aspect_ratios[i].label);
      break;
      }
    }  
  gv_config->put_bool(SECTION, "ForceAspectRatio", force_aspect_ratio);
  gv_config->put_bool(SECTION, "CheckXV", check_xv);
  gv_config->put_bool(SECTION, "ForceXV", force_xv);
  }


void GVOutputVideoX11::print_info(ostream & output)
  {
  switch(video_mode)
    {
    case GV_X11_VIDEO_MODE_XSHM:
      output << "Using XShm";
      break;
    case GV_X11_VIDEO_MODE_XV_SHM:
      {
      char fourcc[5];
      ((int*)fourcc)[0] = xv_format;
      fourcc[4] = '\0';
      output << "Using XVideo and Shm\n"
             << "XVideo Port: " << xv_port << "\n"
             << "Image Format: " << fourcc << "\n";
      }
    }
  }


GVOutputVideoX11::ConfigWindow::ConfigWindow(GVOutputVideoX11 * _plugin) :
  GVConfigWindow(_plugin->attributes->real_name),
  aspect_ratio_menu("Aspect Ratio")
  {
  table = gvgtk_table_new(3, 2, 0);
  plugin = _plugin;

  aspect_ratio_menu.create_menu(x11_aspect_ratios,
                               sizeof(x11_aspect_ratios)/
                               sizeof(GVConfigMenuItemDouble));
  
  force_aspect_ratio = gtk_check_button_new_with_label("Force Aspect Ratio");
  check_xv = gtk_check_button_new_with_label("Check for XVideo");
  force_xv = gtk_check_button_new_with_label("Force XVideo");
  gtk_widget_show(force_aspect_ratio);
  gtk_widget_show(check_xv);
  gtk_widget_show(force_xv);
  gtk_table_attach_defaults(GTK_TABLE(table), force_aspect_ratio, 0, 1, 0, 1);
  aspect_ratio_menu.attach(table, 1, 0);
  
  gtk_table_attach_defaults(GTK_TABLE(table), check_xv, 0, 1, 2, 3);
  gtk_table_attach_defaults(GTK_TABLE(table), force_xv, 1, 2, 2, 3);
  
  gtk_widget_show(table);
  gtk_box_pack_start_defaults(GTK_BOX(mainbox), table);
  }

void GVOutputVideoX11::ConfigWindow::get_values()
  {
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(force_aspect_ratio),
                               plugin->force_aspect_ratio);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check_xv),
                               plugin->check_xv);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(force_xv),
                               plugin->force_xv);
  aspect_ratio_menu.set(plugin->aspect_ratio);
  }

void GVOutputVideoX11::ConfigWindow::apply_values()
  {
  plugin->force_aspect_ratio =
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(force_aspect_ratio));
  plugin->check_xv =
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(check_xv));
  plugin->force_xv =
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(force_xv));
  plugin->aspect_ratio = aspect_ratio_menu.get_double();
  }

GVOutputVideoX11::ConfigWindow::~ConfigWindow()
  {
    
  }

#define PLUGIN_NAME GVOutputVideoX11
#include "interface.cc"
