/*****************************************************************

  ov_x11_shm.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// X11 video driver (XShm subsystem)
#include <iostream>

using std::ostream;
using std::cerr;

#include "ov_x11.h"

#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>

#include <stdio.h>

static int x11_error = 0;

static int __dummy(Display* , XErrorEvent*)
  {
  x11_error=1;
  return true;
  }

void GVOutputVideoX11::destroy_xshm_image(XImage * image, int num, int total)
  {
  if(num == total -1)
    XShmDetach(win->x_display, shminfo );
  XDestroyImage(image);

  if(num == total -1)
    {
    shmdt( shminfo->shmaddr );
    delete shminfo;
    shminfo = (XShmSegmentInfo*)0;
    }
  
  }

void GVOutputVideoX11::alloc_frames_xshm(GVVideoFrame * frame, int num)
  {
  for(int i = 0; i < num; i++)
    {
    XImage * image = create_xshm_image(i, num);
    frame[i].pixels = (guchar*)(image->data);
    frame[i].data = (gpointer)(image);
    frame[i].pixels_rows = new guchar*[video_format.height];
    for(int j = 0; j < video_format.height; j++)
      frame[i].pixels_rows[j]
        = &(frame->pixels[video_format.width*i*video_format.bytes_per_pixel]);
    }
  }

void GVOutputVideoX11::free_frames_xshm(GVVideoFrame * frame, int num)
  {
  for(int i = 0; i < num; i++)
    {
    destroy_xshm_image((XImage *)(frame[i].data), i, num);
    delete frame[i].pixels_rows;
    }
  }

int GVOutputVideoX11::check_shm_support()
  {
/*
 * Check if the X Shared Memory extension is available.
 * Return:  0 = not available
 *          1 = shared XImage support available
 *          2 = shared Pixmap support available also
 */
  int major, minor, ignore;
  Bool pixmaps;
  if (XQueryExtension(win->x_display, "MIT-SHM", &ignore, &ignore, &ignore ))
    {
    if (XShmQueryVersion( win->x_display, &major, &minor, &pixmaps )==True)
      return (pixmaps==True) ? 2 : 1;
    else
      return 0;
    }
  else
    return 0;
  }

void GVOutputVideoX11::init_xshm()
  {
  shm_image = create_xshm_image(0, 1);
  }

void GVOutputVideoX11::put_frame_xshm(GVVideoFrame * frame)
  {
  XSetErrorHandler(__dummy);

  //  cerr << "Put Frame xshm, XImage = "
  //       << ((XImage*)(frame->data)) << "\n";
  
  XShmPutImage(win->x_display,win->x_win, 
               win->x_gc, (XImage*)(frame->data),
               0, 0, 0, 0, video_format.width, video_format.height, False);

  XSync(win->x_display, False);  
  XSetErrorHandler(NULL);
  XFlush(win->x_display);
  }

XImage * GVOutputVideoX11::create_xshm_image(int num, int total)
  {
  int depth = win->get_depth();
  

  XImage * ret;
  if(!num)
    {
    shminfo = new XShmSegmentInfo;
    memset(shminfo, 0, sizeof(XShmSegmentInfo));
    }
  if(!num)
    {
    //    cerr << "Creating Image, depth: " << render_depth << "\n";
    ret = XShmCreateImage(win->x_display,win->x_visual,
                          depth, ZPixmap, NULL, shminfo,
                          video_format.width,
                          video_format.height);
    //    cerr << "Got depth: " << ret->bits_per_pixel << "\n";
    }
  else
    {
    char * data = (char*)(shminfo->shmaddr) + num*video_format.buffer_size;
    //    cerr << "Buffer size: " << video_format.buffer_size << "\n";
    ret = XShmCreateImage(win->x_display,win->x_visual,
                          depth, ZPixmap, data, shminfo,
                          video_format.width,
                          video_format.height);
    ret->data = data;
    }

  if(!ret)
    {
    cerr << "XShmCreateImage failed\n";
    return (XImage*)0;
    }
  
  if(!num)
    {
    //    cerr << "First allocated Frame\n";
    
    // Create the real values

    render_depth = ret->bits_per_pixel;

    // Create shared memory

    shminfo->shmid=shmget(IPC_PRIVATE,
                         total*(video_format.height*ret->bytes_per_line+4),
                         IPC_CREAT|0777);
    
    if(shminfo->shmid<0)
      {
      cerr << "shmget failed\n";
      return (XImage *)0;
      }
    shminfo->shmaddr = (char*)shmat(shminfo->shmid, (const void*)0, 0 );
    if(shminfo->shmaddr == (char *) -1)
      cerr << "shmat failed\n";
    ret->data = shminfo->shmaddr;
    shminfo->readOnly=False;
    }

  //  cerr << "Allocated shared Memory "
  //       << total*(video_format.height*ret->bytes_per_line+4) << " Bytes"
  //       << " Start Adress: " << (void*)(shminfo->shmaddr) << "\n";
  
  
  //  ret->data += num * (video_format.height * ret->bytes_per_line + 4); 
  
  //  cerr << "ret->data: " << (void*)ret->data << "\n";
  
  if(num == total - 1)
    {
    //    cerr << "Last allocated Frame\n";
    XSetErrorHandler(__dummy);
    if(!XShmAttach(win->x_display,shminfo))
      cerr << "XShmAttach failed\n";
    shmctl(shminfo->shmid, IPC_RMID, 0 ); /* nobody else needs it */
    XSync(win->x_display, False);
    XSetErrorHandler(NULL);
    //    XFlush(win->x_display);
    }
  return ret;
  }
