#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <X11/Xatom.h>

#include "remote.h"
#include <stdio.h>
#include <limits.h>

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "config.h"

#define REMOTE_BUFFERSIZE (PATH_MAX*2) /* Should be enough */
#define REMOTE_TIMEOUT    1000   /* Timeout for waiting for confirmation */
#define DONE_TIMEOUT    60000    /* Timeout for waiting for done */

static int x11_error = 0;

/* Forward declarations */

static int gv_remote_send_command(GVRemote * remote,
                                  const char * command, int wait);

static int gv_remote_send_command_with_arg(GVRemote * remote,
                                            const char * command,
                                            const char * arg, int wait);

static void gv_remote_init(GVRemote * remote, int only_if_exists);

static int gv_remote_ok(GVRemote * remote);

static int gv_remote_wait_string(GVRemote * remote, const char * str,
                                 int timeout);

static int gv_remote_wait(GVRemote * remote, int wait);

/* X11 error handler */

static int __dummy(Display* diaply, XErrorEvent* event)
  {
  x11_error = 1;
  fprintf(stderr, "X11 error caught\n");
  return TRUE;
  }


/*
 *   Send command to the remote, return 1 if succeeded and 0
 *   if error or timeout
 */

static int gv_remote_send_command(GVRemote * remote,
                                  const char * _command, int wait)
  {
  int (*old_handler)(Display* , XErrorEvent*) = XSetErrorHandler(__dummy);

  char * command = g_strdup(_command);
  char * arg = g_strdup(GV_REMOTE_DUMMY_ARG);
    
  x11_error = 0;
  gdk_property_change(remote->window, remote->arg_atom,
                      XA_STRING, 8, GDK_PROP_MODE_REPLACE,
                      (guchar*)(arg), strlen(GV_REMOTE_DUMMY_ARG)+1);
  gdk_property_change(remote->window, remote->command_atom,
                      XA_STRING, 8, GDK_PROP_MODE_REPLACE,
                      (guchar*)(command), strlen(command)+1);
  gdk_flush();
  XSetErrorHandler(old_handler);

  g_free(command);
  g_free(arg);
  
  if(x11_error)
    return 0;

  return gv_remote_wait(remote, wait);
  }

static int gv_remote_send_command_with_arg(GVRemote * remote,
                                            const char * _command,
                                            const char * _arg, int wait)
  {
  int (*old_handler)(Display* , XErrorEvent*) = XSetErrorHandler(__dummy);

  char * command = g_strdup(_command);
  char * arg = g_strdup(_arg);
  
  x11_error = 0;

  strcpy(remote->command, command);
  strcpy(remote->arg, arg);
    
  // We change first the argument and then the command
  
  gdk_property_change(remote->window, remote->arg_atom,
                      XA_STRING, 8, GDK_PROP_MODE_REPLACE,
                      (guchar*)(arg), strlen(arg)+1);

  gdk_property_change(remote->window, remote->command_atom,
                      XA_STRING, 8, GDK_PROP_MODE_REPLACE,
                      (guchar*)(command), strlen(command)+1);
  gdk_flush();
  XSetErrorHandler(old_handler);

  g_free(command);
  g_free(arg);
  
  if(x11_error)
    return 0;

  if(!wait)
    return 1;
  
  /* wait for confirmation */

  if(gv_remote_ok(remote))
    return 1;
  return 0;
  }


/*
 *   Get a command and arg from the remote,
 *   return TRUE on success
 */

int gv_remote_get_command(GVRemote * remote)
  {
  GdkAtom actual_property_type = GDK_NONE;
  gint actual_length = 0;
  gint actual_format = 0;
  
  gdk_property_get(remote->window,
                   remote->command_atom,
                   XA_STRING,
                   0,
                   REMOTE_BUFFERSIZE-1,
                   FALSE,
                   &actual_property_type,
                   &actual_format,
                   &actual_length,
                   (guchar**)(&(remote->command)));
  if((actual_property_type != XA_STRING) ||
     (actual_format != 8))
    {
    fprintf(stderr, "Actual property type: %ld (should be %ld)\n",
            actual_property_type, XA_STRING);
    fprintf(stderr, "Actual format: %d (should be 8)\n",
            actual_format);
    return 0;
    }
  
  actual_property_type = GDK_NONE;
  actual_length = 0;
  actual_format = 0;
  
  gdk_property_get(remote->window,
                   remote->arg_atom,
                   XA_STRING,
                   0,
                   REMOTE_BUFFERSIZE-1,
                   FALSE,
                   &actual_property_type,
                   &actual_format,
                   &actual_length,
                   (guchar**)(&(remote->arg)));

  gdk_flush();
  
  if((actual_property_type != XA_STRING) ||
     (actual_format != 8))
    {
    fprintf(stderr, "Actual property type: %ld (should be %ld)\n",
            actual_property_type, XA_STRING);
    fprintf(stderr, "Actual format: %d (should be 8)\n",
            actual_format);
    return 0;
    }

  /* Send confirmation and return 0 if we got a test request */
  
  if(!strcmp(remote->command, GV_REMOTE_TEST_REQUEST))
    {
    gv_remote_confirm(remote);
    return 0;
    }
  else if(!strcmp(remote->command, GV_REMOTE_ANSWER) ||
          !strcmp(remote->command, GV_REMOTE_DONE))
    return 0;
  return 1;
  }

static void gv_remote_init(GVRemote * remote, int only_if_exists)
  {
  remote->command_atom = gdk_atom_intern(GV_REMOTE_COMMAND_ATOM,
                                         only_if_exists);

  remote->arg_atom = gdk_atom_intern(GV_REMOTE_ARG_ATOM,
                                     only_if_exists);

  remote->command = g_new(char, REMOTE_BUFFERSIZE);
  remote->arg = g_new(char, REMOTE_BUFFERSIZE);
  }

// Wait until the command_atom get the string str
// timeout is in milliseconds

static int gv_remote_wait_string(GVRemote * remote, const char * str,
                                 int timeout)
  {
  struct timeval delay_duration;
  int answer = 0;
  int i;

  GdkAtom actual_property_type;
  gint actual_length;
  gint actual_format;
    
  int (*old_handler)(Display* , XErrorEvent*);

  old_handler = XSetErrorHandler(__dummy);
  
  delay_duration.tv_sec = 0;
  delay_duration.tv_usec = 1000;

  x11_error = 0;
  fprintf(stderr, "Waiting for confirmation string %s...", str);
  for(i = 0; i < timeout; i++)
    {
    gdk_property_get(remote->window,
                     remote->command_atom,
                     XA_STRING,
                     0,
                     REMOTE_BUFFERSIZE-1,
                     FALSE,
                     &actual_property_type,
                     &actual_format,
                     &actual_length,
                     (guchar**)(&(remote->command)));

    if(x11_error)
      break;
    
    if((actual_property_type != XA_STRING) ||
       (actual_format != 8))
      break;
    
    remote->command[actual_length] = '\0';

    if(!strcmp(remote->command, str))
      {
      answer = 1;
      break;
      }
    select(0,  NULL,  NULL, NULL, &delay_duration);
    }
  if(answer)
    fprintf(stderr, "Ok\n");
  else
    fprintf(stderr, "Failed\n");
    
  XSetErrorHandler(old_handler);
  return answer;
  
  }

// Wait until the command_atom get the string "OK"

static int gv_remote_ok(GVRemote * remote)
  {
  return gv_remote_wait_string(remote, GV_REMOTE_ANSWER,
                               REMOTE_TIMEOUT);
  }

GVRemote * gv_remote_find()
  {
  char * window_id_file;

  guint32 window_id;
  
  FILE * input;

  GVRemote * ret = 0;

  window_id_file = g_strdup_printf("%s/.%s/window_id",
                                   g_get_home_dir(), PACKAGE);
  
  input = fopen(window_id_file, "r");
  if(!input)
    return (GVRemote*)0;
  if(fscanf(input, "%d", &window_id) != 1)
    return (GVRemote*)0;

  ret = g_new(GVRemote, 1);

  fprintf(stderr, "Window ID: %d\n" ,window_id);
  ret->window = gdk_window_foreign_new(window_id);

  gv_remote_init(ret, 1);

  /* Atoms not defined => no other instance is running */
  
  if((ret->command_atom == GDK_NONE) || (ret->arg_atom == GDK_NONE))
    {
    //    remove(window_id_file);
    g_free(window_id_file);
    gv_remote_destroy(ret);
    return (GVRemote*)0;
    }

  /* Try to talk to the other instance, send the string "TEST" */

  if(!gv_remote_send_command(ret, GV_REMOTE_TEST_REQUEST, 1))
    {
    //   remove(window_id_file);
    g_free(window_id_file);
    gv_remote_destroy(ret);
    return (GVRemote*)0;
    }
  g_free(window_id_file);
  ret->receiver = 0;
  return ret;
  }

void gv_remote_destroy(GVRemote * remote)
  {
  char * window_id_file;
  if(remote->receiver)
    {
    window_id_file = g_strdup_printf("%s/.%s/window_id",
                                     g_get_home_dir(), PACKAGE);
    remove(window_id_file);
    g_free(window_id_file);
    }
  g_free(remote->command);
  g_free(remote->arg);
  g_free(remote);
  
  }

GVRemote * gv_remote_create(GdkWindow * window)
  {
  GVRemote * ret;
  char * window_id_file = g_strdup_printf("%s/.%s/window_id",
                                          g_get_home_dir(), PACKAGE);
  FILE * output = fopen(window_id_file, "w");

  fprintf(stderr, "Creating Remote, Window ID: %ld\n",
          GDK_WINDOW_XWINDOW(window));
  
  fprintf(output, "%ld\n", GDK_WINDOW_XWINDOW(window));
  fclose(output);
  
  ret = g_new(GVRemote, 1);
  ret->window = window;
  
  gv_remote_init(ret, 0);
  ret->receiver = 1;
  return ret;
  }

/* Start playing */

void gv_remote_play(GVRemote * remote)
  {
  gv_remote_send_command(remote, GV_REMOTE_PLAY, 2);
  gv_remote_get_done(remote);
  }

/* Stop playing */

void gv_remote_stop(GVRemote * remote)
  {
  gv_remote_send_command(remote, GV_REMOTE_STOP, 2);
  gv_remote_get_done(remote);
  }

/* Next track */

void gv_remote_next(GVRemote * remote)
  {
  gv_remote_send_command(remote, GV_REMOTE_NEXT, 2);
  gv_remote_get_done(remote);
  }

/* Previous track */

void gv_remote_previous(GVRemote * remote)
  {
  gv_remote_send_command(remote, GV_REMOTE_PREVIOUS, 2);
  gv_remote_get_done(remote);
  }

/* Add a file to the playlist */

void gv_remote_add_file(GVRemote * remote,
                        const char * filename)
  {
  gv_remote_send_command_with_arg(remote, GV_REMOTE_ADDFILE, filename, 2);
  gv_remote_get_done(remote);
  }

void gv_remote_play_file(GVRemote * remote,
                         const char * filename)
  {
  gv_remote_send_command_with_arg(remote, GV_REMOTE_PLAYFILE, filename, 2);
  gv_remote_get_done(remote);
  }


/* Load a plugin */

void gv_remote_set_input(GVRemote * remote,
                         const char * plugin_name)
  {
  gv_remote_send_command_with_arg(remote, GV_REMOTE_SET_INPUT,
                                  plugin_name, 2);
  gv_remote_get_done(remote);
  }

/* Load a plugin */

void gv_remote_set_video_output(GVRemote * remote,
                                const char * plugin_name)
  {
  gv_remote_send_command_with_arg(remote, GV_REMOTE_SET_VIDEO_OUTPUT,
                                  plugin_name, 2);
  gv_remote_get_done(remote);
  }

/* Load a plugin */

void gv_remote_set_audio_output(GVRemote * remote,
                                const char * plugin_name)
  {
  gv_remote_send_command_with_arg(remote, GV_REMOTE_SET_AUDIO_OUTPUT,
                                  plugin_name, 2);
  gv_remote_get_done(remote);
  }

void gv_remote_confirm(GVRemote * remote)
  {
  fprintf(stderr, "Sending confirmation\n");
  gv_remote_send_command(remote, GV_REMOTE_ANSWER, 0);
  }

// Send the done confirmation
  
void gv_remote_send_done(GVRemote * remote)
  {
  fprintf(stderr, "Remote command done\n");
  gv_remote_send_command(remote, GV_REMOTE_DONE, 0);
  }

// Wait for the done confirmation

int gv_remote_get_done(GVRemote * remote)
  {
  return gv_remote_wait_string(remote, GV_REMOTE_DONE,
                               DONE_TIMEOUT);

  }

static int gv_remote_wait(GVRemote * remote, int wait)
  {
  switch(wait)
    {
    case 0:
      return 1;
      break;
    case 1:
      return gv_remote_ok(remote);
      break;
    case 2:
      if(!gv_remote_ok(remote))
        {
        return 0;
        break;
        }
      if(!gv_remote_get_done(remote))
        {
        return 0;
        break;
        }
    }
  return 1;
  }
