/*****************************************************************

  videoconvert.h

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Typedefs for conversion functions

// Conversion between packed formats

typedef void (*GVConvert_PACKED_PACKED)(guchar * in_pixels,
                                        guchar * out_pixels,
                                        int num_pixels);

typedef void (*GVConvert_PACKED_ALPHA_PACKED)(guchar * in_pixels,
                                              guchar * out_pixels,
                                              int num_pixels,
                                              guchar * background);

// Conversion from packed to planar YUV

typedef void (*GVConvert_PACKED_PLANAR)(guchar * in_pixels,
                                        guchar * out_y,
                                        guchar * out_u,
                                        guchar * out_v,
                                        int width,
                                        int height);

typedef void (*GVConvert_PACKED_ALPHA_PLANAR)(guchar * in_pixels,
                                              guchar * out_y,
                                              guchar * out_u,
                                              guchar * out_v,
                                              int width,
                                              int height,
                                              guchar * background);

// Conversion from planar to packed YUV

typedef void(* GVConvert_PLANAR_PACKED)(guchar * in_y,
                                        guchar * in_u,
                                        guchar * in_v,
                                        guchar * out_pixels,
                                        int width,
                                        int height);

// Conversion routines planar to planar

typedef void(* GVConvert_PLANAR_PLANAR)(guchar ** in_y_rows,
                                        guchar ** in_u_rows,
                                        guchar ** in_v_rows,
                                        guchar ** out_y_rows,
                                        guchar ** out_u_rows,
                                        guchar ** out_v_rows,
                                        int width,
                                        int height);

// Gmerlin video format converter
// Converts between gmerlin colorspaces
// In the future, maybe scaling and cropping will also be supported

class GVVideoConverter
  {
  int do_scale;

  // Background color from globalconfig (RGB)
  
  guchar background[3];
  
  enum ConvertMode
  {
    Convert_PACKED_PACKED,
    Convert_PLANAR_PACKED,
    Convert_PACKED_PLANAR,
    Convert_PLANAR_PLANAR,
    Convert_PACKED_ALPHA_PACKED,
    Convert_PACKED_ALPHA_PLANAR,
    Convert_NONE
  };

  ConvertMode convert_mode;
    
  GVVideoFormat input_format;
  GVVideoFormat output_format;

  int input_num_pixels;
  int output_num_pixels;

  // n*(n-1) Pointers to conversion functions
  // (n: Number of colorspaces)
  
  GVConvert_PACKED_PACKED RGB16_to_BGR16;
  GVConvert_PACKED_PACKED RGB16_to_RGB24;
  GVConvert_PACKED_PACKED RGB16_to_BGR24;
  GVConvert_PACKED_PACKED RGB16_to_RGB32;
  GVConvert_PACKED_PACKED RGB16_to_BGR32;
  GVConvert_PACKED_PACKED RGB16_to_YUY2;
  GVConvert_PACKED_PLANAR RGB16_to_YUV420_P;
  GVConvert_PACKED_PLANAR RGB16_to_YUV422_P;

  GVConvert_PACKED_PACKED BGR16_to_RGB16;
  GVConvert_PACKED_PACKED BGR16_to_RGB24;
  GVConvert_PACKED_PACKED BGR16_to_BGR24;
  GVConvert_PACKED_PACKED BGR16_to_RGB32;
  GVConvert_PACKED_PACKED BGR16_to_BGR32;
  GVConvert_PACKED_PACKED BGR16_to_YUY2;
  GVConvert_PACKED_PLANAR BGR16_to_YUV420_P;
  GVConvert_PACKED_PLANAR BGR16_to_YUV422_P;

  GVConvert_PACKED_PACKED RGB24_to_RGB16;
  GVConvert_PACKED_PACKED RGB24_to_BGR16;
  GVConvert_PACKED_PACKED RGB24_to_BGR24;
  GVConvert_PACKED_PACKED RGB24_to_RGB32;
  GVConvert_PACKED_PACKED RGB24_to_BGR32;
  GVConvert_PACKED_PACKED RGB24_to_YUY2;
  GVConvert_PACKED_PLANAR RGB24_to_YUV420_P;
  GVConvert_PACKED_PLANAR RGB24_to_YUV422_P;

  GVConvert_PACKED_PACKED BGR24_to_RGB16;
  GVConvert_PACKED_PACKED BGR24_to_BGR16;
  GVConvert_PACKED_PACKED BGR24_to_RGB24;
  GVConvert_PACKED_PACKED BGR24_to_RGB32;
  GVConvert_PACKED_PACKED BGR24_to_BGR32;
  GVConvert_PACKED_PACKED BGR24_to_YUY2;
  GVConvert_PACKED_PLANAR BGR24_to_YUV420_P;
  GVConvert_PACKED_PLANAR BGR24_to_YUV422_P;

  GVConvert_PACKED_PACKED RGB32_to_RGB16;
  GVConvert_PACKED_PACKED RGB32_to_BGR16;
  GVConvert_PACKED_PACKED RGB32_to_RGB24;
  GVConvert_PACKED_PACKED RGB32_to_BGR24;
  GVConvert_PACKED_PACKED RGB32_to_BGR32;
  GVConvert_PACKED_PACKED RGB32_to_RGBA32;
  GVConvert_PACKED_PACKED RGB32_to_YUY2;
  GVConvert_PACKED_PLANAR RGB32_to_YUV420_P;
  GVConvert_PACKED_PLANAR RGB32_to_YUV422_P;
  
  GVConvert_PACKED_PACKED BGR32_to_RGB16;
  GVConvert_PACKED_PACKED BGR32_to_BGR16;
  GVConvert_PACKED_PACKED BGR32_to_RGB24;
  GVConvert_PACKED_PACKED BGR32_to_BGR24;
  GVConvert_PACKED_PACKED BGR32_to_RGB32;
  GVConvert_PACKED_PACKED BGR32_to_YUY2;
  GVConvert_PACKED_PLANAR BGR32_to_YUV420_P;
  GVConvert_PACKED_PLANAR BGR32_to_YUV422_P;

  GVConvert_PACKED_ALPHA_PACKED RGBA32_to_RGB16;
  GVConvert_PACKED_ALPHA_PACKED RGBA32_to_BGR16;
  GVConvert_PACKED_ALPHA_PACKED RGBA32_to_RGB24;
  GVConvert_PACKED_ALPHA_PACKED RGBA32_to_BGR24;
  GVConvert_PACKED_ALPHA_PACKED RGBA32_to_RGB32;
  GVConvert_PACKED_ALPHA_PACKED RGBA32_to_BGR32;
  GVConvert_PACKED_ALPHA_PACKED RGBA32_to_YUY2;
  GVConvert_PACKED_ALPHA_PLANAR RGBA32_to_YUV420_P;
  GVConvert_PACKED_ALPHA_PLANAR RGBA32_to_YUV422_P;
  
  GVConvert_PLANAR_PACKED YUV420_P_to_RGB16;
  GVConvert_PLANAR_PACKED YUV420_P_to_BGR16;
  GVConvert_PLANAR_PACKED YUV420_P_to_RGB24;
  GVConvert_PLANAR_PACKED YUV420_P_to_BGR24;
  GVConvert_PLANAR_PACKED YUV420_P_to_RGB32;
  GVConvert_PLANAR_PACKED YUV420_P_to_BGR32;
  GVConvert_PLANAR_PACKED YUV420_P_to_YUY2;
  GVConvert_PLANAR_PLANAR YUV420_P_to_YUV422_P;

  GVConvert_PLANAR_PACKED YUV422_P_to_RGB16;
  GVConvert_PLANAR_PACKED YUV422_P_to_BGR16;
  GVConvert_PLANAR_PACKED YUV422_P_to_RGB24;
  GVConvert_PLANAR_PACKED YUV422_P_to_BGR24;
  GVConvert_PLANAR_PACKED YUV422_P_to_RGB32;
  GVConvert_PLANAR_PACKED YUV422_P_to_BGR32;
  GVConvert_PLANAR_PACKED YUV422_P_to_YUY2;
  GVConvert_PLANAR_PLANAR YUV422_P_to_YUV420_P;
  
  GVConvert_PACKED_PACKED YUY2_to_RGB16;
  GVConvert_PACKED_PACKED YUY2_to_BGR16;
  GVConvert_PACKED_PACKED YUY2_to_RGB24;
  GVConvert_PACKED_PACKED YUY2_to_BGR24;
  GVConvert_PACKED_PACKED YUY2_to_RGB32;
  GVConvert_PACKED_PACKED YUY2_to_BGR32;
  GVConvert_PACKED_PLANAR YUY2_to_YUV420_P;
  GVConvert_PACKED_PLANAR YUY2_to_YUV422_P;
  
  // Actual functions which will be called
  // This eliminates the need of a huge switch statement during the conversion

  GVConvert_PACKED_PACKED convert_packed_packed; // packed -> packed
  GVConvert_PLANAR_PACKED convert_planar_packed; // planar -> packed
  GVConvert_PACKED_PLANAR convert_packed_planar; // packed -> planar
  GVConvert_PLANAR_PLANAR convert_planar_planar; // planar -> planar

  // packed (with alpha) -> packed
  
  GVConvert_PACKED_ALPHA_PACKED convert_packed_alpha_packed; 

  // packed (with alpha) -> planar
  
  GVConvert_PACKED_ALPHA_PLANAR convert_packed_alpha_planar; 
  
  void init_c_funcs();   // Set all function pointers to the C functions
  void init_mmx_funcs(); // Set the functons to MMX versions if avaliable

  void reset(); // Set all functions to NULL
  
public:

  void convert(GVVideoFrame * input, GVVideoFrame * output);

  void init(const GVVideoFormat * _input_format,
            const GVVideoFormat * _output_format);
  
  };
