/*****************************************************************
                                                                  
  videoconvert_c.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// C versions of gmerlins video conversion routines

#include "config.h"
#include <glib.h>
#include <iostream>
#include <iomanip.h>
#include <string.h>
#include "gvtime.h"
#include "videoformat.h"
#include "videoconvert.h"

// YUV tables

static int * r_to_y = (int*)0;
static int * g_to_y = (int*)0;
static int * b_to_y = (int*)0;

static int * r_to_u = (int*)0;
static int * g_to_u = (int*)0;
static int * b_to_u = (int*)0;

static int * r_to_v = (int*)0;
static int * g_to_v = (int*)0;
static int * b_to_v = (int*)0;

static int * y_to_rgb = (int*)0;
static int * v_to_r = (int*)0;
static int * u_to_g = (int*)0;
static int * v_to_g = (int*)0;
static int * u_to_b = (int*)0;

static guchar * map_5_to_8 = (guchar*)0;
static guchar * map_6_to_8 = (guchar*)0;

void __init_c()
  {
  if(r_to_y)
    return;
  
  // Set up the tables

  r_to_y = new int[0x100];
  g_to_y = new int[0x100];
  b_to_y = new int[0x100];

  r_to_u = new int[0x100];
  g_to_u = new int[0x100];
  b_to_u = new int[0x100];

  r_to_v = new int[0x100];
  g_to_v = new int[0x100];
  b_to_v = new int[0x100];

  y_to_rgb = new int[0x100];
  v_to_r   = new int[0x100];
  u_to_g   = new int[0x100];
  v_to_g   = new int[0x100];
  u_to_b   = new int[0x100];
  
  for(int i = 0; i < 0x100; i++)
    {
    // RGB to YUV conversion

    r_to_y[i] = (int)(0.2989*0x10000 * i);
    g_to_y[i] = (int)(0.5866*0x10000 * i);
    b_to_y[i] = (int)(0.1145*0x10000 * i);
    
    r_to_u[i] = (int)(-0.1687*0x10000 * i);
    g_to_u[i] = (int)(-0.3312*0x10000 * i);
    b_to_u[i] = (int)( 0.5000*0x10000 * i + 0x800000);
    
    r_to_v[i] = (int)( 0.5000*0x10000 * i);
    g_to_v[i] = (int)(-0.4183*0x10000 * i);
    b_to_v[i] = (int)(-0.0816*0x10000 * i + 0x800000);

    // YUV to RGB conversion

    y_to_rgb[i] = (i << 16);
    v_to_r[i]   = (int)( 1.4022 * (i - 0x80) * 0x10000);
    u_to_g[i]   = (int)(-0.3456 * (i - 0x80) * 0x10000);
    v_to_g[i]   = (int)(-0.7145 * (i - 0x80) * 0x10000);
    u_to_b[i]   = (int)( 1.7710 * (i - 0x80) * 0x10000);
    }

  map_6_to_8 = new guchar[64];
  map_5_to_8 = new guchar[32];

  for(int i = 0; i < 64; i++)
    map_6_to_8[i] = (int)((float)i/63.0*255.0+0.5);

  for(int i = 0; i < 32; i++)
    map_5_to_8[i] = (int)((float)i/31.0*255.0+0.5);
  }

// Conversions between RGB formats

// Some macros

#define SWAP_RGB_24(src,dst) dst[2]=src[0];dst[1]=src[1];dst[0]=src[2]

#define COPY_RGB_24(src,dst) dst[0]=src[0];dst[1]=src[1];dst[2]=src[2]

// Convert RGBA to RGB and blend in a background
// Requires the following variables defines in the calling function:
// int tmp_red;
// int tmp_green;
// int tmp_blue;
// guchar transp;

#define RGBA_TO_RGB(src,dst,bg) transp=0xff-src[3];\
                                tmp_red=(src[0]*src[3]+bg[0]*transp)>>8;\
                                tmp_green=(src[1]*src[3]+bg[1]*transp)>>8;\
                                tmp_blue=(src[2]*src[3]+bg[2]*transp)>>8;\
                                dst[0]=RECLIP(tmp_red);\
                                dst[1]=RECLIP(tmp_green);\
                                dst[2]=RECLIP(tmp_blue);

#define RGB_TO_YUV(r,g,b,y,u,v) y=(r_to_y[r]+g_to_y[g]+b_to_y[b])>>16;\
                                u=(r_to_u[r]+g_to_u[g]+b_to_u[b])>>16;\
                                v=(r_to_v[r]+g_to_v[g]+b_to_v[b])>>16

#define RGB_TO_Y(r,g,b,y) y=(r_to_y[r]+g_to_y[g]+b_to_y[b])>>16

#define RECLIP(color) (guchar)((color>0xFF)?0xff:((color<0)?0:color))

#define YUV_TO_RGB(y,u,v,r,g,b) i_tmp=(y_to_rgb[y]+v_to_r[v])>>16;\
                                r=RECLIP(i_tmp);\
                                i_tmp=(y_to_rgb[y]+u_to_g[u]+v_to_g[v])>>16;\
                                g=RECLIP(i_tmp);\
                                i_tmp=(y_to_rgb[y]+u_to_b[u])>>16;\
                                b=RECLIP(i_tmp);

// Masks for BGR16 and RGB16 formats

#define RGB16_LOWER_MASK  0x001f
#define RGB16_MIDDLE_MASK 0x07e0
#define RGB16_UPPER_MASK  0xf800

// Extract unsigned char RGB values from 16 bit pixels

#define RGB16_TO_R(pixel) map_5_to_8[pixel & RGB16_LOWER_MASK]
#define RGB16_TO_G(pixel) map_6_to_8[(pixel & RGB16_MIDDLE_MASK)>>5]
#define RGB16_TO_B(pixel) map_5_to_8[(pixel & RGB16_UPPER_MASK)>>11]

#define BGR16_TO_B(pixel) map_5_to_8[pixel & RGB16_LOWER_MASK]
#define BGR16_TO_G(pixel) map_6_to_8[(pixel & RGB16_MIDDLE_MASK)>>5]
#define BGR16_TO_R(pixel) map_5_to_8[(pixel & RGB16_UPPER_MASK)>>11]

// Combine r, g and b values to a 16 bit rgb pixel

#define PACK_RGB16(r,g,b,pixel) pixel=(b>>3);\
                                pixel<<=6;\
                                pixel|=(g>>2);\
                                pixel<<=5;\
                                pixel|=(r>>3)


// Combine r, g and b values to a 16 bit bgr pixel

#define PACK_BGR16(r,g,b,pixel) pixel=(r>>3);\
                                pixel<<=6;\
                                pixel|=(g>>2);\
                                pixel<<=5;\
                                pixel|=(b>>3)


// Swap R and B values in 16 bit pixels

#define SWAP_RGB_16(pixel) (((pixel & RGB16_UPPER_MASK) >> 11)|\
                             (pixel & RGB16_MIDDLE_MASK)|\
                            ((pixel & RGB16_LOWER_MASK) << 11))

#define POINTER_TO_GUINT16(a) *((guint16*)a)

// Here follow n*(n-1) functions for converting colorspaces.
// If a neccesary function has not been defined yet, an error is triggered

// Input format: GVColorspaceRGB16

static void RGB16_to_BGR16_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 in_pixel;
  guint16 out_pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    in_pixel = *((guint16*)in_pixels);
    out_pixel = SWAP_RGB_16(in_pixel);
    *((guint16*)out_pixels) = out_pixel;
    in_pixels += 2;
    out_pixels += 2;
    }
  }

static void RGB16_to_RGB24_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  guint16 in_pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    in_pixel = POINTER_TO_GUINT16(in_pixels);

    out_pixels[0] = RGB16_TO_R(in_pixel);
    out_pixels[1] = RGB16_TO_G(in_pixel);
    out_pixels[2] = RGB16_TO_B(in_pixel);

    in_pixels += 2;
    out_pixels += 3;
    }
  }

static void RGB16_to_BGR24_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  guint16 in_pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    in_pixel = POINTER_TO_GUINT16(in_pixels);

    out_pixels[0] = RGB16_TO_B(in_pixel);
    out_pixels[1] = RGB16_TO_G(in_pixel);
    out_pixels[2] = RGB16_TO_R(in_pixel);

    in_pixels += 2;
    out_pixels += 3;
    }
  }

static void RGB16_to_RGB32_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 in_pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    in_pixel = POINTER_TO_GUINT16(in_pixels);

    out_pixels[2] = RGB16_TO_B(in_pixel);
    out_pixels[1] = RGB16_TO_G(in_pixel);
    out_pixels[0] = RGB16_TO_R(in_pixel);

    in_pixels += 2;
    out_pixels += 4;
    }
  }

static void RGB16_to_BGR32_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 in_pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    in_pixel = POINTER_TO_GUINT16(in_pixels);

    out_pixels[0] = RGB16_TO_B(in_pixel);
    out_pixels[1] = RGB16_TO_G(in_pixel);
    out_pixels[2] = RGB16_TO_R(in_pixel);

    in_pixels += 2;
    out_pixels += 4;
    }
  }

static void RGB16_to_YUY2_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once
  guint16 in_pixel;

  // Output: Y0 U0 Y1 V0
  
  for(int i = 0; i < imax; i++)
    {
    in_pixel = POINTER_TO_GUINT16(in_pixels);
    in_pixels += 2;
    
    RGB_TO_YUV(RGB16_TO_R(in_pixel),
               RGB16_TO_G(in_pixel),
               RGB16_TO_B(in_pixel),
               out_pixels[0], out_pixels[1], out_pixels[3]);

    in_pixel = POINTER_TO_GUINT16(in_pixels);
    in_pixels += 2;
    
    RGB_TO_Y(RGB16_TO_R(in_pixel),
             RGB16_TO_G(in_pixel),
             RGB16_TO_B(in_pixel),
             out_pixels[2]);
    out_pixels += 4;
    }
  }

static void RGB16_to_YUV420_P_c(guchar * in_pixels, guchar * out_y,
                                 guchar * out_u, guchar * out_v,
                                 int width, int height)
  {
  int imax = width / 2; // We take 4 pixels at once
  int jmax = height / 2;

  int rgb_offset = width * 2;

  guchar * in_pixels_yuv = in_pixels;
  guchar * in_pixels_y = in_pixels + width*2;
  guchar * out_pixels_yuv = out_y;
  guchar * out_pixels_y = out_y + width;

  guint16 in_pixel;
  
  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      in_pixel = POINTER_TO_GUINT16(in_pixels_yuv);
      in_pixels_yuv += 2;
      
      RGB_TO_YUV(RGB16_TO_R(in_pixel),
                 RGB16_TO_G(in_pixel),
                 RGB16_TO_B(in_pixel),
                 out_pixels_yuv[0], *out_u, *out_v);

      in_pixel = POINTER_TO_GUINT16(in_pixels_yuv);
      in_pixels_yuv += 2;
      
      RGB_TO_Y(RGB16_TO_R(in_pixel),
               RGB16_TO_G(in_pixel),
               RGB16_TO_B(in_pixel),
               out_pixels_yuv[1]);

      in_pixel = POINTER_TO_GUINT16(in_pixels_y);
      in_pixels_y += 2;
      
      RGB_TO_Y(RGB16_TO_R(in_pixel),
               RGB16_TO_G(in_pixel),
               RGB16_TO_B(in_pixel),
               out_pixels_y[0]);

      in_pixel = POINTER_TO_GUINT16(in_pixels_y);
      in_pixels_y += 2;
      
      RGB_TO_Y(RGB16_TO_R(in_pixel),
               RGB16_TO_G(in_pixel),
               RGB16_TO_B(in_pixel),
               out_pixels_y[1]);

      out_pixels_yuv += 2;
      out_pixels_y += 2;
      out_u++;
      out_v++;
      }
    in_pixels_yuv += rgb_offset;
    in_pixels_y += rgb_offset;
    out_pixels_y += width;
    out_pixels_yuv += width;
    }
  }

static void RGB16_to_YUV422_P_c(guchar * in_pixels, guchar * out_y,
                                guchar * out_u, guchar * out_v,
                                int width, int height)
  {
  int imax = width / 2; // We take 2 pixels at once
  
  guint16 in_pixel;
  
  for(int j = 0; j < height; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      in_pixel = POINTER_TO_GUINT16(in_pixels);
      in_pixels += 2;
      
      RGB_TO_YUV(RGB16_TO_R(in_pixel),
                 RGB16_TO_G(in_pixel),
                 RGB16_TO_B(in_pixel),
                 *(out_y++), *(out_u++), *(out_v++));
      
      in_pixel = POINTER_TO_GUINT16(in_pixels);
      in_pixels += 2;
      
      RGB_TO_Y(RGB16_TO_R(in_pixel),
               RGB16_TO_G(in_pixel),
               RGB16_TO_B(in_pixel),
               *(out_y++));
      }
    }
  }


// Input format: GVColorspaceBGR16

static void BGR16_to_RGB16_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  guint16 in_pixel;
  guint16 out_pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    in_pixel = *((guint16*)in_pixels);
    out_pixel = SWAP_RGB_16(in_pixel);
    *((guint16*)out_pixels) = out_pixel;
    }
  }

static void BGR16_to_RGB24_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  guint16 in_pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    in_pixel = POINTER_TO_GUINT16(in_pixels);

    out_pixels[0] = BGR16_TO_R(in_pixel);
    out_pixels[1] = BGR16_TO_G(in_pixel);
    out_pixels[2] = BGR16_TO_B(in_pixel);

    in_pixels += 2;
    out_pixels += 3;
    }
  }

static void BGR16_to_BGR24_c(guchar * in_pixels, guchar * out_pixels,
                               int num_pixels)
  {
  guint16 in_pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    in_pixel = POINTER_TO_GUINT16(in_pixels);

    out_pixels[2] = BGR16_TO_R(in_pixel);
    out_pixels[1] = BGR16_TO_G(in_pixel);
    out_pixels[0] = BGR16_TO_B(in_pixel);

    in_pixels += 2;
    out_pixels += 3;
    }
  }

static void BGR16_to_RGB32_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  guint16 in_pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    in_pixel = POINTER_TO_GUINT16(in_pixels);

    out_pixels[0] = BGR16_TO_R(in_pixel);
    out_pixels[1] = BGR16_TO_G(in_pixel);
    out_pixels[2] = BGR16_TO_B(in_pixel);
   
    in_pixels += 2;
    out_pixels += 4;
    }
  }

static void BGR16_to_BGR32_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  guint16 in_pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    in_pixel = POINTER_TO_GUINT16(in_pixels);

    out_pixels[2] = BGR16_TO_R(in_pixel);
    out_pixels[1] = BGR16_TO_G(in_pixel);
    out_pixels[0] = BGR16_TO_B(in_pixel);
   
    in_pixels += 2;
    out_pixels += 4;
    }
  }

static void BGR16_to_YUY2_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once
  guint16 in_pixel;

  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    in_pixel = POINTER_TO_GUINT16(in_pixels);
    in_pixels += 2;
    
    RGB_TO_YUV(BGR16_TO_R(in_pixel),
               BGR16_TO_G(in_pixel),
               BGR16_TO_B(in_pixel),
               out_pixels[0], out_pixels[1], out_pixels[3]);

    in_pixel = POINTER_TO_GUINT16(in_pixels);
    in_pixels += 2;
    
    RGB_TO_Y(BGR16_TO_R(in_pixel),
             BGR16_TO_G(in_pixel),
             BGR16_TO_B(in_pixel),
             out_pixels[2]);
    out_pixels += 4;
    }
  }

static void BGR16_to_YUV420_P_c(guchar * in_pixels, guchar * out_y,
                                guchar * out_u, guchar * out_v,
                                int width, int height)
  {
  int imax = width / 2; // We take 4 pixels at once
  int jmax = height / 2;

  int rgb_offset = width * 2;

  guchar * in_pixels_yuv = in_pixels;
  guchar * in_pixels_y = in_pixels + width*2;
  guchar * out_pixels_yuv = out_y;
  guchar * out_pixels_y = out_y + width;

  guint16 in_pixel;
  
  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      in_pixel = POINTER_TO_GUINT16(in_pixels_yuv);
      in_pixels_yuv += 2;
      
      RGB_TO_YUV(BGR16_TO_R(in_pixel),
                 BGR16_TO_G(in_pixel),
                 BGR16_TO_B(in_pixel),
                 out_pixels_yuv[0], *out_u, *out_v);

      in_pixel = POINTER_TO_GUINT16(in_pixels_yuv);
      in_pixels_yuv += 2;
      
      RGB_TO_Y(BGR16_TO_R(in_pixel),
               BGR16_TO_G(in_pixel),
               BGR16_TO_B(in_pixel),
               out_pixels_yuv[1]);

      in_pixel = POINTER_TO_GUINT16(in_pixels_y);
      in_pixels_y += 2;
      
      RGB_TO_Y(BGR16_TO_R(in_pixel),
               BGR16_TO_G(in_pixel),
               BGR16_TO_B(in_pixel),
               out_pixels_y[0]);

      in_pixel = POINTER_TO_GUINT16(in_pixels_y);
      in_pixels_y += 2;
      
      RGB_TO_Y(BGR16_TO_R(in_pixel),
               BGR16_TO_G(in_pixel),
               BGR16_TO_B(in_pixel),
               out_pixels_y[1]);

      out_pixels_yuv += 2;
      out_pixels_y += 2;
      out_u++;
      out_v++;
      }
    in_pixels_yuv += rgb_offset;
    in_pixels_y += rgb_offset;
    out_pixels_y += width;
    out_pixels_yuv += width;
    }
  }

static void BGR16_to_YUV422_P_c(guchar * in_pixels, guchar * out_y,
                                guchar * out_u, guchar * out_v,
                                int width, int height)
  {
  int imax = width / 2; // We take 2 pixels at once
  
  guint16 in_pixel;
  
  for(int j = 0; j < height; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      in_pixel = POINTER_TO_GUINT16(in_pixels);
      in_pixels += 2;
      
      RGB_TO_YUV(BGR16_TO_R(in_pixel),
                 BGR16_TO_G(in_pixel),
                 BGR16_TO_B(in_pixel),
                 *(out_y++), *(out_u++), *(out_v++));
      
      in_pixel = POINTER_TO_GUINT16(in_pixels);
      in_pixels += 2;
      
      RGB_TO_Y(BGR16_TO_R(in_pixel),
               BGR16_TO_G(in_pixel),
               BGR16_TO_B(in_pixel),
               *(out_y++));
      }
    }
  }


// Input format: GVColorspaceRGB24

static void RGB24_to_RGB16_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    PACK_RGB16(in_pixels[0],in_pixels[1],in_pixels[2],pixel);
    *((guint16*)out_pixels) = pixel;
    in_pixels += 3;
    out_pixels += 2;
    }
  }

static void RGB24_to_BGR16_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    PACK_BGR16(in_pixels[0],in_pixels[1],in_pixels[2],pixel);
    *((guint16*)out_pixels) = pixel;
    in_pixels += 3;
    out_pixels += 2;
    }
  }

static void RGB24_to_BGR24_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    SWAP_RGB_24(in_pixels,out_pixels);
    in_pixels += 3;
    out_pixels += 3;
    }
  }

static void RGB24_to_BGR32_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    SWAP_RGB_24(in_pixels,out_pixels);
    in_pixels += 3;
    out_pixels += 4;
    }
  }

static void RGB24_to_RGB32_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    COPY_RGB_24(in_pixels,out_pixels);
    in_pixels += 3;
    out_pixels += 4;
    }
  }

static void RGB24_to_YUV420_P_c(guchar * in_pixels, guchar * out_y,
                                 guchar * out_u, guchar * out_v,
                                 int width, int height)
  {
  int imax = width / 2; // We take 4 pixels at once
  int jmax = height / 2;

  int rgb_offset = width * 3;

  guchar * in_pixels_yuv = in_pixels;
  guchar * in_pixels_y = in_pixels + width*3;
  guchar * out_pixels_yuv = out_y;
  guchar * out_pixels_y = out_y + width;
  
  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      RGB_TO_YUV(in_pixels_yuv[0], in_pixels_yuv[1], in_pixels_yuv[2],
                 out_pixels_yuv[0], *out_u, *out_v);

      RGB_TO_Y(in_pixels_yuv[3], in_pixels_yuv[4], in_pixels_yuv[5],
               out_pixels_yuv[1]);
      RGB_TO_Y(in_pixels_y[0], in_pixels_y[1], in_pixels_y[2],
               out_pixels_y[0]);
      RGB_TO_Y(in_pixels_y[3], in_pixels_y[4], in_pixels_y[5],
               out_pixels_y[1]);

      out_pixels_yuv += 2;
      out_pixels_y += 2;
      out_u++;
      out_v++;
            
      in_pixels_y += 6;
      in_pixels_yuv += 6;

      }
    in_pixels_yuv += rgb_offset;
    in_pixels_y += rgb_offset;
    out_pixels_y += width;
    out_pixels_yuv += width;
    }
  }

static void RGB24_to_YUV422_P_c(guchar * in_pixels, guchar * out_y,
                                guchar * out_u, guchar * out_v,
                                int width, int height)
  {
  int imax = width / 2; // We take 2 pixels at once
  
  
  for(int j = 0; j < height; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      RGB_TO_YUV(in_pixels[0], in_pixels[1], in_pixels[2],
                 *(out_y++), *(out_u++), *(out_v++));
      
      in_pixels += 3;
      
      RGB_TO_Y(in_pixels[0],
               in_pixels[1],
               in_pixels[2],
               *(out_y++));

      in_pixels += 3;
      }
    }
  }


static void RGB24_to_YUY2_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once

  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    RGB_TO_YUV(in_pixels[0], in_pixels[1], in_pixels[2],
               out_pixels[0], out_pixels[1], out_pixels[3]);
    RGB_TO_Y(in_pixels[3], in_pixels[4], in_pixels[5],
             out_pixels[2]);
    in_pixels += 6;
    out_pixels += 4;
    }
  }

// Input format: GVColorspaceBGR24

static void BGR24_to_RGB16_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    PACK_RGB16(in_pixels[2],in_pixels[1],in_pixels[0],pixel);
    *((guint16*)out_pixels) = pixel;
    in_pixels += 3;
    out_pixels += 2;
    }
  }

static void BGR24_to_BGR16_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    PACK_BGR16(in_pixels[2],in_pixels[1],in_pixels[0],pixel);
    *((guint16*)out_pixels) = pixel;
    in_pixels += 3;
    out_pixels += 2;
    }
  }

static void BGR24_to_RGB24_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    SWAP_RGB_24(in_pixels,out_pixels);
    in_pixels += 3;
    out_pixels += 3;
    }
  }

static void BGR24_to_RGB32_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    SWAP_RGB_24(in_pixels,out_pixels);
    in_pixels += 3;
    out_pixels += 4;
    }
  }


static void BGR24_to_BGR32_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    COPY_RGB_24(in_pixels,out_pixels);
    in_pixels += 3;
    out_pixels += 4;
    }
  }

static void BGR24_to_YUV420_P_c(guchar * in_pixels, guchar * out_y,
                                 guchar * out_u, guchar * out_v,
                                 int width, int height)
  {
  int imax = width / 2; // We take 4 pixels at once
  int jmax = height / 2;

  int rgb_offset = width * 3;

  guchar * in_pixels_yuv = in_pixels;
  guchar * in_pixels_y = in_pixels + width*3;
  guchar * out_pixels_yuv = out_y;
  guchar * out_pixels_y = out_y + width;
  
  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      RGB_TO_YUV(in_pixels_yuv[2], in_pixels_yuv[1], in_pixels_yuv[0],
                 out_pixels_yuv[0], *out_u, *out_v);

      RGB_TO_Y(in_pixels_yuv[5], in_pixels_yuv[4], in_pixels_yuv[3],
               out_pixels_yuv[1]);
      RGB_TO_Y(in_pixels_y[2], in_pixels_y[1], in_pixels_y[0],
               out_pixels_y[0]);
      RGB_TO_Y(in_pixels_y[5], in_pixels_y[4], in_pixels_y[3],
               out_pixels_y[1]);

      out_pixels_yuv += 2;
      out_pixels_y += 2;
      out_u++;
      out_v++;
            
      in_pixels_y += 6;
      in_pixels_yuv += 6;

      }
    in_pixels_yuv += rgb_offset;
    in_pixels_y += rgb_offset;
    out_pixels_y += width;
    out_pixels_yuv += width;
    }
  }

static void BGR24_to_YUV422_P_c(guchar * in_pixels, guchar * out_y,
                                guchar * out_u, guchar * out_v,
                                int width, int height)
  {
  int imax = width / 2; // We take 2 pixels at once
  
  
  for(int j = 0; j < height; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      RGB_TO_YUV(in_pixels[2], in_pixels[1], in_pixels[0],
                 *(out_y++), *(out_u++), *(out_v++));
      
      in_pixels += 3;
      
      RGB_TO_Y(in_pixels[2], in_pixels[1], in_pixels[0],
               *(out_y++));

      in_pixels += 3;
      }
    }
  }


static void BGR24_to_YUY2_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once

  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    RGB_TO_YUV(in_pixels[2], in_pixels[1], in_pixels[0],
               out_pixels[0], out_pixels[1], out_pixels[3]);
    RGB_TO_Y(in_pixels[5], in_pixels[4], in_pixels[3],
             out_pixels[2]);
    in_pixels += 6;
    out_pixels += 4;
    }
  }

// Input format: GVColorspaceRGB32

static void RGB32_to_RGB16_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    PACK_RGB16(in_pixels[0],in_pixels[1],in_pixels[2],pixel);
    *((guint16*)out_pixels) = pixel;
    in_pixels += 4;
    out_pixels += 2;
    }
  }

static void RGB32_to_BGR16_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    PACK_BGR16(in_pixels[0],in_pixels[1],in_pixels[2],pixel);
    *((guint16*)out_pixels) = pixel;
    in_pixels += 4;
    out_pixels += 2;
    }
  }

static void RGB32_to_RGB24_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    COPY_RGB_24(in_pixels,out_pixels);
    in_pixels += 4;
    out_pixels += 3;
    }
  }

static void RGB32_to_BGR24_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    SWAP_RGB_24(in_pixels,out_pixels);
    in_pixels += 4;
    out_pixels += 3;
    }
  }

static void RGB32_to_BGR32_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    SWAP_RGB_24(in_pixels,out_pixels);
    in_pixels += 4;
    out_pixels += 4;
    }
  }

static void RGB32_to_YUV420_P_c(guchar * in_pixels, guchar * out_y,
                                 guchar * out_u, guchar * out_v,
                                 int width, int height)
  {
  int imax = width / 2; // We take 2 pixels at once
  int jmax = height / 2;

  int rgb_offset = width * 4;

  guchar * in_pixels_yuv = in_pixels;
  guchar * in_pixels_y = in_pixels + width*4;
  guchar * out_pixels_yuv = out_y;
  guchar * out_pixels_y = out_y + width;
  
  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      RGB_TO_YUV(in_pixels_yuv[0], in_pixels_yuv[1], in_pixels_yuv[2],
                 out_pixels_yuv[0], *out_u, *out_v);

      RGB_TO_Y(in_pixels_yuv[4], in_pixels_yuv[5], in_pixels_yuv[6],
               out_pixels_yuv[1]);
      RGB_TO_Y(in_pixels_y[0], in_pixels_y[1], in_pixels_y[2],
               out_pixels_y[0]);
      RGB_TO_Y(in_pixels_y[4], in_pixels_y[5], in_pixels_y[6],
               out_pixels_y[1]);

      out_pixels_yuv += 2;
      out_pixels_y += 2;
      out_u++;
      out_v++;
      
      in_pixels_y += 8;
      in_pixels_yuv += 8;
      }
    in_pixels_yuv += rgb_offset;
    in_pixels_y += rgb_offset;
    out_pixels_y += width;
    out_pixels_yuv += width;
    }
  }


static void RGB32_to_YUV422_P_c(guchar * in_pixels, guchar * out_y,
                                guchar * out_u, guchar * out_v,
                                int width, int height)
  {
  int imax = width / 2; // We take 2 pixels at once
  
  
  for(int j = 0; j < height; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      RGB_TO_YUV(in_pixels[0],
                 in_pixels[1],
                 in_pixels[2],
                 *(out_y++), *(out_u++), *(out_v++));
      
      in_pixels += 4;
      
      RGB_TO_Y(in_pixels[0],
               in_pixels[1],
               in_pixels[2],
               *(out_y++));

      in_pixels += 4;
      }
    }
  }

static void RGB32_to_YUY2_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once

  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    RGB_TO_YUV(in_pixels[0], in_pixels[1], in_pixels[2],
               out_pixels[0], out_pixels[1], out_pixels[3]);
    RGB_TO_Y(in_pixels[4], in_pixels[5], in_pixels[6],
             out_pixels[2]);
    in_pixels += 8;
    out_pixels += 4;
    }
  }

// Input format: GVColorspaceRGBA32


static void RGBA32_to_RGB16_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels, guchar * background)
  {
  guint16 pixel;

  int tmp_red;
  int tmp_green;
  int tmp_blue;
  guchar transp;
  guchar tmp_pixel[3];
  
  for(int i = 0; i < num_pixels; i++)
    {
    RGBA_TO_RGB(in_pixels,tmp_pixel,background);
    PACK_RGB16(tmp_pixel[0],tmp_pixel[1],tmp_pixel[2],pixel);
    *((guint16*)out_pixels) = pixel;
    in_pixels += 4;
    out_pixels += 2;
    }
  }

static void RGBA32_to_BGR16_c(guchar * in_pixels, guchar * out_pixels,
                              int num_pixels, guchar * background)
  {
  guint16 pixel;

  int tmp_red;
  int tmp_green;
  int tmp_blue;
  guchar transp;
  guchar tmp_pixel[3];

  //  cerr << "Converting RGBA32_to_BGR16_c\n";
  
  for(int i = 0; i < num_pixels; i++)
    {
    RGBA_TO_RGB(in_pixels,tmp_pixel,background);
    PACK_BGR16(tmp_pixel[0],tmp_pixel[1],tmp_pixel[2],pixel);
    *((guint16*)out_pixels) = pixel;
    in_pixels += 4;
    out_pixels += 2;
    }
  }

static void RGBA32_to_RGB24_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels, guchar * background)
  {
  int tmp_red;
  int tmp_green;
  int tmp_blue;
  guchar transp;
  //  guchar tmp_pixel[3];

  for(int i = 0; i < num_pixels; i++)
    {
    RGBA_TO_RGB(in_pixels,out_pixels,background);
    in_pixels += 4;
    out_pixels += 3;
    }
  }

static void RGBA32_to_BGR24_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels, guchar * background)
  {
  int tmp_red;
  int tmp_green;
  int tmp_blue;
  guchar transp;
  guchar tmp_pixel[3];
  
  for(int i = 0; i < num_pixels; i++)
    {
    RGBA_TO_RGB(in_pixels,tmp_pixel,background);
    SWAP_RGB_24(tmp_pixel,out_pixels);
    in_pixels += 4;
    out_pixels += 3;
    }
  }

static void RGBA32_to_RGB32_c(guchar * in_pixels, guchar * out_pixels,
                              int num_pixels, guchar * background)
  {
  int tmp_red;
  int tmp_green;
  int tmp_blue;
  guchar transp;
  
  for(int i = 0; i < num_pixels; i++)
    {
    RGBA_TO_RGB(in_pixels,out_pixels,background);
    in_pixels += 4;
    out_pixels += 4;
    }
  }


static void RGBA32_to_BGR32_c(guchar * in_pixels, guchar * out_pixels,
                              int num_pixels, guchar * background)
  {
  int tmp_red;
  int tmp_green;
  int tmp_blue;
  guchar transp;
  guchar tmp_pixel[3];

  for(int i = 0; i < num_pixels; i++)
    {
    RGBA_TO_RGB(in_pixels,tmp_pixel,background);
    SWAP_RGB_24(tmp_pixel,out_pixels);
    in_pixels += 4;
    out_pixels += 4;
    }
  }

static void RGBA32_to_YUV420_P_c(guchar * in_pixels, guchar * out_y,
                                 guchar * out_u, guchar * out_v,
                                 int width, int height, guchar * background)
  {
  int tmp_red;
  int tmp_green;
  int tmp_blue;
  guchar transp;
  guchar tmp_pixel[3];
  
  int imax = width / 2; // We take 2 pixels at once
  int jmax = height / 2;

  int rgb_offset = width * 4;

  guchar * in_pixels_yuv = in_pixels;
  guchar * in_pixels_y = in_pixels + width*4;
  guchar * out_pixels_yuv = out_y;
  guchar * out_pixels_y = out_y + width;
  
  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      RGBA_TO_RGB(in_pixels_yuv,tmp_pixel,background);
      RGB_TO_YUV(tmp_pixel[0], tmp_pixel[1], tmp_pixel[2],
                 out_pixels_yuv[0], *out_u, *out_v);

      in_pixels_yuv += 4;
      
      RGBA_TO_RGB(in_pixels_yuv,tmp_pixel,background);
      RGB_TO_Y(tmp_pixel[0], tmp_pixel[1], tmp_pixel[2],
               out_pixels_yuv[1]);
      
      RGBA_TO_RGB(in_pixels_y,tmp_pixel,background);
      RGB_TO_Y(tmp_pixel[0], tmp_pixel[1], tmp_pixel[2],
               out_pixels_y[0]);

      in_pixels_y += 4;
      
      RGBA_TO_RGB(in_pixels_y,tmp_pixel,background);
      RGB_TO_Y(tmp_pixel[0], tmp_pixel[1], tmp_pixel[2],
               out_pixels_y[1]);

      out_pixels_yuv += 2;
      out_pixels_y += 2;
      out_u++;
      out_v++;
      
      in_pixels_y += 4;
      in_pixels_yuv += 4;
      }
    in_pixels_yuv += rgb_offset;
    in_pixels_y += rgb_offset;
    out_pixels_y += width;
    out_pixels_yuv += width;
    }
  }

static void RGBA32_to_YUV422_P_c(guchar * in_pixels, guchar * out_y,
                                guchar * out_u, guchar * out_v,
                                 int width, int height, guchar * background)
  {
  int tmp_red;
  int tmp_green;
  int tmp_blue;
  guchar transp;
  guchar tmp_pixel[3];
  
  int imax = width / 2; // We take 2 pixels at once
  
  
  for(int j = 0; j < height; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      RGBA_TO_RGB(in_pixels,tmp_pixel,background);
      RGB_TO_YUV(tmp_pixel[0],
                 tmp_pixel[1],
                 tmp_pixel[2],
                 *(out_y++), *(out_u++), *(out_v++));
      
      in_pixels += 4;
      
      RGBA_TO_RGB(in_pixels,tmp_pixel,background);
      RGB_TO_Y(tmp_pixel[0],
               tmp_pixel[1],
               tmp_pixel[2],
               *(out_y++));

      in_pixels += 4;
      }
    }
  }

static void RGBA32_to_YUY2_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels, guchar * background)
  {
  int tmp_red;
  int tmp_green;
  int tmp_blue;
  guchar transp;
  guchar tmp_pixel[3];

  int imax = num_pixels / 2; // We take 2 pixels at once

  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    RGBA_TO_RGB(in_pixels, tmp_pixel,background);
    RGB_TO_YUV(tmp_pixel[0], tmp_pixel[1], tmp_pixel[2],
               out_pixels[0], out_pixels[1], out_pixels[3]);

    in_pixels += 4;
    RGBA_TO_RGB(in_pixels, tmp_pixel,background);
    RGB_TO_Y(tmp_pixel[0], tmp_pixel[1], tmp_pixel[2],
             out_pixels[2]);
    in_pixels += 4;
    out_pixels += 4;
    }
  }

// Input format: GVColorspaceBGR32

static void BGR32_to_RGB16_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    PACK_RGB16(in_pixels[2],in_pixels[1],in_pixels[0],pixel);
    *((guint16*)out_pixels) = pixel;
    in_pixels += 4;
    out_pixels += 2;
    }
  }

static void BGR32_to_BGR16_c(guchar * in_pixels, guchar * out_pixels,
                             int num_pixels)
  {
  guint16 pixel;
  for(int i = 0; i < num_pixels; i++)
    {
    PACK_BGR16(in_pixels[2],in_pixels[1],in_pixels[0],pixel);
    *((guint16*)out_pixels) = pixel;
    in_pixels += 4;
    out_pixels += 2;
    }
  }

static void BGR32_to_RGB24_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    SWAP_RGB_24(in_pixels,out_pixels);
    in_pixels += 4;
    out_pixels += 3;
    }
  }

static void BGR32_to_BGR24_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    COPY_RGB_24(in_pixels,out_pixels);
    in_pixels += 4;
    out_pixels += 3;
    }
  }

static void BGR32_to_RGB32_c(guchar * in_pixels, guchar * out_pixels,
                                  int num_pixels)
  {
  for(int i = 0; i < num_pixels; i++)
    {
    SWAP_RGB_24(in_pixels,out_pixels);
    in_pixels += 4;
    out_pixels += 4;
    }
  }

static void BGR32_to_YUV420_P_c(guchar * in_pixels, guchar * out_y,
                                 guchar * out_u, guchar * out_v,
                                 int width, int height)
  {
  int imax = width / 2; // We take 2 pixels at once
  int jmax = height / 2;

  int rgb_offset = width * 4;

  guchar * in_pixels_yuv = in_pixels;
  guchar * in_pixels_y = in_pixels + width*4;
  guchar * out_pixels_yuv = out_y;
  guchar * out_pixels_y = out_y + width;
  
  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      RGB_TO_YUV(in_pixels_yuv[2], in_pixels_yuv[1], in_pixels_yuv[0],
                 out_pixels_yuv[0], *out_u, *out_v);

      RGB_TO_Y(in_pixels_yuv[6], in_pixels_yuv[5], in_pixels_yuv[4],
               out_pixels_yuv[1]);
      RGB_TO_Y(in_pixels_y[2], in_pixels_y[1], in_pixels_y[0],
               out_pixels_y[0]);
      RGB_TO_Y(in_pixels_y[6], in_pixels_y[5], in_pixels_y[4],
               out_pixels_y[1]);

      out_pixels_yuv += 2;
      out_pixels_y += 2;
      out_u++;
      out_v++;
      
      in_pixels_y += 8;
      in_pixels_yuv += 8;
      }
    in_pixels_yuv += rgb_offset;
    in_pixels_y += rgb_offset;
    out_pixels_y += width;
    out_pixels_yuv += width;
    }
  }


static void BGR32_to_YUV422_P_c(guchar * in_pixels, guchar * out_y,
                                guchar * out_u, guchar * out_v,
                                int width, int height)
  {
  int imax = width / 2; // We take 2 pixels at once
  
  
  for(int j = 0; j < height; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      RGB_TO_YUV(in_pixels[2], in_pixels[1], in_pixels[0],
                 *(out_y++), *(out_u++), *(out_v++));
      
      in_pixels += 4;
      
      RGB_TO_Y(in_pixels[2], in_pixels[1], in_pixels[0],
               *(out_y++));

      in_pixels += 4;
      }
    }
  }

static void BGR32_to_YUY2_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once

  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    RGB_TO_YUV(in_pixels[2], in_pixels[1], in_pixels[0],
               out_pixels[0], out_pixels[1], out_pixels[3]);
    RGB_TO_Y(in_pixels[6], in_pixels[5], in_pixels[4],
             out_pixels[2]);
    in_pixels += 8;
    out_pixels += 4;
    }
  }

// Input format: GVColorspaceYUV420_P

static void YUV420_P_to_RGB16_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;
  int height2 = height / 2;

  guchar * u_start = in_u;
  guchar * v_start = in_v;

  int i_tmp;

  guchar tmp_r, tmp_g, tmp_b;

  guint16 pixel;
  
  for(int i = 0; i < height2; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_RGB16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;

      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_RGB16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;
      in_u++;
      in_v++;
      }

    in_u = u_start;
    in_v = v_start;

    // Do odd row

    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_RGB16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;

      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_RGB16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;
      in_u++;
      in_v++;
      }

    // Increment chroma pointers

    u_start += width2;
    v_start += width2;
    }
  }


static void YUV420_P_to_BGR16_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;
  int height2 = height / 2;

  guchar * u_start = in_u;
  guchar * v_start = in_v;

  int i_tmp;

  guchar tmp_r, tmp_g, tmp_b;

  guint16 pixel;
  
  for(int i = 0; i < height2; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_BGR16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;

      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_BGR16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;
      in_u++;
      in_v++;
      }

    in_u = u_start;
    in_v = v_start;

    // Do odd row

    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_BGR16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;

      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_BGR16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;
      in_u++;
      in_v++;
      }

    // Increment chroma pointers

    u_start += width2;
    v_start += width2;
    }
  }


static void YUV420_P_to_RGB24_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;
  int height2 = height / 2;

  guchar * u_start = in_u;
  guchar * v_start = in_v;

  int i_tmp;
  
  for(int i = 0; i < height2; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[0],out_pixels[1],out_pixels[2]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[4],out_pixels[5],out_pixels[6]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 6;
      }

    in_u = u_start;
    in_v = v_start;

    // Do odd row

    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[0],out_pixels[1],out_pixels[2]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[4],out_pixels[5],out_pixels[6]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 6;
      }

    // Increment chroma pointers

    u_start += width2;
    v_start += width2;
    }
  }


static void YUV420_P_to_BGR24_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;
  int height2 = height / 2;

  guchar * u_start = in_u;
  guchar * v_start = in_v;

  int i_tmp;
  
  for(int i = 0; i < height2; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[2],out_pixels[1],out_pixels[0]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[6],out_pixels[5],out_pixels[4]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 6;
      }

    in_u = u_start;
    in_v = v_start;

    // Do odd row

    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[2],out_pixels[1],out_pixels[0]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[6],out_pixels[5],out_pixels[4]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 6;
      }

    // Increment chroma pointers

    u_start += width2;
    v_start += width2;
    }
  }

static void YUV420_P_to_RGB32_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;
  int height2 = height / 2;

  guchar * u_start = in_u;
  guchar * v_start = in_v;

  int i_tmp;
  
  for(int i = 0; i < height2; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[0],out_pixels[1],out_pixels[2]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[4],out_pixels[5],out_pixels[6]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 8;
      }

    in_u = u_start;
    in_v = v_start;

    // Do odd row

    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[0],out_pixels[1],out_pixels[2]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[4],out_pixels[5],out_pixels[6]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 8;
      }

    // Increment chroma pointers

    u_start += width2;
    v_start += width2;
    }
  }

static void YUV420_P_to_BGR32_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;
  int height2 = height / 2;

  guchar * u_start = in_u;
  guchar * v_start = in_v;

  int i_tmp;
  
  for(int i = 0; i < height2; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[2],out_pixels[1],out_pixels[0]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[6],out_pixels[5],out_pixels[4]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 8;
      }

    in_u = u_start;
    in_v = v_start;

    // Do odd row

    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[2],out_pixels[1],out_pixels[0]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[6],out_pixels[5],out_pixels[4]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 8;
      }

    // Increment chroma pointers

    u_start += width2;
    v_start += width2;
    }
  }

static void YUV420_P_to_YUV422_P_c(guchar ** in_y_rows,
                                   guchar ** in_u_rows,
                                   guchar ** in_v_rows,
                                   guchar ** out_y_rows,
                                   guchar ** out_u_rows,
                                   guchar ** out_v_rows,
                                   int width, int height)
  {
  int imax = height/2;
  int width2 = width/2;
  
  memcpy(*out_y_rows, *in_y_rows, width*height);

  for(int i = 0; i < imax; i++)
    {
    memcpy(out_u_rows[2*i],   in_u_rows[i], width2);
    memcpy(out_u_rows[2*i+1], in_u_rows[i], width2);

    memcpy(out_v_rows[2*i],   in_v_rows[i], width2);
    memcpy(out_v_rows[2*i+1], in_v_rows[i], width2);
    }
  }

static void YUV420_P_to_YUY2_c(guchar * in_y, guchar * in_u, guchar * in_v,
                               guchar * out_pixels,
                               int width, int height)
  {
  // TODO: Fix odd numbers of rows/columns

  int width2 = width/2;
  int height2 = height/2;
  
  guchar * u_start = in_u;
  guchar * v_start = in_v;
  
  for(int i = 0; i < height2; i++)
    {
    // Do even row

    for(int j = 0; j < width2; j++)
      {
      *(out_pixels++) = *(in_y++);
      *(out_pixels++) = *(in_u++);
      *(out_pixels++) = *(in_y++);
      *(out_pixels++) = *(in_v++);
      }

    in_u = u_start;
    in_v = v_start;

    // Do odd row

    for(int j = 0; j < width2; j++)
      {
      *(out_pixels++) = *(in_y++);
      *(out_pixels++) = *(in_u++);
      *(out_pixels++) = *(in_y++);
      *(out_pixels++) = *(in_v++);
      }

    // Increment chroma pointers

    u_start += width2;
    v_start += width2;
    }
  }

// Input format: GVColorspaceYUV422_P

static void YUV422_P_to_BGR16_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;

  int i_tmp;

  guchar tmp_r, tmp_g, tmp_b;

  guint16 pixel;
  
  for(int i = 0; i < height; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_BGR16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;

      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_BGR16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;
      in_u++;
      in_v++;
      }

    }
  }

static void YUV422_P_to_RGB16_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;

  int i_tmp;

  guchar tmp_r, tmp_g, tmp_b;

  guint16 pixel;
  
  for(int i = 0; i < height; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_RGB16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;

      YUV_TO_RGB(*in_y,*in_u,*in_v,tmp_r,tmp_g,tmp_b);

      PACK_RGB16(tmp_r,tmp_g,tmp_b,pixel);
      *((guint16*)out_pixels) = pixel;
      out_pixels += 2;

      in_y++;
      in_u++;
      in_v++;
      }

    }
  }


static void YUV422_P_to_RGB24_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;

  int i_tmp;
  
  for(int i = 0; i < height; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[0],out_pixels[1],out_pixels[2]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[3],out_pixels[4],out_pixels[5]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 6;
      }
    }
  }

static void YUV422_P_to_BGR24_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;

  int i_tmp;
  
  for(int i = 0; i < height; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[2],out_pixels[1],out_pixels[0]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[5],out_pixels[4],out_pixels[3]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 6;
      }
    }
  }

static void YUV422_P_to_RGB32_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;

  int i_tmp;
  
  for(int i = 0; i < height; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[0],out_pixels[1],out_pixels[2]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[4],out_pixels[5],out_pixels[6]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 8;
      }
    }
  }


static void YUV422_P_to_BGR32_c(guchar * in_y, guchar * in_u, guchar * in_v,
                                guchar * out_pixels,
                                int width, int height)
  {
  int width2 = width / 2;

  int i_tmp;
  
  for(int i = 0; i < height; i++)
    {
    // Do even row

    //    YUV_TO_RGB(y,u,v,r,g,b)
    
    for(int j = 0; j < width2; j++)
      {
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[2],out_pixels[1],out_pixels[0]);
      in_y++;
      YUV_TO_RGB(*in_y,*in_u,*in_v,out_pixels[6],out_pixels[5],out_pixels[4]);
      in_y++;
      in_u++;
      in_v++;
      out_pixels += 8;
      }

    }
  }


static void YUV422_P_to_YUV420_P_c(guchar ** in_y_rows,
                                   guchar ** in_u_rows,
                                   guchar ** in_v_rows,
                                   guchar ** out_y_rows,
                                   guchar ** out_u_rows,
                                   guchar ** out_v_rows,
                                   int width, int height)
  {
  memcpy(*out_y_rows, *in_y_rows, width*height);

  int width2 = width/2;
  
  for(int i = 0; i < height; i+=2)
    {
    memcpy(out_u_rows[i/2],in_u_rows[i], width2);
    memcpy(out_v_rows[i/2],in_v_rows[i], width2);
    }
  }

static void YUV422_P_to_YUY2_c(guchar * in_y, guchar * in_u, guchar * in_v,
                               guchar * out_pixels,
                               int width, int height)
  {
  // TODO: Fix odd numbers of rows/columns

  int width2 = width/2;
  
  for(int i = 0; i < height; i++)
    {
    // Do even row

    for(int j = 0; j < width2; j++)
      {
      *(out_pixels++) = *(in_y++);
      *(out_pixels++) = *(in_u++);
      *(out_pixels++) = *(in_y++);
      *(out_pixels++) = *(in_v++);
      }
    }
  }


// Input format: GVColorspaceYUY2

static void YUY2_to_RGB16_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once

  guchar tmp_r, tmp_g, tmp_b;

  guint16 pixel;
  int i_tmp;
  
  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    YUV_TO_RGB(in_pixels[0], in_pixels[1], in_pixels[3],
               tmp_r, tmp_g, tmp_b);

    PACK_RGB16(tmp_r,tmp_g,tmp_b,pixel);
    *((guint16*)out_pixels) = pixel;
    out_pixels += 2;
    
    YUV_TO_RGB(in_pixels[2], in_pixels[1], in_pixels[3],
               out_pixels[3], out_pixels[4], out_pixels[5]);

    PACK_RGB16(tmp_r,tmp_g,tmp_b,pixel);
    *((guint16*)out_pixels) = pixel;
    out_pixels += 2;

    in_pixels += 4;
    }
  }


static void YUY2_to_BGR16_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once

  guchar tmp_r, tmp_g, tmp_b;
  guint16 pixel;
  
  // Output: Y0 U0 Y1 V0

  int i_tmp;
  
  for(int i = 0; i < imax; i++)
    {
    YUV_TO_RGB(in_pixels[0], in_pixels[1], in_pixels[3],
               tmp_r, tmp_g, tmp_b);

    PACK_BGR16(tmp_r,tmp_g,tmp_b,pixel);
    *((guint16*)out_pixels) = pixel;
    out_pixels += 2;
    
    YUV_TO_RGB(in_pixels[2], in_pixels[1], in_pixels[3],
               tmp_r, tmp_g, tmp_b);

    PACK_BGR16(tmp_r,tmp_g,tmp_b,pixel);
    *((guint16*)out_pixels) = pixel;
    out_pixels += 2;

    in_pixels += 4;
    }
  }

static void YUY2_to_RGB24_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once
  int i_tmp;

  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    YUV_TO_RGB(in_pixels[0], in_pixels[1], in_pixels[3],
               out_pixels[0], out_pixels[1], out_pixels[2]);
    YUV_TO_RGB(in_pixels[2], in_pixels[1], in_pixels[3],
               out_pixels[3], out_pixels[4], out_pixels[5]);
    in_pixels += 4;
    out_pixels += 6;
    }
  }

static void YUY2_to_BGR24_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once
  int i_tmp;

  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    YUV_TO_RGB(in_pixels[0], in_pixels[1], in_pixels[3],
               out_pixels[2], out_pixels[1], out_pixels[0]);
    YUV_TO_RGB(in_pixels[2], in_pixels[1], in_pixels[3],
               out_pixels[5], out_pixels[4], out_pixels[3]);
    in_pixels += 4;
    out_pixels += 6;
    }
  }

static void YUY2_to_RGB32_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once
  int i_tmp;

  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    YUV_TO_RGB(in_pixels[0], in_pixels[1], in_pixels[3],
               out_pixels[0], out_pixels[1], out_pixels[2]);
    YUV_TO_RGB(in_pixels[2], in_pixels[1], in_pixels[3],
               out_pixels[4], out_pixels[5], out_pixels[6]);
    in_pixels += 4;
    out_pixels += 8;
    }
  }

static void YUY2_to_BGR32_c(guchar * in_pixels, guchar * out_pixels,
                            int num_pixels)
  {
  int imax = num_pixels / 2; // We take 2 pixels at once
  int i_tmp;

  // Output: Y0 U0 Y1 V0

  for(int i = 0; i < imax; i++)
    {
    YUV_TO_RGB(in_pixels[0], in_pixels[1], in_pixels[3],
               out_pixels[2], out_pixels[1], out_pixels[0]);
    YUV_TO_RGB(in_pixels[2], in_pixels[1], in_pixels[3],
               out_pixels[6], out_pixels[5], out_pixels[4]);
    in_pixels += 4;
    out_pixels += 8;
    }
  }

static void YUY2_to_YUV420_P_c(guchar * in_pixels, guchar * out_y,
                                guchar * out_u, guchar * out_v,
                                int width, int height)
  {
  // TODO: Fix odd numbers of rows/columns

  int width2 = width/2;
  int height2 = height/2;
  
  for(int i = 0; i < height2; i++)
    {
    // Do even row (width chroma components)

    for(int j = 0; j < width2; j++)
      {
      *(out_y++) = *(in_pixels++);
      *(out_u++) = *(in_pixels++);
      *(out_y++) = *(in_pixels++);
      *(out_v++) = *(in_pixels++);
      }

    // Do odd row (ignore chroma)

    for(int j = 0; j < width2; j++)
      {
      *(out_y++) = *(in_pixels++);
      in_pixels++;
      *(out_y++) = *(in_pixels++);
      in_pixels++;
      }
    }
  }

static void YUY2_to_YUV422_P_c(guchar * in_pixels, guchar * out_y,
                                guchar * out_u, guchar * out_v,
                                int width, int height)
  {
  // TODO: Fix odd numbers of rows/columns

  int width2 = width/2;
//   int height2 = height/2;
  
  for(int i = 0; i < height; i++)
    {
    // Do even row (width chroma components)

    for(int j = 0; j < width2; j++)
      {
      *(out_y++) = *(in_pixels++);
      *(out_u++) = *(in_pixels++);
      *(out_y++) = *(in_pixels++);
      *(out_v++) = *(in_pixels++);
      }
    }
  }


void GVVideoConverter::init_c_funcs()
  {
  __init_c();

  RGB16_to_BGR16 = RGB16_to_BGR16_c;
  RGB16_to_RGB24 = RGB16_to_RGB24_c;
  RGB16_to_BGR24 = RGB16_to_BGR24_c;
  RGB16_to_RGB32 = RGB16_to_RGB32_c;
  RGB16_to_BGR32 = RGB16_to_BGR32_c;
  RGB16_to_YUY2 = RGB16_to_YUY2_c;
  RGB16_to_YUV420_P = RGB16_to_YUV420_P_c;
  RGB16_to_YUV422_P = RGB16_to_YUV422_P_c;

  BGR16_to_RGB16 = BGR16_to_RGB16_c;
  BGR16_to_RGB24 = BGR16_to_RGB24_c;
  BGR16_to_BGR24 = BGR16_to_BGR24_c;
  BGR16_to_RGB32 = BGR16_to_RGB32_c;
  BGR16_to_BGR32 = BGR16_to_BGR32_c;
  BGR16_to_YUY2 = BGR16_to_YUY2_c;
  BGR16_to_YUV420_P = BGR16_to_YUV420_P_c;
  BGR16_to_YUV422_P = BGR16_to_YUV422_P_c;

  RGB24_to_RGB16    = RGB24_to_RGB16_c;
  RGB24_to_BGR16    = RGB24_to_BGR16_c;
  RGB24_to_BGR24    = RGB24_to_BGR24_c;
  RGB24_to_RGB32    = RGB24_to_RGB32_c;
  RGB24_to_BGR32    = RGB24_to_BGR32_c;
  RGB24_to_YUY2     = RGB24_to_YUY2_c;
  RGB24_to_YUV420_P = RGB24_to_YUV420_P_c;
  RGB24_to_YUV422_P = RGB24_to_YUV422_P_c;

  BGR24_to_RGB16    = BGR24_to_RGB16_c;
  BGR24_to_BGR16    = BGR24_to_BGR16_c;
  BGR24_to_RGB24    = BGR24_to_RGB24_c;
  BGR24_to_RGB32    = BGR24_to_RGB32_c;
  BGR24_to_BGR32    = BGR24_to_BGR32_c;
  BGR24_to_YUY2     = BGR24_to_YUY2_c;
  BGR24_to_YUV420_P = BGR24_to_YUV420_P_c;
  BGR24_to_YUV422_P = BGR24_to_YUV422_P_c;

  RGB32_to_RGB16    = RGB32_to_RGB16_c;
  RGB32_to_BGR16    = RGB32_to_BGR16_c;
  RGB32_to_RGB24    = RGB32_to_RGB24_c;
  RGB32_to_BGR24    = RGB32_to_BGR24_c;
  RGB32_to_BGR32    = RGB32_to_BGR32_c;
  RGB32_to_YUY2     = RGB32_to_YUY2_c;
  RGB32_to_YUV420_P = RGB32_to_YUV420_P_c;
  RGB32_to_YUV422_P = RGB32_to_YUV422_P_c;

  RGBA32_to_RGB16    = RGBA32_to_RGB16_c;
  RGBA32_to_BGR16    = RGBA32_to_BGR16_c;
  RGBA32_to_RGB24    = RGBA32_to_RGB24_c;
  RGBA32_to_BGR24    = RGBA32_to_BGR24_c;
  RGBA32_to_RGB32    = RGBA32_to_RGB32_c;
  RGBA32_to_BGR32    = RGBA32_to_BGR32_c;
  RGBA32_to_YUY2     = RGBA32_to_YUY2_c;
  RGBA32_to_YUV420_P = RGBA32_to_YUV420_P_c;
  RGBA32_to_YUV422_P = RGBA32_to_YUV422_P_c;
  
  BGR32_to_RGB16    = BGR32_to_RGB16_c;
  BGR32_to_BGR16    = BGR32_to_BGR16_c;
  BGR32_to_RGB24    = BGR32_to_RGB24_c;
  BGR32_to_BGR24    = BGR32_to_BGR24_c;
  BGR32_to_RGB32    = BGR32_to_RGB32_c;
  BGR32_to_YUY2     = BGR32_to_YUY2_c;
  BGR32_to_YUV420_P = BGR32_to_YUV420_P_c;
  BGR32_to_YUV422_P = BGR32_to_YUV422_P_c;
  
  YUV420_P_to_RGB16    = YUV420_P_to_RGB16_c;
  YUV420_P_to_BGR16    = YUV420_P_to_BGR16_c;
  YUV420_P_to_RGB24    = YUV420_P_to_RGB24_c;
  YUV420_P_to_BGR24    = YUV420_P_to_BGR24_c;
  YUV420_P_to_RGB32    = YUV420_P_to_RGB32_c;
  YUV420_P_to_BGR32    = YUV420_P_to_BGR32_c;
  YUV420_P_to_YUY2     = YUV420_P_to_YUY2_c;
  YUV420_P_to_YUV422_P = YUV420_P_to_YUV422_P_c;

  YUV422_P_to_RGB16    = YUV422_P_to_RGB16_c;
  YUV422_P_to_BGR16    = YUV422_P_to_BGR16_c;
  YUV422_P_to_RGB24    = YUV422_P_to_RGB24_c;
  YUV422_P_to_BGR24    = YUV422_P_to_BGR24_c;
  YUV422_P_to_RGB32    = YUV422_P_to_RGB32_c;
  YUV422_P_to_BGR32    = YUV422_P_to_BGR32_c;
  YUV422_P_to_YUY2     = YUV422_P_to_YUY2_c;
  YUV422_P_to_YUV420_P = YUV422_P_to_YUV420_P_c;
  
  YUY2_to_RGB16    = YUY2_to_RGB16_c;
  YUY2_to_BGR16    = YUY2_to_BGR16_c;
  YUY2_to_RGB24    = YUY2_to_RGB24_c;
  YUY2_to_BGR24    = YUY2_to_BGR24_c;
  YUY2_to_RGB32    = YUY2_to_RGB32_c;
  YUY2_to_BGR32    = YUY2_to_BGR32_c;
  YUY2_to_YUV420_P = YUY2_to_YUV420_P_c;
  YUY2_to_YUV422_P = YUY2_to_YUV422_P_c;
  }
