/*****************************************************************

  videoconvert_mmx.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// MMX versions of gmerlins video conversion routines

#include "config.h"

#include <glib.h>
#include <iostream>

#include "gvmmx.h"
#include "gvtime.h"
#include "videoformat.h"
#include "videoconvert.h"

extern "C" {
#include "mmx/attributes.h"
#include "mmx/mmx.h"
}

//     Conversion formulae from the Colorspace FAQ
//     (http://www.neuro.sfc.keio.ac.jp/~aly/polygon/info/color-space-faq.html)

//     Usually, you'll need the following conversions based on Rec 601-1
//     for TIFF and JPEG works:

//     RGB -> YCbCr (with Rec 601-1 specs)
//     Y=  0.2989*Red+0.5866*Green+0.1145*Blue
//     Cb=-0.1687*Red-0.3312*Green+0.5000*Blue
//     Cr= 0.5000*Red-0.4183*Green-0.0816*Blue

//     YCbCr (with Rec 601-1 specs) -> RGB
//     Red=  Y+0.0000*Cb+1.4022*Cr
//     Green=Y-0.3456*Cb-0.7145*Cr
//     Blue= Y+1.7710*Cb+0.0000*Cr

//     I assume Y is within the range [0;1], and Cb and Cr are within the range
//     [-0.5;0.5].

// Some Macros

static mmx_t rgb16_r_mask; // Mask for the lower bits in RGB16 
static mmx_t rgb16_g_mask; // Mask for the middle bits in RGB16
static mmx_t rgb16_b_mask; // Mask for the upprt bits in RGB16

static mmx_t bgr16_r_mask; // Mask for the lower bits in RGB16 
static mmx_t bgr16_g_mask; // Mask for the middle bits in RGB16
static mmx_t bgr16_b_mask; // Mask for the upprt bits in RGB16

static mmx_t rgb24_l_mask; // Mask for the lower of  2 RGB24 pixels
static mmx_t rgb24_u_mask; // Mask for the upper of  2 RGB24 pixels

static mmx_t rgb32_l_mask; // Mask for the lower of  2 RGB24 pixels
static mmx_t rgb32_u_mask; // Mask for the upper of  2 RGB24 pixels

static mmx_t rgb32_r_mask; // Mask for the reds of   2 RGB32 pixels
static mmx_t rgb32_g_mask; // Mask for the greens of 2 RGB32 pixels
static mmx_t rgb32_b_mask; // Mask for the blues  of 2 RGB32 pixels

static mmx_t bgr32_r_mask; // Mask for the reds of   2 RGB32 pixels
static mmx_t bgr32_g_mask; // Mask for the greens of 2 RGB32 pixels
static mmx_t bgr32_b_mask; // Mask for the blues  of 2 RGB32 pixels

static mmx_t rgb24_r_mask; // Mask for the reds of   2 RGB24 pixels
static mmx_t rgb24_g_mask; // Mask for the greens of 2 RGB24 pixels
static mmx_t rgb24_b_mask; // Mask for the blues  of 2 RGB24 pixels

static mmx_t bgr24_r_mask; // Mask for the reds of   2 RGB24 pixels
static mmx_t bgr24_g_mask; // Mask for the greens of 2 RGB24 pixels
static mmx_t bgr24_b_mask; // Mask for the blues  of 2 RGB24 pixels

static mmx_t lower_word_mask; // Mask for the lower words
static mmx_t upper_word_mask; // Mask for the upper words

static mmx_t lower_dword_mask; // Mask for the lower doublewords
static mmx_t upper_dword_mask; // Mask for the upper doublewords

static mmx_t lower_byte_mask; // Mask for the lower bytes
static mmx_t upper_byte_mask; // Mask for the upper bytes

static mmx_t uv_sign_mask;  // Mask for the sign

static mmx_t mmx_pack_mask_32;

static mmx_t mmx_r_to_y;
static mmx_t mmx_r_to_u;
static mmx_t mmx_r_to_v;
static mmx_t mmx_g_to_y;
static mmx_t mmx_g_to_u;
static mmx_t mmx_g_to_v;
static mmx_t mmx_b_to_y;
static mmx_t mmx_b_to_u;
static mmx_t mmx_b_to_v;

// Set up some variables

void __init_mmx()
  {
  rgb16_r_mask.q = 0x001f001f001f001fLL; // Mask for the red   bits in RGB16 
  rgb16_g_mask.q = 0x07e007e007e007e0LL; // Mask for the green bits in RGB16
  rgb16_b_mask.q = 0xf800f800f800f800LL; // Mask for the blue  bits in RGB16

  bgr16_b_mask.q = 0x001f001f001f001fLL; // Mask for the lower  bits in RGB16 
  bgr16_g_mask.q = 0x07e007e007e007e0LL; // Mask for the middle bits in RGB16
  bgr16_r_mask.q = 0xf800f800f800f800LL; // Mask for the upper  bits in RGB16
  
  rgb24_l_mask.q = 0x0000000000FFFFFFLL; // Mask for the lower of 2 RGB24 pixels
  rgb24_u_mask.q = 0x0000FFFFFF000000LL; // Mask for the upper of 2 RGB24 pixels

  rgb32_l_mask.q = 0x0000000000FFFFFFLL; // Mask for the lower of 2 RGB32 pixels
  rgb32_u_mask.q = 0x00FFFFFF00000000LL; // Mask for the upper of 2 RGB32 pixels
  
  rgb32_r_mask.q = 0x000000FF000000FFLL; // Mask for the reds of   2 RGB32 pixels
  rgb32_g_mask.q = 0x0000FF000000FF00LL; // Mask for the greens of 2 RGB32 pixels
  rgb32_b_mask.q = 0x00FF000000FF0000LL; // Mask for the blues  of 2 RGB32 pixels

  bgr32_r_mask.q = 0x00FF000000FF0000LL; // Mask for the reds of   2 RGB32 pixels
  bgr32_g_mask.q = 0x0000FF000000FF00LL; // Mask for the greens of 2 RGB32 pixels
  bgr32_b_mask.q = 0x000000FF000000FFLL; // Mask for the blues  of 2 RGB32 pixels

  rgb24_r_mask.q = 0x00000000FF0000FFLL; // Mask for the reds of   2 RGB32 pixels
  rgb24_g_mask.q = 0x000000FF0000FF00LL; // Mask for the greens of 2 RGB32 pixels
  rgb24_b_mask.q = 0x0000FF0000FF0000LL; // Mask for the blues  of 2 RGB32 pixels

  bgr24_r_mask.q = 0x0000FF0000FF0000LL; // Mask for the reds of   2 RGB32 pixels
  bgr24_g_mask.q = 0x000000FF0000FF00LL; // Mask for the greens of 2 RGB32 pixels
  bgr24_b_mask.q = 0x00000000FF0000FFLL; // Mask for the blues  of 2 RGB32 pixels
  
  mmx_pack_mask_32.q = 0x00FFFFFF00FFFFFFLL;

  lower_dword_mask.q = 0x00000000FFFFFFFFLL;
  upper_dword_mask.q = 0xFFFFFFFF00000000LL;
  
  lower_word_mask.q = 0x0000FFFF0000FFFFLL;
  upper_word_mask.q = 0xFFFF0000FFFF0000LL;

  lower_byte_mask.q = 0x00FF00FF00FF00FFLL;
  upper_byte_mask.q = 0xFF00FF00FF00FF00LL;

  uv_sign_mask.q = 0x0080008000800080LL;
  
  // Y=  0.2989*Red+0.5866*Green+0.1145*Blue     | Red=  Y+0.0000*Cb+1.4022*Cr
  // Cb=-0.1687*Red-0.3312*Green+0.5000*Blue     | Green=Y-0.3456*Cb-0.7145*Cr
  // Cr= 0.5000*Red-0.4183*Green-0.0816*Blue     | Blue= Y+1.7710*Cb+0.0000*Cr
  
  for(int i = 0; i < 4; i++)
    {
    mmx_r_to_y.uw[i] = (guint16)(0.2989*65535.0);
    mmx_g_to_y.uw[i] = (guint16)(0.5866*65535.0*0.5);
    mmx_b_to_y.uw[i] = (guint16)(0.1145*65535.0);
    mmx_r_to_u.uw[i] = (guint16)(0.1687*65535.0);
    mmx_g_to_u.uw[i] = (guint16)(0.3312*65535.0);
    mmx_b_to_u.uw[i] = (guint16)(0.5000*65535.0);
    mmx_r_to_v.uw[i] = (guint16)(0.5000*65535.0);
    mmx_g_to_v.uw[i] = (guint16)(0.4183*65535.0);
    mmx_b_to_v.uw[i] = (guint16)(0.0816*65535.0);
    }
  }

// This is cool: testreg(<string>, reg) prints
// <string>: req.q

// static mmx_t test_mmx;

#define TESTREG(str, reg) movq_r2m(reg,test_mmx);\
                          cerr.setf(ios::hex, ios::dec|ios::hex|ios::oct);\
                          cerr << str << test_mmx.q << "\n"

// Load 4 pixels from the pixel pointer into mm0 (and mm1)

#define LOAD_16(in) movq_m2r(*in, mm0);\
                    in += 8

#define LOAD_24(in) movq_m2r(*in, mm0);\
                    in += 6;\
                    movq_m2r(*in, mm1);\
                    in += 6;

#define LOAD_32(in) movq_m2r(*in, mm0);\
                    in += 8;\
                    movq_m2r(*in, mm1);\
                    in += 8

// Write values to pixel pointers


#define WRITE_16(out) movq_r2m(mm0, *out);\
                      out+=8;

#define WRITE_24(out) movq_r2m(mm0, *out);\
                      out+=6;\
                      movq_r2m(mm1, *out);\
                      out+=6;

#define WRITE_32(out) movq_r2m(mm0, *out);\
                      out+=8;\
                      movq_r2m(mm1, *out);\
                      out+=8;

#define WRITE_YUY2(out) movq_r2m(mm0, *out);\
                        out+=8

#define WRITE_YUV420_P_Y(out) movd_r2m(mm0, *out);\
                              out+=4

#define WRITE_YUV420_P_UV(out_u, out_v) movq_r2m(mm1, write_16);\
                                        *((guint16*)out_u)=write_16.uw[0];\
                                        out_u+=2;\
                                        movq_r2m(mm2, write_16);\
                                        *((guint16*)out_v)=write_16.uw[0];\
                                        out_v+=2

// Convert 4 RGB/BGR24 pixels in mm0 and mm1 to 4 RGB/BGR32 in mm0 and mm1 
// Used registers: mm0, mm1, mm2

#define PACK_24_TO_32 movq_r2r(mm0, mm2);\
                      pand_m2r(rgb24_l_mask,mm0);\
                      pand_m2r(rgb24_u_mask,mm2);\
                      psllq_i2r(8, mm2);\
                      por_r2r(mm2,mm0);\
                      movq_r2r(mm1, mm2);\
                      pand_m2r(rgb24_l_mask,mm1);\
                      pand_m2r(rgb24_u_mask,mm2);\
                      psllq_i2r(8, mm2);\
                      por_r2r(mm2,mm1);

// Convert 4 RGB/BGR16 pixels in mm0 to 4 RGB/BGR32 pixels in mm0 and mm1

#define PACK_16_TO_32 pxor_r2r(mm7, mm7);\
                      movq_r2r(mm0, mm2);\
                      movq_r2r(mm0, mm3);\
                      pand_m2r(lower_dword_mask, mm2);\
                      pand_m2r(upper_dword_mask, mm3);\
                      psrlq_i2r(32, mm3);\
                      punpcklwd_r2r(mm7, mm2);\
                      punpcklwd_r2r(mm7, mm3);\
                      movq_r2r(mm2, mm7);\
                      pand_m2r(rgb16_r_mask, mm7);\
                      psllw_i2r(3, mm7);\
                      movq_r2r(mm7, mm0);\
                      movq_r2r(mm2, mm7);\
                      pand_m2r(rgb16_g_mask, mm7);\
                      psllw_i2r(5, mm7);\
                      por_r2r(mm7, mm0);\
                      movq_r2r(mm2, mm7);\
                      pand_m2r(rgb16_b_mask, mm7);\
                      pslld_i2r(8, mm7);\
                      por_r2r(mm7, mm0);\
                      movq_r2r(mm3, mm7);\
                      pand_m2r(rgb16_r_mask, mm7);\
                      psllw_i2r(3, mm7);\
                      movq_r2r(mm7, mm1);\
                      movq_r2r(mm3, mm7);\
                      pand_m2r(rgb16_g_mask, mm7);\
                      psllw_i2r(5, mm7);\
                      por_r2r(mm7, mm1);\
                      movq_r2r(mm3, mm7);\
                      pand_m2r(rgb16_b_mask, mm7);\
                      pslld_i2r(8, mm7);\
                      por_r2r(mm7, mm1);

// Convert 4 RGB/BGR32 pixels in mm0 and mm1 to 4 RGB/BGR16 pixels in mm0

// #define PACK_32_to_16 

#define PACK_32_TO_24 movq_r2r(mm0, mm2);\
                      pand_m2r(rgb32_l_mask,mm0);\
                      pand_m2r(rgb32_u_mask,mm2);\
                      psrlq_i2r(8, mm2);\
                      por_r2r(mm2,mm0);\
                      movq_r2r(mm1, mm2);\
                      pand_m2r(rgb32_l_mask,mm1);\
                      pand_m2r(rgb32_u_mask,mm2);\
                      psrlq_i2r(8, mm2);\
                      por_r2r(mm2,mm1);

// #define PACK_24_to_16

// Do RGB32 <-> BGR32 conversion of 4 pixels in mm0 and mm1
// Uses mm0, mm1, mm2, mm3

#define SWAP_RGB_32 movq_r2r(mm0, mm2);\
                    movq_r2r(mm0, mm3);\
                    pand_m2r(rgb32_g_mask,mm0);\
                    pand_m2r(rgb32_r_mask,mm2);\
                    pand_m2r(rgb32_b_mask,mm3);\
                    psllq_i2r(16, mm2);\
                    psrlq_i2r(16, mm3);\
                    por_r2r(mm2,mm0);\
                    por_r2r(mm3,mm0);\
                    movq_r2r(mm1, mm2);\
                    movq_r2r(mm1, mm3);\
                    pand_m2r(rgb32_g_mask,mm1);\
                    pand_m2r(rgb32_r_mask,mm2);\
                    pand_m2r(rgb32_b_mask,mm3);\
                    psllq_i2r(16, mm2);\
                    psrlq_i2r(16, mm3);\
                    por_r2r(mm2,mm1);\
                    por_r2r(mm3,mm1)

// Do RGB24 <-> BGR24 conversion of 2 pixels in mm0
// Uses mm0, mm1, mm2, mm3

#define SWAP_RGB_24 movq_r2r(mm0, mm2);\
                    movq_r2r(mm0, mm3);\
                    pand_m2r(rgb24_g_mask,mm0);\
                    pand_m2r(rgb24_r_mask,mm2);\
                    pand_m2r(rgb24_b_mask,mm3);\
                    psllq_i2r(16, mm2);\
                    psrlq_i2r(16, mm3);\
                    por_r2r(mm2,mm0);\
                    por_r2r(mm3,mm0);\
                    movq_r2r(mm1, mm2);\
                    movq_r2r(mm1, mm3);\
                    pand_m2r(rgb24_g_mask,mm1);\
                    pand_m2r(rgb24_r_mask,mm2);\
                    pand_m2r(rgb24_b_mask,mm3);\
                    psllq_i2r(16, mm2);\
                    psrlq_i2r(16, mm3);\
                    por_r2r(mm2,mm1);\
                    por_r2r(mm3,mm1)

// Do RGB16 <-> BGR16 conversion of 4 pixels in mm0
// Uses mm0, mm1, mm2

#define SWAP_RGB_16 movq_r2r(mm0, mm1);\
                    movq_r2r(mm0, mm2);\
                    pand_m2r(rgb16_r_mask, mm1);\
                    pand_m2r(rgb16_g_mask, mm0);\
                    pand_m2r(rgb16_b_mask, mm2);\
                    psllq_i2r(11, mm1);\
                    psrlq_i2r(11, mm2);\
                    por_r2r(mm1,mm0);\
                    por_r2r(mm2,mm0)

// Convert 2 RGB32 pixels in mm0 and 2 RGB32 pixels in mm1 to
// separated R G and B Registers in mm4 and mm5 and mm6
// The RGB Values are expanded to 15 bit accuracy
//
// Before: mm0: XX B1 G1 R1 XX B0 G0 R0
//         mm1: XX B3 G3 R3 XX B2 G2 R2
//
// After:  mm4: R3 00 R2 00 R1 00 R0 00
//         mm5: G3 00 G2 00 G1 00 G0 00
//         mm6: B3 00 B2 00 B1 00 B0 00

#define SEPARATE_RGB32 movq_r2r(mm0, mm2);\
                       movq_r2r(mm1, mm3);\
                       pand_m2r(rgb32_r_mask, mm2);\
                       pand_m2r(rgb32_r_mask, mm3);\
                       packssdw_r2r(mm3, mm2);\
                       movq_r2r(mm2, mm4);\
                       movq_r2r(mm0, mm2);\
                       movq_r2r(mm1, mm3);\
                       pand_m2r(rgb32_g_mask, mm2);\
                       pand_m2r(rgb32_g_mask, mm3);\
                       psrld_i2r(8, mm2);\
                       psrld_i2r(8, mm3);\
                       packssdw_r2r(mm3, mm2);\
                       movq_r2r(mm2, mm5);\
                       movq_r2r(mm0, mm2);\
                       pand_m2r(rgb32_b_mask, mm2);\
                       movq_r2r(mm1, mm3);\
                       pand_m2r(rgb32_b_mask, mm3);\
                       psrld_i2r(16, mm2);\
                       psrld_i2r(16, mm3);\
                       packssdw_r2r(mm3, mm2);\
                       movq_r2r(mm2, mm6);\
                       movq_r2r(mm4, mm7);\
                       pslld_i2r(8, mm7);\
                       por_r2r(mm7, mm4);\
                       movq_r2r(mm5, mm7);\
                       pslld_i2r(8, mm7);\
                       por_r2r(mm7, mm5);\
                       movq_r2r(mm6, mm7);\
                       pslld_i2r(8, mm7);\
                       por_r2r(mm7, mm6);\
                       psrlw_i2r(1, mm4);\
                       psrlw_i2r(1, mm5);\
                       psrlw_i2r(1, mm6)

// Same as SEPARATE_RGB32 but for BGR32

#define SEPARATE_BGR32 movq_r2r(mm0, mm2);\
                       movq_r2r(mm1, mm3);\
                       pand_m2r(bgr32_r_mask, mm2);\
                       pand_m2r(bgr32_r_mask, mm3);\
                       psrld_i2r(16, mm2);\
                       psrld_i2r(16, mm3);\
                       packssdw_r2r(mm3, mm2);\
                       movq_r2r(mm2, mm4);\
                       movq_r2r(mm0, mm2);\
                       movq_r2r(mm1, mm3);\
                       pand_m2r(bgr32_g_mask, mm2);\
                       pand_m2r(bgr32_g_mask, mm3);\
                       psrld_i2r(8, mm2);\
                       psrld_i2r(8, mm3);\
                       packssdw_r2r(mm3, mm2);\
                       movq_r2r(mm2, mm5);\
                       movq_r2r(mm0, mm2);\
                       movq_r2r(mm1, mm3);\
                       pand_m2r(bgr32_b_mask, mm2);\
                       pand_m2r(bgr32_b_mask, mm3);\
                       packssdw_r2r(mm3, mm2);\
                       movq_r2r(mm2, mm6);\
                       movq_r2r(mm4, mm7);\
                       pslld_i2r(8, mm7);\
                       por_r2r(mm7, mm4);\
                       movq_r2r(mm5, mm7);\
                       pslld_i2r(8, mm7);\
                       por_r2r(mm7, mm5);\
                       movq_r2r(mm6, mm7);\
                       pslld_i2r(8, mm7);\
                       por_r2r(mm7, mm6);\
                       psrlw_i2r(1, mm4);\
                       psrlw_i2r(1, mm5);\
                       psrlw_i2r(1, mm6)

// Convert 4 RGB16 pixels in mm0 to
// separated R G and B Registers in mm4 and mm5 and mm6
// The RGB Values are expanded to 15 bit
//
// Before: mm0: P3 P3 P2 P2 P1 P1 P0 P0
//
// After:  mm4: R3 00 R2 00 R1 00 R0 00
//         mm5: G3 00 G2 00 G1 00 G0 00
//         mm6: B3 00 B2 00 B1 00 B0 00

// TODO: increase accuracy by mapping 5/6 bity to 8 bits

#define SEPARATE_RGB16 movq_r2r(mm0, mm4);\
                       pand_m2r(rgb16_r_mask, mm4);\
                       psllw_i2r(10, mm4);\
                       movq_r2r(mm4, mm7);\
                       psrlw_i2r(5, mm7);\
                       por_r2r(mm7, mm4);\
                       psrlw_i2r(5, mm7);\
                       por_r2r(mm7, mm4);\
                       movq_r2r(mm0, mm5);\
                       pand_m2r(rgb16_g_mask, mm5);\
                       psllw_i2r(4, mm5);\
                       movq_r2r(mm5, mm7);\
                       psrlw_i2r(6, mm7);\
                       por_r2r(mm7, mm5);\
                       psrlw_i2r(6, mm7);\
                       por_r2r(mm7, mm5);\
                       movq_r2r(mm0, mm6);\
                       pand_m2r(rgb16_b_mask, mm6);\
                       psrlw_i2r(1, mm6);\
                       movq_r2r(mm6, mm7);\
                       psrlw_i2r(5, mm7);\
                       por_r2r(mm7, mm6);\
                       psrlw_i2r(5, mm7);\
                       por_r2r(mm7, mm6);

// Same as above, but for BGR16

#define SEPARATE_BGR16 movq_r2r(mm0, mm4);\
                       pand_m2r(bgr16_r_mask, mm4);\
                       psrlw_i2r(1, mm4);\
                       movq_r2r(mm4, mm7);\
                       psrlw_i2r(5, mm7);\
                       por_r2r(mm7, mm4);\
                       psrlw_i2r(5, mm7);\
                       por_r2r(mm7, mm4);\
                       movq_r2r(mm0, mm5);\
                       pand_m2r(bgr16_g_mask, mm5);\
                       psllw_i2r(4, mm5);\
                       movq_r2r(mm5, mm7);\
                       psrlw_i2r(6, mm7);\
                       por_r2r(mm7, mm5);\
                       psrlw_i2r(6, mm7);\
                       por_r2r(mm7, mm5);\
                       movq_r2r(mm0, mm6);\
                       pand_m2r(bgr16_b_mask, mm6);\
                       psllw_i2r(10, mm6);\
                       movq_r2r(mm6, mm7);\
                       psrlw_i2r(5, mm7);\
                       por_r2r(mm7, mm6);\
                       psrlw_i2r(5, mm7);\
                       por_r2r(mm7, mm6);

// Convert 4 RGB separated pixels in mm4, mm5 and mm6 to
// Y values in mm0

// Before:  mm4: 00 R3 00 R2 00 R1 00 R0
//          mm5: 00 G3 00 G2 00 G1 00 G0
//          mm6: 00 B3 00 B2 00 B1 00 B0

// After:   mm0: 00 Y3 00 Y2 00 Y1 00 Y0

// Y=  0.2989*Red+0.5866*Green+0.1145*Blue     | Red=  Y+0.0000*Cb+1.4022*Cr
// Cb=-0.1687*Red-0.3312*Green+0.5000*Blue     | Green=Y-0.3456*Cb-0.7145*Cr
// Cr= 0.5000*Red-0.4183*Green-0.0816*Blue     | Blue= Y+1.7710*Cb+0.0000*Cr

#define RGB_TO_Y movq_r2r(mm4, mm7);\
                 pmulhw_m2r(mmx_r_to_y, mm7);\
                 movq_r2r(mm7, mm0);\
                 movq_r2r(mm5, mm7);\
                 pmulhw_m2r(mmx_g_to_y, mm7);\
                 psllw_i2r(1, mm7);\
                 paddusw_r2r(mm7, mm0);\
                 movq_r2r(mm6, mm7);\
                 pmulhw_m2r(mmx_b_to_y, mm7);\
                 paddusw_r2r(mm7, mm0);\
                 psrlw_i2r(7, mm0);\
                 pand_m2r(lower_byte_mask, mm0)

// TESTREG("Blue Part: ", mm7)


// Convert 4 RGB separated pixels in mm4, mm5 and mm6 to
// U and V values in mm1 and mm2

// Before:  mm4: 00 R3 00 R2 00 R1 00 R0
//          mm5: 00 G3 00 G2 00 G1 00 G0
//          mm6: 00 B3 00 B2 00 B1 00 B0

// After:   mm1: 00 U3 00 U2 00 U1 00 U0
//          mm2: 00 V3 00 V2 00 V1 00 V0

// U=-0.1687*Red-0.3312*Green+0.5000*Blue
// V= 0.5000*Red-0.4183*Green-0.0816*Blue

// TODO: Optimize by calculating U and V in one single step if
//       Chroma is taken only for every second pixel

#define RGB_TO_UV movq_r2r(mm6, mm7);\
                  pmulhw_m2r(mmx_b_to_u, mm7);\
                  movq_r2r(mm7, mm1);\
                  movq_r2r(mm4, mm7);\
                  pmulhw_m2r(mmx_r_to_u, mm7);\
                  psubsw_r2r(mm7, mm1);\
                  movq_r2r(mm5, mm7);\
                  pmulhw_m2r(mmx_g_to_u, mm7);\
                  psubsw_r2r(mm7, mm1);\
                  pmulhw_m2r(mmx_r_to_v, mm4);\
                  movq_r2r(mm4, mm2);\
                  pmulhw_m2r(mmx_g_to_v, mm5);\
                  psubsw_r2r(mm5, mm2);\
                  pmulhw_m2r(mmx_b_to_v, mm6);\
                  psubsw_r2r(mm6, mm2);\
                  psraw_i2r(7, mm1);\
                  psraw_i2r(7, mm2);\
                  pand_m2r(lower_byte_mask, mm1);\
                  pand_m2r(lower_byte_mask, mm2);\
                  pxor_m2r(uv_sign_mask, mm1);\
                  pxor_m2r(uv_sign_mask, mm2);

// Combine Separate YUV registers to YUY2 in mm0

// Before:   mm0: 00 Y3 00 Y2 00 Y1 00 Y0
//           mm1: 00 U3 00 U2 00 U1 00 U0
//           mm2: 00 V3 00 V2 00 V1 00 V0
//
// After:    mm0: V2 Y3 U2 Y2 V0 Y1 U0 Y0

#define COMBINE_YUY2 pand_m2r(lower_word_mask, mm1);\
                     pand_m2r(lower_word_mask, mm2);\
                     pslld_i2r(8, mm1);\
                     pslld_i2r(24, mm2);\
                     por_r2r(mm1, mm0);\
                     por_r2r(mm2, mm0)

// Combine Separate YUV registers to planar YUV420 in mm0, mm1 and mm2

// Before:   mm0: 00 Y3 00 Y2 00 Y1 00 Y0
//
// After:    mm0: 00 00 00 00 Y3 Y2 Y1 Y0


#define COMBINE_YUV420_P_Y pxor_r2r(mm7, mm7);\
                           packuswb_r2r(mm7, mm0) 


// Before:   mm1: 00 U3 00 U2 00 U1 00 U0
//           mm2: 00 V3 00 V2 00 V1 00 V0
//
// After:    mm1: 00 00 00 00 00 00 U2 U0
//           mm2: 00 00 00 00 00 00 V2 V0


#define COMBINE_YUV420_P_UV pand_m2r(lower_word_mask, mm1);\
                            pand_m2r(lower_word_mask, mm2);\
                            packssdw_r2r(mm7, mm1);\
                            packssdw_r2r(mm7, mm2);\
                            packuswb_r2r(mm7, mm1);\
                            packuswb_r2r(mm7, mm2) 

// Combine Separate YUV registers to planar YUV422 in mm0, mm1 and mm2

// Before:   mm0: 00 Y3 00 Y2 00 Y1 00 Y0
//           mm1: 00 U3 00 U2 00 U1 00 U0
//           mm2: 00 V3 00 V2 00 V1 00 V0
//
// After:    mm0: 00 00 00 00 Y3 Y2 Y1 Y0
//           mm1: 00 00 00 00 00 00 U2 U0
//           mm2: 00 00 00 00 00 00 V2 V0

#define COMBINE_YUV422_P COMBINE_YUV420_P_Y;\
                         COMBINE_YUV420_P_UV


// Here follow n*(n-1) functions for converting colorspaces.

// Input format: GVColorspaceRGB16

static void RGB16_to_BGR16_mmx(guchar * in_pixels, guchar * out_pixels,
                               int num_pixels)
  {

//   // We can take 4 pixels at once

  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_16(in_pixels);

    SWAP_RGB_16;

    WRITE_16(out_pixels);
    }
  emms();
  }

static void RGB16_to_RGB24_mmx(guchar * in_pixels, guchar * out_pixels,
                               int num_pixels)
  {

//   // We can take 4 pixels at once

  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_16(in_pixels);
    PACK_16_TO_32;
    PACK_32_TO_24;
    WRITE_24(out_pixels);
    }
  emms();
  }

static void RGB16_to_BGR24_mmx(guchar * in_pixels, guchar * out_pixels,
                               int num_pixels)
  {

//   // We can take 4 pixels at once

  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_16(in_pixels);
    SWAP_RGB_16;
    PACK_16_TO_32;
    PACK_32_TO_24;
    WRITE_24(out_pixels);
    }
  emms();
  }

static void RGB16_to_BGR32_mmx(guchar * in_pixels, guchar * out_pixels,
                               int num_pixels)
  {

//   // We can take 4 pixels at once

  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_16(in_pixels);
    SWAP_RGB_16;
    PACK_16_TO_32;
    WRITE_32(out_pixels);
    }
  emms();
  }

static void RGB16_to_YUV420_P_mmx(guchar * in_pixels,
                                  guchar * y,
                                  guchar * u,
                                  guchar * v,
                                  int width, int height)
  {
  int imax = width / 4;
  int jmax = height / 2;

  guchar * in0 = in_pixels;
  guchar * in1 = in_pixels + 2 * width;

  guchar * y0 = y;
  guchar * y1 = y + width;
  mmx_t write_16;
  
  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      LOAD_16(in0);
      
      SEPARATE_RGB16;
      RGB_TO_Y;
      RGB_TO_UV;
      
      COMBINE_YUV420_P_Y;
      COMBINE_YUV420_P_UV;
      
      WRITE_YUV420_P_Y(y0);
      WRITE_YUV420_P_UV(u, v);

      LOAD_16(in1);
      
      SEPARATE_RGB16;
      RGB_TO_Y;
            
      COMBINE_YUV420_P_Y;
            
      WRITE_YUV420_P_Y(y1);
      }
    in0 += width * 2;
    in1 += width * 2;
    y0 += width;
    y1 += width;
    }
  emms();
  } 

static void RGB16_to_YUY2_mmx(guchar * in_pixels,
                              guchar * out_pixels,
                              int num_pixels)
  {
  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_16(in_pixels);
    SEPARATE_RGB16;
    RGB_TO_Y;
    RGB_TO_UV;

    COMBINE_YUY2;

    WRITE_YUY2(out_pixels);
    }
  emms();
  } 


// Input format: GVColorspaceBGR16

static void BGR16_to_RGB24_mmx(guchar * in_pixels, guchar * out_pixels,
                               int num_pixels)
  {

//   // We can take 4 pixels at once

  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_16(in_pixels);
    SWAP_RGB_16;
    PACK_16_TO_32;
    PACK_32_TO_24;
    WRITE_24(out_pixels);
    }
  emms();
  }

static void BGR16_to_BGR24_mmx(guchar * in_pixels, guchar * out_pixels,
                               int num_pixels)
  {

//   // We can take 4 pixels at once

  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_16(in_pixels);
    PACK_16_TO_32;
    PACK_32_TO_24;
    WRITE_24(out_pixels);
    }
  emms();
  }

static void BGR16_to_BGR32_mmx(guchar * in_pixels, guchar * out_pixels,
                               int num_pixels)
  {

//   // We can take 4 pixels at once

  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_16(in_pixels);
    PACK_16_TO_32;
    WRITE_32(out_pixels);
    }
  emms();
  }

static void BGR16_to_YUV420_P_mmx(guchar * in_pixels,
                                  guchar * y,
                                  guchar * u,
                                  guchar * v,
                                  int width, int height)
  {
  int imax = width / 4;
  int jmax = height / 2;

  guchar * in0 = in_pixels;
  guchar * in1 = in_pixels + 2 * width;

  guchar * y0 = y;
  guchar * y1 = y + width;
  mmx_t write_16;
  
  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      LOAD_16(in0);
      
      SEPARATE_BGR16;
      RGB_TO_Y;
      RGB_TO_UV;
      
      COMBINE_YUV420_P_Y;
      COMBINE_YUV420_P_UV;
      
      WRITE_YUV420_P_Y(y0);
      WRITE_YUV420_P_UV(u, v);

      LOAD_16(in1);
      
      SEPARATE_RGB16;
      RGB_TO_Y;
            
      COMBINE_YUV420_P_Y;
            
      WRITE_YUV420_P_Y(y1);
      }
    in0 += width * 2;
    in1 += width * 2;
    y0 += width;
    y1 += width;
    }
  emms();
  } 


static void BGR16_to_YUY2_mmx(guchar * in_pixels,
                              guchar * out_pixels,
                              int num_pixels)
  {
  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_16(in_pixels);
    SEPARATE_BGR16;
    RGB_TO_Y;
    RGB_TO_UV;

    COMBINE_YUY2;

    WRITE_YUY2(out_pixels);
    }
  emms();
  } 


// Input format: GVColorspaceRGB24


static void RGB24_to_BGR24_mmx(guchar * in_pixels, guchar * out_pixels,
                               int num_pixels)
  {
  //  mmx_t in_mmx;
  //  mmx_t out_mmx;

  int imax = num_pixels / 4;

  for(int i = 0; i < imax; i++)
    {
    //    in_mmx = *((mmx_t*)in_pixels);

    LOAD_24(in_pixels);
    
    SWAP_RGB_24;
        
    movq_r2m(mm0, *((mmx_t*)out_pixels));
    out_pixels += 6;
    movq_r2m(mm1, *((mmx_t*)out_pixels));
    out_pixels += 6;

    //    *((mmx_t*)out_pixels) = out_mmx;

    }
  emms();
  }

static void RGB24_to_BGR32_mmx(guchar * in_pixels, guchar * out_pixels,
                               int num_pixels)
  {
  //  mmx_t in_mmx;
  //  mmx_t out_mmx;

  int imax = num_pixels / 4;

  for(int i = 0; i < imax; i++)
    {
    //    in_mmx = *((mmx_t*)in_pixels);

    LOAD_24(in_pixels);
    
    PACK_24_TO_32;
    SWAP_RGB_32;
      
    WRITE_32(out_pixels);
    }
  emms();
  }

static void RGB24_to_YUV420_P_mmx(guchar * in_pixels,
                                  guchar * y,
                                  guchar * u,
                                  guchar * v,
                                  int width, int height)
  {
  int imax = width / 4;
  int jmax = height / 2;

  guchar * in0 = in_pixels;
  guchar * in1 = in_pixels + 3 * width;

  guchar * y0 = y;
  guchar * y1 = y + width;
  mmx_t write_16;
  
  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      LOAD_24(in0);
      PACK_24_TO_32;
      SEPARATE_RGB32;
      RGB_TO_Y;
      RGB_TO_UV;
      
      COMBINE_YUV420_P_Y;
      COMBINE_YUV420_P_UV;
      
      WRITE_YUV420_P_Y(y0);
      WRITE_YUV420_P_UV(u, v);

      LOAD_24(in1);
      PACK_24_TO_32;
      SEPARATE_RGB32;
      RGB_TO_Y;
            
      COMBINE_YUV420_P_Y;
            
      WRITE_YUV420_P_Y(y1);
      }
    in0 += width * 3;
    in1 += width * 3;
    y0 += width;
    y1 += width;
    }
  emms();
  } 


static void RGB24_to_YUY2_mmx(guchar * in_pixels,
                              guchar * out_pixels,
                              int num_pixels)
  {
  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_24(in_pixels);
    PACK_24_TO_32;
    SEPARATE_RGB32;
    RGB_TO_Y;
    RGB_TO_UV;

    COMBINE_YUY2;

    WRITE_YUY2(out_pixels);
    }
  emms();
  } 

// Input format: GVColorspaceBGR24

static void BGR24_to_BGR32_mmx(guchar * in_pixels,
                               guchar * out_pixels,
                               int num_pixels)
  {
  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_24(in_pixels);
    PACK_24_TO_32;
    WRITE_32(out_pixels);
    }
  emms();
  }

static void BGR24_to_YUV420_P_mmx(guchar * in_pixels,
                                  guchar * y,
                                  guchar * u,
                                  guchar * v,
                                  int width, int height)
  {
  int imax = width / 4;
  int jmax = height / 2;

  guchar * in0 = in_pixels;
  guchar * in1 = in_pixels + 3 * width;

  guchar * y0 = y;
  guchar * y1 = y + width;
  mmx_t write_16;

  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      LOAD_24(in0);
      PACK_24_TO_32;
      SEPARATE_BGR32;
      RGB_TO_Y;
      RGB_TO_UV;
      
      COMBINE_YUV420_P_Y;
      COMBINE_YUV420_P_UV;
      
      WRITE_YUV420_P_Y(y0);
      WRITE_YUV420_P_UV(u, v);

      LOAD_24(in1);
      PACK_24_TO_32;
      SEPARATE_BGR32;
      RGB_TO_Y;
            
      COMBINE_YUV420_P_Y;
            
      WRITE_YUV420_P_Y(y1);
      }
    in0 += width * 3;
    in1 += width * 3;
    y0 += width;
    y1 += width;
    }
  emms();
  } 

static void BGR24_to_YUY2_mmx(guchar * in_pixels,
                              guchar * out_pixels,
                              int num_pixels)
  {
  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_24(in_pixels);
    PACK_24_TO_32;
    SEPARATE_BGR32;
    RGB_TO_Y;
    RGB_TO_UV;

    COMBINE_YUY2;
    WRITE_YUY2(out_pixels);
    }
  emms();
  } 

// Input format: GVColorspaceBGR32

static void BGR32_to_RGB24_mmx(guchar * in_pixels,
                               guchar * out_pixels,
                               int num_pixels)
  {
  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_32(in_pixels);
    SWAP_RGB_32;
    PACK_32_TO_24;
    WRITE_24(out_pixels);
    }
  emms();
  }

static void BGR32_to_BGR24_mmx(guchar * in_pixels,
                               guchar * out_pixels,
                               int num_pixels)
  {
  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_32(in_pixels);
    PACK_32_TO_24;
    WRITE_24(out_pixels);
    }
  emms();
  }

static void BGR32_to_YUV420_P_mmx(guchar * in_pixels,
                                  guchar * y,
                                  guchar * u,
                                  guchar * v,
                                  int width, int height)
  {
  int imax = width / 4;
  int jmax = height / 2;

  guchar * in0 = in_pixels;
  guchar * in1 = in_pixels + 4 * width;

  guchar * y0 = y;
  guchar * y1 = y + width;
  mmx_t write_16;

  for(int j = 0; j < jmax; j++)
    {
    for(int i = 0; i < imax; i++)
      {
      LOAD_32(in0);
      SEPARATE_BGR32;
      RGB_TO_Y;
      RGB_TO_UV;
      
      COMBINE_YUV420_P_Y;
      COMBINE_YUV420_P_UV;
      
      WRITE_YUV420_P_Y(y0);
      WRITE_YUV420_P_UV(u, v);

      LOAD_32(in1);
      SEPARATE_BGR32;
      RGB_TO_Y;
            
      COMBINE_YUV420_P_Y;
      WRITE_YUV420_P_Y(y1);
      }
    in0 += width * 4;
    in1 += width * 4;
    y0 += width;
    y1 += width;
    }
  emms();
  } 

static void BGR32_to_YUY2_mmx(guchar * in_pixels,
                              guchar * out_pixels,
                              int num_pixels)
  {
  int imax = num_pixels/4;

  for(int i = 0; i < imax; i++)
    {
    LOAD_32(in_pixels);
    SEPARATE_BGR32;
    RGB_TO_Y;
    RGB_TO_UV;
    COMBINE_YUY2;
    WRITE_YUY2(out_pixels);
    }
  emms();
  } 

// Input format: GVColorspaceYUV_420_P

// Input format: GVColorspaceYUV_422_P

// Input format: GVColorspaceYUY2

// Set the appropriate function pointers

#ifndef MMXTEST

void GVVideoConverter::init_mmx_funcs()
  {
  __init_mmx();
  
  RGB16_to_BGR16 = RGB16_to_BGR16_mmx;
  RGB16_to_RGB24 = RGB16_to_RGB24_mmx;
  RGB16_to_BGR24 = RGB16_to_BGR24_mmx;
  RGB16_to_BGR32 = RGB16_to_BGR32_mmx;
  RGB16_to_YUY2 = RGB16_to_YUY2_mmx;
  RGB16_to_YUV420_P = RGB16_to_YUV420_P_mmx;
  //  RGB16_to_YUV422_P = RGB16_to_YUV422_P_mmx;
  
  BGR16_to_RGB16 = RGB16_to_BGR16_mmx;
  BGR16_to_RGB24 = BGR16_to_RGB24_mmx;
  BGR16_to_BGR24 = BGR16_to_BGR24_mmx;
  BGR16_to_BGR32 = BGR16_to_BGR32_mmx;
  BGR16_to_YUY2 = BGR16_to_YUY2_mmx;
  BGR16_to_YUV420_P = BGR16_to_YUV420_P_mmx;
  //  BGR16_to_YUV422_P = BGR16_to_YUV422_P_mmx;

  //  RGB24_to_RGB16 = RGB24_to_RGB16_mmx;
  //  RGB24_to_BGR16 = RGB24_to_BGR16_mmx;
  RGB24_to_BGR24 = RGB24_to_BGR24_mmx;
  RGB24_to_BGR32 = RGB24_to_BGR32_mmx;
  RGB24_to_YUV420_P = RGB24_to_YUV420_P_mmx;
  //  RGB24_to_YUV422_P = RGB24_to_YUV422_P_mmx;
  RGB24_to_YUY2 = RGB24_to_YUY2_mmx;
  
  //  BGR24_to_RGB16 = BGR24_to_RGB16_mmx;
  //  BGR24_to_BGR16 = BGR24_to_BGR16_mmx;
  BGR24_to_RGB24 = RGB24_to_BGR24_mmx;
  BGR24_to_BGR32 = BGR24_to_BGR32_mmx;
  BGR24_to_YUV420_P  = BGR24_to_YUV420_P_mmx;
  //  BGR24_to_YUV422_P  = BGR24_to_YUV422_P_mmx;
  BGR24_to_YUY2  = BGR24_to_YUY2_mmx;

  //  BGR32_to_RGB16 = BGR32_to_RGB16_mmx;
  //  BGR32_to_BGR16 = BGR32_to_BGR16_mmx;
  BGR32_to_RGB24 = BGR32_to_RGB24_mmx;
  BGR32_to_BGR24 = BGR32_to_BGR24_mmx;
  BGR32_to_YUV420_P = BGR32_to_YUV420_P_mmx;
  //  BGR32_to_YUV422_P = BGR32_to_YUV422_P_mmx;
  BGR32_to_YUY2 = BGR32_to_YUY2_mmx;

  //   YUV_420_P_to_RGB16 = YUV420_P_to_RGB16_mmx;
  //   YUV_420_P_to_BGR16 = YUV420_P_to_BGR16_mmx;
  //   YUV_420_P_to_RGB24 = YUV420_P_to_RGB24_mmx;
  //   YUV_420_P_to_BGR24 = YUV420_P_to_BGR24_mmx;
  //   YUV_420_P_to_BGR32 = YUV420_P_to_BGR32_mmx;
  //   YUV_420_P_to_YUV_422_P = YUV420_P_to_YUV_422_P_mmx;
  //   YUV_420_P_to_YUY2  = YUV420_P_to_YUY2_mmx;

  //   YUV_422_P_to_RGB16 = YUV422_P_to_RGB16_mmx;
  //   YUV_422_P_to_BGR16 = YUV422_P_to_BGR16_mmx;
  //   YUV_422_P_to_RGB24 = YUV422_P_to_RGB24_mmx;
  //   YUV_422_P_to_BGR24 = YUV422_P_to_BGR24_mmx;
  //   YUV_422_P_to_BGR32 = YUV422_P_to_BGR32_mmx;
  //   YUV_422_P_to_YUV_420_P = YUV422_P_to_YUV_420_P_mmx;
  //   YUV_422_P_to_YUY2  = YUV422_P_to_YUY2_mmx;
  
  //   YUY2_to_RGB16 = YUY2_to_RGB16_mmx;
  //   YUY2_to_BGR16 = YUY2_to_BGR16_mmx;
  //   YUY2_to_RGB24 = YUY2_to_RGB24_mmx;
  //   YUY2_to_BGR24 = YUY2_to_BGR24_mmx;
  //   YUY2_to_BGR32 = YUY2_to_BGR32_mmx;
  //   YUY2_to_YUV420_P = YUY2_to_YUV420_P_mmx;
  //   YUY2_to_YUV422_P = YUY2_to_YUV422_P_mmx;

  }

#endif
