/*****************************************************************

  videoformat.cc

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Routines for the videoformat class

#include <glib.h>
#include <iostream>
#include <sstream>
#include <iomanip>

using std::stringstream;
using std::ostream;
using std::ends;
using std::setw;

#include "config.h"
#include "gvtime.h"
#include "videoformat.h"

char * gv_colorspace_to_string(GVColorspace & colorspace)
  {
  stringstream line;
  //  line.freeze(0);
  line.seekp(0);
  line << colorspace << ends;
  char * ret = g_strdup(line.str().c_str());
  return ret;
  }

void GVVideoFormat::init()
  {
  switch(colorspace)
    {
    case GVColorspaceRGB16:
    case GVColorspaceBGR16:
      bytes_per_pixel = 2;
      buffer_size = 2*width*height + 4;
      break;
    case GVColorspaceRGB24:
    case GVColorspaceBGR24:
      bytes_per_pixel = 3;
      buffer_size = 3*width*height + 4;
      break;
    case GVColorspaceBGR32:
    case GVColorspaceRGB32:
    case GVColorspaceRGBA32:
      bytes_per_pixel = 4;
      buffer_size = 4*width*height + 4;
      break;
    case GVColorspaceYUV_420_P:
      bytes_per_pixel = 0;
      buffer_size = width*height + (width*height)/2 + 4;
      //      buffer_size = 4*width*height + 4;
      break;
    case GVColorspaceYUV_422_P:
      bytes_per_pixel = 0;
      buffer_size = 2 * width*height + 4;
      //      buffer_size = 4*width*height + 4;
      break;
    case GVColorspaceYUY2:
      bytes_per_pixel = 2;
      buffer_size = width*height*2;
      //      buffer_size = 4*width*height + 4;
      break;
    }

  // Nanoseconds per frame
  
  nanoseconds_per_frame = (int)((1.0/framerate)*1e9+0.5); 
  frame_time.set_from_frames(1, framerate);

  aspect_ratio = (float)(width)/(float)(height);
  }

ostream& operator<<(ostream & output, const GVColorspace & colorspace)
  {
  switch(colorspace)
    {
    case GVColorspaceRGB16:
      output << "RGB 16";
      break;
    case GVColorspaceBGR16:
      output << "BGR 16";
      break;
    case GVColorspaceRGB24:
      output << "RGB 24";
      break;
    case GVColorspaceBGR24:
      output << "BGR 24";
      break;
    case GVColorspaceBGR32:
      output << "BGR 32";
      break;
    case GVColorspaceRGB32:
      output << "RGB 32";
      break;
    case GVColorspaceRGBA32:
      output << "RGBA 32";
      break;
    case GVColorspaceYUV_420_P:
      output << "YUV 420 Planar";
      break;
    case GVColorspaceYUV_422_P:
      output << "YUV 422 Planar";
      break;
    case GVColorspaceYUY2:
      output << "YUY2";
      break;
    default:
      output << "Unknown";
      break;
    }
  return output;
  }

void GVVideoFormat::print(ostream & output) const
  {
  output << "Size: " << width << " x " << height << "\n";

  output << "Framerate: " << framerate << " Frames/sec\n";
  output << "Colorspace: " << colorspace << "\n";
  
  output << "Number of frames: " << num_frames << "\n";

  int hours;
  int minutes;

  double time;
  
  time = (double)(num_frames)/(double)(framerate);

  int time_i = (int)(time);

  time_i /= 60;
  
  minutes = time_i % 60;
  time_i /= 60;
  
  hours = time_i;

  time -= ((minutes*60) + (hours * 3600));
  
  output << "Video Time: ";
  
  if(hours)
    {
    if(hours >= 10)
      output << setw(2) << hours;
    else
      output << setw(1) << hours;
    output << ":";
    }


  if(!(hours) && (minutes < 10))
    output << setw(1) << minutes;
  else
    output << setw(2) << minutes;

  output << ":" << time << "\n";

  //  output << "Nanoseconds per frame: " << nanoseconds_per_frame << "\n";

  output << "Bytes per pixel: " << bytes_per_pixel << "\n";
  output << "Buffer Size (in bytes): " << buffer_size << "\n";
  output << "Aspect Ratio: " << aspect_ratio << "\n";

  output << "Frame time: " << frame_time.get_double() << " seconds";
  }
