/*****************************************************************

  videoformat.h

  Copyright (c) 2001 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de

  http://gmerlin.sourceforge.net

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

// Video format description

// We use the colormodels.h / colormodels.c from quicktime here to
// do the video format conversion

#define GV_VIDEO_ROW_ORDER_UP    1

// Colorspaces (More to be added!!)

// Flags for colorspaces

#define GV_COLORSPACE_IS_RGB     (1<<8)
#define GV_COLORSPACE_IS_YUV     (1<<9)
#define GV_COLORSPACE_IS_PACKED (1<<10)
#define GV_COLORSPACE_IS_PLANAR (1<<11)
#define GV_COLORSPACE_HAS_ALPHA (1<<12)

enum GVColorspace
  {
    GVColorspaceRGB16 = 1|GV_COLORSPACE_IS_RGB|GV_COLORSPACE_IS_PACKED, // 565
    GVColorspaceBGR16 = 2|GV_COLORSPACE_IS_RGB|GV_COLORSPACE_IS_PACKED, // 565
    GVColorspaceRGB24 = 3|GV_COLORSPACE_IS_RGB|GV_COLORSPACE_IS_PACKED, // 888
    GVColorspaceBGR24 = 4|GV_COLORSPACE_IS_RGB|GV_COLORSPACE_IS_PACKED,
    GVColorspaceRGB32 = 7|GV_COLORSPACE_IS_RGB|GV_COLORSPACE_IS_PACKED,
    GVColorspaceBGR32 = 6|GV_COLORSPACE_IS_RGB|GV_COLORSPACE_IS_PACKED,

    // RGB with Alpha channel
    
    GVColorspaceRGBA32 = 1|
                         GV_COLORSPACE_IS_RGB|
                         GV_COLORSPACE_IS_PACKED|
                         GV_COLORSPACE_HAS_ALPHA,
    
    // 4:2:0 planar YUV, pixels[0]: y, pixels[1]: u, pixels[2]: v
    GVColorspaceYUV_420_P = 1|GV_COLORSPACE_IS_YUV|GV_COLORSPACE_IS_PLANAR,

    // 4:2:2 planar YUV (might come out of jpeg files)
    
    GVColorspaceYUV_422_P = 2|GV_COLORSPACE_IS_YUV|GV_COLORSPACE_IS_PLANAR,

    // 4:2:2 packed YUV, Ordering |Y0|U0|Y1|V0| ....
    GVColorspaceYUY2 = 1|GV_COLORSPACE_IS_YUV|GV_COLORSPACE_IS_PACKED,
  };


class ostream;

ostream& operator<<(ostream & output, const GVColorspace & colorspace);

// Convert a colorspace to a string, return value must be
// g_free'd when no longer needed

char * gv_colorspace_to_string(GVColorspace & colorspace);

struct GVVideoFormat
  {
  int flags;               // Special flags
  int width;               // Width and height of the pictures
  int height;  
  double framerate;        // Framerate (could better be a rational)
  int num_frames;          // Total number of frames
  GVColorspace colorspace; // Colorspace

  int nanoseconds_per_frame; // Nanoseconds per frame
  
  int bytes_per_pixel;     // Bytes per pixel
  int buffer_size;         // Size of the pixels buffer

  float aspect_ratio;
  
  GVTime frame_time;

  void print(ostream & output) const;
  
  GVVideoFormat & operator=(const GVVideoFormat & format)
    {
    flags = format.flags;
    width = format.width;
    height = format.height;
    framerate = format.framerate;
    num_frames = format.num_frames;
    colorspace = format.colorspace;
    init();
    return *this;
    }

  int operator!=(const GVVideoFormat & format) const
    {
    return ((flags != format.flags) ||
            (width != format.width) ||
            (height != format.height) ||
            (framerate != format.framerate) ||
            (num_frames != format.num_frames) ||
            (colorspace != format.colorspace) ||
            (bytes_per_pixel != format.bytes_per_pixel));
    }
  void init();
  };

struct GVVideoFrame
  {
  // This flag indicates, that this is the first frame, after skipping

  int skip; 
  GVTime time; // Timestamp
  guchar * pixels;
  guchar ** pixels_rows;

  // Some convenience pointers for planar YUV formats
  
  guchar * y;
  guchar * u;
  guchar * v;

  guchar ** y_rows;
  guchar ** u_rows;
  guchar ** v_rows;
  
  gpointer data;         // To store additional data
  GVVideoFrame * next;

  void alloc(const GVVideoFormat * format);
  void free();

  GVVideoFrame();
  
  };
